<module/rex/
<section/Synopsis/

[[
cheia.load "rex"

start, finish, captures = rex.find("text", "regular expression")
matches = rex.gsub("text", "regular expression",
   function(match, captures) ... end)

regex = rex.comp("regular expression")
start, finish, captures = regex:match("text")
matches = regex:gmatch("text",
   function(match, captures) ... end)
]]

<section/Description/

This module provides facilities for pattern matching using POSIX
extended regular expressions (<q/regexes/). Regular expressions are
templates which match certain classes of string. In this respect, they
are much like shell <q/globs/ (i.e. pathnames containing the wildcard
characters <q/tt[*]/ and <q/tt[?]/). However, they are much more powerful. I
will not attempt to describe them completely here, but to summarise
the main facilities:

<ul>

If no special characters are used, a regular expression is simply a
string which may be searched for inside another, e.g. the regex ["is"]
matches twice in the string ["this is a test"]: first on the last two
characters of <q/this/, and then again on the word <q/is/.

The <q/tt[.]/ character matches any character except a newline, thus
["t."] also matches twice: the first two letters of <q/this/, and the
first two letters of <q/test/, however the final <q/t/ is not matched,
because there is no following character for the <q/./ to match.

The <q/tt[*]/ character causes the preceeding atom (for the moment an atom
is either a single letter or a dot) to be matched by zero or more
repetitions, so ["s.*t"] matches once: ["s is a test"]. Notice that
there are a number of shorter substrings that this expression could
also match. These are not matched because regexes are <q/greedy/: they
take the first match they find and try to make it as long as possible.

The <q/tt[+]/ character is like <q/tt[*]/, but causes the preceeding atom to
match one or more times. The <q/tt[?]/ character causes it to match zero
or one times.

Two expressions can be made alternatives using <q/tt[|]/,
e.g. ["this|test"] matches <q/this/ and <q/test/.

A set of characters can be specified as alternatives using square
brackets. Such a grouping as treated as a single atom, e.g. ["[his]+"]
means <q/one or more occurrences of h, i or s in a row/, and matches 3
times: the last three letters of <q/this/, the word <q/is/, and and
the <q/s/ from <q/test/. If the first character of a bracketed group
is a <q/tt[^]/, then all further characters <e/not/ listed will
match. Character ranges can also be specified, e.g. ["[0-9]"] matches
any digit.

Parts of a regex can become <q/captures/ by placing them in
parentheses. A parenthesised group is also treated as a single
atom. Captured substrings are returned separately by the functions
below.

Any special character can be escaped by preceding it with a
backslash. Be aware that the backslash itself will need escaping
within quoted strings. For this reason the [\,[[...]]] string notation
may be preferable.

</ul>

<section/Examples/

Compile a regex to match identifiers, then print each identifier found
in a string:

[[
ids = rex.comp("[A-Za-z_][A-Za-z0-9_]*")
ids:gmatch("_X6 = 5 * sin(theta)",
   function(s,t) print(s) end) \= _X6
\, \= sin
\, \= theta
]]

<section/Reference/

doc[[

=fn rex.comp
=in regex string	The regular expression string.
=out userdata		The compiled form.

=d Compiles the supplied regular expression. Once it has been
compiled, it may be used repeatedly without incurring the overhead of
recompiling.

=fn rex.find
=in text string		The text to search.
=in regex string	The regular expression to find.
=out number		The 1-based index of the first character of the first matching substring.
=out number		The 1-based index of the last character of the first matching substring.
=out table		An array of captured strings.

=d Searches for [regex] within [text], and returns the start and end
of the first matching substring followed by a table of
captures. Returns [nil]s if no match is found. Largely equivalent to
the [string.find] function included with Lua, but uses the more
standard POSIX notation. Avoid using this function repeatedly with the
same regular expression and consider compiling the expression in
advance instead.

<note>[string.find] returns captures as additional results.
[rex.find] returns a table of captures.</note>

=fn rex.gsub
=in text string		The text to search.
=in regex string	The regular expression to find.
=in fn function		The function to call when a matching substring is found.
=inopt n number		The maximum number of substitutions to perform.
=out number		The number of matches found.

=d Searches for [regex] within [text], and for each matching substring
calls [fn], passing the whole matching substring as the first
argument, and a table (array) of captures as the second. Provides some
of the functionality of the [string.gsub] function included with Lua,
but uses the more standard POSIX notation. Avoid using this function
repeatedly with the same regular expression and consider compiling the
expression in advance instead.

<note>Despite the name, this function does not do substitution. It
does not accept a string as the third argument, and it does not return
a modified string. The supplied function is also called with different
parameters to [string.gsub].</note>

]]

<subsection/Compiled Regular Expressions/

The following methods are available for the compiled regular
expressions returned by [rex.comp]:

doc[[

=fn regex:match
=in text string		The text to search.
=out number		The 1-based index of the first character of the first matching substring.
=out number		The 1-based index of the last character of the first matching substring.
=out table		An array of captured strings.

=d Searches for [regex] within [text], and returns the start and end
of the first matching substring followed by a table of captures.

=fn regex:gmatch
=in text string		The text to search.
=in fn function		The function to call when a matching substring is found.
=inopt n number		The maximum number of substitutions to perform.
=out number		The number of matches found.

=d Searches for [regex] within [text], and for each matching substring
calls [fn], passing the whole matching substring as the first
argument, and a table (array) of captures as the second.

]]

<section/Issues/

The differences between the [rex] and [string] functions of the same
name could be confusing.

<section/See also/

<ref/pcre/.

The UNIX manual page <manpage name="regex" section="7"/>.

<section/Revision history/

Added in LuaCheia 5.0.

<section/Credits/

Based on the [regex] library by Henry Spencer.

Lua module by Reuben Thomas.

Modifications for LuaCheia by Thatcher Ulrich.

Documentation for LuaCheia by Jamie Webb.
