/*
** $Id: loadlib.c,v 1.5 2004/01/06 20:48:22 tulrich Exp $
** Dynamic library loader for Lua
** See Copyright Notice in lua.h
*
* This  Lua library  exports a  single function,  called loadlib,  which is
* called from Lua  as loadlib(lib,init), where lib is the  full name of the
* library to be  loaded (including the complete path) and  init is the name
* of a function  to be called after the library  is loaded. Typically, this
* function will register other functions,  thus making the complete library
* available  to Lua.  The init  function is  *not* automatically  called by
* loadlib. Instead,  loadlib returns  the init function  as a  Lua function
* that the client  can call when it  thinks is appropriate. In  the case of
* errors,  loadlib  returns  nil  and two  strings  describing  the  error.
* The  first string  is  supplied by  the operating  system;  it should  be
* informative and useful  for error messages. The second  string is "open",
* "init", or  "absent" to identify  the error and is  meant to be  used for
* making  decisions without  having to  look into  the first  string (whose
* format is system-dependent).
*
* This module contains  an implementation of loadlib for  Unix systems that
* have dlfcn, an implementation for Windows,  and a stub for other systems.
* See  the list  at  the end  of  this  file for  some  links to  available
* implementations of dlfcn  and interfaces to other  native dynamic loaders
* on top of which loadlib could be implemented.
*
*/

#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"


#undef LOADLIB


#ifdef USE_DLOPEN
#define LOADLIB
/*
* This is an implementation of loadlib based on the dlfcn interface.
* The dlfcn interface is available in Linux, SunOS, Solaris, IRIX, FreeBSD,
* NetBSD, AIX 4.2, HPUX 11, and  probably most other Unix flavors, at least
* as an emulation layer on top of native functions.
*/

#include <dlfcn.h>

/* OpenBSD apparently doesn't define RTLD_NOW */
# ifndef RTLD_NOW
#  define RTLD_NOW RTLD_LAZY
# endif

static int loadlib(lua_State *L)
{
 const char *path=luaL_checkstring(L,1);
 const char *init=luaL_checkstring(L,2);
 void *lib=dlopen(path,RTLD_NOW);
 if (lib!=NULL)
 {
  lua_CFunction f=(lua_CFunction) dlsym(lib,init);
  if (f!=NULL)
  {
   lua_pushlightuserdata(L,lib);
   lua_pushcclosure(L,f,1);
   return 1;
  }
 }
 /* else return appropriate error messages */
 lua_pushnil(L);
 lua_pushstring(L,dlerror());
 lua_pushstring(L,(lib!=NULL) ? "init" : "open");
 if (lib!=NULL) dlclose(lib);
 return 3;
}

#endif



/*
** In Windows, default is to use dll; otherwise, default is not to use dll
*/
#ifndef USE_DLL
#ifdef _WIN32
#define USE_DLL	1
#else
#define USE_DLL	0
#endif
#endif


#if USE_DLL
#define LOADLIB
/*
* This is an implementation of loadlib for Windows using native functions.
*/

#include <windows.h>

static void pusherror(lua_State *L)
{
 int error=GetLastError();
 char buffer[128];
 if (FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
	0, error, 0, buffer, sizeof(buffer), 0))
  lua_pushstring(L,buffer);
 else
  lua_pushfstring(L,"system error %d\n",error);
}

static int loadlib(lua_State *L)
{
 const char *path=luaL_checkstring(L,1);
 const char *init=luaL_checkstring(L,2);
 HINSTANCE lib=LoadLibrary(path);
 if (lib!=NULL)
 {
  lua_CFunction f=(lua_CFunction) GetProcAddress(lib,init);
  if (f!=NULL)
  {
   lua_pushlightuserdata(L,lib);
   lua_pushcclosure(L,f,1);
   return 1;
  }
 }
 lua_pushnil(L);
 pusherror(L);
 lua_pushstring(L,(lib!=NULL) ? "init" : "open");
 if (lib!=NULL) FreeLibrary(lib);
 return 3;
}

#endif


#ifndef LOADLIB
# ifndef USE_DYLD
#  if defined(__APPLE__) && defined(__MACH__) && !defined(USE_DYLD)
#   define USE_DYLD
#  endif
# endif
#endif


#ifdef USE_DYLD
#define LOADLIB
#if 1
/* Eli's code */
#include <mach-o/dyld.h>

static int loadlib(lua_State *L)
{
	const char *path=luaL_checkstring(L,1);
	const char *init=luaL_checkstring(L,2);

	NSObjectFileImage image;
	NSModule module = NULL;
	NSObjectFileImageReturnCode res;
	NSSymbol sym;
	const char *symname;

	res = NSCreateObjectFileImageFromFile(path, &image);
	if (res != NSObjectFileImageSuccess)
	{
		lua_pushnil(L);
		switch (res)
		{
			case NSObjectFileImageFailure:
				lua_pushstring(L,"Unknown error.");
				break;
			case NSObjectFileImageInappropriateFile:
				lua_pushstring(L,"File is invalid type.");
				break;
			case NSObjectFileImageArch:
				lua_pushstring(L,"File is for wrong CPU architecture.");
				break;
			case NSObjectFileImageFormat:
				lua_pushstring(L,"File does not appear to be a Mach-O file.");
				break;
			case NSObjectFileImageAccess:
				lua_pushstring(L,"Permission denied.");
				break;
			default:
				lua_pushstring(L,"Unknown error.");
		}
		lua_pushstring(L,"open");
		return 3;
	}

	module = NSLinkModule(image, path, NSLINKMODULE_OPTION_PRIVATE | NSLINKMODULE_OPTION_RETURN_ON_ERROR);

	if (module == NULL)
	{
		NSLinkEditErrors dylderr;
		int dylderrno;
		const char *dyldfile;
		const char *errstr;
		NSLinkEditError(&dylderr, &dylderrno, &dyldfile, &errstr);

		lua_pushnil(L);
		lua_pushstring(L,errstr);
		lua_pushstring(L,"init");
		return 3;
	}

	NSDestroyObjectFileImage(image);

	lua_pushfstring(L, "_%s", init);
	symname = lua_tostring(L, -1);
	sym = NSLookupSymbolInModule(module, symname);
	if (sym == NULL)
	{
		lua_pushnil(L);
		lua_pushstring(L,symname);
		lua_pushstring(L,"init");
		lua_remove(L, -4);
		return 3;
	}

	lua_pop(L, 1);
	lua_pushlightuserdata(L, module);
	lua_pushcclosure(L, NSAddressOfSymbol(sym), 1);
	return 1;
} 
#else /* 0/1 */
/*
 * This code by Adrian Perez, Asko Kauppi & Thatcher Ulrich contains
 * native OS X support for Lua's loadlib (without the need of the
 * 'dlcompat' library).  The original code by Adrian Perez and Asko
 * Kauppi is more modular and exposes a C API; Thatcher Ulrich has cut
 * it down to the minimum necessary to implement loadlib().
 * 
 * The original code can be found in Asko Kauppi's LuaX project.
 */

/*
 * Small (and fast) wrapper around the Dyld API to dynamically load
 * modules. Due to limitations in OSX's public APIs, only modules
 * with a MH_BUNDLE header can be loaded; you can create one of
 * these by passing the '-bundle' option to Apple's GCC when linking
 * the module.
 */

#include <mach-o/dyld.h>
#include <string.h>   /* strlen() */


static int loadlib(lua_State *L)
{
 const char *path=luaL_checkstring(L,1);
 const char *init=luaL_checkstring(L,2);

 NSObjectFileImageReturnCode ret;
 NSObjectFileImage nsimg;
 NSModule nsmod;

 if (!path) return NULL;

 /* Load image from file:
  */
 ret = NSCreateObjectFileImageFromFile(path, &nsimg);

 if (ret != NSObjectFileImageSuccess)
 {
  /* Unable to load! */
  lua_pushnil(L);
  lua_pushstring(L, "failure in NSCreateObjectFileImageFromFile");
  /* @@ would be good to get a real error msg from the OS */
  lua_pushstring(L, "open");
  return 3;
 }
	
 /* Link into the rest of the executable:
  */
 /* According to Apple's documentation, we can destroy the object image
  * after linking it: although we instruct Dyld to do that, Dyld will
  * keep it in memory as long as it is linked in the executable (i.e:
  * until NSUnLinkModule() is called). 
  *
  * This way we can just return the NSModule after releasing the
  * image.
  */
 nsmod = NSLinkModule(
  nsimg,
  path,
  NSLINKMODULE_OPTION_PRIVATE | NSLINKMODULE_OPTION_RETURN_ON_ERROR);
 NSDestroyObjectFileImage(nsimg);
 if (!nsmod)
 {
  /* Unable to link! */
  lua_pushnil(L);
  lua_pushstring(L, "failure in NSLinkModule");	/* @@ would be good to get a real error msg from the OS */
  lua_pushstring(L, "open");
  return 3;
 }

 {
  NSSymbol nss;
  /* tulrich: what the heck is going on here with the dynamically
   * sized stack string?  Is this relying on a GCC extension? */
  char real_name[ strlen(init) + 2 ]; 	/* '_'+init */

  real_name[0]='_';
  strcpy( real_name+1, init );
    
  nss = NSLookupSymbolInModule(nsmod, real_name);
  if (nss)
  {
   /* Success -- return the function */
   lua_CFunction f = (lua_CFunction) NSAddressOfSymbol(nss);
   lua_pushlightuserdata(L,nsmod);
   lua_pushcclosure(L,f,1);
   return 1;
  }
 }

 /* Couldn't find desired function. */
 lua_pushnil(L);
 lua_pushstring(L, "can't find function in library");	/* @@ would be good to get a real error msg from the OS */
 lua_pushstring(L, "init");
 return 3;
}
#endif /* 0/1 */  
#endif /* USE_DYLD */



#ifndef LOADLIB
/* Fallback for other systems */

/*
** Those systems support dlopen, so they should have defined USE_DLOPEN.
** The default (no)implementation gives them a special error message.
*/
#ifdef linux
#define LOADLIB
#endif

#ifdef sun
#define LOADLIB
#endif

#ifdef sgi
#define LOADLIB
#endif

#ifdef BSD
#define LOADLIB
#endif

#ifdef _WIN32
#define LOADLIB
#endif

#ifdef LOADLIB
#undef LOADLIB
#define LOADLIB	"`loadlib' not installed (check your Lua configuration)"
#else
#define LOADLIB	"`loadlib' not supported"
#endif

static int loadlib(lua_State *L)
{
 lua_pushnil(L);
 lua_pushliteral(L,LOADLIB);
 lua_pushliteral(L,"absent");
 return 3;
}
#endif

LUALIB_API int luaopen_loadlib (lua_State *L)
{
 lua_register(L,"loadlib",loadlib);
 return 0;
}

/*
* Here are some links to available implementations of dlfcn and
* interfaces to other native dynamic loaders on top of which loadlib
* could be implemented. Please send contributions and corrections to us.
*
* AIX
* Starting with AIX 4.2, dlfcn is included in the base OS.
* There is also an emulation package available.
* http://www.faqs.org/faqs/aix-faq/part4/section-21.html
*
* HPUX 
* HPUX 11 has dlfcn. For HPUX 10 use shl_*.
* http://www.geda.seul.org/mailinglist/geda-dev37/msg00094.html
* http://www.stat.umn.edu/~luke/xls/projects/dlbasics/dlbasics.html
*
* Macintosh, Windows
* http://www.stat.umn.edu/~luke/xls/projects/dlbasics/dlbasics.html
*
* Mac OS X/Darwin
* http://www.opendarwin.org/projects/dlcompat/
*
* GLIB has wrapper code for BeOS, OS2, Unix and Windows
* http://cvs.gnome.org/lxr/source/glib/gmodule/
*
*/
