-- cheia_self_test module.  Pure Lua.

cheia_self_test = {}

function cheia_self_test.help ()
	return [[
**********************************************************************
cheia_self_test.help

Run "cheia_self_test.run()" to run the module self-tests.  The list of
modules to test is in the table "cheia_self_test.module_list".

You can test a specific module with
"cheia_self_test.test_module('modulename')"
**********************************************************************
]] end

-- the modules to test
cheia_self_test.module_list =
{
	"bit",
	"cgi",
	"fs",
	"fuzzy",
	"luacom",
	"mapm",
	"md5",
	"pack",
	-- "pcre",
	"posix",
	"rex",
	"SDL",
	"shelve",
	"socket",
	"sqlite",
}
	

function cheia_self_test.run()
-- run self tests on a series of modules.

	local function write_failed(mod, msg)
	-- helper, take module name & message, and print formatted failure message
		if not msg then msg = "no error message!" end

		local name_len = string.len(mod)
		local pad_spaces = 15 - name_len;

		io.write(mod);
		if (pad_spaces > 0) then io.write(string.rep(" ", pad_spaces)) end
		if (msg == "module not found") then
			io.write("skipped - ")
		else
			io.write("failed - ")
		end
		io.write(msg)
		io.write("\n")
	end

	local function write_ok(mod, msg)
	-- helper, take module name & *optional* message, and print
	-- formatted failure message
		local name_len = string.len(mod)
		local pad_spaces = 15 - name_len;

		io.write(mod);
		if (pad_spaces > 0) then io.write(string.rep(" ", pad_spaces)) end
		io.write("OK")
		if msg then io.write(" - ") io.write(msg) end
		io.write("\n")
	end

	local	ok_count = 0
	local	failed_count = 0
	table.foreach(cheia_self_test.module_list, function (_, mod)
		local res, msg = cheia_self_test.test_module(mod)
		if (not res) then
			-- failure
			failed_count = failed_count + 1
			write_failed(mod, msg)
		else
			-- success
			ok_count = ok_count + 1
			write_ok(mod, msg)
		end
	end)

	io.write(string.format("module self-tests: %d ok, %d failed/skipped\n", ok_count, failed_count))
end


function cheia_self_test.test_module(mod)
-- self-test the given module, if possible.

	-- load the module
	local res, msg = cheia.load(mod)
	if not res then
		-- result of test: can't load.
		if not msg then
			-- perhaps the stub loader is not returning a value???
			return false, "stub loader returned false, with no error message!  Stub-loaders must return true on success."
		elseif string.find(msg, "could not locate package") then
			return false, "module not found"
		end
		return false, msg
	end

	-- make sure the module made a global table under its name
	if type(_G[mod]) ~= "table" and type(_G[mod]) ~= "userdata" then
		-- module apparently didn't load properly
		return false, "module did not create its table"
	end

	-- look for a self-test function
	local function get_self_test(mod)
		return _G[mod]._self_test
	end

	local self_test_func
	do
		local res, fn = pcall(get_self_test, mod)
		if (not res) then
			-- error trying to index _G[mod]._self_test
			return false, ("can't get _self_test(): " .. fn)
		end
		self_test_func = fn
	end

	if self_test_func == nil then
		-- failure: no self-test function.
		return false, "no _self_test function"
	end

	-- run protected, in case of inadvertent errors.  E.g. mod._self_test
	-- might not actually be callable.
	-- Also, we want to catch asserts and errors.
	local res0, res_msg1, msg = pcall(self_test_func)
	if not res0 then
		-- couldn't even run the self-test function.
		return false, "can't run _self_test: " .. res_msg1
	end

	-- return result of mod._self_test()
	return res_msg1, msg
end


return true
