-- luacheia.lua
-- implement cheia.load() and cheia.help(), plus the utility cheia.gluaload()

-- create global cheia table.
if cheia == nil then cheia = {} end

-- set up cheia.platform
cheia.platform = os.getenv("LUACHEIA5_PLATFORM")

-- set up cheia.dll_extension
if cheia.platform == "WINDOWS" then
	cheia.dll_extension = ".dll"
else
	cheia.dll_extension = ".so"
end

-- set up cheia.module_paths table, based on environment vars
do
	local path_separator_pattern = "[^:]+"
	if cheia.platform == "WINDOWS" then path_separator_pattern = "[^;]+" end

	cheia.module_paths = {}
	local p
	for p in string.gfind(os.getenv("LUACHEIA5_PATH"), path_separator_pattern) do
		table.insert(cheia.module_paths, p)
	end
end -- cheia.module_paths


do  -- the cheia.load() functionality
	local Public = cheia or {} -- use the 'cheia' table if present
	cheia = Public
	if cheia.load then return end 
	local Private= {}
	Private.loaded = {}

	-- turn this on to get progress messages during module loading.
	local verbose = false

	function Private.locate(name)
		local _, path
		local tried = ""
		for _ , path in cheia.module_paths do
			path = path.."/"..name..".lua"
			if (verbose) then print("trying "..path) end
			local chunk, err = loadfile(path)
			if chunk then 
				if (verbose) then print("found "..path, chunk) end
				return chunk, path 
			else
				-- keep looking...
				if tried ~= "" then tried = tried .. ", " end
				tried = tried .. path
			end
		end
		return nil, tried
	end 
	
	function Public.load(name, force)
		local doload = false
		if (not Private.loaded[name]) then doload = true end 
		if force then doload = true end 
		if not doload then return true end
		
		local chunk, path = Private.locate(name)
		if not chunk then 
			return false, string.format("could not locate package `%s'; tried %s", name, path)
		else
			if (verbose) then print('chunk: ', chunk) end
			local pcall_result, result, err = pcall(chunk) -- execute the chunk
			if not pcall_result then
				return false, result
			elseif not result then 
				return false, err
			else 
				Private.loaded[name] = true -- mark as already loaded
				return true
			end
		end
	end

	-- Attempts to load a gluax module
	function Public.gluaload(name)
		-- make sure we have the gluahost module
		local result, err = cheia.load("gluahost")
		if not result then
			return result, err
		end

		-- locate the .dll or .so file...
		local _, path
		for _, path in cheia.module_paths do
			path = path .. "/lua" .. name .. cheia.dll_extension
			if (verbose) then print("trying "..path) end
			local result, err = pcall(gluahost.gluahost, path, name)
			if result then 
				if (verbose) then print("loaded "..path) end
				return true
			else
				return false, err
			end
		end
		return false, path or ""
	end

	-- a version of loadlib that takes a luacheia module name,
	-- adds the appropriate prefixes and suffixes, and looks for the actual
	-- library in cheia.module_paths[].
	--
	-- for example:
	--
	--   cheia.loadlib("mylib", "my_entry_function")
	--
	-- on Windows, this tries to load .../luacheia5/luamylib.dll,
	-- looks for the exported function "my_entry_function", and on
	-- success, returns a lua function that can be used to call
	-- "my_entry_function".
	--
	-- This is a convenience for the module loaders of
	-- loadlib-style modules.
	function Public.loadlib(libname, functionname)
		-- locate the .dll or .so file
		local _, path
		local tried = ""
		for _, path in cheia.module_paths do
			path = path .. "/lua" .. libname .. cheia.dll_extension
			if (verbose) then print("trying "..path) end
			local init_fn, err1, err2 = loadlib(path, functionname)
			if init_fn == nil then
				if err2 == "absent" or err2 == "open" then
					-- Didn't find it in this path; keep iterating.
					if tried ~= "" then tried = tried .. ", " end
					tried = tried .. path
				else
					return false, err1, err2
				end
			else
				-- return the init function
				return init_fn
			end
		end
		return false, string.format("tried %s", tried), "absent"
	end

	-- @@ quick and dirty loadmodule -- just uses cheia.loadlib() -- doesn't do version checking yet!
	function Public.loadmodule(name)
		local init_fn, err1, err2 = cheia.loadlib(name, "luaLM_import")
		if init_fn == nil then
			return false, "[" .. err2 .. "] " .. err1
		end

		-- try the init function
		local result, err = pcall(init_fn)
		if result then 
			if (verbose) then print("loaded "..path) end
			return true
		end

		return false, err
	end

end -- cheia.load stuff


-- Some minimal help.
function cheia.help()
	local exit_key = "Ctrl-D"
	if cheia.platform == "WINDOWS" then exit_key = "Ctrl-Z Return" end

	return [[
**********************************************************************
* cheia.help
* 
* LuaCheia is a full-featured, modular Lua distribution, using the
* standard Lua distribution as its core.
* 
* * See http://luacheia.lua-users.org for more info.
* 
* To exit, type 'os.exit()' Return or ]] .. exit_key .. [[.
* Run with "luacheia5 -h" for a summary of command-line options.
* Type =cheia.load("modulename") to load a module.
* Type =cheia.self_test() to run self-tests on LuaCheia.
**********************************************************************
]]
end

function cheia.self_test()
	assert(cheia.load("cheia_self_test"))
	cheia_self_test.run()
end


-- welcome message
function cheia.print_welcome()
	print(string.rep('*', 70))
	print([[* Hello and welcome to LuaCheia 5.0.1a5]])
	print ("* Platform: ".. cheia.platform)
	print([[* type =cheia.help() for more info]])
	print(string.rep('*', 70))
end


-- print the welcome message if the wrapper thought we should.
if os.getenv("LUACHEIA5_PRINT_WELCOME") == "1" then
	cheia.print_welcome()
end

-- set the prompt to something distinctive
_PROMPT = "LC>"


return true
