/* luacheia_wrapper.c -- Copyrighted under the "MIT License", see LICENSE.TXT for details */

/* A wrapper program for invoking a standard Lua 5.x interpreter with
 * some variables and flags set so that luacheia support is available.
 * 
 * All this program does is set up some configuration for the luacheia
 * scripts and code, and then try to call the standard Lua.  It
 * doesn't contain a Lua interpreter itself.
 *
 * As a user, you can skip this wrapper entirely and still use
 * luacheia features, by setting the proper environment variables as
 * listed below.
 */

/* We need to set the following environment variables for the benefit
 * of the Lua interpreter:

 LUACHEIA5_PATH
 
   -- on Windows, set this to the location of this executable, so you
      can easily copy the distro tree anywhere on your disk, and it
      will work in-place.

   -- on other systems (POSIX, DARWIN, etc) set this to a fixed
      config-driven default, typically something like
      "/usr/lib/luacheia5".

   -- on all platforms, if this var had a previous value, then append
      the previous value to the platform-specific value.

  LUACHEIA5_PLATFORM

   -- set this to the platform string (from LC_PLATFORMNAME)

  LUA_INIT

   -- on all platforms, set this to '@' plus the location of the
      luacheia bootscript.  I.e. probably "@${LUACHEIA5_PATH}/luacheia.lua"

  LUACHEIA5_PRINT_WELCOME

   -- set this to "1" if the luacheia boot script should print a
      welcome message.  We try to determine this the same way the Lua
      interpreter does: if there are no args, and stdin is a tty, then
      we print the welcome.  Also print a welcome if the user passes
      -v as a command line arg.

  Then we need to invoke the Lua interpreter executable.  This involves:

   1. finding the interpreter.  If the environment var LUACHEIA5_LUA
      is set, then use that as the name of the interpreter program.
      Otherwise, on Windows, look in the same directory we are for
      "lua5.exe".  On other platforms, use a config-driven default,
      typically something like "/usr/bin/lua5"
 
   2. invoking the interpreter, passing all our args through to it.

*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#ifdef PLATFORM_WINDOWS
#	include <process.h>
#	define putenv _putenv
#	define execv _execv
#	define stdin_is_tty() 1 /* assume stdin is a tty */  /* this should match lua/src/lua/lua.c */
#else /* not PLATFORM_WINDOWS */
#	include <unistd.h>
#	include <errno.h>
#	define stdin_is_tty() isatty(0) /* this should match lua/src/lua/lua.c */
#endif /* not PLATFORM_WINDOWS */

/* Define symbols for env-var names, so the compiler catches typos. */
#define LUACHEIA5_LUA "LUACHEIA5_LUA"
#define LUACHEIA5_PATH "LUACHEIA5_PATH"
#define LUACHEIA5_PLATFORM "LUACHEIA5_PLATFORM"
#define LUACHEIA5_PRINT_WELCOME "LUACHEIA5_PRINT_WELCOME"
#define LUA_INIT "LUA_INIT"

/* Define a string for our platform name. */
#define POSIX_PLATFORM_NAME     "POSIX"
#define WINDOWS_PLATFORM_NAME   "WINDOWS"
#define OSX_PLATFORM_NAME       "OSX"
#define UNKNOWN_PLATFORM_NAME   "UNKNOWN"

#ifdef PLATFORM_POSIX
#	define LC_PLATFORMNAME POSIX_PLATFORM_NAME
#elif defined(PLATFORM_WINDOWS) 
#	define LC_PLATFORMNAME WINDOWS_PLATFORM_NAME
#elif defined(PLATFORM_OSX) 
#	define LC_PLATFORMNAME OSX_PLATFORM_NAME
#else
#	define LC_PLATFORMNAME UNKNOWN_PLATFORM_NAME
#endif


char* concat(const char* a, const char* b)
/* concatenate the two strings into a newly malloc'd buffer.
 * 
 * NOTE: to avoid leaks, caller must free the returned buffer!
 */
{
	int	len_a, len_b;
	char* result;

	assert(a);
	assert(b);

	len_a = strlen(a);
	len_b = strlen(b);
	result = malloc(len_a + len_b + 1);
	
	memcpy(result, a, len_a);
	memcpy(result + len_a, b, len_b);
	result[len_a + len_b] = 0;

	return result;
}


char* concat3(const char* a, const char* b, const char* c)
/* concatenate the three strings into a newly malloc'd buffer.
 * 
 * NOTE: to avoid leaks, caller must free the returned buffer!
 */
{
	int	len_a, len_b, len_c;
	char* result;

	assert(a);
	assert(b);
	assert(c);

	len_a = strlen(a);
	len_b = strlen(b);
	len_c = strlen(c);
	result = malloc(len_a + len_b + len_c + 1);
	
	memcpy(result, a, len_a);
	memcpy(result + len_a, b, len_b);
	memcpy(result + len_a + len_b, c, len_c);
	result[len_a + len_b + len_c] = 0;

	return result;
}


char* append(char* a, const char* b)
/* Combine a and b into a new buffer, freeing a's buffer. */
{
	int	len_a, len_b;
	char* result;

	assert(a);
	assert(b);

	len_a = strlen(a);
	len_b = strlen(b);
	result = realloc(a, len_a + len_b + 1);
	
	memcpy(result + len_a, b, len_b);
	result[len_a + len_b] = 0;

	return result;
}


void my_setenv(const char* var, const char* value)
/* Similar to setenv(), but uses putenv() since Win32 apparently does
 * not have setenv(). */
{
	char* str0 = concat3(var, "=", value);
	putenv(str0);
	/* don't free str0 -- this causes a leak, but the SUSv2 spec requires that */
	/* str0 become a part of the environment, so we really can't destroy it. */
}


#ifdef PLATFORM_WINDOWS


/* Windows needs more careful hand-holding to spawn the
 * interpreter process */

#include <windows.h>


char*	append_with_windows_escapes_and_add_space(char* so_far, const char* new_string)
{
	int	len_a, len_b;
	char* result;
	const char* p;
	char* q;

	assert(so_far);
	assert(new_string);

	len_a = strlen(so_far);

	/* Scan new_string, looking for characters we need to escape (double-quote). */
	p = new_string;
	len_b = 0;
	while (*p)
	{
		if (*p == '\"' /* || *p == '\\' */)
		{
			len_b++;
		}
		p++;
		len_b++;
	}
	len_b += 3;	/* need room for leading and trailing double-quotes, and a final trailing space. */
	result = realloc(so_far, len_a + len_b + 1);
	
	/* result already contains so_far, due to realloc. */

	/* append new_string, quoted and escaped. */
	p = new_string;
	q = result + len_a;
	*q++ = '\"';	/* leading quote. */
	while (*p)
	{
		if (*p == '\"' /* || *p == '\\' */)
		{
			/* Escape the following char. */
			*q++ = '\\';
		}
		*q++ = *p++;
	}
	*q++ = '\"';	/* trailing quote. */
	*q++ = ' ';	/* trailing space. */
	*q++ = 0;	/* terminate. */

	assert(q - result == len_a + len_b + 1);

	return result;
}


void windows_exec(const char* program, int argc, const char* argv[])
{
	int i;
	DWORD result = 1;
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	char*	cmd_line;

	memset(&si, 0, sizeof(si));
	si.cb = sizeof(si);
	memset(&pi, 0, sizeof(pi));

	/* Build the command line. */
	cmd_line = malloc(1);
	cmd_line[0] = 0;
	for (i = 0; i < argc; i++)
	{
		cmd_line = append_with_windows_escapes_and_add_space(cmd_line, argv[i]);
	}

	/* Start the child process. */
	if (!CreateProcess(
		program,
		cmd_line,		/* Command line. */
		NULL,			/* Process handle not inheritable. */
		NULL,			/* Thread handle not inheritable. */
		TRUE,			/* Set handle inheritance to TRUE. */
		0,			/* No creation flags. */
		NULL,			/* Use parent's environment block. */
		NULL,			/* Use parent's starting directory. */
		&si,			/* Pointer to STARTUPINFO structure. */
		&pi)			/* Pointer to PROCESS_INFORMATION structure. */
	    ) 
	{
		/* Something didn't work. */
		free(cmd_line);
		return;
	}

	free(cmd_line);

	/* Wait until child process exits. */
	WaitForSingleObject( pi.hProcess, INFINITE );

	/* Get the exit code. */
	GetExitCodeProcess(pi.hProcess, &result);

	/* Close process and thread handles. */
	CloseHandle( pi.hProcess );
	CloseHandle( pi.hThread );

	exit(result);
}
#endif /* PLATFORM_WINDOWS*/


int should_print_welcome(int argc, const char* argv[])
/* return true we want the bootscript to print a welcome message.  Try
 * to determine this the same way that the Lua interpreter does. */
{
	/* this code lifted and pruned from lua-5.0/src/lua/lua.c, should
	 * be kept in sync. */
	if (argv[1] == NULL) {  /* no more arguments? */
		if (stdin_is_tty()) {
			return 1;
		}
	}
	else {  /* other arguments; loop over them */
		int i;
		for (i = 1; argv[i] != NULL; i++) {
			if (argv[i][0] != '-') break;  /* not an option? */
			switch (argv[i][1]) {  /* option */
			case '-': {  /* `--' */
				if (argv[i][2] != '\0') {
					return 0;
				}
				i++;  /* skip this argument */
				return 0;  /* stop handling arguments */
			}
			case '\0': {
				break;
			}
			case 'i': {
				return 1;
				break;
			}
			case 'v': {
				return 1;
				break;
			}
			case 'e': {
				const char *chunk = argv[i] + 2;
				if (*chunk == '\0') chunk = argv[++i];
				if (chunk == NULL) {
					return 0;
				}
				break;
			}
			case 'l': {
				const char *filename = argv[i] + 2;
				if (*filename == '\0') filename = argv[++i];
				if (filename == NULL) {
					return 0;
				}
				break;
			}
			case 'c': {
				break;
			}
			case 's': {
				break;
			}
			default: {
				return 0;
			}
			}
		}
	}
	return 0;
}


int main(int argc, const char *argv[])
{
	int	result = 0;
	char* luacheia5_default_path = NULL;
#if PLATFORM_WINDOWS
	/* Construct a string containing this executable's path. */
	char*	current_path = NULL;
	const char* last_slash;

	last_slash = strrchr(_pgmptr, '\\');
	if (last_slash == NULL)
	{
		/* Fallback -- use current path! */
		current_path = malloc(1);
		current_path[0] = 0;
	} else {
		int	path_len = last_slash - _pgmptr;
		current_path = malloc(path_len + 1);
		memcpy(current_path, _pgmptr, path_len);
		current_path[path_len] = 0;
	}
#endif /* PLATFORM_WINDOWS */

	/* Set up LUACHEIA5_PATH */
	{
		char* path = NULL;
		char* path_separator = NULL;
		const char*	existing_path = getenv(LUACHEIA5_PATH);

#if PLATFORM_WINDOWS
		/* On Windows, use the directory this .exe is contained in, plus /luacheia5. */
		luacheia5_default_path = concat(current_path, "/luacheia5");
		path_separator = ";";
#else /* not PLATFORM_WINDOWS */
		/* On POSIX, DARWIN etc, we use LC_BASEDIRECTORY, from
		 * ./configure */
		luacheia5_default_path = concat("", LC_BASEDIRECTORY);
		path_separator = ":";
#endif /* not PLATFORM_WINDOWS */

		assert(luacheia5_default_path);
		assert(path_separator);

		path = concat("", luacheia5_default_path);

		/* append existing path if any*/
		if (existing_path && existing_path[0])
		{
			path = append(path, path_separator);
			path = append(path, existing_path);
		}

		assert(path);

		my_setenv(LUACHEIA5_PATH, path);

		free(path);
	}
	assert(getenv(LUACHEIA5_PATH));

	/* set up LUACHEIA5_PLATFORM */
	{
		my_setenv(LUACHEIA5_PLATFORM, LC_PLATFORMNAME);
	}
	assert(getenv(LUACHEIA5_PLATFORM));

	/* Set up LUA_INIT, so that lua runs the luacheia bootscript. */
	{
		char*	value = NULL;
		assert(luacheia5_default_path != NULL);

		value = concat3("@", luacheia5_default_path, "/luacheia.lua");
		my_setenv(LUA_INIT, value);
		free(value);
	}
	assert(getenv(LUA_INIT));

	/* Set up LUACHEIA5_PRINT_WELCOME */
	if (should_print_welcome(argc, argv))
	{
		my_setenv(LUACHEIA5_PRINT_WELCOME, "1");
	}

	/* Find Lua executable, and invoke it. */
	{
		char*	lua_exe = getenv(LUACHEIA5_LUA);
		int free_lua_exe = 0;

		if (lua_exe == NULL
		    || lua_exe[0] == 0)
		{
			/* Null interpreter, so use our platform-specific
			 * defaults. */

#if PLATFORM_WINDOWS
			lua_exe = concat(current_path, "/lua5.exe");
			free_lua_exe = 1;
#else /* not PLATFORM_WINDOWS */
			lua_exe = LC_LUAINTERPRETER;
#endif /* not PLATFORM_WINDOWS */
		}

		assert(lua_exe && lua_exe[0]);	/* we should have some value for the interpreter executable */

		/* Invoke the interpreter. */
		{
#if PLATFORM_WINDOWS
			windows_exec(lua_exe, argc, argv);
#else /* not PLATFORM_WINDOWS */
			int result = execv(lua_exe, argv);
#endif /* not PLATFORM_WINDOWS */

			/* We shouldn't get to here.  Something went wrong with the exec. */
			fprintf(stderr, "luacheia_wrapper can't exec '%s', errno = %d\n", lua_exe, errno);
			exit(1);
		}

		/* Clean up */
		if (free_lua_exe)
		{
			free(lua_exe);
		}
	}

#if PLATFORM_WINDOWS
	/* Clean up */
	free(current_path);
#endif /* PLATFORM_WINDOWS */

	return result;
}
