//---
// SDL_MODULE.C   (18-Mar-2003, asko.kauppi@fi.flextronics.com)  
//
// Lua interface for Simple DirectMedia Layer.
//
// PARAMETER USAGE:
//
//   The SDL API has been reproduced in Lua as closely to the original
//   as possible. This should ease porting code C-to-Lua or vise versa
//   and also reduce the amount of Lua-specific documentation required.
//
//   However, in some cases s.a. reference variables, buffers, etc. it is
//   natural for the API to change a little (in syntax, not in concept).
//
//   In addition to the C-style usage, some functions offer more structured
//   variants (that are not C-compatible). E.g. colors or rect's can be
//   provided as a bunch of 'int's or enclosed into table parameters. This
//   can happen alongside the original C-style parameters since we can query
//   the Lua parameter types in runtime.
//
// RETURN VALUES:
//
//   Returning 'boolean' (0/1) return values is troublesome, and we cannot
//   provide 100% C compatibility here. The reason is, C uses numbers (zero/
//   nonzero) for booleans whereas Lua uses types (nil/non-nil in Lua 4.0,
//   or false/true in Lua 5.0). 
//
//   We've chosen to return C-style numbers (0/1) but this means Lua scripts
//   cannot (shouldn't!) directly use them in 'if' cases etc. The most 
//   notable example is the events loop that in C would be:
//
//		"while( SDL_PollEvent(&event) ) { ... }"
//
//   in Lua needs to be:
//
//		"while( SDL_PollEvent(event) ~= 0 ) do"
//
//   The reason is, Lua would handle '0' as true' (since it is not a 'nil')
//	 and would never exit the loop without the "~= 0".
//
//   N.B. We could of course return a Lua boolean (nil/non-nil) from this
//		function but that might lead to problems as well (if not here, then
//		perhaps with structure members that are 0/1 in C...). So it seems 
//		simplest to stick to the C convention of using 0/1 only. If you want,
//		you can easily build wrappers around these base functions in Lua 
//		to convert to booleans instead.
//
// NON-SUPPORTED FEATURES:
//
//   The following features have not been implemented since they involve
//	 function pointers in a multithreaded scenario, and it is not clear
//	 how this (if at all!) could be implemented in Lua. Suggestions are
//	 welcome!
//
//      - Timer callbacks ('SDL_AddTimer()' etc.)
//		- Event filters
//		- Audio subsystem
//		- OpenGL runtime loading ('SDL_GL_GetProcAddress()')
//
//   The audio subsystem would probably be best covered by a higher-level 
//   'SDLlua_...' API that would simply be used for playing/stopping WAVs,
//	 not for runtime mixing of the input streams (as the C SDL API is).
//
// LINKS:
//
//	 'http://www.libsdl.org' - Simple DirectMedia Layer
//   'http://www.lua.org'    - Lua scripting language
//   'http://...'            - GluaX dynamic linkage (Lua/C)
//
// TO-DO:
//    - Separate higher-level audio subsystem.
//    - Separate (or within this module) multithreading subsystem.
//    - ...
//
// 30-Jun-03 (AK): Changed rectangle implementation from tables to userdata.
// 28-Nov-02 (AK): Started the module.
//---

#ifndef __cplusplus
  #error "This module must be compiled as C++ code!"
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>	// strcmp()
#include <stdio.h>
#include <time.h>	// time()

#include "gluax.h"

extern "C" {
#include "SDL/SDL.h"

#include "SDL/SDL_mutex.h"
#include "SDL/SDL_syswm.h"  // platform specific window manager events
#include "SDL/SDL_thread.h"
}

//---
// This can be changed into 'nil'/non-nil later if required:
//
#define glua_pushInteger_01(v) \
        glua_pushInteger(v)     // <-- now pushes 0/1 as Lua return values.


/*----- Local data -----*/


/*----- Userdata tags -----*/
//
// These identify embedded data types that are accessible to Lua.
// (see Lua documentation for more info)

static int tag_surface= 0;		// 'SDL_Surface*'
//
#define glua_getSurface(i)  ((SDL_Surface*)glua_getUserdata(i,tag_surface))
#define glua_pushSurface(p) glua_pushUserdata(tag_surface,0,p)
#define glua_pushSurfaceGc(p,gc_func) \
							glua_pushUserdataGc(p,tag_surface,gc_func)

static int tag_format= 0;		// 'SDL_PixelFormat*'
//
#define glua_getFormat(i)   ((SDL_PixelFormat*)glua_getUserdata(i,tag_format))
#define glua_pushFormat(p)  glua_pushUserdata(tag_format,0,p)

static int tag_event= 0;		// 'SDL_Event*'
//
#define glua_getEvent(i)    ((SDL_Event*)glua_getUserdata(i,tag_event))
#define glua_pushEvent(p)   glua_pushUserdata(tag_event,0,p)

static int tag_joystick= 0;		// 'SDL_Joystick*'
//
#define glua_getJoystick(i)  ((SDL_Joystick*)glua_getUserdata(i,tag_joystick))
#define glua_pushJoystick(p) glua_pushUserdata(tag_joystick,0,p)

static int tag_cdrom= 0;		// 'SDL_CD*'
//
#define glua_getCdrom(i)    ((SDL_CD*)glua_getUserdata(i,tag_cdrom))
#define glua_pushCdrom(p)   glua_pushUserdata(tag_cdrom,0,p)

static int tag_cursor= 0;		// 'SDL_Cursor*'
//
#define glua_getCursor(i)   ((SDL_Cursor*)glua_getUserdata(i,tag_cursor))
#define glua_pushCursor(p)  glua_pushUserdata(tag_cursor,0,p)

static int tag_overlay= 0;		// 'SDL_Overlay*'
//
#define glua_getOverlay(i)  ((SDL_Overlay*)glua_getUserdata(i,tag_overlay))
#define glua_pushOverlay(p) glua_pushUserdata(tag_overlay,0,p)

static int tag_thread= 0;		// 'SDL_Thread*'
//
#define glua_getThread(i)  ((SDL_Thread*)glua_getUserdata(i,tag_thread))
#define glua_pushThread(p) glua_pushUserdata(tag_thread,0,p)

static int tag_mutex= 0;		// 'SDL_mutex*'
//
#define glua_getMutex(i)   ((SDL_mutex*)glua_getUserdata(i,tag_mutex))
#define glua_pushMutex(p)  glua_pushUserdata(tag_mutex,0,p)

static int tag_semaphore= 0;	// 'SDL_sem*'
//
#define glua_getSemaphore(i)   ((SDL_sem*)glua_getUserdata(i,tag_semaphore))
#define glua_pushSemaphore(p)  glua_pushUserdata(tag_semaphore,0,p)

static int tag_condvar= 0;		// 'SDL_cond*'
//
#define glua_getCondVar(i)   ((SDL_cond*)glua_getUserdata(i,tag_condvar))
#define glua_pushCondVar(p)  glua_pushUserdata(tag_condvar,0,p)


//---
// Two ways of dealing with rectangles (userdata and tables).
//
// Rough tests (with the MeteorShower demo) seem to show that tables are 
// actually faster than userdata in this respect. Worth more study..?
//
// Note: We read in 'num', allowing also doubles to be entered into the 'rect'
//       fields (without decimals, of course). This allows calculation results
//       to be used as coordinates.
//

// Rect as tables..
//
static SDL_Rect _rect_store={0};    // Note: this is _not_ multithreading safe!

#define /* SDL_Rect* */ glua_getRectTbl(i)  ( glua_getTable(i) ? \
    					  ( _rect_store.x= (int)(glua_getTable_num("x")+0.5), \
                            _rect_store.y= (int)(glua_getTable_num("y")+0.5), \
                            _rect_store.w= (int)(glua_getTable_num("w")+0.5), \
                            _rect_store.h= (int)(glua_getTable_num("h")+0.5), \
				            &_rect_store ) : NULL /*'nil' table*/ )

#define glua_pushRectTbl(p) glua_pushTable(); \
                              { glua_setTable_int("x",(p)->x); \
                                glua_setTable_int("y",(p)->y); \
                                glua_setTable_int("w",(p)->w); \
                                glua_setTable_int("h",(p)->h); \
                              }
// Rect as userdata
//
static int tag_rect= 0;     

#define /* SDL_Rect* */ glua_getRect(i) \
      ( glua_isUserdata(i) ? (SDL_Rect*)glua_getUserdata(i,tag_rect) \
                           : glua_getRectTbl(i) )

#define glua_pushRect(p)  glua_pushUserdata(tag_rect,0,p)


/*----- SDL functions -----*/

//---
// tbl= SDL_Info( )
//
// This is a collection of rarely needed API bits and pieces. Instead of polluting
// the namespace & making lots to learn, we just return a single table with all
// the necessary information.
//
// .version			e.g. "1.2.5"
// .linked_version	e.g. "1.2.7"
// .endian			"little" / "big"
//
GLUA_FUNC( SDL_Info )
{
SDL_version ver;
const SDL_version* ver2;
char buf[80];
const char* str;

    glua_pushTable();
		{
		//---
		SDL_VERSION( &ver );
		sprintf( buf, "%d.%d.%d", ver.major, ver.minor, ver.patch );
		//
		glua_setTable_str( "version", buf );

		//---
		ver2= SDL_Linked_Version();
		sprintf( buf, "%d.%d.%d", ver2->major, ver2->minor, ver2->patch );
		//
		glua_setTable_str( "linked_version", buf );

		//---
		str= (SDL_BYTEORDER == SDL_LIL_ENDIAN) ? "little" : "big";
		//
		glua_setTable_str( "endian", str );
		}
}
GLUA_END


//---
// void= SDL_SetSurface( surf_ref, pixels_tbl )
//
// Using this function takes locking issues etc. away from the Lua scripts.
//
// 'tbl' = 2D table, [0..height-1][0..width-1] with pixel contents
//
GLUA_FUNC( SDL_SetSurface )
{
SDL_Surface* surf= glua_getSurface(1);
uint w,h;
int tmp;

	w= surf->w;
	h= surf->h;

	glua_getTable(2);
		{
		tmp= SDL_LockSurface( surf );
			{
			Uint8* row= (Uint8*)surf->pixels;
			uint y,x;

			ASSUME( tmp >= 0 );	  // locking shouldn't fail, right?!?

			for( y=0; y<h; y++ )
				{
				for( x=0; x<w; x++ )
					{
					uint value= glua_getTable_int( y, x );

					switch (surf->format->BytesPerPixel)
						{
						case 1:	row[x]= (Uint8)value;
								break;

						case 2: ((Uint16*)row)[x]= (Uint16)value;
								break;

						case 4: ((Uint32*)row)[x]= (Uint32)value;
								break;

						default:
							ASSUME_NEVER();	  // what?!?
						}
					}
				row += surf->pitch;	// next row
				}
			}
		SDL_UnlockSurface(surf);
		}

	SDL_UpdateRect(surf,0,0,0,0);
}
GLUA_END


/*----- SDL general functions -----*/

// WARNING: This counter _should_ be treated as atomic but currently is not.
//
static volatile bool_int initialized= FALSE;

/**
static void QuitAtExit( void )
{
    if (initialized)    // If application didn't clean up, we'll do it:
        SDL_Quit();
}
**/

//---
// err_int= SDL_Init( init_flags )
//
// Note: Unlike the SDL C API, this function also works as 'InitSubSystem()',
//       further calls adding to the initialized subsystems.
//
GLUA_FUNC( SDL_Init )
{
Uint32 flags= glua_getInteger(1);
int err;

    if (!initialized)  // this _should_ be made atomic!
        {
        // First time, normal 'SDL_Init()' initialization.
        //
        err= SDL_Init( flags );

        if (err >= 0)
            initialized= TRUE;
                
        // Make sure 'SDL_Quit()' gets called even if the application
        // would not do it.
        //
		// NOTE: FOR SOME REASON HAVING THIS MAKES A CRASH AT PROGRAM EXIT.
		//
        //atexit( QuitAtExit );

		#if 1
		{
		// Help the Lua side by really randomizing the random number generator
		// here. Note: we cannot use 'SDL_GetTicks()' for this since it starts
		// with 0 each first call (at least on Win2000 & SDL 1.2.5).
		//
		srand( time(NULL) );	// secs since 1.1.1970 (or some other date)
		//
		// NOTE: THE ABOVE DOES NOT SEEM TO WORK - LUA STILL RETURNS SAME
		//       RANDOM NUMBERS AFTER LAUNCHING. WHERE'S THE TROUBLE?!?
		}
		#endif
        }
    else    // Later times, use 'SDL_InitSubSystem()':
        {
        err= SDL_InitSubSystem( flags );
        }
    
    glua_pushInteger( err );
}
GLUA_END


//---
// void= SDL_Quit( [init_flags] )
//
// Note: Unlike SDL C API, this function can take flag parameters (optional)
//       to work as 'QuitSubSystem()'. No param means unconditional quit.
//
GLUA_FUNC( SDL_Quit )
{
    if (glua_isInteger(1))     // having optional flags
        {
        SDL_QuitSubSystem( glua_getInteger(1) );
        }
    else    // total quit
        {
        SDL_Quit( );
        
        initialized= FALSE;
        }
}
GLUA_END


//---
// str= SDL_GetError( )
//
GLUA_FUNC( SDL_GetError )
{
    glua_pushString( SDL_GetError() );
}
GLUA_END


//---
// str= SDL_VERSION( )
//
GLUA_FUNC( SDL_VERSION )
{
SDL_version ver;
char buf[80];

    SDL_VERSION( &ver );
	sprintf( buf, "%d.%d.%d", ver.major, ver.minor, ver.patch );
	
	glua_pushString( buf );
}
GLUA_END


//---
// tbl= SDL_Linked_Version( )
//
GLUA_FUNC( SDL_Linked_Version )
{
const SDL_version* ver;
char buf[80];

    ver= SDL_Linked_Version();
    sprintf( buf, "%d.%d.%d", ver->major, ver->minor, ver->patch );

    glua_pushString( buf );
}
GLUA_END


/*** (deemed as unnecessary) ***
//---
// err_int= SDL_InitSubsystem( init_flags )
//
GLUA_FUNC( SDL_InitSubSystem )
{
	glua_pushInteger( SDL_InitSubSystem( glua_getInteger(1) ) );
}
GLUA_END


//---
// void= SDL_QuitSubsystem( init_flags )
//
GLUA_FUNC( SDL_QuitSubSystem )
{
    SDL_QuitSubSystem( glua_getInteger(1) );
}
GLUA_END


//---
// uint= SDL_WasInit( init_flags )
//
GLUA_FUNC( SDL_WasInit )
{
    glua_pushInteger( SDL_WasInit( glua_getInteger(1) ) );
}
GLUA_END
***/


/*----- SDL video functions -----*/

//---
// surf_ref= SDL_GetVideoSurface( void )
//
GLUA_FUNC( SDL_GetVideoSurface )
{
    glua_pushSurface( SDL_GetVideoSurface() );
}
GLUA_END


//---
// [info_tbl]= SDL_GetVideoInfo( void )
//
GLUA_FUNC( SDL_GetVideoInfo )
{
const SDL_VideoInfo* info;

    info= SDL_GetVideoInfo();

    if (info)
        {
        glua_pushTable();
            {
			// The field names are same as in SDL C code:
			//
            glua_setTable_int( "hw_available", info->hw_available ); // 0/1
            glua_setTable_int( "wm_available", info->wm_available ); // 0/1
            //
            glua_setTable_int( "blit_hw",      info->blit_hw );      // 0/1
            glua_setTable_int( "blit_hw_CC",   info->blit_hw_CC );   // 0/1
            glua_setTable_int( "blit_hw_A",    info->blit_hw_A );    // 0/1
            //
            glua_setTable_int( "blit_sw",      info->blit_sw );      // 0/1
            glua_setTable_int( "blit_sw_CC",   info->blit_sw_CC );   // 0/1
            glua_setTable_int( "blit_sw_A",    info->blit_sw_A );    // 0/1
            //
            glua_setTable_int( "blit_fill",    info->blit_fill );    // 0/1
            glua_setTable_int( "video_mem",    info->video_mem );    // 0/1
            //
            glua_setTable_ud( "vfmt", info->vfmt, tag_format );
            }
        }
}
GLUA_END


//---
// [str]= SDL_VideoDriverName( void )
//
GLUA_FUNC( SDL_VideoDriverName )
{
char buf[99];

    if (SDL_VideoDriverName( buf, sizeof(buf) ))
        {
        glua_pushString( buf );
        }
}
GLUA_END


//---
// [rect_tbl]= SDL_ListModes( format_ref, surf_flags )
//
// Returns: - non-empty table, listing the available modes
//          - empty table '{ }', if any mode is okay
//          - 'nil' if no modes are supported
//
GLUA_FUNC( SDL_ListModes )
{
SDL_PixelFormat* format= glua_getFormat(1);
Uint32 flags= glua_getInteger(2);
//
SDL_Rect** modes;
unsigned int i;

    modes= SDL_ListModes( format, flags );

    if (modes == (SDL_Rect**)0)  // no modes available?
        {
        glua_pushNil();
        }
    else if (modes == (SDL_Rect**)-1)    // anything goes
        {
        glua_pushTable();   // '{ }' = empty table
        }
    else
        {
        // 'modes' is an array of 'SDL_Rect' pointers, ending at a NULL entry.

        glua_pushTable();
            {
            for( i=0; modes[i]; i++ )
                {
				// set subtable 'tbl[i].x' etc. entries:
				//
                glua_setTable_int( i, "x", modes[i]->x );
                glua_setTable_int( i, "y", modes[i]->y );
                glua_setTable_int( i, "w", modes[i]->w );
                glua_setTable_int( i, "h", modes[i]->h );
                }
            }
        }
}
GLUA_END


//---
// int= SDL_VideoModeOK( width_int, height_int, bpp_int, surf_flags )
//
// Returns:	- 0 if requested mode is not supported
//			- bits-per-pixel of the closest available mode
//
GLUA_FUNC( SDL_VideoModeOK )
{
int w= glua_getInteger(1);
int h= glua_getInteger(2);
int bpp= glua_getInteger(3);
Uint32 flags= glua_getInteger(4);
int ret;

    ret= SDL_VideoModeOK( w, h, bpp, flags );

    glua_pushInteger(ret);
}
GLUA_END


//---
// surf_ref= SDL_SetVideoMode( w_int, h_int, bpp_int, surf_flags )
//
GLUA_FUNC( SDL_SetVideoMode )	// (used by MeteorShower)
{
int w=   glua_getInteger(1);
int h=   glua_getInteger(2);
int bpp= glua_getInteger(3);
Uint32 flags= glua_getUnsigned(4);
SDL_Surface* ref;

    ref= SDL_SetVideoMode( w, h, bpp, flags );

    glua_pushSurface( ref );
}
GLUA_END


//---
// void= SDL_UpdateRect( surf_ref, x_int, y_int, w_int, h_int )
// void= SDL_UpdateRect( surf_ref )
// void= SDL_UpdateRect( surf_ref, rect1_tbl [,rect2_tbl [, ...]] )
//
// Note: Just 'surf_ref' means 'update whole surface'.
//
GLUA_FUNC( SDL_UpdateRect )		// (used by MeteorShower)
{
SDL_Surface* ref= glua_getSurface(1);
int x,y,w,h;

	switch( glua_getType(2) )
		{
		case GLUA_TYPE_INT:
		case GLUA_TYPE_FLOAT:
			{
			x= glua_getInteger(2);
			y= glua_getInteger(3);
			w= glua_getInteger(4);
			h= glua_getInteger(5);

			SDL_UpdateRect( ref, x,y,w,h );
			}
			break;

		case GLUA_TYPE_TABLE:
			{
			SDL_Rect* rects;
			uint i,count;

			count= glua_getArgn()-1;	  // 2..N are rects

			rects= new SDL_Rect[count];
				{
				ASSUME_NONZERO( rects );
				
				for( i=0; i<count; i++ )
					{
					memcpy( &rects[i], glua_getRect(2+i), sizeof(SDL_Rect) );
					}

				SDL_UpdateRects( ref, count, rects );
				}
			delete[] rects;
			}
			break;

		//case GLUA_TYPE_NONE:
		case GLUA_TYPE_NIL:
			{
			SDL_UpdateRect( ref, 0,0,0,0 );
			}
			break;

		default:
			glua_error( "Bad params to 'SDL_UpdateRect()'!" );
			break;
		}
}
GLUA_END


//---
// void= SDL_UpdateRects( surf_ref, rect1_tbl [,rect2_tbl [, ...]] )
//
// Note: Implemented in 'SDL_UpdateRect()' alternative parameter lists.
//		 This function is just a stub for C look-alike API.
//
GLUA_FUNC( SDL_UpdateRects )
{
    glua_Forward( SDL_UpdateRect );
}
GLUA_END


//---
// err_int= SDL_Flip( surf_ref )
//
GLUA_FUNC( SDL_Flip )
{
    glua_pushInteger( SDL_Flip( glua_getSurface(1) ) );
}
GLUA_END


//---
// int_01= SDL_SetColors( surf_ref, { rgb_tbl, ... } )
//
GLUA_FUNC( SDL_SetColors )
{
SDL_Surface* ref= glua_getSurface(1);
int ret;

    glua_getTable(2);	// table of rgb's
        {
        SDL_Color* array;
        uint i, n;

        n= glua_getTableN();

        array= new SDL_Color[n];
			{
		    ASSUME_NONZERO( array );

			for( i=0; i<n; i++ )
				{
		        // Note: 'i+1' because Lua tables normally start with [1].
				//
			    array[i].r= glua_getTable_int( i+1, "r" );
				array[i].g= glua_getTable_int( i+1, "g" );
	            array[i].b= glua_getTable_int( i+1, "b" );
		        }

			ret= SDL_SetColors( ref, array, 0 /*firstcolor*/, n /*ncolors*/ );
			}
        delete[] array;
        }

    glua_pushInteger_01( ret );
}
GLUA_END


//---
// int_01= SDL_SetPalette( surf_ref, flags_int, { rgb_tbl, ... } )
//
GLUA_FUNC( SDL_SetPalette )
{
SDL_Surface* ref= glua_getSurface(1);
int flags= glua_getInteger(2);
int ret;

    glua_getTable(3);   // table of rgb entries
        {
        SDL_Color* array;
        unsigned int i, n;

        n= glua_getTableN();

        array= new SDL_Color[n];
			{
            ASSUME_NONZERO( array );
            
		    for( i=0; i<n; i++ )
			    {
				array[i].r= glua_getTable_int( i+1, "r" );
	            array[i].g= glua_getTable_int( i+1, "g" );
		        array[i].b= glua_getTable_int( i+1, "b" );
			    }

			ret= SDL_SetPalette( ref, flags, array, 0, n );
			}
        delete[] array;
        }

    glua_pushInteger_01( ret );
}
GLUA_END


//---
// err_int= SDL_SetGamma( r_num, g_num, b_num )
// err_int= SDL_SetGamma( rgb_tbl )
//
GLUA_FUNC( SDL_SetGamma )
{
double r,g,b;

    if (!glua_isTable(1))
        {
        r= glua_getNumber(1);
        g= glua_getNumber(2);
        b= glua_getNumber(3);
        }
    else
        {
        glua_getTable(1);
        
        r= glua_getTable_num("r");
        g= glua_getTable_num("g");
        b= glua_getTable_num("b");
        }

    glua_pushInteger( SDL_SetGamma( (float)r, (float)g, (float)b ) );
}
GLUA_END


//---
// [r_tbl, g_tbl, b_tbl]= SDL_GetGammaRamp( )
//
// Returns:	- nothing on error
//			- three separate integer arrays, indexes [1]..[256].
//
GLUA_FUNC( SDL_GetGammaRamp )
{
Uint16 ramp[3][256];
int tmp;
uint i,j;

    tmp= SDL_GetGammaRamp( ramp[0], ramp[1], ramp[2] );

    if (tmp>=0)     // success
        {
        for( i=0; i<3; i++ )
            {
            glua_pushTable();   // r,g,b each get their own tables
                {
                for( j=0; j<256; j++ )
                    {
                    glua_setTable_int( j+1, ramp[i][j] );
                    }
                }
            }
        }
}
GLUA_END


//---
// err_int= SDL_SetGammaRamp( r_tbl, g_tbl, b_tbl )
//
GLUA_FUNC( SDL_SetGammaRamp )
{
Uint16 ramp[3][256];
int ret;
uint i,j;

    for( i=0; i<3; i++ )
        {
        glua_getTable(i+1);
            {
            for( j=0; j<256; j++ )
                {
                ramp[i][j]= glua_getTable_int( j+1 );
                }
            }
        }

    ret= SDL_SetGammaRamp( ramp[0], ramp[1], ramp[2] );

    glua_pushInteger(ret);
}
GLUA_END


//---
// pixel_uint= SDL_MapRGB( format_ref, r_int, g_int, b_int )
// pixel_uint= SDL_MapRGB( format_ref, rgb_tbl )
// 
GLUA_FUNC( SDL_MapRGB )		// (used by MeteorShower)
{
SDL_PixelFormat* format= glua_getFormat(1);
Uint8 r,g,b;
Uint32 ret;

    if (!glua_isTable(2))
        {
        r= glua_getInteger(2);
        g= glua_getInteger(3);
        b= glua_getInteger(4);
        }
    else
        {
        glua_getTable(2);
        
        r= glua_getTable_int("r");
        g= glua_getTable_int("g");
        b= glua_getTable_int("b");
        }

    ret= SDL_MapRGB( format, r,g,b );

    glua_pushInteger( ret );
}
GLUA_END


//---
// pixel_uint= SDL_MapRGBA( format_ref, r_int, g_int, b_int, a_int )
// pixel_uint= SDL_MapRGBA( format_ref, rgba_tbl )
// 
GLUA_FUNC( SDL_MapRGBA )
{
SDL_PixelFormat* format= glua_getFormat(1);
Uint8 r,g,b,a;
Uint32 ret;

    if (!glua_isTable(2))
        {
        r= glua_getInteger(2);
        g= glua_getInteger(3);
        b= glua_getInteger(4);
        a= glua_getInteger(5);
        }
    else
        {
        glua_getTable(2);
        
        r= glua_getTable_int("r");
        g= glua_getTable_int("g");
        b= glua_getTable_int("b");
        a= glua_getTable_int("a");
        }

    ret= SDL_MapRGBA( format, r,g,b,a );

    glua_pushInteger( ret );
}
GLUA_END


//---
// rgb_tbl= SDL_GetRGB( pixel_uint, format_ref )
// 
GLUA_FUNC( SDL_GetRGB )
{
Uint32 pixel= glua_getInteger(1);
SDL_PixelFormat* format= glua_getFormat(2);
Uint8 r,g,b;

    SDL_GetRGB( pixel, format, &r,&g,&b );

    glua_pushTable();
        {
        glua_setTable_int( "r", r );
        glua_setTable_int( "g", g );
        glua_setTable_int( "b", b );
        }
}
GLUA_END


//---
// rgba_tbl= SDL_GetRGBA( pixel_uint, format_ref )
// 
GLUA_FUNC( SDL_GetRGBA )
{
Uint32 pixel= glua_getInteger(1);
SDL_PixelFormat* format= glua_getFormat(2);
Uint8 r,g,b,a;

    SDL_GetRGBA( pixel, format, &r,&g,&b,&a );

    glua_pushTable();
        {
        glua_setTable_int( "r", r );
        glua_setTable_int( "g", g );
        glua_setTable_int( "b", b );
        glua_setTable_int( "a", a );
        }
}
GLUA_END


//---
// surf_ref= SDL_CreateRGBSurface( flags_uint, w_int, h_int, depth_int, 
//								   r_mask, g_mask, b_mask, a_mask )
//
// surf_ref= SDL_CreateRGBSurface( flags_uint, w_int, h_int, depth_int,
//								   rgba_mask_tbl )
// 
GLUA_FUNC( SDL_CreateRGBSurface )
{
Uint32 flags= glua_getInteger(1);
int w= glua_getInteger(2);
int h= glua_getInteger(3);
int depth= glua_getInteger(4);
//
Uint32 r,g,b,a;
SDL_Surface* ref;

    if (!glua_isTable(5))
        {
        r= glua_getInteger(5);
        g= glua_getInteger(6);
        b= glua_getInteger(7);
        a= glua_getInteger(8);
        }
    else
        {
        glua_getTable(5);

        r= glua_getTable_int( "r" );
        g= glua_getTable_int( "g" );
        b= glua_getTable_int( "b" );
        a= glua_getTable_int( "a" );
        }

    ref= SDL_CreateRGBSurface( flags, w,h,depth, r,g,b,a );

    glua_pushSurface(ref);
}
GLUA_END


//---
// surf_ref= SDL_CreateRGBSurfaceFrom( pixels_uint_tbl, w_int, h_int, depth_int,
//									   pitch_int, r_mask,g_mask,b_mask,a_mask )
//
// surf_ref= SDL_CreateRGBSurfaceFrom( pixels_uint_tbl, w_int, h_int, depth_int,
//									   pitch_int, rgba_mask_tbl )
// 
// NOTE: The SDL documentation is somewhat fuzzy on this function.
//		 Please do not rely on it until tested & confirmed!
//
GLUA_FUNC( SDL_CreateRGBSurfaceFrom )
{
int w= glua_getInteger(2);
int h= glua_getInteger(3);
int depth= glua_getInteger(4);
int pitch= glua_getInteger(5);
//
Uint32 r,g,b,a;
void* block;
SDL_Surface* ref;

    if (!glua_isTable(6))
        {
        r= glua_getInteger(6);
        g= glua_getInteger(7);
        b= glua_getInteger(8);
        a= glua_getInteger(9);
        }
    else
        {
        glua_getTable(6);

        r= glua_getTable_int( "r" );
        g= glua_getTable_int( "g" );
        b= glua_getTable_int( "b" );
        a= glua_getTable_int( "a" );
        }

    // Now tackle the 'pixels_tbl', which is most work...
    //
    glua_getTable(1);
        {
        uint i, items;

        // ?!? Is converting to 'void*' block done in a proper way ?!?

        items= glua_getTableN();

        block= new Uint8[items * depth];  // is 'depth' bytes per pixel?!
		
		for( i=0; i<items; i++ )
			{
			int tmp= glua_getTable_int( i+1 );

	        switch( depth )
				{
			    case 1:	((Uint8*)block)[i]= tmp;  break;
	            case 2:	((Uint16*)block)[i]= tmp; break;
	            case 4:	((Uint32*)block)[i]= tmp; break;

	            default:
		        case 0:	ASSUME_NEVER();     // Something's wrong!
						break;
	            }
			}
        }

    ref= SDL_CreateRGBSurfaceFrom( block, w, h, depth, pitch, r,g,b,a );
	
    glua_pushSurface( ref );

	//---
	// NOTE: 'block' needs to remain available until 'FreeSurface()' is called.
	//		 We use the Lua engine's registry service to attach 'block' to its
	//		 parent userdata, and reading it back in 'FreeSurface()'.
	//
	glua_setAttachment( ref, block );	 // remember that 'block' belongs to 'ref'.
}
GLUA_END


//---
// void= SDL_FreeSurface( surf_ref )
// 
GLUA_FUNC( SDL_FreeSurface )	// (used by MeteorShower)
{
SDL_Surface* ref= glua_getSurface(1);
void* block;

    if (ref)
        {
        SDL_FreeSurface( ref );
        
        // Also free the attached 'block' (if any):
        //
		block= glua_getAttachment( ref, TRUE /*remove*/ );

		if (block)
            free(block);  // note: this must be 'freed', not 'deleted'.
        }
}
GLUA_END


//---
// err_int= SDL_LockSurface( surf_ref )
// 
GLUA_FUNC( SDL_LockSurface )
{
    glua_pushInteger( SDL_LockSurface( glua_getSurface(1) ) );
}
GLUA_END


//---
// void= SDL_UnlockSurface( surf_ref )
// 
GLUA_FUNC( SDL_UnlockSurface )
{
    glua_pushInteger( SDL_LockSurface( glua_getSurface(1) ) );
}
GLUA_END


//---
// surface_ref= SDL_LoadBMP( file_str )
//
// With plain SDL, we can open Windows bitmap files only. With 'sdl_image',
// we can open a wide variety 
//
GLUA_FUNC( SDL_LoadBMP )	// (used by MeteorShower)
{
const char* fname= glua_getString(1);
SDL_Surface* s;
  
  #ifdef HAVE_SDL_IMAGE   // got SDL_Image available?
    // Similar to 'SDL_LoadBMP()' but can open any common graphics format
    // (s.a. BMP, PNM, XPM, PCX, GIF, JPG, TIF, PNG, LBM, see 'sdl_image.h').
    //
    // Requires external 'sdl_image' library (and jpeg & png support libraries)
    // to be installed.
    //
    s= IMG_Load( fname );
  #else   
    s= SDL_LoadBMP( fname );
  #endif
  
    glua_pushSurface(s);
}
GLUA_END


//---
// err_int= SDL_SaveBMP( surf_ref, file_str )
//
GLUA_FUNC( SDL_SaveBMP )	// (used by MeteorShower)
{
SDL_Surface* ref= glua_getSurface(1);
const char* filename= glua_getString(2);

    glua_pushInteger( SDL_SaveBMP( ref, filename ) );
}
GLUA_END


//---
// err_int= SDL_SetColorKey( surface_ref, flag_uint, key_uint )
//
GLUA_FUNC( SDL_SetColorKey )	// (used by MeteorShower)
{
SDL_Surface* ref= glua_getSurface(1);
Uint32 flags= glua_getInteger(2);
Uint32 key= glua_getInteger(3);
int err;

    err= SDL_SetColorKey( ref, flags, key );

	glua_pushInteger( err );
}
GLUA_END


//---
// err_int= SDL_SetAlpha( surface_ref, flag_uint, alpha_uint )
//
GLUA_FUNC( SDL_SetAlpha )
{
SDL_Surface* ref= glua_getSurface(1);
Uint32 flags= glua_getInteger(2);
Uint8 alpha= glua_getInteger(3);
int err;

    err= SDL_SetAlpha( ref, flags, alpha );

	glua_pushInteger( err );
}
GLUA_END


//---
// void= SDL_SetClipRect( surf_ref, rect_tbl )
//
GLUA_FUNC( SDL_SetClipRect )
{
SDL_Surface* ref= glua_getSurface(1);
SDL_Rect *rect= glua_getRect(2);

    SDL_SetClipRect( ref, rect );
}
GLUA_END


//---
// rect_tbl= SDL_GetClipRect( surf_ref )
//
GLUA_FUNC( SDL_GetClipRect )
{
SDL_Surface* ref= glua_getSurface(1);
SDL_Rect rect;

    SDL_GetClipRect( ref, &rect );

    glua_pushRect(&rect);
}
GLUA_END


//---
// surf_ref= SDL_ConvertSurface( surf_ref, format_ref, flags_uint )
//
GLUA_FUNC( SDL_ConvertSurface )
{
SDL_Surface* src= glua_getSurface(1);
SDL_PixelFormat* fmt= glua_getFormat(2);
Uint32 flags= glua_getInteger(3);
SDL_Surface* ret;

    ret= SDL_ConvertSurface( src, fmt, flags );

    glua_pushSurface(ret);
}
GLUA_END


//---
// err_int, final_rect_tbl= SDL_BlitSurface( src_surf_ref, src_rect_tbl, 
//                                           dst_surf_ref, dst_rect_tbl )
//
// Note: In the C API, 'dst_rect' is modified and holds the "final blit rectangle"
//       after return. We don't do this *nasty* ;) parameter mangling but rather
//       return the final rect as a second return value (it is up to the caller
//		 to use that or not!).
//
GLUA_FUNC( SDL_BlitSurface )	// (used by MeteorShower - as 'SDL_UpperBlit()')
{
SDL_Surface* src_surf= glua_getSurface(1);
SDL_Rect *src_rect;
SDL_Surface* dst_surf= glua_getSurface(3);
SDL_Rect *dst_rect;
int err;

	src_rect= glua_getRect(2);
    dst_rect= glua_getRect(4);

    err= SDL_BlitSurface( src_surf, src_rect, dst_surf, dst_rect );

    glua_pushInteger(err);

    glua_pushRect( dst_rect ); // provide modified 'dst_rect' back to caller
}
GLUA_END


//---
// err_int= SDL_FillRect( surf_ref, rect_tbl, pixel_uint )
//
GLUA_FUNC( SDL_FillRect )	// (used by MeteorShower)
{
SDL_Surface* surf= glua_getSurface(1);
SDL_Rect *rect= glua_getRect(2); 
Uint32 color= glua_getInteger(3);
int ret;

    ret= SDL_FillRect( surf, rect, color );

    glua_pushInteger(ret);
}
GLUA_END


//---
// surf_ref= SDL_DisplayFormat( surf_ref )
//
GLUA_FUNC( SDL_DisplayFormat )	// (used by MeteorShower)
{
    glua_pushSurface( SDL_DisplayFormat( glua_getSurface(1) ) );
}
GLUA_END


//---
// surf_ref= SDL_DisplayFormatAlpha( surf_ref )
//
GLUA_FUNC( SDL_DisplayFormatAlpha )
{
    glua_pushSurface( SDL_DisplayFormatAlpha( glua_getSurface(1) ) );
}
GLUA_END


//---
// void= SDL_WarpMouse( x_uint, y_uint )
//
GLUA_FUNC( SDL_WarpMouse )
{
Uint16 x= glua_getInteger(1);
Uint16 y= glua_getInteger(2);

    SDL_WarpMouse( x, y );
}
GLUA_END


//---
// cursor_ref= SDL_CreateCursor( data_uint_tbl, mask_uint_tbl, 
//								 w_int, h_int, hotx_int, hoty_int )
//
GLUA_FUNC( SDL_CreateCursor )
{
int w= glua_getInteger(3);
int h= glua_getInteger(4);
int hotx= glua_getInteger(5);
int hoty= glua_getInteger(6);
//
Uint8* data= NULL;
Uint8* mask= NULL;
//
SDL_Cursor* ret;
unsigned int i,n;

    if (glua_getTable(1))   // 'nil' will leave the block pointer as NULL
        {
        n= glua_getTableN();    // table indices 1..N

        data= new Uint8[n];

        for( i=0; i<n; i++ )
            data[i]= glua_getTable_int( i+1 );
        }

    if (glua_getTable(2))
        {
        n= glua_getTableN();    // table indices 1..N

        mask= new Uint8[n];

        for( i=0; i<n; i++ )
            mask[i]= glua_getTable_int( i+1 );
        }

    ret= SDL_CreateCursor( data, mask, w,h, hotx,hoty );

    if (data) delete[] data;
    if (mask) delete[] mask;

    glua_pushCursor( ret );
}
GLUA_END


//---
// void= SDL_FreeCursor( cursor_ref )
//
GLUA_FUNC( SDL_FreeCursor )
{
SDL_Cursor* ref= glua_getCursor(1);

    SDL_FreeCursor(ref);
}
GLUA_END


//---
// void= SDL_SetCursor( cursor_ref )
//
GLUA_FUNC( SDL_SetCursor )
{
SDL_Cursor* ref= glua_getCursor(1);

    SDL_SetCursor(ref);
}
GLUA_END


//---
// cursor_ref= SDL_GetCursor( void )
//
GLUA_FUNC( SDL_GetCursor )
{
    glua_pushCursor( SDL_GetCursor() );
}
GLUA_END


//---
// state_int= SDL_ShowCursor( state_int )
//
// Param:   SDL_DISABLE / SDL_ENABLE / SDL_QUERY
// Returns: SDL_DISABLE / SDL_ENABLE
//
GLUA_FUNC( SDL_ShowCursor )	// (used by MeteorShower)
{
    glua_pushInteger( SDL_ShowCursor( glua_getInteger(1) ) );
}
GLUA_END


//---
// err_int= SDL_GL_LoadLibrary( path_str )
//
GLUA_FUNC( SDL_GL_LoadLibrary )
{
    glua_pushInteger( SDL_GL_LoadLibrary( glua_getString(1) ) );
}
GLUA_END


//---
// func_ptr= SDL_GL_GetProcAddress( funcname_str )
//
// NOT IMPLEMENTED since having raw C functions at the Lua side makes no sense.
//
GLUA_FUNC( SDL_GL_GetProcAddress )
{
    ASSUME_NEVER();
}
GLUA_END


//---
// [value_int]= SDL_GL_GetAttribute( attr_int )
//
GLUA_FUNC( SDL_GL_GetAttribute )
{
SDL_GLattr attr= (SDL_GLattr)glua_getInteger(1);
int value;

    if (SDL_GL_GetAttribute( attr, &value ) >= 0)   // failure gives 'nil'
        {
        glua_pushInteger( value );
        }
}
GLUA_END


//---
// err_int= SDL_GL_SetAttribute( attr_int, value_int )
//
GLUA_FUNC( SDL_GL_SetAttribute )
{
SDL_GLattr attr= (SDL_GLattr)glua_getInteger(1);
int value= glua_getInteger(2);
int err;

    err= SDL_GL_SetAttribute( attr, value );

    glua_pushInteger( err );
}
GLUA_END


//---
// void= SDL_GL_SwapBuffers( void )
//
GLUA_FUNC( SDL_GL_SwapBuffers )
{
    SDL_GL_SwapBuffers();
}
GLUA_END


//---
// overlay_ref= SDL_CreateYUVOverlay( w_int, h_int, format_uint, surf_ref )
//
GLUA_FUNC( SDL_CreateYUVOverlay )
{
int w= glua_getInteger(1);
int h= glua_getInteger(2);
Uint32 format= glua_getInteger(3);
SDL_Surface* surf= glua_getSurface(4);
SDL_Overlay* ret;

    ret= SDL_CreateYUVOverlay( w,h, format, surf );

    glua_pushOverlay(ret);
}
GLUA_END


//---
// err_int= SDL_LockYUVOverlay( overlay_ref )
//
GLUA_FUNC( SDL_LockYUVOverlay )
{
    glua_pushInteger( SDL_LockYUVOverlay( glua_getOverlay(1) ) );
}
GLUA_END


//---
// SDL_UnlockYUVOverlay( overlay_ref )
//
GLUA_FUNC( SDL_UnlockYUVOverlay )
{
    SDL_LockYUVOverlay( glua_getOverlay(1) );
}
GLUA_END


//---
// int= SDL_DisplayYUVOverlay( overlay_ref, rect_tbl )
//
GLUA_FUNC( SDL_DisplayYUVOverlay )
{
SDL_Overlay* overlay= glua_getOverlay(1);
SDL_Rect* rect;
int ret;

    rect= glua_getRect(2);

    ret= SDL_DisplayYUVOverlay( overlay, rect );

    glua_pushInteger(ret);
}
GLUA_END


//---
// SDL_FreeYUVOverlay( overlay_ref )
//
GLUA_FUNC( SDL_FreeYUVOverlay )
{
    SDL_FreeYUVOverlay( glua_getOverlay(1) );
}
GLUA_END



/*----- SDL window manager functions -----*/

//---
// void= SDL_WM_SetCaption( title_str, icon_str )
//
GLUA_FUNC( SDL_WM_SetCaption )
{
const char* str1= glua_getString(1);
const char* str2= glua_getString(2);

    SDL_WM_SetCaption( str1, str2 );
}
GLUA_END


//---
// title_str, icon_str= SDL_WM_GetCaption( void )
//
GLUA_FUNC( SDL_WM_GetCaption )
{
char* title;
char* icon;

    SDL_WM_GetCaption( &title, &icon );

    glua_pushString(title);
    glua_pushString(icon);
}
GLUA_END


//---
// void= SDL_WM_SetIcon( surface_ref [,mask_tbl] )
//
// Notes: 'mask_tbl' is an array of 8-bit unsigned numbers (0..255) and needs
//        to be long enough to match the surface.
//
GLUA_FUNC( SDL_WM_SetIcon )
{
SDL_Surface* surf= glua_getSurface(1);
Uint8* mask= NULL;
uint i,n;

    if (glua_getTable(2))    // FALSE if nil
        {
        n= glua_getTableN();   // largest numerical index (1..N)

        mask= new Uint8[n];

        ASSUME_NONZERO( mask );

        for( i=0; i<n; i++ )
            mask[i]= glua_getTable_int(i+1);
        }

    SDL_WM_SetIcon( surf, mask );

    if (mask) delete[] mask;
}
GLUA_END


//---
// int= SDL_WM_IconifyWindow( void )
//
// Returns: Non-zero on success, 0 if refused to minimize.
//
GLUA_FUNC( SDL_WM_IconifyWindow )
{
    glua_pushInteger( SDL_WM_IconifyWindow( ) );
}
GLUA_END


//---
// int= SDL_WM_ToggleFullScreen( surface_ref )
//
GLUA_FUNC( SDL_WM_ToggleFullScreen )
{
    glua_pushInteger_01( SDL_WM_ToggleFullScreen( glua_getSurface(1) ) );
}
GLUA_END


//---
// int= SDL_WM_GrabInput( mode_int )
//
// Returns: SDL_GRAB_OFF or SDL_GRAB_ON
//
GLUA_FUNC( SDL_WM_GrabInput )
{
    glua_pushInteger( SDL_WM_GrabInput( (SDL_GrabMode)glua_getInteger(1) ) );
}
GLUA_END



/*----- SDL event functions -----*/

//---
// void= SDL_PumpEvents( void )
//
GLUA_FUNC( SDL_PumpEvents )
{
    SDL_PumpEvents();
}
GLUA_END


//---
// [events_tbl]= SDL_PeepEvents( event_tbl, numevents_int, action_int, mask_uint )
//
// Note: This function really is three different, based on 'eventaction_int':
//
//		 SDL_ADDEVENT:  'event_tbl' used, 'numevents' ignored. 'nil' returned.
//		 SDL_PEEKEVENT: 'event_tbl' nil, 'numevents' used, table returned.
//		 SDL_GETEVENT:  (as with 'PEEKEVENT' but removes the events from queue)
//
GLUA_FUNC( SDL_PeepEvents )
{
SDL_eventaction action= (SDL_eventaction)glua_getInteger(3);
Uint32 mask=            glua_getInteger(4);
SDL_Event* event_buf= NULL;
int i, n;

	if (glua_getTable(1))	// non-nil? (table of userdata entries 1..N)
		{
		SDL_Event* tmp;

		ASSUME( action == SDL_ADDEVENT );

		n= glua_getTableN();

		event_buf= new SDL_Event[n];
		//
		ASSUME_NONZERO(mask);

		for( i=0; i<n; i++ )
			{
			tmp= (SDL_Event*)glua_getTable_ud( i+1, tag_event );

			if (tmp)
				memcpy( &event_buf[i], tmp, sizeof(SDL_Event) );
			else
				memset( &event_buf[i], 0, sizeof(SDL_Event) );
			}

		// Now 'events' has the contents of the input table.
		}
	else	// Need a buffer for reading in events
		{
		ASSUME( action != SDL_ADDEVENT );

		n= glua_getInteger(2);

		event_buf= new SDL_Event[n];
		
		ASSUME_NONZERO( event_buf );
		}

    n= SDL_PeepEvents( event_buf, n, action, mask );

	if (action != SDL_ADDEVENT)
		{
		if (n >= 0)		// failure (-1) gives 'nil'
			{
	        glua_pushTable();
		        {
			    for( i=0; i<n; i++ )
					{
	                glua_setTable_ud( i+1, &event_buf[i], tag_event );
	                }
		        }
			}
		}
}
GLUA_END


//---
// int_01= SDL_PollEvent( event_ref )
//
GLUA_FUNC( SDL_PollEvent )	// (used by MeteorShower)
{
    glua_pushInteger_01( SDL_PollEvent( glua_getEvent(1) ) );
}
GLUA_END


//---
// int_01= SDL_WaitEvent( event_ref )
//
GLUA_FUNC( SDL_WaitEvent )
{
    glua_pushInteger_01( SDL_WaitEvent( glua_getEvent(1) ) );
}
GLUA_END


//---
// err_int= SDL_PushEvent( event_tbl )
//
// Note: The read-in event is a table, not userdata. This is easier to the 
//       Lua programmer (no need to init userdata) and safer for everyone
//       (we don't need to allow Lua code to be able to change 'event' 
//       internals).
//
GLUA_FUNC( SDL_PushEvent )
{
SDL_Event e= { 0 };
int err;

    glua_getTable(1);
        {
        e.type= glua_getTable_int( "type" );

        switch( e.type )
            {
            case SDL_ACTIVEEVENT:
                {
                e.active.gain=  glua_getTable_int( "active.gain" );
                e.active.state= glua_getTable_int( "active.state" );
                }
                break;

            case SDL_KEYDOWN:
            case SDL_KEYUP:
                {
                e.key.which= glua_getTable_int( "key.which" );
                e.key.state= glua_getTable_int( "key.state" );

                e.key.keysym.scancode= glua_getTable_int( "key.keysym.scancode" );
                e.key.keysym.sym= (SDLKey)glua_getTable_int( "key.keysym.sym" );
                e.key.keysym.mod= (SDLMod)glua_getTable_int( "key.keysym.mod" );
                e.key.keysym.unicode=  glua_getTable_int( "key.keysym.unicode" );
                }
                break;

            case SDL_MOUSEMOTION:
                {
                e.motion.which= glua_getTable_int( "motion.which" );
                e.motion.state= glua_getTable_int( "motion.state" );
                e.motion.x=     glua_getTable_int( "motion.x" );
                e.motion.y=     glua_getTable_int( "motion.y" );
                e.motion.xrel=  glua_getTable_int( "motion.xrel" );
                e.motion.yrel=  glua_getTable_int( "motion.yrel" );
                }
                break;

            case SDL_MOUSEBUTTONDOWN:
            case SDL_MOUSEBUTTONUP:
                {
                e.button.which=  glua_getTable_int( "button.which" );
                e.button.button= glua_getTable_int( "button.button" );
                e.button.state=  glua_getTable_int( "button.state" );
                e.button.x=      glua_getTable_int( "button.x" );
                e.button.y=      glua_getTable_int( "button.y" );
                }
                break;

            case SDL_JOYAXISMOTION:
                {
                e.jaxis.which=  glua_getTable_int( "jaxis.which" );
                e.jaxis.axis=   glua_getTable_int( "jaxis.axis" );
                e.jaxis.value=  glua_getTable_int( "jaxis.value" );
                }
                break;

            case SDL_JOYBALLMOTION:
                {
                e.jball.which= glua_getTable_int( "jball.which" );
                e.jball.ball=  glua_getTable_int( "jball.ball" );
                e.jball.xrel=  glua_getTable_int( "jball.xrel" );
                e.jball.yrel=  glua_getTable_int( "jball.yrel" );
                }
                break;

            case SDL_JOYHATMOTION:
                {
                e.jhat.which= glua_getTable_int( "jhat.which" );
                e.jhat.hat=   glua_getTable_int( "jhat.hat" );
                e.jhat.value= glua_getTable_int( "jhat.value" );
                }
                break;

            case SDL_JOYBUTTONUP:
            case SDL_JOYBUTTONDOWN:
                {
                e.jbutton.which=  glua_getTable_int( "jbutton.which" );
                e.jbutton.button= glua_getTable_int( "jbutton.button" );
                e.jbutton.state=  glua_getTable_int( "jbutton.state" );
                }
                break;

            case SDL_VIDEORESIZE:
                {
                e.resize.w= glua_getTable_int( "resize.w" );
                e.resize.h= glua_getTable_int( "resize.h" );
                }
                break;

            case SDL_VIDEOEXPOSE:
            case SDL_QUIT:
                {
                // no fields
                }
                break;

            case SDL_USEREVENT:     // (should also handle USEREVENT+n...)
                {
                e.user.code=  glua_getTable_int( "user.code" );
                e.user.data1= (void*)glua_getTable_int( "user.data1" );
                e.user.data2= (void*)glua_getTable_int( "user.data2" );
                }
                break;

            case SDL_SYSWMEVENT:
                {
                ASSUME_NEVER();  // no sense in sending Window Manager events!
                }
                break;

            default:
                glua_errorN( "Unknown event type %d!", e.type );
                break;
            }
        }

    err= SDL_PushEvent( &e );

    glua_pushInteger(err);  // <0: error, 0: success
}
GLUA_END


//---
// void= SDL_SetEventFilter( filter_func )
//
// NOT IMPLEMENTED since callback func would be run in a separate thread.
//
GLUA_FUNC( SDL_SetEventFilter )
{
    ASSUME_NEVER();     
}
GLUA_END


//---
// filter_func= SDL_GetEventFilter( void )
//
// NOT IMPLEMENTED since 'SDL_SetEventFilter()' is neither.
//
GLUA_FUNC( SDL_GetEventFilter )
{
    ASSUME_NEVER();     
}
GLUA_END


//---
// state_int= SDL_EventState( event_type_int, state_int )
//
GLUA_FUNC( SDL_EventState )
{
Uint8 type= glua_getInteger(1);
Uint8 state= glua_getInteger(2);    // SDL_IGNORE / SDL_ENABLE / SDL_QUERY
Uint8 ret;

    ret= SDL_EventState( type, state );

    glua_pushInteger(ret);  // SDL_IGNORE / SDL_ENABLE
}
GLUA_END


//---
// keystate_tbl= SDL_GetKeyState( void )
//
// NOTE: Returned table contains 'nil' and '1' (can be checked as boolean).
//		 Indices are same as in C (keyboard constants).
//
GLUA_FUNC( SDL_GetKeyState )
{
Uint8* ptr;
int i, numkeys;

    ptr= SDL_GetKeyState( &numkeys );

    if (ptr)    // NULL gives 'nil'
        {
        glua_pushTable();
            {
            for( i=0; i<numkeys; i++ )
                {
                // NOTE: Lua tables are sparse so we don't have to push non-
                //       existing entries. This will speed things up but table
                //       will have 'nil' where C table has zero.
                //
                if (ptr[i] != 0)
                    {
                    // Use same indices as in C (0..N-1) since they will be
                    // indexed directly with key codes (s.a. 'SDLK_RETURN').

                    glua_setTable_int( i, 1 );
                    }
                }
            }
        }
}
GLUA_END


//---
// modstate_int= SDL_GetModState( void )
//
GLUA_FUNC( SDL_GetModState )
{
    glua_pushInteger( SDL_GetModState() );
}
GLUA_END


//---
// void= SDL_SetModState( modstate_int )
//
GLUA_FUNC( SDL_SetModState )
{
    SDL_SetModState( (SDLMod)glua_getInteger(1) );
}
GLUA_END


//---
// str= SDL_GetKeyName( key_int )
//
GLUA_FUNC( SDL_GetKeyName )
{
    glua_pushString( SDL_GetKeyName( (SDLKey)glua_getInteger(1) ) );
}
GLUA_END


//---
// oldstate_int= SDL_EnableUnicode( newstate_int )
//
GLUA_FUNC( SDL_EnableUNICODE )
{
    glua_pushInteger( SDL_EnableUNICODE( glua_getInteger(1) ) );
}
GLUA_END


//---
// err_int= SDL_EnableKeyRepeat( [delay_int [,interval_int]] )
//
GLUA_FUNC( SDL_EnableKeyRepeat )
{
int delay= SDL_DEFAULT_REPEAT_DELAY;
int interval= SDL_DEFAULT_REPEAT_INTERVAL;
int err;

    switch( glua_getArgn() )
        {
        default:
        case 2:
            interval= glua_getInteger(2);
			// pass through...
        case 1:
            delay= glua_getInteger(1);
            break;

        case 0:     // use default values
            break;  
        }

    err= SDL_EnableKeyRepeat( delay, interval );

    glua_pushInteger( err );
}
GLUA_END


//---
// buttons_int, x_int, y_int= SDL_GetMouseState( void )
//
GLUA_FUNC( SDL_GetMouseState )	// (used by MeteorShower)
{
Uint8 buttons;
int x,y;

    buttons= SDL_GetMouseState( &x, &y );

    glua_pushInteger( buttons );
    glua_pushInteger( x );
    glua_pushInteger( y );
}
GLUA_END


//---
// buttons_int, x_int, y_int= SDL_GetRelativeMouseState( void )
//
GLUA_FUNC( SDL_GetRelativeMouseState )
{
Uint8 buttons;
int dx,dy;

    buttons= SDL_GetRelativeMouseState( &dx, &dy );

    glua_pushInteger( buttons );
    glua_pushInteger( dx );
    glua_pushInteger( dy );
}
GLUA_END


//---
// state_int= SDL_GetAppState( void )
//
GLUA_FUNC( SDL_GetAppState )
{
    glua_pushInteger( SDL_GetAppState() );
}
GLUA_END


//---
// state_int= SDL_JoystickEventState( state_int )
//
GLUA_FUNC( SDL_JoystickEventState )
{
    glua_pushInteger( SDL_JoystickEventState( glua_getInteger(1) ) );
}
GLUA_END


//---
// int= SDL_BUTTON(int)
//
GLUA_FUNC( SDL_BUTTON )
{
int a= glua_getInteger(1);  // Use a variable since macros may have side-effects

    glua_pushInteger( SDL_BUTTON(a) );  // 'SDL_BUTTON()' is a macro
}
GLUA_END



/*----- Joystick functions -----*/
//

//---
// int= SDL_NumJoysticks( void )
//
GLUA_FUNC( SDL_NumJoysticks )
{
    glua_pushInteger( SDL_NumJoysticks() );
}
GLUA_END


//---
// str= SDL_JoystickName( index_int )
//
GLUA_FUNC( SDL_JoystickName )
{
int index= glua_getInteger(1);  // 0..NumJoysticks()-1

    glua_pushString( SDL_JoystickName(index) );
}
GLUA_END


//---
// joy_ref= SDL_JoystickOpen( index_int )
//
GLUA_FUNC( SDL_JoystickOpen )
{
int index= glua_getInteger(1);  // 0..NumJoysticks()-1

    glua_pushJoystick( SDL_JoystickOpen(index) );
}
GLUA_END


//---
// int_01= SDL_JoystickOpened( index_int )
//
GLUA_FUNC( SDL_JoystickOpened )
{
int index= glua_getInteger(1);  // 0..NumJoysticks()-1

    glua_pushInteger_01( SDL_JoystickOpened(index) );
}
GLUA_END


//---
// int= SDL_JoystickIndex( joy_ref )
// int= SDL_JoystickNumAxes( joy_ref )
// int= SDL_JoystickNumBalls( joy_ref )
// int= SDL_JoystickNumHats( joy_ref )
// int= SDL_JoystickNumButtons( joy_ref )
//
// Note: These are implemented as ".index" etc. fields of the joystick objects.
//       Lua wrapper can cover them in original C API look-alike functions.
//---


//---
// void= SDL_JoystickUpdate( void )
//
GLUA_FUNC( SDL_JoystickUpdate )
{
    SDL_JoystickUpdate();
}
GLUA_END


//---
// int= SDL_JoystickGetAxis( joy_ref, axis_index_int )
//
GLUA_FUNC( SDL_JoystickGetAxis )
{
SDL_Joystick* ref= glua_getJoystick(1);
int index= glua_getInteger(2);

    glua_pushInteger( SDL_JoystickGetAxis(ref,index) );
}
GLUA_END


//---
// int= SDL_JoystickGetHat( joy_ref, hat_index_int )
//
GLUA_FUNC( SDL_JoystickGetHat )
{
SDL_Joystick* ref= glua_getJoystick(1);
int index= glua_getInteger(2);

    glua_pushInteger( SDL_JoystickGetHat(ref,index) );
}
GLUA_END


//---
// int= SDL_JoystickGetButton( joy_ref, button_index_int )
//
GLUA_FUNC( SDL_JoystickGetButton )
{
SDL_Joystick* ref= glua_getJoystick(1);
int index= glua_getInteger(2);

    glua_pushInteger( SDL_JoystickGetButton(ref,index) );
}
GLUA_END


//---
// dx,dy= SDL_JoystickGetBall( joy_ref, ball_index_int )
//
GLUA_FUNC( SDL_JoystickGetBall )
{
SDL_Joystick* ref= glua_getJoystick(1);
int index= glua_getInteger(2);
int dx,dy;

    SDL_JoystickGetBall( ref, index, &dx, &dy );

    glua_pushInteger(dx);
    glua_pushInteger(dy);
}
GLUA_END


//---
// void= SDL_JoystickClose( joy_ref )
//
GLUA_FUNC( SDL_JoystickClose )
{
SDL_Joystick* ref= glua_getJoystick(1);

    SDL_JoystickClose( ref );
}
GLUA_END



/*----- CD-ROM functions -----*/
//

//---
// int= SDL_CDNumDrives( void )
//
GLUA_FUNC( SDL_CDNumDrives )
{
    glua_pushInteger( SDL_CDNumDrives() );
}
GLUA_END


//---
// str= SDL_CDName( drive_int )
//
GLUA_FUNC( SDL_CDName )
{
int drive= glua_getInteger(1);  // 0..CDNumDrives()-1

    glua_pushString( SDL_CDName(drive) );
}
GLUA_END


//---
// cd_ref= SDL_CDOpen( drive_int )
//
GLUA_FUNC( SDL_CDOpen )
{
int drive= glua_getInteger(1);  // 0..CDNumDrives()-1

    glua_pushCdrom( SDL_CDOpen(drive) );
}
GLUA_END


//---
// status_int= SDL_CDStatus( cd_ref )
//
GLUA_FUNC( SDL_CDStatus )
{
SDL_CD* cd_ref= glua_getCdrom(1);
int status;

    status= SDL_CDStatus( cd_ref );

    glua_pushInteger( status );
}
GLUA_END


//---
// err_int= SDL_CDPlay( cd_ref, start_frame_int, length_frame_int )
//
GLUA_FUNC( SDL_CDPlay )
{
SDL_CD* cd_ref= glua_getCdrom(1);
int start= glua_getInteger(2);
int len= glua_getInteger(3);
int err;

    err= SDL_CDPlay( cd_ref, start, len );

    glua_pushInteger( err );    // 0=success, <0=error
}
GLUA_END


//---
// err_int= SDL_CDPlayTracks( cd_ref, start_track_int, start_frame_int,
//                                    n_tracks_int, n_frames_int )
//
GLUA_FUNC( SDL_CDPlayTracks )
{
SDL_CD* cd_ref= glua_getCdrom(1);
int start_track= glua_getInteger(2);
int start_frame= glua_getInteger(3);
int n_tracks= glua_getInteger(4);
int n_frames= glua_getInteger(5);
int err;

    err= SDL_CDPlayTracks( cd_ref, start_track,start_frame, n_tracks,n_frames );

    glua_pushInteger( err );    // 0=success, <0=error
}
GLUA_END


//---
// err_int= SDL_CDPause( cd_ref )
//
GLUA_FUNC( SDL_CDPause )
{
SDL_CD* cd_ref= glua_getCdrom(1);

    glua_pushInteger( SDL_CDPause( cd_ref ) );  // 0=success, <0=error
}
GLUA_END


//---
// err_int= SDL_CDResume( cd_ref )
//
GLUA_FUNC( SDL_CDResume )
{
SDL_CD* cd_ref= glua_getCdrom(1);

    glua_pushInteger( SDL_CDResume( cd_ref ) );  // 0=success, <0=error
}
GLUA_END


//---
// err_int= SDL_CDStop( cd_ref )
//
GLUA_FUNC( SDL_CDStop )
{
SDL_CD* cd_ref= glua_getCdrom(1);

    glua_pushInteger( SDL_CDStop( cd_ref ) );  // 0=success, <0=error
}
GLUA_END


//---
// err_int= SDL_CDEject( cd_ref )
//
GLUA_FUNC( SDL_CDEject )
{
SDL_CD* cd_ref= glua_getCdrom(1);

    glua_pushInteger( SDL_CDEject( cd_ref ) );  // 0=success, <0=error
}
GLUA_END


//---
// void= SDL_CDClose( cd_ref )
//
GLUA_FUNC( SDL_CDClose )
{
SDL_CD* cd_ref= glua_getCdrom(1);

    SDL_CDClose( cd_ref );
}
GLUA_END



/*----- Multithreading functions -----*/
//

//---
// thread_ref= SDL_CreateThread( cmd_str [,stack_size] )
//
// Params:  'cmd_str' is any Lua command that the created thread will run.
//      
// NOTE:    The thread will start with an _absolutely_clear_, new Lua state.
//          You will need to load any extensions etc. as part of the 'cmd_str'.
//
// To-do:   Lua 5.0 supports inheriting a Lua state's contents from another
//          thread. This could be used to get a duplicate of our state (e.g.
//          an additional flag parameter for this), but currently not 
//          implemented. Also, it won't be usable from Lua 4.0.
//
struct s_ThreadParams {
    char* cmd_str;
    unsigned int stack_size; 
};

static int Loc_NewThread( void* params_v )
{
struct s_ThreadParams* p= (struct s_ThreadParams*)params_v;
int ret;

    ASSUME_NONZERO( p );
    ASSUME_NONZERO( p->cmd_str );

    // 'p' is allocated by the caller, and we are responsible for releasing
    // it. This way, it is guaranteed to be valid throughout our execution.

	ret= glua_exec( p->cmd_str, p->stack_size );

	delete[] p->cmd_str;
	delete p;

    return ret;
}

GLUA_FUNC( SDL_CreateThread )
{
const char* cmd= glua_getString(1);
int stack_size= glua_getInteger(2); // 0 = default stack size
//
struct s_ThreadParams* params;
//
SDL_Thread* thread;

    ASSUME_NONZERO( cmd );

    // Make an allocated copy of the parameters so that their lifespan can last
    // throughout the thread execution.
    //
    // NOTE: The launched thread is responsible for releasing these!
    //
    params= new struct s_ThreadParams;

    params->cmd_str= new char[strlen(cmd)+1];
    strcpy( params->cmd_str, cmd );

    params->stack_size= stack_size;

    // Now launch the thread (creates a whole new, independent Lua state):
    //
    thread= SDL_CreateThread( Loc_NewThread, params );

    glua_pushThread(thread);
}
GLUA_END


//---
// uint= SDL_ThreadId( void )
//
GLUA_FUNC( SDL_ThreadID )
{
    glua_pushInteger( SDL_ThreadID() );
}
GLUA_END


//---
// uint= SDL_GetThreadId( thread_ref )
//
GLUA_FUNC( SDL_GetThreadID )
{
    glua_pushInteger( SDL_GetThreadID( glua_getThread(1) ) );
}
GLUA_END


//---
// retcode_int= SDL_WaitThread( thread_ref )
//
GLUA_FUNC( SDL_WaitThread )
{
int retcode;

    SDL_WaitThread( glua_getThread(1), &retcode );

    glua_pushInteger( retcode );
}
GLUA_END


//---
// void= SDL_KillThread( thread_ref )
//
GLUA_FUNC( SDL_KillThread )
{
    SDL_KillThread( glua_getThread(1) );
}
GLUA_END


//---
// mutex_ref= SDL_CreateMutex( void )
//
GLUA_FUNC( SDL_CreateMutex )
{
    glua_pushMutex( SDL_CreateMutex() );
}
GLUA_END


//---
// void= SDL_DestroyMutex( mutex_ref )
//
GLUA_FUNC( SDL_DestroyMutex )
{
    SDL_DestroyMutex( glua_getMutex(1) );
}
GLUA_END


//---
// err_int= SDL_LockMutex( mutex_ref )
//
GLUA_FUNC( SDL_LockMutex )
{
    glua_pushInteger( SDL_LockMutex( glua_getMutex(1) ) );
}
GLUA_END


//---
// err_int= SDL_UnlockMutex( mutex_ref )
//
GLUA_FUNC( SDL_UnlockMutex )
{
    glua_pushInteger( SDL_UnlockMutex( glua_getMutex(1) ) );
}
GLUA_END


//---
// sema_ref= SDL_CreateSemaphore( initial_value_uint )
//
GLUA_FUNC( SDL_CreateSemaphore )
{
    glua_pushSemaphore( SDL_CreateSemaphore( glua_getInteger(1) ) );
}
GLUA_END


//---
// void= SDL_DestroySemaphore( sema_ref )
//
GLUA_FUNC( SDL_DestroySemaphore )
{
    SDL_DestroySemaphore( glua_getSemaphore(1) );
}
GLUA_END


//---
// err_int= SDL_SemWait( sema_ref )
//
GLUA_FUNC( SDL_SemWait )
{
    glua_pushInteger( SDL_SemWait( glua_getSemaphore(1) ) );
}
GLUA_END


//---
// int= SDL_SemTryWait( sema_ref )
//
// Returns: 0 if succesfully locked
//          SDL_MUTEX_TIMEOUT if the thread would have suspended
//          -1 if there was an error
//
GLUA_FUNC( SDL_SemTryWait )
{
    glua_pushInteger( SDL_SemTryWait( glua_getSemaphore(1) ) );
}
GLUA_END


//---
// int= SDL_SemWaitTimeout( sema_ref, timeout_ms_uint )
//
// Returns: 0 if succesfully locked
//          SDL_MUTEX_TIMEOUT if the timeout period was exceeded
//          -1 if there was an error
//
GLUA_FUNC( SDL_SemWaitTimeout )
{
SDL_sem* sema= glua_getSemaphore(1);
Uint32 timeout_ms= glua_getInteger(2);

    glua_pushInteger( SDL_SemWaitTimeout( sema, timeout_ms ) );
}
GLUA_END


//---
// err_int= SDL_SemPost( sema_ref )
//
GLUA_FUNC( SDL_SemPost )
{
    glua_pushInteger( SDL_SemPost( glua_getSemaphore(1) ) );
}
GLUA_END


//---
// value_uint= SDL_SemValue( sema_ref )
//
GLUA_FUNC( SDL_SemValue )
{
    glua_pushInteger( SDL_SemValue( glua_getSemaphore(1) ) );
}
GLUA_END


//---
// cond_ref= SDL_CreateCond( void )
//
GLUA_FUNC( SDL_CreateCond )
{
    glua_pushCondVar( SDL_CreateCond() );
}
GLUA_END


//---
// void= SDL_DestroyCond( cond_ref )
//
GLUA_FUNC( SDL_DestroyCond )
{
    SDL_DestroyCond( glua_getCondVar(1) );
}
GLUA_END


//---
// err_int= SDL_CondSignal( cond_ref )
//
GLUA_FUNC( SDL_CondSignal )
{
    glua_pushInteger( SDL_CondSignal( glua_getCondVar(1) ) );
}
GLUA_END


//---
// err_int= SDL_CondBroadcast( cond_ref )
//
GLUA_FUNC( SDL_CondBroadcast )
{
    glua_pushInteger( SDL_CondBroadcast( glua_getCondVar(1) ) );
}
GLUA_END


//---
// err_int= SDL_CondWait( cond_ref, mutex_ref )
//
GLUA_FUNC( SDL_CondWait )
{
SDL_cond* cond_ref= glua_getCondVar(1);
SDL_mutex* mutex_ref= glua_getMutex(2);

    glua_pushInteger( SDL_CondWait( cond_ref, mutex_ref ) );
}
GLUA_END


//---
// int= SDL_CondWaitTimeout( cond_ref, mutex_ref, timeout_ms_uint )
//
// Returns: 0 (success)
//          SDL_MUTEX_TIMEDOUT
//          -1 on error
//
GLUA_FUNC( SDL_CondWaitTimeout )
{
SDL_cond* cond_ref= glua_getCondVar(1);
SDL_mutex* mutex_ref= glua_getMutex(2);
Uint32 timeout_ms= glua_getInteger(3);

    glua_pushInteger( SDL_CondWaitTimeout( cond_ref, mutex_ref, timeout_ms ) );
}
GLUA_END



/*----- Timing functions -----*/
//

//---
// int= SDL_GetTicks( )
//
GLUA_FUNC( SDL_GetTicks )	// (used by MeteorShower)
{
    glua_pushInteger( SDL_GetTicks() );
}
GLUA_END


//---
// void= SDL_Delay( ms_uint )
//
GLUA_FUNC( SDL_Delay )
{
    SDL_Delay( glua_getInteger(1) );
}
GLUA_END


//---
// timer_ref= SDL_AddTimer( interval_uint, callback_func, param_ref )
//
// NOT IMPLEMENTED because uses callbacks from another thread.
//
GLUA_FUNC( SDL_AddTimer )
{
    ASSUME_NEVER();
}
GLUA_END


//---
// bool= SDL_RemoveTimer( timer_ref )
//
// NOT IMPLEMENTED because only used with 'AddTimer()'.
//
GLUA_FUNC( SDL_RemoveTimer )
{
    ASSUME_NEVER();
}
GLUA_END


//---
// int= SDL_SetTimer( interval_uint, callback_func )
//
// NOT IMPLEMENTED: backwards compatibility func (replaced by 'AddTimer()').
//
GLUA_FUNC( SDL_SetTimer )
{
    ASSUME_NEVER();
}
GLUA_END



/*----- Userdata constructors -----*/
//

GLUA_FUNC( SDL_Event_new )
{
SDL_Event* obj;

    obj= new SDL_Event;

    if (!obj)
        glua_error( "Out of memory creating 'SDL_Event'!" );

    memset(obj,0,sizeof(SDL_Event));
    
    glua_pushEvent( obj );
}
GLUA_END

//--
GLUA_FUNC( SDL_Rect_new )
{
SDL_Rect* obj;

    obj= new SDL_Rect;

    if (!obj)
        glua_error( "Out of memory creating 'SDL_Rect'!" );

    memset(obj,0,sizeof(SDL_Rect));
    
    glua_pushRect( obj );
}
GLUA_END


/*----- Tag method functions -----*/
//
// The tag methods are not publicized to the Lua namespace (unless separately
// done so). They are called implicitly when e.g. using table-like syntax or 
// arithmetical operations on userdata.
//
// For a complete list of available tag method names, see Lua documentation.

//---
// value= Surface_gettable( obj, index )
//
// Note: 'index' may be any type but we only expect strings.
//
#define FIELD_IS( str ) ( strcmp( (field), (str) ) == 0 )

GLUA_FUNC( _Surface_gettable )
{
SDL_Surface* obj= glua_getSurface(1);
const char* field= glua_getString(2);

    if (!obj) break;    // goes to 'GLUA_END' (pushes no value)

    if (FIELD_IS( "format" ))    // SDL_PixelFormat*
        glua_pushFormat( obj->format );
    else
    if (FIELD_IS( "flags" ))     // Uint32
        glua_pushInteger( obj->flags );
    else
    if (FIELD_IS( "w" ))         // int
        glua_pushInteger( obj->w );
    else
    if (FIELD_IS( "h" ))         // int
        glua_pushInteger( obj->h );
    else    
        {
        // There are other fields but we don't support them ...at least yet.
        //
        glua_error( "Unknown field for 'SDL_Surface' object!" );
        }
}
GLUA_END


//---
// value= Format_gettable( obj, index )
//
// Note: 'index' may be any type but we only expect strings.
//
GLUA_FUNC( _Format_gettable )
{
SDL_PixelFormat* obj= glua_getFormat(1);
const char* field= glua_getString(2);

    if (!obj) break;    // goes to 'GLUA_END' (pushes no value)

	// For compatibility, we've repeated the data layour exactly as it is
	// in the C API. However, the 'mask', 'shift' and 'loss' fields really
	// should be grouped together, and returned as a table from here.
	// Would be much more elegant... :)     Do so in your own source.
	//
    if (FIELD_IS( "palette" ))    // SDL_Palette*
        glua_pushNil();	// NOT IMPLEMENTED. (could be..)
    else
    if (FIELD_IS( "BitsPerPixel" ))  // Uint8
        glua_pushInteger( obj->BitsPerPixel );
    else
    if (FIELD_IS( "BytesPerPixel" ))  // Uint8
        glua_pushInteger( obj->BytesPerPixel );
    else
	if (FIELD_IS( "mask" ))	// LUA-ONLY ADDITION!
		{
		glua_pushTable();
			{
			glua_setTable_int( "r", obj->Rmask );
			glua_setTable_int( "g", obj->Gmask );
			glua_setTable_int( "b", obj->Bmask );
			glua_setTable_int( "a", obj->Amask );
			}
		}
	else
    if (FIELD_IS( "Rmask" ))  // Uint32
        glua_pushInteger( obj->Rmask );
    else
    if (FIELD_IS( "Gmask" ))  // Uint32
        glua_pushInteger( obj->Gmask );
    else
    if (FIELD_IS( "Bmask" ))  // Uint32
        glua_pushInteger( obj->Bmask );
    else
    if (FIELD_IS( "Amask" ))  // Uint32
        glua_pushInteger( obj->Amask );
    else
	if (FIELD_IS( "shift" ))	// LUA-ONLY ADDITION!
		{
		glua_pushTable();
			{
			glua_setTable_int( "r", obj->Rshift );
			glua_setTable_int( "g", obj->Gshift );
			glua_setTable_int( "b", obj->Bshift );
			glua_setTable_int( "a", obj->Ashift );
			}
		}
    else
    if (FIELD_IS( "Rshift" ))  // Uint8
        glua_pushInteger( obj->Rshift );
    else
    if (FIELD_IS( "Gshift" ))  // Uint8
        glua_pushInteger( obj->Gshift );
    else
    if (FIELD_IS( "Bshift" ))  // Uint8
        glua_pushInteger( obj->Bshift );
    else
    if (FIELD_IS( "Ashift" ))  // Uint8
        glua_pushInteger( obj->Ashift );
    else
	if (FIELD_IS( "loss" ))	// LUA-ONLY ADDITION!
		{
		glua_pushTable();
			{
			glua_setTable_int( "r", obj->Rloss );
			glua_setTable_int( "g", obj->Gloss );
			glua_setTable_int( "b", obj->Bloss );
			glua_setTable_int( "a", obj->Aloss );
			}
		}
    else
    if (FIELD_IS( "Rloss" ))  // Uint8
        glua_pushInteger( obj->Rloss );
    else
    if (FIELD_IS( "Gloss" ))  // Uint8
        glua_pushInteger( obj->Gloss );
    else
    if (FIELD_IS( "Bloss" ))  // Uint8
        glua_pushInteger( obj->Bloss );
    else
    if (FIELD_IS( "Aloss" ))  // Uint8
        glua_pushInteger( obj->Aloss );
	else
    if (FIELD_IS( "colorkey" ))	// Uint32
        glua_pushInteger( obj->colorkey );
	else
    if (FIELD_IS( "alpha" ))	// Uint8
        glua_pushInteger( obj->alpha );
    else    
        glua_error( "Unknown field for 'SDL_Format'!" );
}
GLUA_END


//---
// value= Event_gettable( obj, index )
//
// Note: 'index' may be any type but we only expect strings.
//
GLUA_FUNC( _Event_gettable )
{
SDL_Event* e= glua_getEvent(1);
const char* field= glua_getString(2);

    if (!e) break;    // goes to 'GLUA_END' (pushes no value)

    if (FIELD_IS( "type" ))  // event.type
        {
        glua_pushInteger( e->type );    // Uint8
        }
    else
    if (FIELD_IS( "active" ))  // event.active.*
        {
        ASSUME( e->type == SDL_ACTIVEEVENT );

        glua_pushTable();
            {
            glua_setTable_int( "gain", e->active.gain );
            glua_setTable_int( "state", e->active.state );
            }
        }
    else
    if (FIELD_IS( "key" ))       // event.key.*
        {
        ASSUME( (e->type == SDL_KEYDOWN) || (e->type == SDL_KEYUP) );

        glua_pushTable();
            {
            glua_setTable_int( "which", e->key.which );

            // Note: 'state' is actually redundant since it is contained
            //       in the 'KEYDOWN'/'KEYUP' type info already.
            //       (but we keep it here for compatibility reasons)
            //
            glua_setTable_int( "state", e->key.state );

            // Note: 'scancode' is hardware and platform specific!
            //
            glua_setTable_int( "keysym.scancode", e->key.keysym.scancode );
            glua_setTable_int( "keysym.sym",      e->key.keysym.sym );
            glua_setTable_int( "keysym.mod",      e->key.keysym.mod );
            glua_setTable_int( "keysym.unicode",  e->key.keysym.unicode );
            }
        }
    else
    if (FIELD_IS( "motion" ))  // event.motion.*
        {
        ASSUME( e->type == SDL_MOUSEMOTION );

        glua_pushTable();
            {
            glua_setTable_int( "which", e->motion.which );
            glua_setTable_int( "state", e->motion.state );

            glua_setTable_int( "x",     e->motion.x );
            glua_setTable_int( "y",     e->motion.y );
            glua_setTable_int( "xrel",  e->motion.xrel );
            glua_setTable_int( "yrel",  e->motion.yrel );
            }
        }
    else
    if (FIELD_IS( "button" ))  // event.button.*
        {
        ASSUME( (e->type == SDL_MOUSEBUTTONDOWN) || 
                (e->type == SDL_MOUSEBUTTONUP) );

        glua_pushTable();
            {
            glua_setTable_int( "which",  e->button.which );
            glua_setTable_int( "button", e->button.button );
            glua_setTable_int( "state",  e->button.state );
            glua_setTable_int( "x",      e->button.x );
            glua_setTable_int( "y",      e->button.y );
            }
        }
    else
    if (FIELD_IS( "jaxis" ))     // Joystick axis
        {
        ASSUME( e->type == SDL_JOYAXISMOTION );

        glua_pushTable();
            {
            glua_setTable_int( "which", e->jaxis.which );
            glua_setTable_int( "axis",  e->jaxis.axis );
            glua_setTable_int( "value", e->jaxis.value );
            }
        }
    else
    if (FIELD_IS( "jball" ))     // Joystick ball
        {
        ASSUME( e->type == SDL_JOYBALLMOTION );

        glua_pushTable();
            {
            glua_setTable_int( "which", e->jball.which );
            glua_setTable_int( "ball",  e->jball.ball );
            glua_setTable_int( "xrel",  e->jball.xrel );
            glua_setTable_int( "yrel",  e->jball.yrel );
            }
        }
    else
    if (FIELD_IS( "jhat" ))      // Joystick hat
        {
        ASSUME( e->type == SDL_JOYHATMOTION );

        glua_pushTable();
            {
            glua_setTable_int( "which", e->jhat.which );
            glua_setTable_int( "hat",   e->jhat.hat );
            glua_setTable_int( "value", e->jhat.value );
            }
        }
    else
    if (FIELD_IS( "jbutton" ))   // Joystick button
        {
        ASSUME( (e->type == SDL_JOYBUTTONDOWN) || 
                (e->type == SDL_JOYBUTTONUP) );

        glua_pushTable();
            {
            glua_setTable_int( "which",  e->jbutton.which );
            glua_setTable_int( "button", e->jbutton.button );
            glua_setTable_int( "state",  e->jbutton.state );
            }
        }
    else
    if (FIELD_IS( "resize" ))
        {
        ASSUME( e->type == SDL_VIDEORESIZE );

        glua_pushTable();
            {
            glua_setTable_int( "w", e->resize.w );
            glua_setTable_int( "h", e->resize.h );
            }
        }
    else
    if (FIELD_IS( "expose" ))
        {
        ASSUME( e->type == SDL_VIDEOEXPOSE );

        glua_pushNil();     // no fields
        }
    else
    if (FIELD_IS( "quit" ))
        {
        ASSUME( e->type == SDL_QUIT );

        glua_pushNil();     // no fields
        }
    else
    if (FIELD_IS( "user" ))
        {
        ASSUME( e->type == SDL_USEREVENT );

        glua_pushTable();
            {
            glua_setTable_int( "code", e->user.code );

            // Should we push the 'void*' pointers as userdata?
            // but we don't know the tags for them...
            //
            glua_setTable_int( "data1", (int)e->user.data1 );
            glua_setTable_int( "data2", (int)e->user.data2 );
            }
        }
    else
    if (FIELD_IS( "syswm" ))
        {
        ASSUME( e->type == SDL_SYSWMEVENT );

        glua_pushTable();
            {
            // Common fields to all implementations:
            //
            glua_setTable_int( "msg.version.major", e->syswm.msg->version.major );
            glua_setTable_int( "msg.version.minor", e->syswm.msg->version.minor );
            glua_setTable_int( "msg.version.patch", e->syswm.msg->version.patch );

            // Implementation specific fields:
            //
            // NOTE: X11 etc. implementations not here!
            //
        #ifdef PLATFORM_WIN32
            {
            glua_setTable_int( "msg.hwnd", (unsigned long)(e->syswm.msg->hwnd) );
            glua_setTable_int( "msg.msg", e->syswm.msg->msg );   // type of msg
            glua_setTable_int( "msg.wParam", e->syswm.msg->wParam );
            glua_setTable_int( "msg.lParam", e->syswm.msg->lParam );
            }
        #else
           //#error "Implementation for this platform is missing!"
        #endif
            }
        }
    else
        {
        glua_errorN( "Unknown/unsupported 'SDL_Event' field! (%s)", field );
        }
}
GLUA_END


//---
// value= _Joystick_gettable( obj, index )
//
// Note: 'index' may be any type but we only expect strings.
//
// NOTE2: These items are _not_ part of the C SDL API, but rather our Lua
//        wrap-around for several (too many) separate "get-this-and-that"
//        functions. Compatibility to C calls can be made in Lua script level.
//
GLUA_FUNC( _Joystick_gettable )
{
SDL_Joystick* obj= glua_getJoystick(1);
const char* field= glua_getString(2);

    if (!obj) break;    // goes to 'GLUA_END' (pushes no value)

    if (FIELD_IS( "index" ))
        glua_pushInteger( SDL_JoystickIndex(obj) );
    else
    if (FIELD_IS( "axes" ))
        glua_pushInteger( SDL_JoystickNumAxes(obj) );
    else
    if (FIELD_IS( "balls" ))
        glua_pushInteger( SDL_JoystickNumBalls(obj) );
    else
    if (FIELD_IS( "hats" ))
        glua_pushInteger( SDL_JoystickNumHats(obj) );
    else
    if (FIELD_IS( "buttons" ))
        glua_pushInteger( SDL_JoystickNumButtons(obj) );
    else    
        {
        glua_errorN( "Unknown 'SDL_CD' field '%s'!", field );
        }
}
GLUA_END


//---
// value= Cdrom_gettable( obj, index )
//
// Note: 'index' may be any type but we only expect strings.
//
GLUA_FUNC( _Cdrom_gettable )
{
SDL_CD* obj= glua_getCdrom(1);
const char* field= glua_getString(2);
int i;

    if (!obj) break;    // goes to 'GLUA_END' (pushes no value)

    if (FIELD_IS( "id" ))           // int
        glua_pushInteger( obj->id );
    else
    if (FIELD_IS( "status" ))       // CDstatus (enum)
        glua_pushInteger( obj->status );
    else
    if (FIELD_IS( "numtracks" ))    // int
        glua_pushInteger( obj->numtracks );
    else
    if (FIELD_IS( "cur_track" ))         // int
        glua_pushInteger( obj->cur_track );
    else
    if (FIELD_IS( "cur_frame" ))         // int
        glua_pushInteger( obj->cur_frame );
    else
    if (FIELD_IS( "track" ))         // table of tracks
        {
        glua_pushTable();
            {
            for( i=0; i < obj->numtracks; i++ )
                {
				// Note: We're using same indices (0..N-1) as C code.
				//
				glua_setTable_int( i, "id",	   obj->track[i].id );
				glua_setTable_int( i, "type",   obj->track[i].type );
				glua_setTable_int( i, "length", obj->track[i].length );
				glua_setTable_int( i, "offset", obj->track[i].offset );
                }
            }
        }
    else    
        {
        glua_errorN( "Unknown 'SDL_CD' field '%s'!", field );
        }
}
GLUA_END


//---
// value= _Overlay_gettable( obj, index )
//
// Note: 'index' may be any type but we only expect strings.
//
GLUA_FUNC( _Overlay_gettable )
{
SDL_Overlay* obj= glua_getOverlay(1);
const char* field= glua_getString(2);

    if (!obj) break;    // goes to 'GLUA_END' (pushes no value)

    if (FIELD_IS( "format" ))
        glua_pushInteger( /*Uint32*/ obj->format );
    else
    if (FIELD_IS( "w" ))
        glua_pushInteger( /*int*/ obj->w );
    else
    if (FIELD_IS( "h" ))
        glua_pushInteger( /*int*/ obj->h );
    else
    if (FIELD_IS( "planes" ))
        glua_pushInteger( /*int*/ obj->planes );
    else
    if (FIELD_IS( "pitches" ))
		{
		Uint16* pitches= obj->pitches;	// [0]..[planes-1] ?
		int i;

		glua_pushTable();
			{
			for( i=0; i< obj->planes; i++ )
				glua_setTable_int( i, pitches[i] );
			}
		}
    else
/**
    if (FIELD_IS( "pixels" ))
		{
		Uint8** pixels= obj->pixels;

		glua_pushNil();		// NOT IMPLEMENTED yet.
		//
		// Is this needed and if so, how should we present the data?
		}
    else
**/
    if (FIELD_IS( "hw_overlay" ))
        glua_pushInteger( /*0/1*/ obj->hw_overlay );
    else    
        {
        glua_errorN( "Unknown 'SDL_CD' field '%s'!", field );
        }
}
GLUA_END


//---
// value= _Rect_gettable( obj, index )
//
GLUA_FUNC( _Rect_gettable )
{
SDL_Rect* obj= glua_getRect(1);
const char* field= glua_getString(2);   // "x"/"y"/w"/"h"

    if (!obj) break;    // goes to 'GLUA_END' (pushes no value)
    if (!field) break;
    
    if (FIELD_IS( "x" ))
        glua_pushInteger( /*int*/ obj->x );
    else
    if (FIELD_IS( "y" ))
        glua_pushInteger( /*int*/ obj->y );
    else
    if (FIELD_IS( "w" ))
        glua_pushInteger( /*int*/ obj->w );
    else
    if (FIELD_IS( "h" ))
        glua_pushInteger( /*int*/ obj->h );
    else    
        glua_errorN( "Unknown 'SDL_Rect' field '%s'!", field );
}
GLUA_END


//---
// value= _Rect_settable( obj, index, value )
//
// Note: We take in doubles (not integer) in order to allow calculation 
//       results be used (with rounding) as coordinates.
//
GLUA_FUNC( _Rect_settable )
{
SDL_Rect* obj= glua_getRect(1);
const char* field= glua_getString(2);   // "x"/"y"/w"/"h"
int val= (int)( glua_getNumber(3) +0.5 );

    if (!obj) break;    // goes to 'GLUA_END' (pushes no value)
    if (!field) break;
    
    if (FIELD_IS( "x" ))
        obj->x= val;
    else
    if (FIELD_IS( "y" ))
        obj->y= val;
    else
    if (FIELD_IS( "w" ))
        obj->w= val;
    else
    if (FIELD_IS( "h" ))
        obj->h= val;
    else    
        glua_errorN( "Unknown 'SDL_Rect' field '%s'!", field );
}
GLUA_END


/*----- Misc functions -----*/



/*----- Module initialization -----*/

// This block is called once (by gluahost) to declare functions and constants
// to Lua. Make sure all 'GLUA_FUNC' functions are listed here!

GLUA_MODULE( my_path, my_tbl )
{
	//------
    // You can use 'my_path' to read configuration files etc. that would 
	// reside in the same directory as your module.
	// 
    IGNORE_PARAM( my_path );

    //------
    // Module information:
    //
    GLUA_DECL( my_tbl )
        {
        glua_info( "MODULE", "SDL (Simple DirectMedia Layer)" ),
        glua_info( "AUTHOR", "asko.kauppi@fi.flextronics.com" ),
        glua_info( "LICENSE", "LGPL" ),
        glua_info( "PLATFORMS", "Win32, Linux, OS X" ),
        glua_info( "DEPENDENCIES",
                        "sdl (v.1.2.5) by Sam Lantinga (slouken@libsdl.org)\n"
                        "sdl_image (v.1.2.3) by Sam Lantinga\n"
                        "sdl_mixer (v.1.2.5) by Sam Lantinga" ),
        }
    GLUA_DECL_END

	//-----
	// 'my_tbl' is the namespace table suggested by the GluaX host (or the user)
    // and may be ignored if you so wish (passing 0 instead). 
	//
    GLUA_DECL( my_tbl )
        {
    	//---------
        // Additional (Lua only) functions:
        //
		glua_func( SDL_Info ),
		glua_func( SDL_SetSurface ),

		//---------
        // General functions:
        //
        glua_func( SDL_Init ),
        glua_func( SDL_Quit ),
		glua_func( SDL_GetError ),
        glua_func( SDL_VERSION ),
        glua_func( SDL_Linked_Version ),

        // Initialization flags:
        //
        glua_const( SDL_INIT_TIMER ),
        glua_const( SDL_INIT_AUDIO ),
        glua_const( SDL_INIT_VIDEO ),
        glua_const( SDL_INIT_CDROM ),
        glua_const( SDL_INIT_JOYSTICK ),
        //
        glua_const( SDL_INIT_EVERYTHING ),

        // Non-supported functions:
        //
        //glua_func( SDL_InitSubSystem ),
        //glua_func( SDL_QuitSubSystem ),
        //glua_func( SDL_WasInit ),


        //---------
        // Video subsystem functions:
        //
        glua_func( SDL_GetVideoSurface ),
        glua_func( SDL_GetVideoInfo ),
        glua_func( SDL_VideoDriverName ),
        glua_func( SDL_ListModes ),
        glua_func( SDL_VideoModeOK ),
        glua_func( SDL_SetVideoMode ),
        glua_func( SDL_UpdateRect ),
        glua_func( SDL_UpdateRects ),
        glua_func( SDL_Flip ),
        glua_func( SDL_SetColors ),
        glua_func( SDL_SetPalette ),
        glua_func( SDL_SetGamma ),
        glua_func( SDL_GetGammaRamp ),
        glua_func( SDL_SetGammaRamp ),
        glua_func( SDL_MapRGB ),
        glua_func( SDL_MapRGBA ),
        glua_func( SDL_GetRGB ),
        glua_func( SDL_GetRGBA ),
        glua_func( SDL_CreateRGBSurface ),
        glua_func( SDL_CreateRGBSurfaceFrom ),
        glua_func( SDL_FreeSurface ),
        glua_func( SDL_LockSurface ),
        glua_func( SDL_UnlockSurface ),
        glua_func( SDL_LoadBMP ),
        glua_func( SDL_SaveBMP ),
        glua_func( SDL_SetColorKey ),
        glua_func( SDL_SetAlpha ),
        glua_func( SDL_SetClipRect ),
        glua_func( SDL_GetClipRect ),
        glua_func( SDL_ConvertSurface ),
        glua_func( SDL_BlitSurface ),
        glua_func( SDL_FillRect ),
        glua_func( SDL_DisplayFormat ),
        glua_func( SDL_DisplayFormatAlpha ),
        
        // Cursor functions:
        glua_func( SDL_WarpMouse ),
        glua_func( SDL_CreateCursor ),
        glua_func( SDL_FreeCursor ),
        glua_func( SDL_SetCursor ),
        glua_func( SDL_GetCursor ),
        glua_func( SDL_ShowCursor ),

        // OpenGL functions:
        glua_func( SDL_GL_LoadLibrary ),
        glua_func( SDL_GL_GetProcAddress ),     // NOT SUPPORTED!
        glua_func( SDL_GL_GetAttribute ),
        glua_func( SDL_GL_SetAttribute ),
        glua_func( SDL_GL_SwapBuffers ),
        glua_func( SDL_CreateYUVOverlay ),
        glua_func( SDL_LockYUVOverlay ),
        glua_func( SDL_UnlockYUVOverlay ),
        glua_func( SDL_DisplayYUVOverlay ),
        glua_func( SDL_FreeYUVOverlay ),
        
        // Surface flags:
        glua_const( SDL_SWSURFACE ),
        glua_const( SDL_HWSURFACE ),	// (used by MeteorShower)
        glua_const( SDL_ASYNCBLIT ),
        glua_const( SDL_ANYFORMAT ),	// (used by MeteorShower)
        glua_const( SDL_HWPALETTE ),
        glua_const( SDL_DOUBLEBUF ),	// (used by MeteorShower)
        glua_const( SDL_FULLSCREEN ),	// (used by MeteorShower)
        glua_const( SDL_OPENGL ),
        glua_const( SDL_OPENGLBLIT ),
        glua_const( SDL_RESIZABLE ),
	glua_const( SDL_HWACCEL ),
        glua_const( SDL_SRCCOLORKEY ),
        glua_const( SDL_RLEACCEL ),
        glua_const( SDL_SRCALPHA ),
        glua_const( SDL_PREALLOC ),
        glua_const( SDL_NOFRAME ),

        // 'SetPalette()' flags:
        glua_const( SDL_LOGPAL ),
        glua_const( SDL_PHYSPAL ),        


        //---------
        // Windows management:
        //
        glua_func( SDL_WM_SetCaption ),
        glua_func( SDL_WM_GetCaption ),
        glua_func( SDL_WM_SetIcon ),
        glua_func( SDL_WM_IconifyWindow ),
        glua_func( SDL_WM_ToggleFullScreen ),
        glua_func( SDL_WM_GrabInput ),


        //---------
        // Events:
        //
        glua_func( SDL_PumpEvents ),
        glua_func( SDL_PeepEvents ),
        glua_func( SDL_PollEvent ),
        glua_func( SDL_WaitEvent ),
        glua_func( SDL_PushEvent ),
        glua_func( SDL_SetEventFilter ),    // not implemented!
        glua_func( SDL_GetEventFilter ),    // not implemented!
        glua_func( SDL_EventState ),
        glua_func( SDL_GetKeyState ),
        glua_func( SDL_GetModState ),
        glua_func( SDL_SetModState ),
        glua_func( SDL_GetKeyName ),
        glua_func( SDL_EnableUNICODE ),
        glua_func( SDL_EnableKeyRepeat ),
        glua_func( SDL_GetMouseState ),
        glua_func( SDL_GetRelativeMouseState ),
        glua_func( SDL_GetAppState ),
        glua_func( SDL_JoystickEventState ),
        glua_func( SDL_BUTTON ),

        // Event types:
        glua_const( SDL_ACTIVEEVENT ),
        glua_const( SDL_KEYDOWN ),			// (used by MeteorShower)
        glua_const( SDL_KEYUP ),
        glua_const( SDL_MOUSEMOTION ),
        glua_const( SDL_MOUSEBUTTONDOWN ),
        glua_const( SDL_MOUSEBUTTONUP ),
        glua_const( SDL_JOYAXISMOTION ),
        glua_const( SDL_JOYBALLMOTION ),
        glua_const( SDL_JOYHATMOTION ),
        glua_const( SDL_JOYBUTTONDOWN ),
        glua_const( SDL_JOYBUTTONUP ),
        glua_const( SDL_QUIT ),				// (used by MeteorShower)
        glua_const( SDL_SYSWMEVENT ),
        glua_const( SDL_VIDEORESIZE ),
        glua_const( SDL_VIDEOEXPOSE ),
        glua_const( SDL_USEREVENT ),

        // SDL_ACTIVEEVENT states:
        glua_const( SDL_APPMOUSEFOCUS ),
        glua_const( SDL_APPINPUTFOCUS ),
        glua_const( SDL_APPACTIVE ),

        // Button states (KEYUP/DOWN, MOUSEBUTTONUP/DOWN, JOYBUTTONUP/DOWN):
        glua_const( SDL_RELEASED ),
        glua_const( SDL_PRESSED ),

        // Mouse buttons:
        glua_const( SDL_BUTTON_LEFT ),
        glua_const( SDL_BUTTON_MIDDLE ),    
        glua_const( SDL_BUTTON_RIGHT ),

        // Joystick hats:
        glua_const( SDL_HAT_CENTERED ),
        glua_const( SDL_HAT_UP ),
        glua_const( SDL_HAT_RIGHT ),
        glua_const( SDL_HAT_DOWN ),
        glua_const( SDL_HAT_LEFT ),
        glua_const( SDL_HAT_RIGHTUP ),
        glua_const( SDL_HAT_RIGHTDOWN ),
        glua_const( SDL_HAT_LEFTUP ),
        glua_const( SDL_HAT_LEFTDOWN ),

        // 'SDL_PeepEvents' actions:
        glua_const( SDL_ADDEVENT ),
        glua_const( SDL_PEEKEVENT ),
        glua_const( SDL_GETEVENT ),

        // 'SDL_EnableKeyRepeat' constants:
        glua_const( SDL_DEFAULT_REPEAT_DELAY ),
        glua_const( SDL_DEFAULT_REPEAT_INTERVAL ),
    

        //---------
        // Joystick:
        //
        glua_func( SDL_NumJoysticks ),
        glua_func( SDL_JoystickName ),
        glua_func( SDL_JoystickOpen ),
        glua_func( SDL_JoystickOpened ),
        glua_func( SDL_JoystickUpdate ),
        glua_func( SDL_JoystickGetAxis ),
        glua_func( SDL_JoystickGetHat ),
        glua_func( SDL_JoystickGetButton ),
        glua_func( SDL_JoystickGetBall ),
        glua_func( SDL_JoystickClose ),
        

        //---------
        // Audio:
        //
        // (none implemented)

        //---------
        // CD-ROM:
        //
        glua_func( SDL_CDNumDrives ),
        glua_func( SDL_CDName ),
        glua_func( SDL_CDOpen ),
        glua_func( SDL_CDStatus ),
        glua_func( SDL_CDPlay ),
        glua_func( SDL_CDPlayTracks ),
        glua_func( SDL_CDPause ),
        glua_func( SDL_CDResume ),
        glua_func( SDL_CDStop ),
        glua_func( SDL_CDEject ),
        glua_func( SDL_CDClose ),

        // 'SDL_CDstatus' return values:
        glua_const( CD_TRAYEMPTY ),
        glua_const( CD_STOPPED ),
        glua_const( CD_PLAYING ),
        glua_const( CD_PAUSED ),
        glua_const( CD_ERROR ),

        //---------
        // Multithreading:
        //
        glua_func( SDL_CreateThread ),
        glua_func( SDL_ThreadID ),
        glua_func( SDL_GetThreadID ),
        glua_func( SDL_WaitThread ),
        glua_func( SDL_KillThread ),
        glua_func( SDL_CreateMutex ),
        glua_func( SDL_DestroyMutex ),
        glua_func( SDL_LockMutex ),
        glua_func( SDL_UnlockMutex ),
        glua_func( SDL_CreateSemaphore ),
        glua_func( SDL_DestroySemaphore ),
        glua_func( SDL_SemWait ),
        glua_func( SDL_SemTryWait ),
        glua_func( SDL_SemWaitTimeout ),
        glua_func( SDL_SemPost ),
        glua_func( SDL_SemValue ),
        glua_func( SDL_CreateCond ),
        glua_func( SDL_DestroyCond ),
        glua_func( SDL_CondSignal ),
        glua_func( SDL_CondBroadcast ),
        glua_func( SDL_CondWait ),
        glua_func( SDL_CondWaitTimeout ),


        //---------
        // Timing:
        //
        glua_func( SDL_GetTicks ),
        glua_func( SDL_Delay ),
        glua_func( SDL_AddTimer ),      // (not implemented)
        glua_func( SDL_RemoveTimer ),   // (not implemented)
        glua_func( SDL_SetTimer ),      // (not implemented)


        //---------
        // Userdata constructors:
        //
        glua_func( SDL_Event_new ),
        glua_func( SDL_Rect_new ),


        //---------
        // Userdata definitions:
        //
        glua_tag( "SDL_Surface",	 &tag_surface ),
        glua_tag( "SDL_PixelFormat", &tag_format ),
        glua_tag( "SDL_Event",		 &tag_event ),
        glua_tag( "SDL_Joystick",	 &tag_joystick ),
        glua_tag( "SDL_CD", 		 &tag_cdrom ),
        glua_tag( "SDL_Cursor",		 &tag_cursor ),
        glua_tag( "SDL_Overlay",	 &tag_overlay ),
        glua_tag( "SDL_Thread",		 &tag_thread ),
        glua_tag( "SDL_mutex",		 &tag_mutex ),
        glua_tag( "SDL_sem",		 &tag_semaphore ),
        glua_tag( "SDL_cond",		 &tag_condvar ),
        glua_tag( "SDL_Rect",		 &tag_rect ),

        glua_tagmethod( tag_surface, "gettable", _Surface_gettable ),
        glua_tagmethod( tag_format,  "gettable", _Format_gettable ),
        glua_tagmethod( tag_event,   "gettable", _Event_gettable ),
        glua_tagmethod( tag_joystick,"gettable", _Joystick_gettable ),
        glua_tagmethod( tag_cdrom,   "gettable", _Cdrom_gettable ),
        glua_tagmethod( tag_overlay, "gettable", _Overlay_gettable ),
        glua_tagmethod( tag_rect,    "gettable", _Rect_gettable ),
        glua_tagmethod( tag_rect,    "settable", _Rect_settable ),


        //---------
        // SDL constants:

        // Keyboard symbols ('event.key.keysym.sym'):
        //
        glua_const( SDLK_BACKSPACE ),
        glua_const( SDLK_TAB ),
        glua_const( SDLK_CLEAR ),
        glua_const( SDLK_RETURN ),
        glua_const( SDLK_PAUSE ),
        glua_const( SDLK_ESCAPE ),	// (used by MeteorShower)
        glua_const( SDLK_SPACE ),
        glua_const( SDLK_ESCAPE ),
        glua_const( SDLK_QUOTEDBL ),
        glua_const( SDLK_HASH ),
        glua_const( SDLK_DOLLAR ),
        glua_const( SDLK_AMPERSAND ),
        glua_const( SDLK_QUOTE ),
        glua_const( SDLK_LEFTPAREN ),
        glua_const( SDLK_RIGHTPAREN ),
        glua_const( SDLK_ASTERISK ),
        glua_const( SDLK_PLUS ),
        glua_const( SDLK_COMMA ),
        glua_const( SDLK_MINUS ),
        glua_const( SDLK_PERIOD ),
        glua_const( SDLK_SLASH ),
        glua_const( SDLK_0 ),
        glua_const( SDLK_1 ),
        glua_const( SDLK_2 ),
        glua_const( SDLK_3 ),
        glua_const( SDLK_4 ),
        glua_const( SDLK_5 ),
        glua_const( SDLK_6 ),
        glua_const( SDLK_7 ),
        glua_const( SDLK_8 ),
        glua_const( SDLK_9 ),
        glua_const( SDLK_COLON ),
        glua_const( SDLK_SEMICOLON ),
        glua_const( SDLK_LESS ),
        glua_const( SDLK_EQUALS ),
        glua_const( SDLK_GREATER ),
        glua_const( SDLK_QUESTION ),
        glua_const( SDLK_AT ),
        glua_const( SDLK_LEFTBRACKET ),
        glua_const( SDLK_BACKSLASH ),
        glua_const( SDLK_RIGHTBRACKET ),
        glua_const( SDLK_CARET ),
        glua_const( SDLK_UNDERSCORE ),
        glua_const( SDLK_BACKQUOTE ),
        glua_const( SDLK_a ),
        glua_const( SDLK_b ),
        glua_const( SDLK_c ),
        glua_const( SDLK_d ),
        glua_const( SDLK_e ),
        glua_const( SDLK_f ),
        glua_const( SDLK_g ),
        glua_const( SDLK_h ),
        glua_const( SDLK_i ),
        glua_const( SDLK_j ),
        glua_const( SDLK_k ),
        glua_const( SDLK_l ),
        glua_const( SDLK_m ),
        glua_const( SDLK_n ),
        glua_const( SDLK_o ),
        glua_const( SDLK_p ),
        glua_const( SDLK_q ),	// (used by MeteorShower)
        glua_const( SDLK_r ),
        glua_const( SDLK_s ),	// (used by MeteorShower)
        glua_const( SDLK_t ),
        glua_const( SDLK_u ),
        glua_const( SDLK_v ),
        glua_const( SDLK_w ),
        glua_const( SDLK_x ),
        glua_const( SDLK_y ),
        glua_const( SDLK_z ),
        glua_const( SDLK_DELETE ),
        glua_const( SDLK_KP0 ),
        glua_const( SDLK_KP1 ),
        glua_const( SDLK_KP2 ),
        glua_const( SDLK_KP3 ),
        glua_const( SDLK_KP4 ),
        glua_const( SDLK_KP5 ),
        glua_const( SDLK_KP6 ),
        glua_const( SDLK_KP7 ),
        glua_const( SDLK_KP8 ),
        glua_const( SDLK_KP9 ),
        glua_const( SDLK_KP_PERIOD ),
        glua_const( SDLK_KP_DIVIDE ),
        glua_const( SDLK_KP_MULTIPLY ),
        glua_const( SDLK_KP_MINUS ),
        glua_const( SDLK_KP_PLUS ),
        glua_const( SDLK_KP_ENTER ),
        glua_const( SDLK_KP_EQUALS ),
        glua_const( SDLK_UP ),
        glua_const( SDLK_DOWN ),
        glua_const( SDLK_RIGHT ),
        glua_const( SDLK_LEFT ),
        glua_const( SDLK_INSERT ),
        glua_const( SDLK_HOME ),
        glua_const( SDLK_END ),
        glua_const( SDLK_PAGEUP ),
        glua_const( SDLK_PAGEDOWN ),
        glua_const( SDLK_F1 ),
        glua_const( SDLK_F2 ),
        glua_const( SDLK_F3 ),
        glua_const( SDLK_F4 ),
        glua_const( SDLK_F5 ),
        glua_const( SDLK_F6 ),
        glua_const( SDLK_F7 ),
        glua_const( SDLK_F8 ),
        glua_const( SDLK_F9 ),
        glua_const( SDLK_F10 ),
        glua_const( SDLK_F11 ),
        glua_const( SDLK_F12 ),
        glua_const( SDLK_F13 ),
        glua_const( SDLK_F14 ),
        glua_const( SDLK_F15 ),
        glua_const( SDLK_NUMLOCK ),
        glua_const( SDLK_CAPSLOCK ),
        glua_const( SDLK_SCROLLOCK ),
        glua_const( SDLK_LSHIFT ),
        glua_const( SDLK_RSHIFT ),
        glua_const( SDLK_LCTRL ),
        glua_const( SDLK_RCTRL ),
        glua_const( SDLK_LALT ),
        glua_const( SDLK_RALT ),
        glua_const( SDLK_LMETA ),
        glua_const( SDLK_RMETA ),
        glua_const( SDLK_LSUPER ),
        glua_const( SDLK_RSUPER ),
        glua_const( SDLK_MODE ),
        glua_const( SDLK_HELP ),
        glua_const( SDLK_PRINT ),
        glua_const( SDLK_SYSREQ ),
        glua_const( SDLK_BREAK ),
        glua_const( SDLK_MENU ),
        glua_const( SDLK_POWER ),
        glua_const( SDLK_EURO ),
		//
		glua_const( SDLK_FIRST ),
		glua_const( SDLK_LAST ),

        // Keyboard modifiers ('event.key.keysym.mod'):
        //
        glua_const( KMOD_NONE ),
        glua_const( KMOD_NUM ),
        glua_const( KMOD_CAPS ),
		//
        glua_const( KMOD_SHIFT ),
        glua_const( KMOD_LSHIFT ),
        glua_const( KMOD_RSHIFT ),
		//
        glua_const( KMOD_CTRL ),
        glua_const( KMOD_LCTRL ),
        glua_const( KMOD_RCTRL ),
		//
        glua_const( KMOD_ALT ),
        glua_const( KMOD_LALT ),
        glua_const( KMOD_RALT ),
		//
        glua_const( KMOD_META ),
        glua_const( KMOD_LMETA ),
        glua_const( KMOD_RMETA ),

        // Mouse buttons:
        //
        glua_const( SDL_BUTTON_LEFT ),
        glua_const( SDL_BUTTON_MIDDLE ),
        glua_const( SDL_BUTTON_RIGHT ),
// These appear first in SDL 1.2.5
#ifdef SDL_BUTTON_WHEELUP
        glua_const( SDL_BUTTON_WHEELUP ),
        glua_const( SDL_BUTTON_WHEELDOWN ),
#endif

        // SDL GL attributes:
        glua_const( SDL_GL_RED_SIZE ),
        glua_const( SDL_GL_GREEN_SIZE ),
        glua_const( SDL_GL_BLUE_SIZE ),
        glua_const( SDL_GL_ALPHA_SIZE ),
        glua_const( SDL_GL_DOUBLEBUFFER ),
        glua_const( SDL_GL_BUFFER_SIZE ),
        glua_const( SDL_GL_DEPTH_SIZE ),
        glua_const( SDL_GL_STENCIL_SIZE ),
        glua_const( SDL_GL_ACCUM_RED_SIZE ),
        glua_const( SDL_GL_ACCUM_GREEN_SIZE ),
        glua_const( SDL_GL_ACCUM_BLUE_SIZE ),
        glua_const( SDL_GL_ACCUM_ALPHA_SIZE ),

		// Misc constants:
        // ...
        }
    GLUA_DECL_END

    // Initialization ok - Lua can call us now. :)
}
GLUA_MODULE_END

