
-----------------------------------------------------------------------------
-- Lua CGI module and utilities
-- Copyright (C) 2003 Adrian Perez <mobius@iespana.es>
--
-- $Id: cgi.lua,v 1.3 2003/07/01 15:55:36 tulrich Exp $                     
-- $RCSfile: cgi.lua,v $
-- 
-- This library is free software, it is release under the same license as Lua 5
-- Lua is licensed under the terms of the MIT license
-- This means that Lua is free software and can be used for both academic
-- and commercial purposes at absolutely no cost,
-- as long as the original copyright notice is retained
-- For details and rationale, see http://www.lua.org/license.html .
----------------------------------------------------------------------------- 

-- Namespace table
cgi = {}
	

cgi.__header_items =
{
	content		= "Content-type" ;
	redirect	= "Location" ;
}

cgi.__default_header = 
{
	content = "text/html" ;
}


------
-- Prints the content-type header
------
function cgi.put_header(hdr, fd)
	fd  = fd  or io.stdout
	hdr = hdr or cgi.__default_header

	for key,val in hdr do
		fd:write(cgi.__header_items[key]..": "..val.."\r\n")
	end
	fd:write("\r\n") -- end header with an extra newline
end


------
-- This strips the '\r' characters from input (if present) as nearly
-- all browsers send \r\n endlines when submitting data to the server.
------
function cgi.strip_rr(txt)
	return string.gsub(txt, "\r", "")
end	

function cgi.strip_nn(txt)
	return string.gsub(txt, "\n", "")
end

function cgi.strip_rn(txt)
	return string.gsub(txt, "[\r\n]", "")
end	

------ 
-- Processes incoming data in a multipart message sent to the server
-- using the POST method. Returns table where keys are item names and
-- values are contents of that part of the message.
------
function cgi.process_multipart_post(fd)
	fd = fd or io.stdin
	-- guess what's the delimiter used by the browser
	local delim = cgi.strip_rr(fd:read("*l"))
	local t = {}

	-- extract parts
	while true do
		local info = cgi.strip_rr(fd:read("*l"))
		local text = ""
		local name = string.gfind(info, "name=\"([%a_][%w_]*)\"")()
		local line = cgi.strip_rr(fd:read("*l"))

		while (line and (line ~= delim) and (line ~= (delim.."--"))) do
			text = text..line.."\n"
			line = fd:read("*l")
			line = line and cgi.strip_rr(line)
		end
		t[name] = text

		if (not line) or (line == (delim.."--")) then break end
	end
	return t
end	


------
-- Decodes an encoded URL
------
function cgi.url_decode(text, use_plus)
	use_plus = use_plus or true
	if use_plus then
		text = string.gsub(text, "+", " ")
	end	
	return cgi.strip_rr(string.gsub(
			text, "%%(%x%x)", 
			function (num)
				return string.char(tonumber(num, 16))
			end))
end


------
-- Encodes an URL
------
function cgi.url_encode(text, use_plus)
	use_plus = use_plus or true
	if use_plus then
		text = string.gsub(text, " ", "+")
	end	
	return string.gsub(
			text, "[^%w_%.%-+]",
			function (ch)
				return string.format("%%%X", string.byte(ch))
			end)	
end


------
-- Decodes a query string into a table.
------
function cgi.decode_qs(qs, t)
	t = t or {}
	qs = qs or ""
	for item in string.gfind(qs.."&", "([^&]+)&") do
		local key, val = string.gfind(item, "([^=]+)=?(.*)")()
		t[key] = (val and cgi.url_decode(val)) or true
	end
	return t
end


------
-- Encodes a table into a query string. The table cannot contain 
-- nested tables.
------
function cgi.encode_qs(tbl)
	local qs = ""
	for key,val in tbl do
		qs = qs.."&"..key.."="..cgi.url_encode(tostring(val))
	end
	return string.sub(qs, 2)
end

------
-- Processes incoming data sent to the server using the POST method
-- and it's encoded in a big string encoded as the query strings used
-- in the GET method. Returns a table keyed by names with its values
-- in the query string.
------
function cgi.process_urlenc_post(fd)
	fd = fd or io.stdin
	return cgi.decode_qs(fd:read("*a"))
end


------
-- Processes incoming data sent to the server using the GET method.
-- Data is encoded the usual way in the query string: 
-- 		name1=value1&name2=value2& ... &nameN=valueN
------
function cgi.process_get()
	return cgi.decode_qs(os.getenv("QUERY_STRING"))
end


------
-- Processes the data sent to a CGI and returns a table with the 
-- parsed contents.
------
function cgi.process(fd)
	fd = fd or io.stdin
	for _,i in {"QUERY_STRING", "PATH_INFO", "REQUEST_METHOD", "CONTENT_TYPE"} do
		cgi[i] = os.getenv(i)
	end
	-- guess method or assume GET if no available
	cgi.REQUEST_METHOD = cgi.REQUEST_METHOD or "GET"

	if cgi.REQUEST_METHOD=="POST" then
		local encoding = string.gfind(cgi.CONTENT_TYPE, "[^;%s]+")()
		if encoding=="multipart/form-data" then
			return cgi.process_multipart_post(fd)
		elseif encoding=="application/x-www-form-urlencoded" then
			return cgi.process_urlenc_post(fd)
		end
	else
		return cgi.process_get()	
	end
end


------
-- self-test function, for cheia_self_test module.
------
function cgi._self_test()
	-- TODO: Fill this in with real tests.
	return true;
end

-- LuaCheia module stub loaders must return true on successful load!
-- (on failure, should return false and an error message)
return true

-- vim: ts=2

