/* Portable file system handling routines. */

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <lua.h>
#include <lauxlib.h>


#if defined(PLATFORM_POSIX) || defined(PLATFORM_OSX)

#  include <sys/types.h>
#  include <dirent.h>
#  include <errno.h>

#elif defined(PLATFORM_WINDOWS)

#  include <errno.h>
#  include "gusu.h"

#  define opendir gusu_opendir
#  define readdir gusu_readdir
#  define closedir gusu_closedir

#else

#  error "Sorry, your platform is not yet supported"

#endif


#include <sys/stat.h>
#ifdef PLATFORM_WINDOWS
#   define stat(a,b) _stat((a),(b))
#   define lstat(a,b) _stat((a),(b))
    typedef struct _stat xpstat;
#else
    typedef struct stat xpstat;
#endif


/* return true if the file is a directory */
static int is_dir(lua_State *L, const char *path, const char *file)
{
  const char *filename;
  xpstat st;

  lua_pushstring(L, path);
  lua_pushstring(L, "/");
  lua_pushstring(L, file);
  lua_concat(L, 3);

  filename = lua_tostring(L, -1);

  if(stat(filename, &st) != 0)
  {
    /* is a dangling link so we have to lstat() it */
    if(lstat(filename, &st) != 0)
    {
      /* ooops... */
      lua_pop(L, 1);
      return 0;
    }
  }
  lua_pop(L, 1);

  if(S_ISDIR(st.st_mode))
  {
    return 1;
  }

  return 0;
}


int fs_listdir(lua_State *L)
{
  int index;
  int dirlen; 
  int len;
  DIR *dir;
  struct dirent *handle;
  const char *dirpath;
  char *entryname;

  /* check parameter type */
  luaL_checkstring (L, -1);
  dirpath = lua_tostring(L, -1);
  dirlen = strlen(dirpath);
  dir = opendir(dirpath);
  
  if(dir == NULL) 
  {  
    /* signal an error. */
    lua_pushnil(L);
    lua_pushfstring(L, "'%s' is not a directory", dirpath);
    return 2; 
  }
  lua_newtable(L); /* New table to hold the contents. */
  index = 0;  
  for (handle = readdir(dir); handle != NULL; handle = readdir(dir))
  {
     entryname = handle->d_name;

     /*  name of current entry */
     if(strcmp(".", entryname) == 0 ||
        strcmp("..", entryname) == 0)
     {
       continue;
     }
     /* ignore those two */

     index++;

     /* check if entry is a dir */
     if (is_dir(L, dirpath, entryname))
     {
       lua_pushnumber(L,index);
       lua_pushstring(L,entryname);
       lua_pushstring(L,"/");
       lua_concat(L, 2);
       lua_rawset(L,-3); 
     }
     else 
     {
       lua_pushnumber(L,index);
       lua_pushstring(L,entryname);
       lua_rawset(L,-3); 
     }

     /* 
        Get all file system entries for this directory, 
        and store them in a table. 
     */  
  }

  closedir(dir);
  
  return 1; /* we leave the result table on the stack. */        
}

int fs_stat(lua_State *L)
/* return a table containing common info about a file/path */
{
  const char *entrypath;
  const char *entrytype;
  xpstat st;
  luaL_checkstring (L, -1);
  entrypath = lua_tostring(L, -1);
  /* stat like normal */
  if(stat(entrypath, &st) != 0)
  /* but if we can't find it then it */
  /* is a dangling link so we have to lstat() it */
  {
    if(lstat(entrypath, &st) != 0)
    /* ooops... */
    {
      perror(entrypath);
      return 0;
    }
  }

  /* Now return the info */
  { /* brackets only for visualisation ;) */
    lua_newtable(L); 
    /* New table to hold the info. */

	lua_pushstring(L,"fullpath");
	lua_pushstring(L,entrypath);
	lua_settable(L, -3);

    lua_pushstring(L,"ctime");
    lua_pushnumber(L,st.st_ctime);
    lua_settable (L, -3);
          
    lua_pushstring(L,"mtime");
    lua_pushnumber(L,st.st_mtime);
    lua_settable (L, -3);

    lua_pushstring(L,"atime");
    lua_pushnumber(L,st.st_atime);
    lua_settable (L, -3);

    lua_pushstring(L,"size");
    lua_pushnumber(L,st.st_size);
    lua_settable (L, -3);

    lua_pushstring(L,"rawmode");
    /* this needs some processing 
    user-permissions etc, lie here
    see 'type' below
    _S_IREAD Read permission, owner 
    _S_IWRITE Write permission, owner 
    _S_IEXEC Execute/search permission, owner 
    */
    lua_pushnumber(L,st.st_mode);
    lua_settable (L, -3);
          
    lua_pushstring(L,"type");
    /* check if entry is a dir */
    if(S_ISDIR(st.st_mode))
    {
      lua_pushstring(L,"dir");
    } 
    else 
    {
      lua_pushstring(L,"file");
    }
    lua_settable (L, -3);
          
    /* only useful on POSIX */
    lua_pushstring(L,"uid");
    lua_pushnumber(L,st.st_uid);
    lua_settable (L, -3);

    /* only useful on POSIX */
    lua_pushstring(L,"gid");
    lua_pushnumber(L,st.st_gid);
    lua_settable (L, -3);
    
    /* TODO (traumwind#1#): Add user permissions (calculate from mode) */
  }
  return 1;
}

int fs_listdir2(lua_State *L)
/* proposed alternate implementation */
/* 
* this returns a table with
* entryname -> sub-table
*      type of entry (file, dir)
*      fullpath
*      creationtime
*      modification time
*      filesize
* it needs stat() or equivalent 
* on systems w/o we try to provide a subset
*/
{
  int index;
  int dirlen; 
  int len;
  DIR *dir;
  struct dirent *handle;
  const char *dirpath;
  char *entryname;
  char *entrypath;
  
  /* check parameter type */
  luaL_checkstring (L, -1);
  dirpath = lua_tostring(L, -1);
  dirlen = strlen(dirpath);
  dir = opendir(dirpath);
  
  if(dir == NULL) 
  {  
     /* signal an error. */
     lua_pushnil(L);
     lua_pushfstring(L, "'%s' is not a directory", dirpath);
     return 2;
  }

  /* 
     Get all file system entries for this directory, 
     and store them in a table. 
  */  

  lua_newtable(L); /* New table to hold the contents. */
  index = 0;  
  for (handle = readdir(dir); handle != NULL; handle = readdir(dir))
  {
    entryname = handle->d_name;
    /*  name of current entry */
    
    if(strcmp(".", entryname) == 0 ||
      strcmp("..", entryname) == 0)
    {
      continue;
    }
    /*  get rid of those two */
    
    index++;
    /* set string to path of current entry */
    /* !! this path is relative to the current working dir !! */
    lua_pushfstring(L, "%s/%s", dirpath, entryname);
    fs_stat(L);
	lua_pushstring(L, entryname);
	lua_replace(L, -3); /* use bare filename as the index to overall table */
    if(lua_istable(L, -1))
    {
      lua_settable(L,-3); 
    }
  }

  closedir(dir);
  
  return 1; /* we leave the result table on the stack. */        
}

/*
** This function changes the working (current) directory
*/
static int auxL_chdir (lua_State *L)
{
  const char *path;
  char newpath[255+2];  /* @@ truncation possible! */
  /* check parameter type */
  luaL_checkstring (L, -1);
  path = lua_tostring(L, -1);
   
  if (chdir(path)) {
    lua_pushnil(L);
    lua_pushfstring  (L, 
                      "Unable to change working directory to '%s'\n%s\n", 
                      path, strerror(errno));
    return 2;
  } 
  else 
  {
    getcwd(newpath, 255);
    lua_pushstring(L, newpath);
    return 1;
  }
}

/*
** This function returns the current directory
** If unable to get the current directory, it returns nil
**  and a string describing the error
*/
static int auxL_getcwd (lua_State *L)
{
  char path[255+2]; /* @@ truncation possible! */
  if (! getcwd(path, 255)) {
    lua_pushnil(L);
    lua_pushstring(L, strerror(errno));
    return 2;
  }
  else
  {
    lua_pushstring(L, path);
    return 1;
  }
}


/* functions to export */
static const struct luaL_reg fs_funcs[] = {
  {"listdir", fs_listdir},
  {"listdir2", fs_listdir2},
  {"fileinfo", fs_stat},
  {"getcwd", auxL_getcwd},
  {"chdir", auxL_chdir},
  {NULL, NULL}
};


/* open 'fs' library sub-part */
LUAMODULE_API int luaLM_import (lua_State *L) {
  luaL_openlib (L, "fs", fs_funcs, 0);
  return 1;
}
