/*
 * gusu.c
 *
 * Angus' UNIX lib for windows
 */

/********** includes ************/

#include "config.h"

#ifdef PLATFORM_WINDOWS

#include <errno.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "gusu.h"

/********** static functions ************/

static void tv2tb(const struct timeval *tv, struct _timeb *tb)
{
  tb->time = tv->tv_sec;
  tb->millitm = tv->tv_usec / 1000;
}

static void tb2tv(const struct _timeb *tb, struct timeval *tv)
{
  tv->tv_sec = tb->time;
  tv->tv_usec = tb->millitm * 1000;
}

/********** global functions ************/

int gusu_gettimeofday(struct timeval *tv, struct timezone *tz)
{
  struct _timeb tb;

  /* we don't support the timezone */
  if(tz != 0)
  {
    return EINVAL;
  }

  _ftime(&tb);

  tb2tv(&tb, tv);

  return 0;
}

void gusu_usleep(unsigned long usec)
{
  Sleep(usec / 1000);
}

unsigned int gusu_sleep(unsigned int seconds)
{
  time_t t1;
  time_t t2;

  t1 = time(0);
  Sleep(seconds * 1000);
  t2 = time(0);

  return((unsigned int)(seconds - (t2 - t1)));
}

enum { DIR_INIT, DIR_READING, DIR_CLOSED };

DIR *gusu_opendir(const char *dirname)
{
  DIR *d;
  WIN32_FIND_DATA data;
  DWORD attr;

  attr = GetFileAttributes(dirname);
  if (attr == INVALID_FILE_ATTRIBUTES
	  || (attr & FILE_ATTRIBUTE_DIRECTORY) == 0)
  {
    /* not a dir */
    return NULL;
  }

  d = (DIR *)malloc(sizeof(DIR));
  if(d == NULL)
  {
    return(NULL);
  }
  d->state = DIR_INIT;
  d->dirname = (char *)malloc(strlen(dirname)+1+2);
  if(d->dirname == NULL)
  {
    free(d);
    return(NULL);
  }
  strcpy(d->dirname, dirname);
  strcat(d->dirname, "/*");

  return(d);
}

struct dirent *gusu_readdir(DIR *d)
{
  WIN32_FIND_DATA data;
  static struct dirent s_dirent;

  switch(d->state)
  {
    default:
    case DIR_CLOSED:
      return(NULL);
    case DIR_INIT:
	  if((d->h=FindFirstFile(d->dirname, &data)) == NULL)
	  {
		return(NULL);
	  }
      d->state = DIR_READING;
	  s_dirent.d_ino = 0;
	  s_dirent.d_off = -1;
	  s_dirent.d_reclen = 0;
	  strncpy(s_dirent.d_name, data.cFileName, MAX_PATH);
	  s_dirent.d_name[MAX_PATH - 1] = 0;
      return &s_dirent;
    case DIR_READING:
      if(FindNextFile(d->h, &data) == FALSE)
      {
        return(NULL);
      }
      s_dirent.d_ino = 0;
      s_dirent.d_off = -1;
      s_dirent.d_reclen = 0;
      strncpy(s_dirent.d_name, data.cFileName, MAX_PATH);
	  s_dirent.d_name[MAX_PATH - 1] = 0;
      return(&s_dirent);
  }
  return(NULL); // avoid warnings
}

int gusu_closedir(DIR *d)
{
  FindClose(d->h);
  d->state = DIR_CLOSED;
  free(d->dirname);
  free(d);

  return 0;
}

clock_t gusu_times(struct tms *t)
{
    t->tms_utime = clock() / (CLOCKS_PER_SEC / 100);
    t->tms_stime = 0;
    t->tms_cutime = 0;
    t->tms_cstime = 0;

    return(0);
}

#endif
