/*
** GLUAHOST.C  Lua 4.0/5.0 standalone interpreter with GluaX support.
**
** Notes:   #define GLUAHOST_NOMAIN in project options to embed this
**          code within a larger framework.
**
** by asko.kauppi@fi.flextronics.com
*/

/*
 * AP: Use autoconf's generated "config.h" if available
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "gluahost.h"
#include "gluax.h"

#include "loadlib.h"	// Dynamic linkage functions

#include <string.h> // strlen() etc.
#include <stdio.h>	// sprintf()
#include <stdlib.h> // min()

#ifdef GLUA_DYNAMIC
  #error "This module must be compiled with static GluaX support."
#endif

#if defined(WIN32)
  #define IS_SLASH(c) strchr("/\\",(c))
  #define SLASH_STR   "\\"
  #define DL_SUFFIX   ".dll"
  //
#elif defined(PLATFORM_OSX) 
	// OS X...
	#define IS_SLASH		((c)=='/')
  #define SLASH_STR		"/"
	#define DL_SUFFIX		".dylib"
#else   // Linux etc. (default)
  #define IS_SLASH(c) ((c)=='/')
  #define SLASH_STR   "/"
  #define DL_SUFFIX   ".so"
#endif

#define MIN(a,b) ( (a)<(b) ? (a) : (b) )

#ifndef MAXPATHLEN  // Some #includes may have defined this.
  #define MAXPATHLEN 1024
#endif


/* =======================================================
**   Dynamic linkage (platform specific)
** =======================================================
*/
//--------------------- MS Windows --------------------
#ifdef WIN32
  //
  #include <windows.h>
  //
  typedef HINSTANCE h_Module;
  //
  #define Loc_OpenDll(path)   LoadLibrary(path)
  #define Loc_CloseDll(h)     FreeLibrary( h )
  #define Loc_DllFunc(h,name) GetProcAddress( h, name )
  //
  static const char* Loc_DllError( void )
  { static char buf[200];
    //
    if ( FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
                       0, GetLastError(), 0, buf, sizeof(buf)-1, 0 ) )
        return buf;
    else
        return NULL;
  }
  //
//--------------------- Darwin / Mac OS X --------------------
#elif defined(PLATFORM_OSX)
  //
  // Thanks to Adrian Prez <mobius@iespana.es> for providing this
  // code and allowing it to be used in GluaX.
  //
  /*
   * Small (and fast) wrapper around the Dyld API to dynamically load
   * modules. Due to limitations in OSX's public APIs, only modules
   * with a MH_BUNDLE header can be loaded; you can create one of
   * these by passing the '-bundle' option to Apple's GCC when linking
   * the module.
   *
   * At this moment, this module loader does nothing more than providing
   * a clearer API to the obscure one that Dyld includes (splitted over
   * three different sets of functions, argh!).
   */
  #include <mach-o/dyld.h>
  //
  //typedef struct
  //{
	//	NSModule module;
	//	NSObjectFileImage image;	
  //} osx_hModule;
	//typedef osx_hModule* h_Module;
	typedef NSModule h_Module;
  //
  h_Module Loc_OpenDll(const char *path)
  {
  	NSObjectFileImageReturnCode ret;
  	NSObjectFileImage nsimg;
 	 	NSModule nsmod;
  	//osx_hModule* h;

    if (!path) return NULL;

		// Load image from file:
    //
		ret = NSCreateObjectFileImageFromFile(path, &nsimg);

    if (ret != NSObjectFileImageSuccess)
        return NULL;    // Unable to load!
	
	// Link into the rest of the executable:
    //
		nsmod = NSLinkModule(nsimg, path,
			NSLINKMODULE_OPTION_PRIVATE | NSLINKMODULE_OPTION_RETURN_ON_ERROR);
		NSDestroyObjectFileImage(nsimg);
		if (!nsmod)
			return NULL;    // Unable to link!

    // Success :)
    //	
		//h = (osx_hModule*) malloc(sizeof(osx_hModule));
  	//ASSUME_NONZERO(h);

		//h->image  = nsimg;
		//h->module = nsmod;

		return nsmod;
  }
  //
  void Loc_CloseDll( h_Module h )
  {
		ASSUME_NONZERO(h);
		NSUnLinkModule(h /*->module*/, NSUNLINKMODULE_OPTION_NONE);
		//NSDestroyObjectFileImage(h->image);
    //free(h);
  }
  //
  void* Loc_DllFunc( h_Module h, const char *symbol )
  {
  	NSSymbol nss;
  	char real_symbol[strlen(symbol) + 2];
  	*real_symbol = '_';
  	strcpy(real_symbol + 1, symbol);
	
		ASSUME_NONZERO(h);
		nss = NSLookupSymbolInModule(h /*->module*/, real_symbol);
    return (nss) ? NSAddressOfSymbol(nss) : NULL;
  }
  //
  #define Loc_DllError() (NULL)   // no explanations for now
  //
//--------------------- Linux & most Unixes --------------------
#else   // default
  //
  //---  
  // Note: We use 'RTLD_NOW' (and not 'RTLD_LAZY' to know right away that all
  //       necessary symbols really are there. Also, since there's just a 
  //       single entry point in GluaX modules, the optimization effect of
  //       'LAZY' would most likely be none.
  //
  #include <dlfcn.h>
  typedef void* h_Module;

  #define Loc_OpenDll(filename)  dlopen( filename, (RTLD_NOW | RTLD_LOCAL) )
  #define Loc_CloseDll(lib)      dlclose( lib )
  #define Loc_DllFunc(lib,name)  dlsym( lib, name )
  #define Loc_DllError()         dllerror()
  //
#endif

#ifndef MAXPATHLEN  // Some #includes may have defined this.
  #define MAXPATHLEN 1024
#endif


//---------------------------------

static bool_int
Loc_FileExists( const char* filename )
{
    FILE* f= fopen( filename, "rb" );
    if (f) fclose(f);

    return (f != NULL);
}


static bool_int
Loc_PrepareFilename( char buf[MAXPATHLEN], const char* fn, const char* def_suffix )
{
    // Try first with filename as it is:
    //
    strcpy_safe( buf, fn, MAXPATHLEN );
    
    if (Loc_FileExists(buf))
        return TRUE;
        
    // Then with default suffix appended:
    //
    if (def_suffix)
       {
       strcat_safe( buf, def_suffix, MAXPATHLEN );

       if (Loc_FileExists(buf))
          return TRUE;
       }

   return FALSE;   // No such file!
}


/* =======================================================
**   Connecting to Lua
**
**      int= Gluahost( filename_str [,namespace_tbl] )
**      int= Gluahost( filename_str, prefix_str )
**
**   You can use 'namespace_tbl' to force the module's definitions under 
**   a certain namespace (s.a. "MyLib.MyFunc()") in Lua. No second param
**   places them in global namespace.
** =======================================================
*/
static struct s_GluaFuncs _gluafuncs= { GLUA_LINKS };

GLUA_FUNC( Gluahost )
{
const char* orig_filename;
const char* err_param= NULL;
h_LoadLib h=0;
t_GluaModuleFunc* func= NULL;
int ret= 0;
const char* err= NULL;
uint target_tbl=0;
char fn_buf[ MAXPATHLEN ];
bool_int found;

    orig_filename= glua_getString(1);
    if (!orig_filename)
        {
		ret= 0;
		err= "Usage: Gluahost( filename_str [,namespace_tbl] )\n"
             "       Gluahost( filename_str, prefix_str )\n"
			 "\n"
             "Loads GluaX extension modules (.dll/.so) to be used from Lua.";
		goto Error;
        }

    if (glua_isTable(2))
        {
        target_tbl= 2;  // index of the namespace table
        }
    else  // Old way: create the table as a global name
        {
        const char* prefix= glua_getString(2);  // (may be NULL)

        if (prefix != NULL)        // NULL puts into global namespace
           {
           STACK_CHECK(_lua_)
              {
              lua_getglobal( _lua_, prefix );  // An existing table?
              
              if ( !lua_istable( _lua_, -1 ) )
                 {
                 // Create a new table at top-of-stack & name it
                 //
                 lua_settop(_lua_,-2);  // remove the 'nil'
                 lua_newtable(_lua_);
                 
                 lua_pushvalue( _lua_,-1 );  // Local duplicate for
                                             // 'setglobal()' to eat.
                 lua_setglobal( _lua_, prefix );
                 
                 // The new table left at top-of-stack (-1).
                 }
              }
           STACK_END(+1)  // One item (the table) created.
           
           target_tbl= lua_gettop(_lua_);  // absolute index
           }
        }

// tulrich: PrepareFilename junk not necessary for luacheia
#if 0        
    found= Loc_PrepareFilename( fn_buf, orig_filename, DL_SUFFIX );

    if (!found)
        {
        ret= -1;
        err= "Module '%s' not found!";
        err_param= orig_filename;
        goto Error;
        }
#endif // 0

    err_param= orig_filename;

    h= loadlib_open( orig_filename );
    //
    if (!h)
        {
        ret= -2;
        err= "Unable to open '%s'! (used by another process or missing some library?)";
        goto Error;
        }

    func= (const t_GluaModuleFunc*) /*Loc_DllFunc*/ loadlib_func( h, "GluaModule" );
    //
    if (!func)
        {
        ret= -3;
        err= "No 'GluaModule()' entry point!";
        goto Error;
        }

    // Check that our function table has no holes (safety check).
    {
    void** vptr= (void**)(&_gluafuncs);
    uint i;

    for( i=0; i < (sizeof(_gluafuncs) / sizeof(void*)); i++ )
        {
        if ( vptr[i] == NULL )
            {
            ret= -4;
            err= "Holes in function table! (check GLUA_LINKS definition)";
            goto Error;
            }
        }
    }

    // Seems like a valid Glua extension module: initialise it!
    //
    ret= glua_link( _lua_,   // <-- (hidden parameter by 'GLUA_FUNC')
                    _gluafuncs,
					func,
                    orig_filename,
                    target_tbl );
    //
    // Negative returns are interpreted as error.
    // Positive ones only as warnings (will be passed on to Lua).
    //
    if (ret<0)
        err= "'GluaModule('%s')' initialization failed! (code=%d)";

Error:
    if (err)
        {
		char buf[999];  // protected (see below)
        char loadlib_buf[256];
        const char* err2;

        if (h)
            {
            loadlib_close( h );
            h= 0;   // don't use any more!
            }

		// We could use 'lua_pushfstring()' to bypass the need for 
		// a fixed buffer, but it only works on 5.0.
		//
		//  lua_pushstring( _lua_, "Gluahost: " );
		//	lua_pushfstring( _lua_, err, err_param, ret );
		//	lua_concat( _lua_, 2 );	  // merge the strings
		//	glua_error( lua_tostring( _lua_, -1 ) );
		
		// This works for both 4.0 and 5.0:
		//
	    strcpy( buf, "Gluahost: " );
	    
	    // Buffer overrun check:
	    ASSUME( sizeof(buf) > strlen(buf) + strlen(err) + strlen(err_param) 
                              + 10 /*integer space*/ );
	    
		sprintf( buf+strlen(buf), err, err_param, ret );

        // Note: 'loadlib_error()' does not necessarily write its 
        //       string to the buffer. This is platform dependent.
        //
        err2= loadlib_error( loadlib_buf, sizeof(loadlib_buf) );
        if (err2)
            {
            strcat_safe( buf, "\n", sizeof(buf) );
            strcat_safe( buf, err2, sizeof(buf) );
            }

		glua_error( buf );  // does not return
		
		ASSUME( NEVER );
		}

    glua_pushInteger( ret );

    // Note: Currently, we don't store the module handles anywhere but rely
    //       on the OS to automatically unlink them at program exit.
    //
    //       If necessary, Lua registry can be used to store the pointers.

    // Done! :)
}
GLUA_END


/* =======================================================
**   Compatibility helpers
**
**      tag_int= newtag( )
**      tag_int= tag( userdata_ref )
**      void= settag( table, tag_int )
**      old_func= settagmethod( tag_int, event_str, new_func )
**
**   These funcs help running 4.0 coded Lua scripts with a 5.0 Lua engine.
** =======================================================
*/
#ifndef LUA_V40     // Only need these funcs on a 5.0 engine.
GLUA_FUNC( newtag )
{
    glua_pushInteger( _glua_newtag( _lua_ ) );
}
GLUA_END
//
GLUA_FUNC( tag )    // tag(ud)
{
    glua_pushInteger( Loc50_Tag( _lua_, 1 ) );
}
GLUA_END
//
GLUA_FUNC( settag )
{
    Loc50_SetTag( _lua_, 1 /*table ref*/, glua_getInteger(2) );

    // Note: If returning a copy of the table reference is required (as 
    // described in the Lua 4.0 manual) that can easily be achieved by 
    // a small Lua wrapper around this function.
}
GLUA_END
//
// Note: Since 'settagmethod' needs some hacking anyhow, we use the
//       native Lua C API to define it. 
//
static int fn_settagmethod(lua_State* L)
{
int tag= (int)lua_tonumber(L,1);
const char* event= lua_tostring(L,2);

    ASSUME( lua_isfunction(L,3) || lua_isnil(L,3) );

    STACK_CHECK(L)
        {
        lua_pushvalue( L, 3 );  // temp copy at top-of-stack

        Loc50_SetTagMethod( L, tag, event, TRUE /*push old*/ );
        }
    STACK_END(+1)   // should have pushed the old entry

    return 1;   // one return value
}
//
#endif  // Lua 5.0


//-----
#undef GLUA_MODULE_NAME
#define GLUA_MODULE_NAME GluaModuleGluahost
GLUA_MODULE( my_path, my_tbl )
{
    // Functions & constants to be accessible from Lua:
    //
    GLUA_DECL( my_tbl )
        {
        glua_func2( "gluahost", Gluahost ),

        // Lua engine information:
        glua_str2( "version", LUA_VERSION ),
        glua_str2( "copyright", LUA_COPYRIGHT ),
        glua_str2( "authors", LUA_AUTHORS ),

        // Support for 4.0 source on 5.0 engine:
        #ifndef LUA_V40   // >=5.0
          glua_func( newtag ),
          glua_func( tag ),
          glua_func( settag ),
          glua_func( settagmethod ),
        #endif
        }
    GLUA_DECL_END
}
GLUA_MODULE_END


//---
// Pushes a table reference on the Lua stack (if needed, creating the table)
// and returns the absolute stack index to it.
//
static int Loc_TableRef( lua_State* lua, const char* table_name )
{
const char* ptr= table_name;
const char* dot;

    if (!table_name) return 0;  // global namespace

    STACK_CHECK(lua)
        {
        #ifdef LUA_V40
            lua_getglobals( lua );  // pushes ref to global table
        #else   // 5.0 & later
            lua_pushvalue( lua, LUA_GLOBALSINDEX );   // does this work?
        #endif

        // Each loop goes one stage further, replacing the parent
        // table ref with the child's.
        //
        while( ptr )
            {
            dot= strchr( ptr, '.' );    // end of current stage

            if (!dot)
                {
                lua_pushstring( lua, ptr ); // last stage
                ptr= NULL;
                }
            else
                {
                lua_pushlstring( lua, ptr, dot-ptr );
                ptr= dot+1;
                }

            lua_pushvalue( lua, -1 );   // 2nd ref to stage name

            // [-1]: stage name
            // [-2]: 2nd ref to stage name
            // [-3]: ref to parent

            lua_gettable( lua, -3 );    // pops key, adds value

            if (lua_istable( lua, -1 ))
                {
                // Subtable existed, proceed to next stage
                //
                lua_remove( lua, -2 );  // stage name
                lua_remove( lua, -2 );  // parent table
                }
            else
                {
                // Create a new table & tie it.
                //
                lua_remove(lua,-1);     // 'nil' away

                lua_newtable(lua);
                //
                // [-1]: empty table
                // [-2]: stage name
                // [-3]: ref to parent

                lua_pushvalue( lua, -1 );   // 2nd ref to new table
                lua_insert( lua, -4 );

                lua_settable( lua, -3 );        
                lua_remove( lua, -1 );      // parent away
                }

            ASSUME( lua_istable(lua,-1) );  // new stage
            }
        }
    STACK_END(+1)   // should have pushed one

    return lua_gettop(lua);   // top of stack
}


/* Lua loadmodule support */

LUAMODULE_API int	luaLM_import(lua_State* L)
{
	return gluahost_open(L);
}

LUAMODULE_API const char* luaLM_version(void)
{
	return LUA_VERSION;
}


//-----
extern t_GluaModuleFunc GluaModule;     // This module
// tulrich: commented out
//extern t_GluaModuleFunc GluaModule_sys; // Filesystem support

int gluahost_open (lua_State* lua)
{
int tbl_index;

    // 'Gluahost' initialization:
    //
    tbl_index= Loc_TableRef( lua, "gluahost" );
    GluaModuleGluahost( lua, 0,0,0, tbl_index );

    // Init of other built-in modules:
    //
// tulrich: commented out
//    tbl_index= Loc_TableRef( lua, "gluahost.static.sys" );
//    GluaModule_sys( lua, 0,0,0, tbl_index );
    
    return 0;
}

//-----
int gluahost_close (lua_State* lua)
{
    // Could unload any loaded modules here...
    // (if we kept track of them, which we don't - yet)
    //
	// NOTE: This call is being done also from possible created threads
	//       so we need a reference counter not to clean up too early!
	//
    return 0;
}
