/*---------------------
//
// GLUAX.C   by asko.kauppi@fi.flextronics.com
//
// This module needs to be compiled into any program using 'gluax.h'.
//
// TO DO: 
//      ...
----------------------*/

#include "gluax.h"

#include <string.h>		// strlen() etc.
#include <stdio.h>		// vsprintf()
#include <stdlib.h>		// min()
#include <ctype.h>		// isdigit()
#include <stdarg.h>
#include <math.h>		// floor()

//----
#define GLUA_VERSION         0x0217   // Version of this GluaX host/module

#ifdef GLUA_DYNAMIC
  #define HOST_VERSION_REQ   0x0202   // Oldest host we can still live with
#else  // static
  #define MODULE_VERSION_REQ 0x0202   // Oldest modules we can still support
#endif

//----
// First Aid for non-__VA_ARGS__ compilers:
// (Thread Specific Data needed or multithreading banned):
//
#ifndef _HAS_VA_ARGS    // poor compilers without __VA_ARGS__
  //
  #ifdef COMPILER_MSC   // MSC (using Thread Local Storage)
    //
    #define WIN32_LEAN_AND_MEAN
    #include <windows.h>	// Thread Local Storage
    //
    #define _TLS_UNINITIALIZED ((uint)(-1))
    //
    static DWORD tls_L= _TLS_UNINITIALIZED;
    static DWORD tls_T= _TLS_UNINITIALIZED;
    //  
    void _setL( lua_State* L ) { TlsSetValue(tls_L,L); }
    lua_State* _getL(void)	 { void* p= TlsGetValue(tls_L); ASSUME(p); return (lua_State*)p; }
    int _setT( int argn )	 { TlsSetValue(tls_T,(void*)argn); return argn; }
    int _getT(void)			 { void* p= TlsGetValue(tls_T); return (int)p; }
    //
  #elif 0   // (could do similar here for PThreads..)
    // ...to be done...
    //
  #else     // Unknown compiler, single threading only:
    //
    static lua_State* L= NULL;
    static int T= 0;
  
    void _setL( lua_State* lua ) { L=lua; }
    lua_State* _getL(void)	 { ASSUME(L); return L; }
    int _setT(int argn)		 { T=argn; return argn; }
    int _getT(void)			 { return T; }
    //
  #endif
#endif

//----
char* strcpy_safe( char* buf, const char* src, uint buf_size )
{
    // Make sure the buffer is always terminated, even at fill-ups:
    //
    buf[buf_size-1]= '\0';

    return strncpy( buf, src, buf_size-1 /*max charcount*/ );
}

char* strcat_safe( char* buf, const char* src, uint buf_size )
{
    // 'strncat()' always terminates the buffer but the size param
    // is charcount of 'src', not the size of the buffer!
    //
    return strncat( buf, src, buf_size -strlen(buf) -1 );
}


/*--- Interface to Lua engine ---------------------------------------------*/

//----
// Notes:
//		'lua_tonumber()' converts a numeric string parameter into actual number
//		type on Lua 4.0 but _not_ (necessarily) on Lua 5.0 and later. Do not 
//		rely on this behaviour!
//
//		'lua_tostring()' converts a number into string type (modifying the 
//		actual stack item) on both versions.
//----

// lua_...() functions can be used as such in statically linked code.
// HOST_CALL...() works over the DLL gate from dynamic modules.
//
#ifdef GLUA_DYNAMIC     // Accessing Lua host via a function table.
  //
  static const struct s_GluaFuncs* _lua_gate_= NULL;
  //
  #define HOST_CALL_1( func, p1 )			  ((_lua_gate_->func)(p1))
  #define HOST_CALL_2( func, p1,p2 )		  ((_lua_gate_->func)(p1,p2))
  #define HOST_CALL_L( func, L )              ((_lua_gate_->func)(L))
  #define HOST_CALL_L1( func, L,p1 )          ((_lua_gate_->func)(L,p1))
  #define HOST_CALL_L2( func, L,p1,p2 )       ((_lua_gate_->func)(L,p1,p2))
  #define HOST_CALL_L3( func, L,p1,p2,p3 )    ((_lua_gate_->func)(L,p1,p2,p3))
  #define HOST_CALL_L4( func, L,p1,p2,p3,p4 ) ((_lua_gate_->func)(L,p1,p2,p3,p4))
  //
#else   // Static linking, all comes from 'lua.h' directly.
  //
  #include "lauxlib.h"	// 'lua_dostring()' etc. compatibility macros for v5.0.
  #include "lualib.h"	// 'lua_baselibopen()' etc.
  //
/*** (needed?) ***
  #ifndef GLUAHOST_OPEN  // Compiler project may #define this as it likes.
    int gluahost_open( lua_State* L );  // as in 'gluahost.h'
    int gluahost_close( lua_State* L );
    //
    #define GLUAHOST_OPEN(L)  gluahost_open(L)
    #define GLUAHOST_CLOSE(L) gluahost_close(L)
  #endif
***/
  //
  static int fn_Glua_GcHub( lua_State* );
  static int fn_Glua_GcFree( lua_State* );
  //
#endif  // GLUA_DYNAMIC


/*--- Debugging functions ---------------------------------------------*/

void _glua_assert( uint line, const char* file, const char* condition )
{
int i;

    if (!file) file="";
    if (!condition) condition="";
    
    // Path part in 'file' is not so important and better be skipped:
    //
    for( i=strlen(file)-1; i>=0; i-- )
        {
        if (strchr( "/\\", file[i] ))    // Is 'file[i]' a slash?
            { file= &file[i+1]; break; }
        }
    
    fprintf( stderr, "'ASSUME(%s)' failed at '%s' line %u !\n", condition, file, line );
    
    abort();    // Do not return.
}

#if 0  // Stack dump only used for debugging purposes (normally disabled)
//
static const char* Loc_Typename( enum e_Glua_Type type )
{
	switch( type )
		{
		//case GLUA_TYPE_NONE:	 return "none";
		case GLUA_TYPE_NIL:		 return "nil";
		case GLUA_TYPE_BOOLEAN:  return "bool";
		case GLUA_TYPE_FLOAT:    // "num"
		case GLUA_TYPE_INT:		 return "num";
		case GLUA_TYPE_STRING:	 return "str";
		case GLUA_TYPE_TABLE:	 return "tbl";
		case GLUA_TYPE_FUNCTION: return "func";
		case GLUA_TYPE_USERDATA: // "usr";
		case GLUA_TYPE_USERDATA_UNK: return "usr";	// was:"usd"
        //
        default: break;
		}

	return "???";
}
//---
// Like 'sprintf()' but concatenates at the end of the buffer.
//
static void Loc_DumpF( char* buf, const char* msg, ... )
{
va_list args;
char* bufend;

	ASSUME_NONZERO( buf );

	bufend= buf + strlen(buf);

    va_start(args,msg);
        {
        vsprintf( bufend, msg, args );
        }
    va_end(args);
}
//
#define DEBUG_STACK(L) _glua_dump(L)
//
void _glua_dump(lua_State* L)
{
char buf[3999]="";  // Long enough for any stack dump?
//                  // (only used for debugging so protection not vital)
//
int top= _glua_gettop(L);
enum e_Glua_Type type;
int i;

	if (top==0)
		Loc_DumpF( buf, "\t(none)\n" );

	for( i=1; i<=top; i++ )
		{
		type= _glua_type( L, i );

		Loc_DumpF( buf, "\t[%d]= (%s) ", i, Loc_Typename(type) );

		// Print item contents here... (for some types)
		//
		switch( type )
			{
			case GLUA_TYPE_FLOAT:
			case GLUA_TYPE_INT:
				//
				Loc_DumpF( buf, "%lf", (double)_glua_tonumber_safe(L,i) );
				break;

			case GLUA_TYPE_STRING:
				//
				Loc_DumpF( buf, "'%s'", _glua_tostring_safe(L,i) );
				break;

			case GLUA_TYPE_BOOLEAN:
				//
				Loc_DumpF( buf, _glua_toboolean(L,i) ? "TRUE":"FALSE" );
				break;

			default:
				// (don't print the other types' contents)
				break;
			}

		Loc_DumpF( buf, "\n" );
		}

	fprintf( stderr, "\tDEBUG STACK:\n%s\n", buf );
}
#endif  // (debugging)


/*--- Local help functions ---------------------------------------------*/

#define NUM_IS_INT(v) ((v) == floor(v))

static bool_int Loc_IsNumeric( const char* str )
{
char* end;

    strtod( str, &end );

    return (*end == '\0');  // FALSE if there was a non-numeric char
}

static void Loc_PushMerged_int( lua_State* L, const char* str, int nbr )
{
	_glua_pushstring( L, str );     // e.g. "glua_attach_"
    _glua_pushnumber( L, nbr );
	_glua_concat( L, 2 );           // merges them into one string
}

static void Loc_PushMerged_str( lua_State* L, const char* a, const char* b )
{
	_glua_pushstring( L, a );
    _glua_pushstring( L, b );
	_glua_concat( L, 2 );
}

#ifdef GLUA_STATIC
  #ifdef LUA_V40	// 4.0 actually converts in-location
    //
    #define Loc_ConvToNumber(L,argn)  lua_tonumber(L,argn)
	//
  #else
	// 5.0 does normally no conversion so we need to help a bit...
	//
    #define Loc_ConvToNumber(L,argn)  Loc50_ConvToNumber(L,argn)
    //
  #endif
#endif


//----
// Event names differ by Lua engine version:
//
//		GluaX:		Lua 4.0:		Lua 5.0:
//
//		"gettable"	"gettable"		"__index"
//		"settable"	"settable"		"__newindex"
//		"function"	"function"		"__call"
//		"gc"		"gc"			"__gc"
//		"add"		"add"			"__add"
//		"sub"		"sub"			"__sub"
//		"mul"		"mul"			"__mul"
//		"div"		"div"			"__div"
//		"pow"		"pow"			"__pow"
//		"unm"		"unm"			"__unm"
//		"lt"		"lt"			"__lt"
//		"le"		(none)			"__le"		(*
//		"concat"	"concat"		"__concat"
//
//		"index"		"gettable"		"__index"   (**
//		"newindex"	"settable"		"__newindex"
//		"call"		"function"		"__call"
//
// *)  'le' not available on Lua 4.0 engine.
//
// **) There is also a native Lua 4.0 'index' method but due to keeping
//     things simple (= easily documentable) that is _not_ used by gluax.
//     This way, we can treat 'gettable' and 'index' as totally the same.
//
#ifdef GLUA_STATIC
//
struct s_EventNameLookup { const char* gluax; const char* lua; };
//
#ifdef LUA_V40
  #define _LOOKUP( gluax, lua4, lua5 )    { gluax, lua4 }   // ignore lua5
#else
  #define _LOOKUP( gluax, lua4, lua5 )    { gluax, lua5 }   // ignore lua4
#endif
//
static const struct s_EventNameLookup lookup[]= {
    
    _LOOKUP( "gettable", "gettable", "index" ),
    _LOOKUP( "settable", "settable", "newindex" ),
    _LOOKUP( "function", "function", "call" ),
    _LOOKUP( "gc",       "gc",       "gc" ),
    _LOOKUP( "add",      "add",      "add" ),
    _LOOKUP( "sub",	     "sub",      "sub" ),
    _LOOKUP( "mul",	     "mul",      "mul" ),
    _LOOKUP( "div",	     "div",      "div" ),
    _LOOKUP( "pow",      "pow",      "pow" ),
    _LOOKUP( "unm",      "unm",      "unm" ),
    _LOOKUP( "lt",       "lt",       "lt" ),
    _LOOKUP( "le",       NULL,       "le" ),    // does not exist for 4.0!
    _LOOKUP( "concat",   "concat",   "concat" ),
    //
    _LOOKUP( "index",    "gettable", "index" ),     // same as "gettable"
    _LOOKUP( "newindex", "settable", "newindex" ),  // same as "settable"
    _LOOKUP( "call",     "function", "call" ),      // same as "function"
    //
    { NULL }  // terminating entry
};
//
static const char* EVENT_NAMECHANGE( lua_State* L, const char* event )
{
uint i;
const char* ret= NULL;

    // Skip underscores if the application uses them:
    //
    while(event[0]=='_') event++;

    for( i=0; lookup[i].gluax != NULL; i++ )
        {
        if (strcmp( lookup[i].gluax, event ) == 0)
            {
            ret= lookup[i].lua;     // Native, version dependent name.
            break;
            }
        }

    if (!ret)
        _glua_errorN( L, "Unknown event '%s'!", event );
            
    return ret;    // Unknown event name!!
}
#endif	// static


/*--- Lua interfacing stubs ----------------------------------------------*/

void _glua_error( lua_State* L, const char* msg )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_error, L, msg );
	//
#elif (defined LUA_V40)   // v4.0 static
	//
	lua_error( L, msg );
    //
#else	// v5.0 static
	//
    lua_pushstring( L, msg );
	lua_error( L );
    //
#endif
}

//-----
// Note: We cannot use Lua 5.0 'lua_pushvfstring()' since we might be 
//       dynamically linked. Passing 'args' further to the host...
//       perhaps different compilers... Seems too risky!
//
//       const char* tmp= lua_pushvfstring(L, msg, args);
//
static
const char* _glua_vsprintf( lua_State* L, const char* msg, va_list args )
{
char* buf;
uint bufsize= 9999;  // !! Should make an estimate based on 'msg' & 'args' !!

    ASSUME_NONZERO( msg );
    
    // By pushing a userdata, we allocate memory that will later be
    // freed by Lua automatically (since there's no ref to it).
    //
    STACK_CHECK(L)
        {
        buf= (char*)_glua_pushuserdata( L, 0 /*no tag*/, NULL, bufsize, NULL );
        
        _glua_settop( L, -2 );  // pops the reference (but pointer will be 
                                // valid until we return from the C side)
        }
    STACK_END(0)

    ASSUME_NONZERO( buf );
        
    vsprintf( buf, msg, args );
    
    return buf;
}

void _glua_errorL( lua_State* L, const char* msg, va_list args )
{
const char* buf= _glua_vsprintf( L, msg, args );

    _glua_error( L, buf );   // never returns
}

void _glua_errorN( lua_State* L, const char* msg, ... )
{
const char* buf;
va_list args;

    va_start(args,msg);
        {
        buf= _glua_vsprintf( L, msg, args );
        }
    va_end(args);
    
    _glua_error( L, buf );  // never returns
}

#ifndef _HAS_VA_ARGS
  //
  void glua_errorN( const char* msg, ... )
    {
    lua_State* L= _getL();
    const char* buf;
    va_list args;

      va_start(args,msg);
        {
        buf= _glua_vsprintf( L, msg, args );
        }
      va_end(args);
    
      _glua_error( L, buf );  // never returns
    }
#endif

uint _glua_ver( uint module_ver )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_1( _glua_ver, module_ver );
	//
#else	// static (host side)
	//
	if ((module_ver>0) && (module_ver < MODULE_VERSION_REQ))
		return 0;	// We disapprove this module!

	return GLUA_VERSION;	// Gluahost version to the module
							// (also it may disapprove us)
    //
#endif
}

//---
// Note: Do NOT return 'enum' directly since its size may be compiler
//       specific.
//
enum_uint /*e_Glua_Type*/ _glua_type2( lua_State *L, int argn, int ud_tag )
{
#ifdef GLUA_DYNAMIC
	//
	return HOST_CALL_L2( _glua_type2, L, argn, ud_tag );
	//
#else
	int items, tmp;
	enum e_Glua_Type ret;

	// Native Lua type enumerations are different for Lua 4.0 and 5.0.
	//
	enum e_Glua_Type lookup[]=
		//
		#ifdef LUA_V40
		//
		{ GLUA_TYPE_USERDATA,	//#define LUA_TUSERDATA	0
		  GLUA_TYPE_NIL,		//#define LUA_TNIL	    1
		  GLUA_TYPE_FLOAT,		//#define LUA_TNUMBER	2
		  GLUA_TYPE_STRING,		//#define LUA_TSTRING	3
		  GLUA_TYPE_TABLE,		//#define LUA_TTABLE	4
		  GLUA_TYPE_FUNCTION	//#define LUA_TFUNCTION	5
		  // no GLUA_TYPE_BOOLEAN for 4.0
		};	//#define LUA_TNONE (-1)

          #if (! ((LUA_TUSERDATA==0) && (LUA_TNIL==1) && (LUA_TNUMBER==2) && \
                  (LUA_TSTRING==3) && (LUA_TTABLE==4) && (LUA_TFUNCTION==5)))
            #error "Wrong assumption on Lua types!"
          #endif
		#else
		//
		// 5.0 beta:
		{ GLUA_TYPE_NIL,		//#define LUA_TNIL	    0
		  GLUA_TYPE_BOOLEAN,	//#define LUA_TBOOLEAN	1
		  GLUA_TYPE_USERDATA,	//#define LUA_TLIGHTUSERDATA 2
		  GLUA_TYPE_FLOAT,		//#define LUA_TNUMBER	3
		  GLUA_TYPE_STRING,		//#define LUA_TSTRING	4

		  GLUA_TYPE_TABLE,		//#define LUA_TTABLE	5
		  GLUA_TYPE_FUNCTION,	//#define LUA_TFUNCTION	6
		  GLUA_TYPE_USERDATA,	//#define LUA_TUSERDATA	7
		  //GLUA_TYPE_THREAD,	//#define LUA_TTHREAD	8
		  
		  // NOTE: Both normal 'userdata' and 'light userdata' map into
		  //       same GLUA type. (could be separate type, too, if needed)
		  //
          #if (! ((LUA_TNIL==0) && (LUA_TBOOLEAN==1) && (LUA_TLIGHTUSERDATA==2) &&\
                  (LUA_TNUMBER==3) && (LUA_TSTRING==4) && (LUA_TTABLE==5) && \
                  (LUA_TFUNCTION==6) && (LUA_TUSERDATA==7) && (LUA_TTHREAD==8)))
            #error "Wrong assumption on Lua types!"
          #endif

		};	//#define LUA_TNONE (-1)
		#endif

	items= (sizeof(lookup) / sizeof(lookup[0]));

	tmp= lua_type( L, argn );

	// Make 'none' type seem like 'nil' to the caller. This helps us handle
	// variable parameter lists and default parameters:
	//
	//		func( a, nil )	--> 2nd param is 'LUA_TNIL'
	//		func( a )		--> now it is 'LUA_TNONE'
	//
	// However, we do not need to differ between these two, and can return
	// 'GLUA_TYPE_NIL' for both of them. So this function never returns 'none'.
	//
	if (tmp<0)	// LUA_TNONE
		return GLUA_TYPE_NIL;

	ASSUME( (tmp>=0) && (tmp<items) );

	ret= lookup[tmp];

	// Add possible extra flags or do some tests.
	//
	switch( ret )
		{
		case GLUA_TYPE_FLOAT:
			{
			double d= lua_tonumber(L,argn);	// we know it's numeric so
											// no conversions are caused
			if (NUM_IS_INT(d))
				ret= GLUA_TYPE_INT;		// no fractional part
			}
			break;

		case GLUA_TYPE_USERDATA:
			{
			if (ud_tag != 0)	// && (tag != _glua_tagofuserdata( L, argn )))
				{
                int tag2;

				#ifdef LUA_V40
					tag2= lua_tag(L,argn);
				#else
					tag2= Loc50_Tag(L,argn);
				#endif

				if (ud_tag != tag2)
					ret = GLUA_TYPE_USERDATA_UNK;	// tags don't match!
				}
			}
			break;

		default:
			break;	// 'ret' good as it is
		}

	return (uint)ret;
	//
#endif	// GLUA_DYNAMIC
}

//---
bool_int _glua_toboolean( lua_State* L, int argn )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_L1( _glua_toboolean, L,argn );
    //
#elif (defined LUA_V40)
    {
	int type= _glua_type( L, argn );

    // Lua 4.0: anything non-nil goes as 'true':
    //
	if (type == GLUA_TYPE_NIL)
		return FALSE;	// 'nil' (or none)
	else
		return TRUE;	// any non-nil type
	}
#else   // 5.0 statically linked:
    {
	return lua_toboolean( L, argn );
	}
#endif
}

//---
// Note: May convert numeric strings to number type! (on Lua 4.0)
//       Use 'glua_tonumber_safe()' to avoid conversion (but slower).
//
//		 Never rely on the conversion - 5.0 does not do it!
//
double  _glua_tonumber_raw( lua_State* L, int argn )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_L1( _glua_tonumber_raw, L,argn );
#else
	{
    if (_glua_type(L,argn) <= GLUA_TYPE_NIL)
        return 0;   // 'nil' is okay - 0.0 is default

	return lua_tonumber( L, argn );
	}
#endif
}

//---
// Note: Converts numbers to strings!
//       Use 'glua_tostring_safe()' to avoid conversion (but slower).
//
// backwards compatibility:
const char* _glua_tostring_raw( lua_State* L, int argn )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_L2( _glua_tolstring_raw, L,argn, NULL );
#else
	return lua_tostring( L, argn );  // may convert stack[argn]
#endif
}

const char* _glua_tolstring_raw( lua_State* L, int argn, size_t* len_ref )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_L2( _glua_tolstring_raw, L,argn, len_ref );
#else
	const char* p= lua_tostring( L, argn ); // may convert stack[argn]

    if (len_ref) *len_ref= lua_strlen( L, argn );

    return p;
#endif
}

void*   _glua_touserdata( lua_State* L, int argn )
{
#ifdef GLUA_DYNAMIC
	//
    return HOST_CALL_L1( _glua_touserdata, L,argn );
	//
#elif (defined LUA_V40)
	//
	return lua_touserdata( L, argn );
	//
#else   // v.5.0 static
	//
    return Loc50_ToUserdata(L,argn);
	//
#endif
}

void _glua_pushboolean( lua_State* L, bool_int v )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_pushboolean, L,v );
	//
#elif (defined LUA_V40)
	//
    if (v) lua_pushnumber( L, 1.0 /*was 'v'*/ );
    else   lua_pushnil( L );
	//
#else	// 5.0 static
	//
	lua_pushboolean( L, v );
	//
#endif
}

void _glua_pushnumber (lua_State *L, double v)
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_pushnumber, L,v );
#else
	lua_pushnumber( L, v );
#endif
}

void _glua_pushstring (lua_State *L, const char *s)
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_pushstring, L,s );
#else
    if (!s) lua_pushnil(L);     // 'glua_pushstring(NULL)' = push 'nil'
    else    lua_pushstring( L, s );
#endif
}

void _glua_pushlstring( lua_State *L, const char *s, uint len )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L2( _glua_pushlstring, L,s,len );
#else
    if (!s) lua_pushnil(L);     // 'glua_pushstring(NULL)' = push 'nil'
    else    lua_pushlstring( L, s, len );
#endif
}

//---
void* _glua_pushuserdata (lua_State *L, int tag, 
						  const void *vp, size_t size,
						  lua_CFunction gc_func )
{
#ifdef GLUA_DYNAMIC
	//
    return HOST_CALL_L4( _glua_pushuserdata, L, tag, vp, size, gc_func );
	//
#else
    void* ret;
    
	if ((!vp) && (!size))	// NULL userdata
		{
		lua_pushnil( L );
		return NULL;
		}

	// Size-mode currently supported only if 'gc_func'==NULL.
    // (since we need to have someone cleaning up after our userdata - or do we?!)
	//
	if (size>0) 
		{
		ASSUME( !gc_func );

		gc_func= fn_Glua_GcFree;
		}

    STACK_CHECK(L)
    {
	#ifdef LUA_V40
		{
        void* ptr= (void*)vp;

		if (size>0)
			{
			ptr= malloc( size );
			ASSUME_NONZERO( ptr );
			}

		// Actually create the userdata as a Lua object, associating the 
        // object-specific GC-function with it.
		//
        ret= Loc40_PushUserdata( L, ptr, tag, gc_func );
		}
	#else	// v5.0 static
		{
		// Lua 5.0 userdata model is totally different from 4.0. We try to 
		// 'emulate' some of the earlier 4.0 model to allow transparency of
		// the underlying engine.
        //
        ret= Loc50_PushUserdata( L, vp, size, tag, gc_func );
		}
	#endif
    }
    STACK_END(+1)   // should have pushed one
    
    if (size>0)   // Using Lua-allocated memory block?
        {
        if (vp) memcpy( ret, vp, size );
        else    memset( ret, 0, size );
        }

    return ret;
#endif
}

// backwards compatibility:
void _glua_pushcfunction (lua_State *L, lua_CFunction f )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L2( _glua_pushcclosure, L, f, 0 );
#else
	lua_pushcclosure( L, f, 0 );
#endif
}

void _glua_newtable( lua_State *L )   // was: pushtable
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L( _glua_newtable, L );
#else
	lua_newtable( L );	// Creates an empty table at top of stack
#endif
}

#ifdef GLUA_STATIC
//---
// Modify '[tbl_i]' so that it contains its own subtable, name or
// number of which is in '[key_i]'. Leave stack as it was.
//
static void
Loc_TableDive( lua_State* L, int tbl_i, 
			   const char* key, uint key_len )
{
	// Make our table index absolute so that it is tolerant of 
	// pushing/popping stuff on the stack.
	//
	if (tbl_i<0) tbl_i= lua_gettop(L) +1 +tbl_i;  // -1 => top

    STACK_CHECK(L)		// Check that stack remains balanced
		{	
		// Make a temporary string that's the starting part of key.
		//
		lua_pushlstring( L, key, key_len );  // temporary str at [-1]
		
		Loc_ConvToNumber(L,-1);	// in place type conversion

		// Now, is there such a subtable already?
		//
		lua_pushvalue(L,-1);	// make a duplicate of our index
								// for 'gettable()' to eat (*yum-yum*)
		lua_rawget(L,tbl_i);

		// Stack now:
		//    [tbl_i] is still the table we're diving to.
		//	  [-2] is the index we're trying to find
		//	  [-1] is 'nil' or a valid table entry
		//
		if (!lua_istable(L,-1))	
			{
			// Subtable wasn't there (or it was some basic type?!?)
			// Create an empty subtable & attach it to the mother.
			//
			lua_remove(L,-1);	// remove 'nil' or whatever it was

			lua_newtable(L);	// two copies of the same empty table
			lua_pushvalue(L,-1);
			lua_insert(L,-3);

			// Stack now:
			//	[tbl_i] still the table we're diving to.
			//	[-3] is a 2nd (backup) reference to our created table
			//  [-2] is the index name / number
			//	[-1] is the empty table we created

			lua_rawset(L,tbl_i);	// eats up [-1] and [-2] implicitly,
									// sets them as a new subtable.
			// Stack now:
			//  [tbl_i] still there...
			//  [-1] is the backup ref of the created (empty) table
			}
		else
			{
			// We got the subtable already, can remove the index entry.
			//
			lua_remove( L, -2 );

			// Stack now:
			//  [tbl_i] still there...
			//  [-1] is the subtable item we fetched
			}

		// Replace caller's table entry with the subtable we found
		// or created.
		//
		lua_insert( L, tbl_i );	   // Moves [-1] into the place of ['tbl_i']
		lua_remove( L, tbl_i+1 );  // Remove the shifted old table	
		}
	STACK_END(0)	  // Gives runtime error if stack out of balance
}
#endif  // GLUA_STATIC

void _glua_totable( lua_State *L )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L( _glua_totable, L );
#else

    STACK_CHECK(L)
    {
    // [-3] is the table to add to.
    // [-2] is the key name (string or integer)
    // [-1] is the value to add (may be string, numeric, whatever...)
    //
    ASSUME( lua_istable( L, -3 ) );     // table to add to

	// Check if we have a "subkey1.subkey2.key3" dot-notation for
	// subtable access (that needs to be handled separately).
	//
	// Note: 'lua_isstring()' can NOT be used for the type checking
	//       since it reports numbers (which are always convertible 
	//		 into strings) as strings.
	//
    if ( (lua_type(L,-2)!=LUA_TSTRING) || !strchr(lua_tostring(L,-2 ),'.') )
        {
		// Just the simple case: implicitly uses [-1] and [-2] and
		// removes them from the stack.
		//
	    lua_rawset( L, -3 );
        }
	else
		{
		//-----
		// Dot notation.
		//
		// - Make a safety copy of [-3] (the main table) as [-4].
		// - Recursively loop, keeping [-3] as the subtable and [-2] as the
		//	 remainder of the key path, until no dots exist in it.
		// - Finally, set the value & remove [-3] subtable entry.
		//
		// Note: Any only-numeric parts of the key string must be handled as
		//		 integers, not as strings.
		//
		lua_pushvalue( L, -3 );		// changes others temporarily to [-2]..[-4]
		lua_insert( L, -4 );		// now orig table is copied into [-4]

		do {
			const char* key;
			const char* dot;

			key= lua_tostring(L,-2);
			dot= strchr(key,'.');

			if (!dot) break;	// wow, we got there...

			// Go one level down (modifies -3 to be the subtable)
			//
			Loc_TableDive( L, -3, key, dot-key );

			// Modify key name to be one level deeper.
			//
			lua_remove( L, -2 );	// old key string away
			lua_pushstring( L, dot+1 );
			lua_insert( L, -2 );	// new one in place
			}
		while(1);

		Loc_ConvToNumber(L,-2);  // make numerical string actually a number

		lua_rawset( L, -3 );	// set at subtable, eats away [-1] and [-2]

		lua_remove( L,-1 );		// remove the subtable item we used
								// (leaves original main table at [-1])
		}
	}
	STACK_END(-2)	// should have consumed [-1] and [-2]
#endif
}

int _glua_gettop( lua_State *L )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_L( _glua_gettop, L );
#else
	return lua_gettop(L);   // Gets the current stack size
	                        // (= absolute index of [-1] top-of-stack entry)
#endif
}

void _glua_settop( lua_State *L, int n )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_settop, L, n );
#else
	lua_settop(L,n);    
#endif
}

void _glua_setglobal( lua_State *L, const char* s )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_setglobal, L, s );
#else
	lua_setglobal(L,s);     // Pops [-1] and stores it to variable named 's'.
#endif
}

//---
// Checks that there is space for 'extra' more items on the stack.
//
bool_int _glua_checkstack( lua_State* L, int extra )
{
#ifdef GLUA_DYNAMIC
	//
    return HOST_CALL_L1( _glua_checkstack, L, extra );
    //
#elif (defined LUA_V40)
	//
	return (lua_stackspace(L) >= extra);	  
	//
#else	// 5.0 static
	//
	return lua_checkstack( L, extra );
	//
#endif
}

//---
// Get something from Lua registry table:
//		[-1]= key
// Pushes registry value associated with key (if any) to the stack
//
void _glua_getreg( lua_State* L )
{
#ifdef GLUA_DYNAMIC
	//
    HOST_CALL_L( _glua_getreg, L );
    //
#else
	{
	STACK_CHECK(L)
		{
		#ifdef LUA_V40
			{
			// Lua 4.0 doesn't have the 'LUA_REGISTRYINDEX' pseudo-index.

			lua_getregistry(L);	  // pushes ref to registry table to [-1]
			lua_insert(L,-2);     // move it to [-2]

			lua_rawget(L,-2);	  // eats key, pushes value

			lua_remove(L,-2);	  // remove the registry ref
			}
		#else	// Lua 5.0 static
			{
			lua_rawget( L, LUA_REGISTRYINDEX );	  // eats key, pushes value
			}
		#endif
		}
	STACK_END(0)
	}
#endif
}

//---
// Store something in Lua registry table:
//	    [-2]= key
//		[-1]= value
//
void _glua_setreg( lua_State* L )
{
#ifdef GLUA_DYNAMIC
	//
    HOST_CALL_L( _glua_setreg, L );
    //
#else
	{
	STACK_CHECK(L)
		{
		#ifdef LUA_V40
			{
			// Lua 4.0 doesn't have the 'LUA_REGISTRYINDEX' pseudo-index.

			lua_getregistry(L);	  // pushes ref to registry table to [-1]
			lua_insert(L,-3);     // move it to [-3]

			lua_rawset(L,-3);	  // eats key + value

			lua_remove(L,-1);	  // remove the registry ref
			}
		#else	// Lua 5.0 static
			{
			lua_rawset( L, LUA_REGISTRYINDEX );	  // eats key + value
			}
		#endif
		}
	STACK_END(-2)
	}
#endif
}

//---
int _glua_newtag( lua_State* L )
{
#ifdef GLUA_DYNAMIC
	//
    return HOST_CALL_L( _glua_newtag, L );
	//
#elif (defined LUA_V40)
    {
	// In Lua 4.0, the engine itself provides unique tag numbers:
	//
	int tag= lua_newtag(L);

	ASSUME( tag != 0 );		// We expect zero never to be a valid tag.

	// All tags should have 'GcHub' as their garbage collector.
	//
	STACK_CHECK(L)
		{
		lua_pushcfunction( L, fn_Glua_GcHub );
		lua_settagmethod( L, tag, "gc" );  // pops new method, pushes old

		//lua_remove( L, -1 );
		}
	STACK_END(0)

	return tag;
	}
#else	// 5.0 static
	{
	// In Lua 5.0 there are no more 'tags', so we emulate some.
	//
	// NOTE: Tag creation should be done BY A SINGLE THREAD at program
	//	     startup (you do want same tags for all your threads, right?!).
	//		 We rely on this, and skip multithreading protection for now.
	//
	static int next_tag= 1;	// 0 = not a valid tag
	
	if (next_tag==1)	// first time call
		{
		// Make sure the '0' metatable contains 'GcHub' (but nothing more).
		// This is used for tags that don't have a dedicated methods table.
		//
		lua_pushcfunction( L, fn_Glua_GcHub );
	    Loc50_SetTagMethod( L, 0 /*tag*/, "gc", FALSE /*don't push old*/ );
		}

	return next_tag++;	// _NOT_ multithreading safe!
	}
#endif
}

//---
void _glua_insert( lua_State *L, int argn )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_insert, L, argn );
#else
	lua_insert(L,argn);
#endif
}

//---
void _glua_pushcclosure( lua_State *L, lua_CFunction fn, uint n )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L2( _glua_pushcclosure, L, fn, n );
#else
	lua_pushcclosure(L,fn,n);
#endif
}

//---
// Skip these functions unless compiling a host (static linkage) or
// GLUAPORT is explicitly #defined for the application project.
//
#if ((defined GLUA_STATIC) || (defined GLUAPORT))
//
uint _gluaport_tagofuserdata( lua_State* L, int argn )
{
#ifdef GLUA_DYNAMIC
	//
    return HOST_CALL_L1( _gluaport_tagofuserdata, L, argn );
	//
#elif (defined LUA_V40)
    {
	return lua_tag( L, argn );
	}
#else	// 5.0 static
	{
	return Loc50_Tag( L, argn );
	}
#endif
}
//
//---
int _gluaport_strlen( lua_State *L, int argn )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_L1( _gluaport_strlen, L, argn );
#else
	return lua_strlen(L,argn);
#endif
}
//
//---
void _gluaport_rawset( lua_State *L, int argn )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _gluaport_rawset, L, argn );
#else
	lua_rawset(L,argn);
#endif
}
//
//---
void _gluaport_rawseti( lua_State *L, int argn, int n )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L2( _gluaport_rawseti, L, argn, n );
#else
	lua_rawseti(L,argn,n);
#endif
}
//
//---
void _gluaport_call( lua_State *L, int nargs, int nresults )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L2( _gluaport_call, L, nargs, nresults );
#else
	lua_call(L,nargs,nresults);
#endif
}
//
//---
void _gluaport_settable( lua_State *L, int argn )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _gluaport_settable, L, argn );
#else
	lua_settable(L,argn);
#endif
}
//
//---
void _gluaport_settag( lua_State *L, int argn, int tag )
{
#ifdef GLUA_DYNAMIC
    //
    HOST_CALL_L2( _gluaport_settag, L, argn, tag );
    //
#elif (defined LUA_V40)
    {
	lua_settag(L,tag);
	}
#else   // 5.0 static
    {
    Loc50_SetTag( L, argn, tag );
    }
#endif
}
//
//---
// Note: We need a different 'settagmethod' for gluaport support.
//       The Lua native behaviour is to pop new function, push the old one.
//
void _gluaport_settagmethod( lua_State *L, int tag, const char* event )
{
#ifdef GLUA_DYNAMIC
    //
    HOST_CALL_L2( _gluaport_settagmethod, L, tag, event );
    //
#else
	event= EVENT_NAMECHANGE( L, event );

    STACK_CHECK(L)
        {
        ASSUME( lua_isfunction(L,-1) || lua_isnil(L,-1) );

        #if (defined LUA_V40)
        	lua_settagmethod(L,tag,event);
        #else   // 5.0 static
	        Loc50_SetTagMethod( L, tag, event, TRUE /*yes, push old*/ );
        #endif
        }
    STACK_END(0)    // popped one, pushed one
#endif
}
#endif  // GLUAPORT
//
//---( end of GLUAPORT section )---


/***
lua_CFunction _gluaport_gettagmethod( lua_State* L, 
									  int tag, const char* event )
{
#ifdef GLUA_DYNAMIC
	//
    return HOST_CALL_L2( _gluaport_gettagmethod, L, tag, event );
	//
#elif (defined LUA_V40)
    {
	lua_CFunction func;

	event= EVENT_NAMECHANGE( event );

	// Lua 4.0 still has tag methods:
	//
    STACK_CHECK(L)
        {
		lua_gettagmethod( L, tag, event );	// pushes 'nil' or func
		
		func= lua_tocfunction( L, -1 );

		lua_pop( L, -1 );	// remove pushed entry 
		}
    STACK_END(0)

	return func;
	}
#else	// 5.0 static
	{
	event= EVENT_NAMECHANGE( event );

    return Loc50_GetTagMethod( L, tag, event );
	}
#endif
}
***/

//---
void _glua_settagmethod( lua_State* L, int tag, const char* event,
								       lua_CFunction func )
{
#ifdef GLUA_DYNAMIC
	//
    HOST_CALL_L3( _glua_settagmethod, L, tag, event, func );
	//
#elif (defined LUA_V40)
    {
	event= EVENT_NAMECHANGE( L, event );

	// Lua 4.0 still has tag methods:
	//
    STACK_CHECK(L)
        {
		lua_pushcfunction( L, func );
                        
	    lua_settagmethod( L, tag, event );  // pops new method, pushes old

		// Lua 4.0 docs say 'settagmethod()' would push the old function
		// but it seems it does not?!
		//
        //lua_remove( L, -1 );
        }
    STACK_END(0)
	}
#else	// 5.0 static
	{
	event= EVENT_NAMECHANGE( L, event );

    STACK_CHECK(L)
        {
        lua_pushcfunction( L, func );
        Loc50_SetTagMethod( L, tag, event, FALSE /*don't push old func*/ );

	    // Make sure every dedicated method table also has 'GcHub' as its 
	    // garbage collector:
	    //
	    if (strcmp( event, "gc" ) != 0)
		    {
	        lua_pushcfunction( L, _glua_gc_hub );
		    Loc50_SetTagMethod( L, tag, "gc", FALSE );
		    }
        }
    STACK_END(0)
	}
#endif
}


//-----
void _glua_fromtable( lua_State *L, int argn )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_fromtable, L, argn );
#else
    // 'argn': table to get from:
    //
    ASSUME( lua_istable( L, argn ) );     // table to get from

    // stack '-1' is the key name (string or numeric index):
    //
    ASSUME( lua_isstring(L,-1) || lua_isnumber(L,-1) );

    lua_gettable( L, argn );    // pushes tbl entry or 'nil'
#endif
}


bool_int _glua_next( lua_State *L, int argn )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_L1( _glua_next, L, argn );
#else
    ASSUME( lua_istable( L, argn ) );     // table to get from

	// Key name, index or 'nil' (for first round):
	//
    ASSUME( lua_isstring(L,-1) || lua_isnumber(L,-1) || lua_isnil(L,-1) );

    // 'lua_next()' pops a key from stack and pushes the key,value pair
    // following it in the table. key='nil' starts the iteration.
    //    
    return lua_next( L, argn );   // pops one, pushes two
    //
    // FALSE: end of iteration (nothing pushed to stack)
    // TRUE:  valid entry (stack[-2]: next key, stack[-1]: next value)
#endif
}

void _glua_remove( lua_State *L, int argn )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_remove, L, argn );
#else
    lua_remove( L, argn );      // removes '[argn]' from stack (& shifts others)
#endif
}

void _glua_pushvalue( lua_State *L, int argn )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_pushvalue, L, argn );
#else
    lua_pushvalue( L, argn );   // pushes a copy of '[argn]' to stack
#endif
}

const char* _glua_tolstring_safe( lua_State* L, int argn, size_t* len_ref )
{
const char* ptr;

	// To bypass the int->str conversion side-effect, we make a local
	// copy of the stack entry, get a pointer to it, and remove it.
	// Even though the item is removed from stack, our pointer remains
	// valid until Lua does Garbage Collection (which is only after
	// GLUA_FUNC_END so this is safe :).

	_glua_pushvalue( L, argn );		// Makes [-1] a local copy
		{
		ptr= _glua_tolstring_raw(L,-1,len_ref);	// may change the local copy
		}
	_glua_remove( L, -1 );			// Removes the copy from stack

	return ptr;		// Should remain valid until GLUA_FUNC_END.
}

double _glua_tonumber_safe( lua_State* L, int argn )
{
#ifdef LUA_V40	// Lua 4.0 converts numeric strings to actually
	{			// become numbers, if 'lua_tonumber()' is used.
	double val;

	_glua_pushvalue( L, argn );		// Makes [-1] a local copy
		{
		val= _glua_tonumber_raw(L,-1);	// may convert str->num.
		}
	_glua_remove( L, -1 );			// Removes the copy from stack

	return val;
	}
#else	// 5.0 static
	{
	// Lua 5.0 'lua_tonumber()' no longer modifies the entries:
	// 'raw' conversion can be used always!
	//
	return _glua_tonumber_raw(L,argn);
	}
#endif
}


//---
// 'glua_exec()' is used by multithreading gluax extensions, creating 
// a Lua engine, running certain commands on it and cleaning away the 
// engine automatically. 
//
// Calling this function may take considerable time and is intended
// to be called from new threads running parallel with the main thread.
//
// Note: This function is somewhat experimental, and may be replaced by
//       a 5.0 -only function that is able to duplicate an existing Lua stack.
//
int gluahost_open (lua_State* lua);
int gluahost_close (lua_State* lua);

int _glua_exec( const char* cmd_str, int stack_size )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_2( _glua_exec, cmd_str, stack_size );
#else
	{
	lua_State* L2;
	int ret;

	#ifdef LUA_V40
		{
		L2= lua_open( stack_size );    // Creates completely new Lua instance
		}
	#else
		{
		IGNORE_PARAM( stack_size );
		L2= lua_open( );	// (no param in Lua 5.0)
		}
	#endif

	if (!L2)
		return -999;

	// Initialize static libraries for the new thread:
	// (pretty helpless without them)
	//
	lua_baselibopen(L2);
	lua_iolibopen(L2);
	lua_strlibopen(L2);
	lua_mathlibopen(L2);
	lua_dblibopen(L2);

	// Gluahost() needs to be available also for the child threads, otherwise
	// they wouldn't be able to use any extension modules.
	//
	gluahost_open(L2);
		{
		ret= lua_dostring( L2, cmd_str );
		}
	gluahost_close(L2);

    lua_close( L2 );

    return ret;
	}
#endif
}


int _glua_getn( lua_State* L, int index )
{
#ifdef GLUA_DYNAMIC
    return HOST_CALL_L1( _glua_getn, L, index );
#else
	{
	#ifdef LUA_V40
	    return lua_getn( L, index );    // 4.0
	#elif (defined LUA50_BETA)
        return Loc50_GetN( L, index );  // 5.0 beta
	#else
	    return luaL_getn( L, index );    // 5.0 final
	#endif
	}
#endif
}


//---
void _glua_concat( lua_State* L, int items )
{
#ifdef GLUA_DYNAMIC
    HOST_CALL_L1( _glua_concat, L, items );
#else
    lua_concat( L, items );
#endif
}


//---
void* _glua_attach( lua_State* L, const void* obj_ptr, void* new_val )
{
void* old_val;

	STACK_CHECK(L) 
	{
	// Store the object key as string (could also be stored as plain userdata)
	//
    Loc_PushMerged_int( L, "glua_attach_", (uint)obj_ptr );

	// Get the current value (if any)
	//
	_glua_getreg( L );

	old_val= _glua_touserdata(L,-1);	// maybe NULL

	_glua_remove( L, -1 );

	// Set new value:
	//
	if ( new_val != (void*)(-1) )	// -1 = don't change
		{
        Loc_PushMerged_int( L, "glua_attach_", (uint)obj_ptr );

		_glua_pushuserdata( L, 0, new_val, 0, NULL );
		//
		// Stack now has key at [-2], value at [-1].

		_glua_setreg( L );	// eats up key&value
		}
	}
	STACK_END(0)

	return old_val;
}


/*--- Garbage collection support ------------------------------------------*/

//---
// Garbage Collection hub, called by Lua engine.
//
#ifdef GLUA_STATIC
//
GLUA_FUNC( Glua_GcHub )     // void = Glua_GcHub( userdata_obj )
{
void* obj= glua_getUserdata(1,0);
lua_State* L= _lua_;

    if (!obj) break;    // jumps to 'GLUA_END'

	STACK_CHECK(L) 
	{
    void* v_ptr;

	// Find associated garbage-collection func via Lua registry:
	//
    // Note: Normally we cannot use direct 'lua_...()' functions within
    //       'GLUA_FUNC()' but here we're always linking statically, so 
    //       that is okay.
    //
	lua_pushvalue(L,-1);	// temp copy of userdata
	_glua_getreg( L );		// eats up temp copy, pushes value or 'nil'

	v_ptr= lua_touserdata( L, -1 );
	lua_remove(L,-1);

	if (v_ptr)
		{
		// Call application's GC func:
		//
        // Object item is still at top-of-stack (-1) as it was when this
        // function was called.
        //
    	(* ((lua_CFunction)v_ptr) )( L );

	    // Remove the registry entry that's no longer needed:
		//
		lua_pushvalue(L,-1);	// temp copy of userdata (key)
		lua_pushnil(L);			// (value)
		_glua_setreg(L);		// (pops key & value)
		}
	}
	STACK_END(0)	// stack should be untouched
}
GLUA_END
//
// 'Glua_GcHub()' is static - we make a bridge for 'gluax_50.c' etc.
//
int _glua_gc_hub( lua_State* L )  
{ 
	return fn_Glua_GcHub( L );
}
//
#endif	// !GLUA_DYNAMIC


//---
// Simple garbage collection for memory blocks allocated by us:
//
#ifdef GLUA_STATIC
//
GLUA_FUNC( Glua_GcFree )
{
void* p= glua_getUserdata(1,0);

	if (p) free(p);
}
GLUA_END
//
#endif	// !GLUA_DYNAMIC


/*--- Error and tag handling ---------------------------------------------*/

//-----
static int _glua_badparam( lua_State *lua, int argn, char which, int opt_tag )
{
char errbuf[999];   // protected (see below)
const char* msg;
const char* type;
const char* tagname= NULL;
const char* temp_str;

    argn= STACK_ABS(lua,argn);  // just in case (argn may be -1)

	// clean interface stack to prevent worst-case overflow
	//
	_glua_settop( lua, 0 );
  
	if (which == 'U')	// userdata
		{
		STACK_CHECK(lua)
			{
			// read tagname from Lua registry
			//
			Loc_PushMerged_int( lua, "glua_tag_", opt_tag );
			_glua_getreg( lua );
  
			tagname = _glua_tostring_raw( lua, -1 );

			_glua_remove( lua, -1 );
			}
		STACK_END(0)

		if (!tagname)
			tagname= "???";
		}

	msg = "arg #%d ('%s') not a valid %s";
	type= "";

	switch (which)
		{
		case 'N':	type= "number";		break;
		case 'I':	type= "integer";	break;
		case 'S':	type= "string";		break;
		case 'U':	type= "userdata: "; break;
		case 'B':	type= "boolean";	break;
		case 'T':   type= "table";		break;

		case '\0':	// none
			msg = "too many params (arg #%d: '%s')%s";
			break;

		default:
			msg = "problem with arg #%d ('%s')%s";
			break;
		}

    temp_str= _glua_tostring_safe( lua, argn );
    if (!temp_str) temp_str="";
						 
    ASSUME( sizeof(errbuf) > strlen(msg) /*+ strlen(func_name)*/ + 10 /*argn*/
                             + strlen(temp_str) + strlen(type) );
    
	sprintf(errbuf, msg, /*func_name,*/ argn, temp_str, type );
	if (tagname)
		strcat_safe( errbuf, tagname, sizeof(errbuf) );
     
	_glua_error( lua, errbuf );

	return NEVER;
}


int _glua_regtag(lua_State *lua, const char *tagname)
{
int tag= 0;

    ASSUME_NONZERO( tagname );

    // Note: Not multithreading safe!!  (should be atomic)

    // AK(3-Jul-03): Check if tagname already exists and reuse its tag value.
    //               This allows sharing tag types between modules.
    //
    STACK_CHECK(lua)
        {
        Loc_PushMerged_str( lua, "glua_tag_", tagname );  // key
        _glua_getreg( lua );    // eats key, pushes 'nil' or tag value (!=0)

        tag= (int)_glua_tonumber_raw( lua, -1 );    // 0 if 'nil'
        _glua_remove( lua, -1 );
        }
    STACK_END(0)

    if (!tag)   // Unique tagname (create new)
        {
        tag= _glua_newtag( lua );  // ask Lua for a new tag number

        STACK_CHECK(lua)
            {
            // write key-tagname into Lua registry 
	        //
            Loc_PushMerged_int( lua, "glua_tag_", tag );	// key
    		_glua_pushstring( lua, tagname );			// value
            _glua_setreg( lua );	// eats key & value

            // Place also tagname->tag mapping (for duplicate checks):
	        //
            Loc_PushMerged_str( lua, "glua_tag_", tagname );
    		_glua_pushnumber( lua, tag );
            _glua_setreg( lua );    // eats key & value
            }
        STACK_END(0)   // Should be balanced
        }

    return tag;
}


/*--- Table handling ---------------------------------------------*/

//-----
// Given a table object in the stack ('argn') and a key string with optional
// subkeys (s.a. "obj.color.red"), returns the entry beneath that reference.
//
// Stack: Pushes 1 item to stack (the table entry or 'nil' for no such item)
//
// Params: If 'index' is given (nonzero), it is taken to be a numeric first-
//         level name (same as providing "1.more.keys" in the string parameter
//         (but this should be more efficient).
//
static bool_int // TRUE:an entry was pushed to [-1]
                // FALSE: entry already there
                //
Loc_TableEntry( lua_State* L, int argn, 
				 int index,	  // 0 = not used
                 const char* key_with_subkeys )  // e.g. "obj.color.red"
{                     
char key_buf[99];   // key without subkeys
const char* key= NULL;
const char* dot= NULL;
uint len;

	if (index==0)	// skip if first level numeric
		{
		if (key_with_subkeys == NULL)	// Iterating through main table?
			{
			return FALSE;	// Value already at 'stack[-1]' (no remove required)
			}

		// Key given as a parameter - may contain 'subtable diving':
		//
		dot= strchr( key_with_subkeys, '.' );    // NULL if no subkeys
        
		if (!dot)
			key= key_with_subkeys;
		else
			{
			len= (dot - key_with_subkeys);  // 1st level key length

			if (len >= sizeof(key_buf))   // exceeding buffer limit?!
				{
				_glua_pushnil(L);
				return TRUE;
				}
        
			key= strcpy_safe( key_buf, key_with_subkeys, len+1 );
			}
		//
		// 'key' now has the name of this level into the stack

		// If 'key' is totally numeric, use integer index instead of string
		// ( In Lua, 'tbl[1]' and 'tbl['1']' are different entries! )
		//
		// ( Note: With this method, we are unable to access 'tbl['1']' )
		//
		if (Loc_IsNumeric( key ))
			index= atoi(key);
		}
    
    STACK_CHECK(L)
    {
    if (argn==0)
        _glua_pushnil(L);   // no table, so no contents either.
    else
        {
	    if (index != 0)
		    _glua_pushnumber( L, (double)index );
	    else
		    _glua_pushstring( L, key );

        _glua_fromtable( L, argn );  // pops key, pushes entry back (or 'nil')
        }

	if (dot)	// Diving deeper...
        {
		// Subtable reference should now be at [-1].
		//
		if (_glua_type(L,-1) != GLUA_TYPE_TABLE)
			{
			_glua_pushnil(L);    // Bad key name!
			}
		else	// Recurse into the subtable
			{
			// Note: Don't use '-1' notation since that changes when things
			//       are pushed onto the stack. 'gettop()' gives us same entry
			//		 with absolute (positive) index.
			//
			Loc_TableEntry( L, _glua_gettop(L), // STACK_ABS(-1)
							   0,	   // no numeric part (unless in string)
			                   dot+1 );	// Pushes entry or 'nil'

			_glua_remove( L, -2 );	// Subtable entry no longer needed
			}
        }    
    }
    STACK_END(+1)     // Should have pushed one (and only one) onto the stack

	return TRUE;	// Value pushed to [-1] - needs to be removed.
}    

 
//---                       
bool_int _glua_gettablekey( lua_State* L,
                            uint argn,  // 0 = no table
                            const char** key_ref, 
                            enum e_Glua_Type* type_ref )
{
bool_int found;

    if (!argn) return 0;    // 'nil' param - see as an empty table
    
	if (!key_ref) return 0;	// Must have it!

	if( !*key_ref )    // First round, no previous key on stack
	    {
		_glua_pushstring( L, NULL );    // pushes 'nil'
		}
    else
        {
		// stack[-1]= last value (may be removed)
        // stack[-2]= last key (it was left in the stack by us)

		_glua_remove( L, -1 );	// last key becomes [-1]
        }

	STACK_CHECK(L)
	{
	found= _glua_next( L, argn );
	//
	if (!found)
		{
		// stack[-1] popped by the Lua engine
		}
	else
	    {
	    // stack[-1]= next value
	    // stack[-2]= next key
	    
        if (type_ref)
			*type_ref= (enum e_Glua_Type)_glua_type(L,-1);

		// We cannot do a regular 'tostring()' on [-2], since that would
		// (as a side-effect) change an integer index into string, which
		// would make future 'next()' call fail! So, we'll make a local
		// copy and get the conversion via it.
		//
		*key_ref= _glua_tostring_safe( L, -2 );

		// Leave the stack value on the stack. 'getTable_...()' functions
		// with 'NULL' key name will be able to fetch it directly.
		//
        // Also key value is left into the stack at [-2].
        }
	}
	STACK_END( found ? +1 : -1 )
    
	return found;   // 0 = no more entries
}                                    

//---
static void Loc_PushMerged_ijk( lua_State* L, int i, int j, const char* key )
{
uint n=1;

    ASSUME( i!=0 );    // "i.j", "i.k" or "i.j.k"

    STACK_CHECK(L)
        {
        _glua_pushnumber( L, i );

        if (j)
            { 
            _glua_pushstring( L, "." );
            _glua_pushnumber( L, j );
            n += 2;
            }

        if (key)
            {
            _glua_pushstring( L, "." );
            _glua_pushstring( L, key );
            n += 2;
            }

        _glua_concat( L, n );   // Leaves one string on the stack
        }
    STACK_END(+1)
}

//---
// [-2]= table to attach to
// [-1]= value (int/string/userdata/table/whatever..)
//
void _glua_tblkey_i( lua_State* L, int i )
{
    ASSUME( _glua_type(L,-2) == GLUA_TYPE_TABLE );

    STACK_CHECK(L)
        {
        _glua_pushnumber(L,i);
        _glua_insert( L, -2 );  // swap [-1] and [-2]
        _glua_totable( L );
        }
    STACK_END(-1)  // should have popped the value off stack
    //
    // [-1]= table ref
}

void _glua_tblkey_k( lua_State* L, const char* key )
{
    ASSUME( _glua_type(L,-2) == GLUA_TYPE_TABLE );

    STACK_CHECK(L)
        {
        _glua_pushstring(L,key);
        _glua_insert( L, -2 );  // swap [-1] and [-2]
        _glua_totable( L );
        }
    STACK_END(-1)  // should have popped the value off stack
    //
    // [-1]= table ref
}

void _glua_tblkey_ijk( lua_State* L, int i, int j, const char* key )
{
    ASSUME( _glua_type(L,-2) == GLUA_TYPE_TABLE );

    STACK_CHECK(L)
        {
        Loc_PushMerged_ijk( L, i, j, key );    // pushes one

        _glua_insert( L, -2 );  // swap [-1] and [-2]
        _glua_totable( L );
        }
    STACK_END(-1)  // should have popped the value off stack
    //
    // [-1]= table ref
}

//---
// Reading table contents is a common runtime thing, so we've splitted it into
// several specialized functions for speed optimization (& code clarity).
//

// String key:
//
// Note: 'key' may be NULL for member browsing (item already pushed on stack)
//       or ".subtable.path" (starting with a dot) for going into subtables of
//       such an item.
//
void _glua_usekey_k( lua_State* L, int tbl_argn, const char* key )
{
#ifdef COMPILER_MSC
    ASSUME( tbl_argn==0 );
    tbl_argn= _getT();
#endif
    
    ASSUME( (tbl_argn==0) || (_glua_type(L,tbl_argn) == GLUA_TYPE_TABLE) );

    STACK_CHECK(L)
        {
        if ( key == NULL )  // Item already at top-of-stack (doing item browsing)
            {
            // just make a copy so that the original remains safe.
            //
            _glua_pushvalue( L, -1 );   // Done!
            }
        else
            {
            if ( key[0] == '.' )    // Subtable of the top-of-stack item.
                {
                // Use the 'current item' as basis, not the whole big table.
                //
                tbl_argn= _glua_gettop(L);  // abs. index of [-1]

                ASSUME( _glua_type(L,tbl_argn) == GLUA_TYPE_TABLE );

                key++;  // bypass the dot
                }

            // Now push an entry of the table onto the stack.
            //
            Loc_TableEntry( L, tbl_argn, 0, key );  // pushes one
            }
        }
    STACK_END(+1)
    //
    // [-1]= table value  (to be consumed by '_glua_getXxx(0)')
}

// 1-D Arrays ('tbl[n]'):
//
void _glua_usekey_i( lua_State* L, int tbl_argn, int i )
{
#ifdef COMPILER_MSC
    ASSUME( tbl_argn==0 );
    tbl_argn= _getT();
#endif

    ASSUME( (tbl_argn==0) || (_glua_type(L,tbl_argn) == GLUA_TYPE_TABLE) );

    STACK_CHECK(L)
        {
        Loc_TableEntry( L, tbl_argn, i, NULL );  // pushes one
        }
    STACK_END(+1)
}

// Other cases (need merging of keys):
//
void _glua_usekey_ijk( lua_State* L, int tbl_argn, int i, int j, const char* key )
{
#ifdef COMPILER_MSC
    ASSUME( tbl_argn==0 );
    tbl_argn= _getT();
#endif

    ASSUME( (tbl_argn==0) || (_glua_type(L,tbl_argn) == GLUA_TYPE_TABLE) );

    STACK_CHECK(L)
        {
        if (j)  // Use merging only if necessary
            {
            Loc_PushMerged_ijk( L, j,0,key );  // "j.key"

            key= _glua_tostring_raw( L, -1 );
            _glua_settop( L, -2 );  // pop Lua entry but C pointer remains valid
            }

        Loc_TableEntry( L, tbl_argn, i, key );  // always pushes one
        }
    STACK_END(+1)
    //
    // [-1]= table value  (to be consumed by '_glua_getXxx(0)')
}

//---
double _glua_getNumber(lua_State* L, int argn_)
{
int argn= argn_ ? argn_ : -1;
double ret;

    // General parameter:
    //
    if ( !(_glua_type(L,argn) & _GLUA_MASK_FLOAT_INT_NIL) )
        _glua_badparam(L, argn, 'N',0 );

    ret= _glua_tonumber_raw(L, argn);

    if (!argn_) _glua_settop(L,-2);   // pop away
    return ret;
}
   
int _glua_getInteger(lua_State* L, int argn_)
{
int argn= argn_ ? argn_ : -1;
int ret;

    // General parameter:
    //
    if ( !(_glua_type(L,argn) & _GLUA_MASK_INT_NIL) )
        _glua_badparam(L, argn, 'I',0 );

    ret= (int)_glua_tonumber_raw(L, argn);

    if (!argn_) _glua_settop(L,-2);   // pop
    return ret;
}

//-----
// Note: It is important to have a separate 'getUnsigned()' function for
//       big numbers. Otherwise i.e. Gcc implicit double->int conversion
//       converts too big positive (that would still fit into uint) into
//       0x80000000. To read in the whole 0..2^32-1 range, we need this.
//
unsigned long _glua_getUnsigned(lua_State* L, int argn_)
{
int argn= argn_ ? argn_ : -1;
unsigned long ret;  // normally 32 bits

    // General parameter:
    //
    if ( !(_glua_type(L,argn) & _GLUA_MASK_INT_NIL) )
        _glua_badparam(L, argn, 'I',0 );

    ret= (unsigned long)_glua_tonumber_raw(L, argn);

    if (!argn_) _glua_settop(L,-2);   // pop
    return ret;
}

/**
#ifdef _HAS_LONGLONG
//
// Note: By default, Lua stores all numbers internally as 'double's, so it is 
//       not sure the whole 64-bit integer range can be presented with them.
//
//       Consider compiling Lua with 'long double' as the default number type,
//       or rather, to implement 64-bit integers as 'userdata' if you need
//       them a lot.
//
unsigned long long _glua_getUnsignedLL(lua_State* L, int argn_)
{
int argn= argn_ ? argn_ : -1;
unsigned long long ret;  // normally 64 bits

    // General parameter:
    //
    if ( !(_glua_type(L,argn) & _GLUA_MASK_INT_NIL) )
        _glua_badparam(L, argn, 'I',0 );

    ret= (unsigned long long)_glua_tonumber_raw(L, argn);

    if (!argn_) _glua_settop(L,-2);   // pop
    return ret;
}
#endif  // _HAS_LONGLONG
**/

const char* _glua_getLString(lua_State* L, int argn_, size_t* len_ref )
{
int argn= argn_ ? argn_ : -1;
const char* ret;

    if ( !(_glua_type(L,argn) & _GLUA_MASK_STRING_FLOAT_INT_NIL) )
        _glua_badparam(L, argn, 'S',0 );

    ret= _glua_tolstring_raw(L, argn, len_ref);

    if (!argn_) _glua_settop(L,-2); // pop
    return ret;
}

void* _glua_getUserdata(lua_State* L, int argn_, int tag )
{
int argn= argn_ ? argn_ : -1;
void* ret;

    if ( !(_glua_type2(L,argn,tag) & _GLUA_MASK_USERDATA_NIL) )
    	_glua_badparam(L, argn, 'U',tag );

    ret= _glua_touserdata(L, argn);

    if (!argn_) _glua_settop(L,-2); // pop
    return ret;
}

bool_int _glua_getBoolean(lua_State* L, int argn_ )
{
int argn= argn_ ? argn_ : -1;
bool_int ret;

    if (_glua_gettop(L) < argn)
        _glua_badparam(L, argn, 'B',0 );

    ret= _glua_toboolean(L, argn);

    if (!argn_) _glua_settop(L,-2);   // pop away
    return ret;
}

int /*tbl_argn*/ _glua_getTable(lua_State* L, int argn)
{
enum e_Glua_Type type;

    ASSUME( argn != 0 );    // use dot notation to access subtables

    type= (enum e_Glua_Type)_glua_type(L,argn);

    return (type==GLUA_TYPE_TABLE) ? argn :
           (type==GLUA_TYPE_NIL) ? 0 /*empty*/ :
           (int)_glua_badparam(L, argn, 'T',0 );  // (never returns)
}

void _glua_getDone(lua_State* L, int argn)
{
    if (_glua_gettop(L) >= argn)
        _glua_badparam(L, argn, 0,0 ); \
}


/*--- Initialisation ---------------------------------------------*/

//---
// New registration function (>=2.07), using table index instead of 
// string table name.
//
// Params:  'tbl_index' > 0: absolute index of a namespace table (any name)
//                      == 0: global namespace (no table)
//
#define LOC_PUSH_ITEM( lua, item ) \
    {      if (item.func) _glua_pushcfunction( lua, item.func ); \
      else if (item.str)  _glua_pushstring( lua, item.str ); \
      else                _glua_pushnumber( lua, item.num ); \
    }

void _glua_register2( lua_State *lua,
                     struct s_GluaItem list[], 
                     uint list_items,
                     int tbl_index )    // namespace (0 = global)
{
uint n;
bool_int pop_tbl_index= FALSE;

    if (!list_items) return;

    ASSUME( tbl_index >= 0 );   // Must be an absolute index (or 0)

    STACK_CHECK(lua)
    {
    // Make a second reference to the table (at top of stack)
    // just in case the index we got wouldn't be topmost.
    //
    if (tbl_index > 0)
        {
        _glua_pushvalue( lua, tbl_index );

        ASSUME( _glua_type(lua,-1) == GLUA_TYPE_TABLE );
        }

    for( n=0; n<list_items; n++ )
        {
        if (list[n].tag_ref)    // userdata tag or tag method
            {
            if (list[n].name)  // Tag creation
                {
                *(list[n].tag_ref)= _glua_regtag( lua, list[n].name );
                }
            else    // tag methods
                {
                int tag= *(list[n].tag_ref);

                if ( tag == 0 )
                    {
                    _glua_errorN( lua, "GLUA_DECL error: "
                                       "Uninitialized tag for method '%s'!", list[n].str );
                    }

                _glua_settagmethod( lua, tag, list[n].str /*event*/, list[n].func );
                }
            }
        else
            {   // string, function, or number
            if (tbl_index==0)
                {
                // No namespace: just push the func / constant & set as global.
                //
                LOC_PUSH_ITEM( lua, list[n] );

                _glua_setglobal( lua, list[n].name );
                }
            else
                {
                // Using namespace: push key & value onto the stack, then merge
                // with the table.
                //

                _glua_pushstring( lua, list[n].name );   // pushes the key

                LOC_PUSH_ITEM( lua, list[n] );      // pushes the value
            
                _glua_totable( lua );   // merges with table (pops key & val)
                }
            }
        }
        
    if (tbl_index > 0)  // Still need to remove the 2nd ref we pushed.
        {
        _glua_settop( lua, -2 );
        }

    if (pop_tbl_index)
        _glua_remove( lua, tbl_index );
    }
    STACK_END(0)
}


//-----
// This func is called by Gluahost at module initialization (see 'GluaModule()'
// macro in 'glua_x.h' for details). We check that the host is good enough
// for us.
//
int // 0=success, negative = errors
	_glua_linktohost( lua_State* lua, const void* v_funcs, uint funcs_size )
{
#define ERR_NOFUNC   (-1)
#define ERR_OLDHOST  (-2)
#define ERR_ALIGN    (-3)
//...(more?)

const struct s_GluaFuncs* funcs= (const struct s_GluaFuncs*)v_funcs;

#if ( (!defined _HAS_VA_ARGS) && (defined COMPILER_MSC) )
	//
	if (tls_L == _TLS_UNINITIALIZED) tls_L= TlsAlloc();
	if (tls_T == _TLS_UNINITIALIZED) tls_T= TlsAlloc();
	//
#endif

#ifdef GLUA_STATIC
    {
    ASSUME( funcs == NULL );
    return 0;	// ok
    }
#else   // Dynamic linkage    
    {
    const void** ptr_tbl;
	uint i, host_ver;

    ASSUME_NONZERO( funcs );
    
    if (funcs_size < sizeof(struct s_GluaFuncs))
        {
		return ERR_OLDHOST;  // Function table smaller than expected! (old host)
		}

    // Check data type sizes used in compilation of this module (must be done
    // before passing 'funcs' to the host!)
    //
    if (!GLUAITEM_SIZECHECK())
        return ERR_ALIGN;   // Change data alignment to 4 bytes or less!

    // Check that all the function entries we need are non-NULL.
    //        
    ptr_tbl= (const void**)funcs;

    for( i=0; i< (sizeof(struct s_GluaFuncs) / sizeof(void*)); i++ )
        {
        if (ptr_tbl[i] == NULL)
            return ERR_NOFUNC;  // Some entry is NULL! (bug in the host)
        }

	// Remember for 'LUA_CALL()' macros to use:
    //
    _lua_gate_= funcs;     // Gate now open! :)

    // Check Gluahost version (may need host-side bug fixes etc.):
    //
    // NOTE: This can only be done _after_ the '_lua_gate_' table has been set!
    //
    host_ver= _glua_ver( GLUA_VERSION );   // We tell our version
                                           // ...and get host version back
                                                    
    if ( (host_ver & 0xffff) < HOST_VERSION_REQ )
        return ERR_OLDHOST;  // Too old host for us!

	return 0;	// success! :)
	}
#endif    
}


//---
// Help function used by 'gluahost.c'. Not available for dynamic modules.
// 
// Pushes a table reference on the Lua stack (if needed, creating the table)
// and returns the absolute stack index to it.
//
#ifdef GLUA_STATIC
int _glua_create_tbl( lua_State* lua, const char* table_name )
{
const char* ptr= table_name;
const char* dot;

    if (!table_name) return 0;  // global namespace

    STACK_CHECK(lua)
        {
        #ifdef LUA_V40
            lua_getglobals( lua );  // pushes ref to global table
        #else   // 5.0 & later
            lua_pushvalue( lua, LUA_GLOBALSINDEX );   // does this work?
        #endif

        // Each loop goes one stage further, replacing the parent
        // table ref with the child's.
        //
        while( ptr )
            {
            dot= strchr( ptr, '.' );    // end of current stage

            if (!dot)
                {
                lua_pushstring( lua, ptr ); // last stage
                ptr= NULL;
                }
            else
                {
                lua_pushlstring( lua, ptr, dot-ptr );
                ptr= dot+1;
                }

            lua_pushvalue( lua, -1 );   // 2nd ref to stage name

            // [-1]: stage name
            // [-2]: 2nd ref to stage name
            // [-3]: ref to parent

            lua_gettable( lua, -3 );    // pops key, adds value

            if (lua_istable( lua, -1 ))
                {
                // Subtable existed, proceed to next stage
                //
                lua_remove( lua, -2 );  // stage name
                lua_remove( lua, -2 );  // parent table
                }
            else
                {
                // Create a new table & tie it.
                //
                lua_remove(lua,-1);     // 'nil' away

                lua_newtable(lua);
                //
                // [-1]: empty table
                // [-2]: stage name
                // [-3]: ref to parent

                lua_pushvalue( lua, -1 );   // 2nd ref to new table
                lua_insert( lua, -4 );

                lua_settable( lua, -3 );        
                lua_remove( lua, -1 );      // parent away
                }

            ASSUME( lua_istable(lua,-1) );  // new stage
            }
        }
    STACK_END(+1)   // should have pushed one

    return lua_gettop(lua);   // top of stack
}
#endif  // GLUA_STATIC
