/*---------------------
// 
// Project:     GluaX   
//
// Purpose:     Interfacing C code to LUA scripting language.
//
// License:     
//    Zlib license (same as original 'glua10', see 'license.txt')
//
// Credits:
//    Original 'glua10' package by Enrico Colombini can be found at:
//    'http://www.cityline.it/personal/erix/progutil.html'. Based on his
//    request, I've renamed these files to 'gluax.h' to keep them apart from 
//    the original.
//
// Notes:
//    Apart from the 'argn' and 'tag' parameters (which should be constant
//    values, without arithmetics), the macros within this header file have been
//    designed to be without side effects. This especially applies to the 'push'
//    macros, where it is often handy to place arithmetic operations or
//    function calls right within the macro parameters.
//
// Bugs:
//    (none known)
//
----------------------*/
/*
  GluaX: Copyright (c) 2002-03 Asko Kauppi. All rights reserved.

  Glua 1.0: Copyright (c) 2002 Enrico Colombini. All rights reserved.

  This software, including its documentation, is provided 'as-is',
  without any express or implied warranty. In no event will the author
  be held liable for any damages arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.

  2. Altered source versions must be plainly marked as such, and must not
     be misrepresented as being the original software.

  3. This notice may not be removed or altered from any source distribution.

  http://www.opensource.org/licenses/zlib-license.php
*/

#ifndef GLUAX_H
#define GLUAX_H

#ifdef __cplusplus
  extern "C" {
#endif

#include <limits.h>     // 'MAXINT' or 'INT_MAX'
#include <stdlib.h>	// 'abort()'
#include <stdarg.h>     // 'va_list'

#include <sys/types.h>   // (declares 'uint', at least on Linux)

//---
// Declare 'PLATFORM_xxx' and 'COMPILER_xxx' defines:
//
#ifdef _WIN32
  #define PLATFORM_WIN32
  #define PLATFORM_NAME "win32"
#elif (defined __APPLE__) && (defined __MACH__)
  #ifndef PLATFORM_DARWIN
    #define PLATFORM_DARWIN
  #endif
  #define PLATFORM_NAME "darwin"
#elif (defined __unix__)
  #ifndef PLATFORM_POSIX
    #define PLATFORM_POSIX
  #endif
  #define PLATFORM_NAME "posix"
#else
  #error "Unknown platform!"
  #define PLATFORM_NAME NULL
#endif

#ifdef _MSC_VER   // Visual C++
  #define COMPILER_MSC
  #define COMPILER_NAME "msc"
  #undef _HAS_VA_ARGS   // no __VA_ARGS__ or anything similar on MSC, right?
  //
#elif (defined __GNUC__)    // Gcc
  #if (__GNUC__ < 3)   // __GNUC_MINOR__ would give .x info
    //#error "GCC 3.x required for binary compatibility"
    #undef _HAS_VA_ARGS
  #else
    #define _HAS_VA_ARGS
  #endif
  //
  #define COMPILER_GCC
  #define COMPILER_NAME "gcc"
  //
#else
  //#error "Unknown compiler! (just comment this line - it may still work)
  #define COMPILER_NAME "unknown"

  #if (__STDC_VERSION__ >= 199901L)  // C99
    #define _HAS_VA_ARGS	// Variadic preprocessor support (#define macro(...))
  #else
    //#error "No __VA_ARGS__ support - compile at your own risk!"
  #endif
#endif

#ifndef MAXINT
  #define MAXINT INT_MAX
  //
  #ifndef MAXINT
    #error "'MAXINT' or 'INT_MAX' needs to be #defined!"
  #endif
#endif

#ifndef TRUE
  #define TRUE  1
  #define FALSE 0
#endif

// Using 'bool_int' instead of plain 'bool' (which is a built-in type in C++)
// may be useful, since we are using the type in the binary interface.
//
typedef unsigned int bool_int;   // for TRUE and FALSE contents

#ifdef PLATFORM_WIN32
  typedef unsigned int uint;    // Linux and OS X have this
#endif
#if (defined PLATFORM_WIN32) || (defined PLATFORM_DARWIN)
  typedef unsigned long ulong;  // Linux has this (in 'sys/types.h')
#endif

typedef unsigned int enum_uint;


//---
// Default mode is 'GLUA_DYNAMIC' (unless application #defines 'GLUA_STATIC').
//
#ifdef GLUA_STATIC  // from project options
  #undef GLUA_DYNAMIC
#else
  #define GLUA_DYNAMIC  // default: building a dynamic module
#endif

//---
// To enable dynamic modules, we do NOT #include "lua.h" here.
// All interfacing to Lua engine happens via 'gluax.c'.
//
#ifdef GLUA_DYNAMIC
  //
  typedef struct lua_State lua_State;
  typedef int (*lua_CFunction) (lua_State *L);
  //
#else
  // Gcc 2.96 complains if we use both one-line definitions (above) and 
  // include 'lua.h' later. That's why we include it here, although the above
  // definitions would be sufficient.
  //
  #include "lua.h"  
  
  // Differ between the version 4.0 and 5.0:
  //
  #ifdef LUA_ANYTAG	// (only defined in Lua 4.0)
    #define LUA_V40
  #else
    #define LUA_V50
    //
    #ifndef LUA_HOOKTAILRET    // introduced in 5.0 final
      #define LUA50_BETA
    #endif
  #endif
  //
#endif

#ifdef __cplusplus
  #define _EXTERN_C extern "C"
#else
  #define _EXTERN_C /*nothing*/
#endif


/*--- Macros etc. ----------------------------------------------------*/

#define /*int abs_index*/ STACK_ABS( L, index ) \
	( (index) >= 0 ? (index) /*absolute*/ : (_glua_gettop(L) +(index) +1) )

#ifdef GLUA_OPT_NOCHECK     // All runtime checks removed
  //
  #define ASSUME(c)         /*nothing*/
  #define ASSUME_NONZERO(v) /*nothing*/
  #define ASSUME_NEVER()    /*nothing*/     // could also be 'abort()'
  //
  #define STACK_CHECK(L)    /*nothing*/
  #define STACK_END(c)      /*nothing*/
  //
#else   // Runtime safety checks:
  //
  // Note: Using 'assert.h' would only apply to debug compilations!
  //
  void _glua_assert( uint line, const char* file, const char* cond_str );

  #define ASSUME(c) { if (!(c)) _glua_assert( __LINE__, __FILE__, #c ); }

  #define ASSUME_NONZERO(v) ASSUME(v)
  #define ASSUME_NEVER()    ASSUME(0)
  //
  #define STACK_CHECK(L)     { lua_State* _luastack_= (L); \
                               int _oldtop_= _glua_gettop(L);
  #define STACK_END(change)  ASSUME( _glua_gettop(_luastack_) == _oldtop_+(change) ); }
#endif

#ifndef NEVER
  #define NEVER 0		// 'return NEVER;' etc.
#endif

#ifndef IGNORE_PARAM
  #define IGNORE_PARAM(var) ( (void)(var) )
#endif

//---
// WARNING! Notes on 'strncpy()' and 'strncat()':
//
// If a buffer is filled with 'strncpy()', it does _not_ add the termination
// character in the last slot.
//
// The size param to 'strncat()' is the count of chars to append, _not_ the
// size of the buffer! Curiously enough, this function _does_ terminate the
// buffer at fill-ups.
//
// For these reason, we have our own variants instead of the standard 
// 'strncpy()' and 'strncat()'. These are defined as functions (not macros)
// to keep clear of side-effects.
//
// Note: Also modules using 'gluax.h' may use these for their own purposes.
//
char* strcpy_safe( char* buf, const char* src, uint buf_size );
char* strcat_safe( char* buf, const char* src, uint buf_size );


/*--- Data structures ----------------------------------------------------*/

//---
// These constants don't match the 'LUA_Txxx' constants in values. They use
// mask bits to allow easy testing against 'either real or integer number',
// or 'either string or nil' and so on...
//
// DON'T CHANGE THESE VALUES or you'll break backwards compatibility with
// existing modules!
//
enum e_Glua_Type
{
	GLUA_TYPE_NIL=			0x0001,	  // 'nil' or none (empty, default value)
	//
	// Note: 'GLUA_TYPE_NIL' covers both Lua types 'NIL' and 'NONE'.

	GLUA_TYPE_BOOLEAN=		0x0002,	  // (only by Lua 5.0 engine)

	GLUA_TYPE_FLOAT=	    0x0004,	  // floating-point number
	GLUA_TYPE_INT=		    0x0008,   // integer number

	GLUA_TYPE_STRING=		0x0010,	  // text string (may be numeric)
	GLUA_TYPE_TABLE=		0x0020,
	GLUA_TYPE_FUNCTION=		0x0040,
	//...(available)		0x0080,

	GLUA_TYPE_USERDATA=     0x0100,   // userdata matching our tag
	GLUA_TYPE_USERDATA_UNK= 0x0200,   // userdata NOT matching our tag
	//...

    //---
    // Mask constants make defining macros simpler. Not for application code!
    //
    _GLUA_MASK_FLOAT_INT=           (GLUA_TYPE_FLOAT | GLUA_TYPE_INT),
    _GLUA_MASK_INT_NIL=             (GLUA_TYPE_INT | GLUA_TYPE_NIL),
    _GLUA_MASK_FLOAT_INT_NIL=       (GLUA_TYPE_FLOAT | _GLUA_MASK_INT_NIL),
    _GLUA_MASK_STRING_FLOAT_INT_NIL= (GLUA_TYPE_STRING | _GLUA_MASK_FLOAT_INT_NIL),
    //
    _GLUA_MASK_STRING_NIL=          (GLUA_TYPE_STRING | GLUA_TYPE_NIL),
    _GLUA_MASK_TABLE_NIL=           (GLUA_TYPE_TABLE | GLUA_TYPE_NIL),
    _GLUA_MASK_USERDATA_NIL=        (GLUA_TYPE_USERDATA | GLUA_TYPE_NIL)
};

//-----
struct s_GluaItem
{
    const char* name;

    int (*func)(lua_State *);   // function entries only
    const char* str;            // constant strings (if 'func' == NULL)
	int* tag_ref;				// tag entries: where to store the created tag.
    double num;                 // constant numbers (if others are NULL)
};

#define GLUAITEM_SIZECHECK() \
    ( ( sizeof(void*) == 4 ) && \
      ( sizeof(double) == 8 ) && \
      ( sizeof(struct s_GluaItem) == 24 ) )


/*--- Host interface functions -----------------------------*/

void _glua_error( lua_State* L, const char* msg );

void _glua_errorN( lua_State* L, const char* msg, ... );
void _glua_errorL( lua_State* L, const char* msg, va_list args );

uint /*host_ver*/ _glua_ver( uint module_ver );

#define /*e_GluaType*/ _glua_type( L, argn )  _glua_type2( L, argn, 0 )
enum_uint /*e_GluaType*/ _glua_type2( lua_State* L, int argn, int ud_tag );

bool_int    _glua_toboolean( lua_State* L, int argn );
double      _glua_tonumber_safe( lua_State* L, int argn );
double      _glua_tonumber_raw( lua_State* L, int argn );
const char* _glua_tostring_raw( lua_State* L, int argn );
const char* _glua_tolstring_raw( lua_State* L, int argn, size_t* len_ref );
const char* _glua_tolstring_safe( lua_State* L, int argn, size_t* len_ref );
void*       _glua_touserdata( lua_State* L, int argn );

#define _glua_tostring_safe( L, argn )  _glua_tolstring_safe( L, argn, NULL )

void _glua_pushboolean( lua_State* L, bool_int v );
void _glua_pushnumber( lua_State* L, double n );
void _glua_pushstring( lua_State* L, const char* s );
void _glua_pushlstring( lua_State* L, const char* s, uint len );
void* _glua_pushuserdata( lua_State* lua, int tag, 
						 const void* p,		// if NULL, 'size' is used
						 size_t size,		// if 'p' given, 'size' must be 0
                         lua_CFunction gc_func );

void _glua_newtable( lua_State *L );    // was: _glua_pushtable()
void _glua_totable( lua_State *L );

int  _glua_gettop (lua_State *L);

void _glua_dump(lua_State* L);  // (can be used for dumping Lua stack also 
                                //  by application code - not documented ;)

void _glua_pushcfunction (lua_State *L, lua_CFunction f );
void _glua_settop( lua_State *L, int n );
void _glua_setglobal( lua_State *L, const char* s );
bool_int _glua_checkstack( lua_State* L, int extra );
void _glua_pushregtbl( lua_State* L );
int _glua_newtag( lua_State* L );

#define _glua_pushnil(L)  _glua_pushlstring(L,NULL,0)

void _glua_fromtable( lua_State *L, int argn );
bool_int _glua_next( lua_State *L, int argn );
void _glua_remove( lua_State* L, int argn );
void _glua_pushvalue( lua_State* L, int argn );

int _glua_exec( const char* cmd_str, int stack_size );
#define glua_exec( cmd_str, stack_size )  _glua_exec( cmd_str, stack_size )


/*--- Internal Functions -----------------------------*/

extern void _glua_register2(lua_State *lua, 
                            struct s_GluaItem list[], uint items, 
                            int tbl_index );

int /*0=success*/ _glua_linktohost( lua_State* lua,
                              const void* v_func_tbl, 
                              uint func_tbl_size );

extern int _glua_regtag( lua_State *lua, const char *tagname );

int _glua_getn( lua_State* L, int argn );

void _glua_concat( lua_State* L, int argn );

void _glua_insert( lua_State* L, int argn );

void _glua_pushcclosure( lua_State* L, lua_CFunction fn, uint n );

#ifdef _HAS_VA_ARGS   // Gcc etc.
  #define _setL(lua)   /*ignore*/
  #define _setT(argn)  (_tbl_argn_=(argn))
  #define _getL()      _lua_
  #define _getT()      _tbl_argn_
  //
#else   // MSC etc.
  void _setL( lua_State* L );
  lua_State* _getL(void);
  int _setT( int tbl_argn );
  int _getT(void);
  //
#endif


/*--- Compatibility functions (for Lua C API emulation) -----*/

#if ((defined GLUA_STATIC) || (defined GLUAPORT))
  //  
  int _gluaport_strlen( lua_State* L, int argn );
  uint _gluaport_tagofuserdata( lua_State* L, int index );
  void _gluaport_rawset( lua_State* L, int argn );
  void _gluaport_rawseti( lua_State* L, int argn, int n );
  void _gluaport_call( lua_State* L, int nargs, int nresults );
  void _gluaport_settable( lua_State* L, int argn );
  void _gluaport_settag( lua_State* L, int argn, int tag );
  void _gluaport_settagmethod( lua_State* L, int tag, const char* event );
#endif


/*--- Public macros -----------------------------------------*/

#define glua_error(msg)  _glua_error( _lua_, (msg) )

#ifdef _HAS_VA_ARGS	// gcc etc.
  #define glua_errorN(...)  _glua_errorN( _lua_, __VA_ARGS__ )
#else
  void glua_errorN( const char* msg, ... );   // needs '_getL()'!
#endif

#define glua_getArgn()      _glua_gettop(_lua_)
#define glua_getType(argn)	( (enum e_Glua_Type)_glua_type(_lua_,argn) )

//---
// 'glua_is...())' macros:
//
// These can be used to query the precise types of input parameters. This is
// needed in variable parameter list functions and also if strict non-converting
// parameter lists are wanted to be used. E.g. 'glua_getString()' will 
// convert numbers automatically to strings. To avoid this, 'glua_isString()'
// returns TRUE only for actual strings (and 'nil'), not for numbers.
//
// Note: 'isNumber()' gives TRUE for both real and integer numbers.
//       'isInteger()' gives TRUE only for integers.
//
// Note: 'nil' (NULL) is accepted as a string and/or userdata value.
//
#define glua_isInteger(argn)  ( _glua_type(_lua_,argn) == GLUA_TYPE_INT )
#define glua_isNumber(argn)	  ( _glua_type(_lua_,argn) & _GLUA_MASK_FLOAT_INT )
#define glua_isString(argn)	  ( _glua_type(_lua_,argn) &  GLUA_TYPE_STRING )
#define glua_isUserdata(argn) ( _glua_type(_lua_,argn) &  GLUA_TYPE_USERDATA )
#define glua_isTable(argn)	  ( _glua_type(_lua_,argn) &  GLUA_TYPE_TABLE )
#define glua_isNil(argn)	  ( _glua_type(_lua_,argn) == GLUA_TYPE_NIL )

// Regular parameters have 'argn' != 0.
// Table fields have 'argn'==0.
//
double _glua_getNumber(lua_State* L, int argn);
int    _glua_getInteger(lua_State* L, int argn);
unsigned long _glua_getUnsigned(lua_State* L, int argn);   // normally 32-bit
const char* _glua_getLString(lua_State* L, int argn, size_t* len_ref );
void* _glua_getUserdata(lua_State* L, int argn, int tag );
bool_int _glua_getBoolean(lua_State* L, int argn );
int _glua_getTable(lua_State* L, int argn);
//void _glua_getDone(lua_State* L, int argn);

#ifdef _HAS_LONGLONG
  unsigned long long _glua_getUnsignedLL(lua_State* L, int argn);   // 64-bit
  #define glua_getUnsigned64(argn)  _glua_getUnsignedLL(_lua_,argn)
#endif

#define _glua_getString(L,argn)  _glua_getLString(L,argn,NULL)
#define glua_getNumber(argn)    _glua_getNumber(_lua_,argn)
#define glua_getInteger(argn)   _glua_getInteger(_lua_,argn)
#define glua_getUnsigned(argn)  _glua_getUnsigned(_lua_,argn)
#define glua_getString(argn)    _glua_getString(_lua_,argn)
#define glua_getLString(argn,len_ref)   _glua_getLString(_lua_,argn,len_ref)
#define glua_getUserdata(argn, tag)     _glua_getUserdata(_lua_,argn,tag)
#define glua_getBoolean(argn)   _glua_getBoolean(_lua_,argn)

// TRUE for a table (even empty ones), FALSE for 'nil':
#define /* bool_int */ glua_getTable(argn)\
  ( (bool_int)_setT(_glua_getTable(_lua_,argn)) )

//#define glua_getDone()  _glua_getDone( _lua_, (int)(_getcount_+1) )

//-----
// Note: The 'do...while(0)'s are dummies that eat up semicolons after them.
//       This helps using the macros in if-else constructs.
//
#define glua_pushNumber(v)      ( _glua_pushnumber(_lua_, v), _retcount_++ )
#define glua_pushString(s)      ( _glua_pushstring(_lua_,s), _retcount_++ )
#define glua_pushLString(s,len) ( _glua_pushlstring(_lua_,s,len), _retcount_++ )
#define glua_pushBoolean(b)     ( _glua_pushboolean(_lua_,b), _retcount_++ )
#define glua_pushTable()        ( _glua_newtable(_lua_), _retcount_++ )
//
#define glua_pushInteger(v)     glua_pushNumber(v)
#define glua_pushUnsigned(v)    glua_pushNumber(v)
#define glua_pushNil()          glua_pushString(NULL)

//---
// In C++, object specific garbage collection funcs are an added extra
// (note - you can also use 'glua_tagmethod()' to cover all objects):
//
//  glua_pushUserdata( tag, size, ptr [,gc_func] )
//
#if ((!defined __cplusplus) || (!defined _HAS_VA_ARGS))  // ANSI C or MSC
  //
  #define /* void* */ glua_pushUserdata( tag, size, ptr ) \
  ( _retcount_++, _glua_pushuserdata( _lua_, tag, ptr, size, NULL ) )
  //
  #define /* void* */ glua_pushUserdata_gc( tag, size, ptr, gc_func ) \
  ( _retcount_++, _glua_pushuserdata( _lua_, tag, ptr, size, gc_func ) )
  //
#else   // C++ (not MSC):
  //
  } // extern "C"

  inline void* _glua_pushUserdata( lua_State* L, int tag, size_t size, const void* ptr, lua_CFunction gc_func=NULL )
    { return _glua_pushuserdata( L, tag, ptr, size, gc_func ); }
  //
  #define /* void* */ glua_pushUserdata( ... )  ( _retcount_++, _glua_pushUserdata( _lua_, __VA_ARGS__ ) )

  extern "C" {
  //
#endif  // C++


//-----
// Attachments to userdata entries (provided by their associated pointers)
// may be required to pass data between the creator & releaser of an object.
// This system uses Lua registry internally, and allows a single pointer
// to be attached to each userdata. New settings override older ones.
//
void _glua_getreg( lua_State* L );
void _glua_setreg( lua_State* L );

void _glua_settagmethod( lua_State* L, int tag, const char* event,
									   lua_CFunction func );

void* _glua_attach( lua_State* L, const void* obj, void* attachment );

#define glua_setAttachment( obj_ptr, attachment_ptr ) \
		(void)_glua_attach( _lua_, obj_ptr, attachment_ptr )

#define /* void* */ glua_getAttachment( obj_ptr, remove ) \
		_glua_attach( _lua_, obj_ptr, (remove) ? NULL:(void*)(-1) ) \


//-----
// Functions for adding entries into a table (which is at top of stack).
//
void _glua_tblkey_i( lua_State* L, int i );
void _glua_tblkey_k( lua_State* L, const char* k );
void _glua_tblkey_ijk( lua_State* L, int i, int j, const char* k );
//
#define _glua_tblkey_ik(L,i,k)  _glua_tblkey_ijk(L,i,0,k)
#define _glua_tblkey_ij(L,i,j)  _glua_tblkey_ijk(L,i,j,NULL)

#ifndef __cplusplus  // ANSI C only gets the basic forms:
  //
  #define glua_tblKey(k)   ( _glua_tblkey_k(_lua_,k), _retcount_-- )
  #define glua_tblKey_i(i) ( _glua_tblkey_i(_lua_,i), _retcount_-- )
  //
  #define glua_setTable_num(k,v)    ( glua_pushNumber(v), glua_tblKey(k) )
  #define glua_setTable_str(k,s)    ( glua_pushString(s), glua_tblKey(k) )
  #define glua_setTable_ud(k,p,tag) ( glua_pushUserdata(p,tag), glua_tblKey(k) )
  #define glua_setTable_bool(k,v)   ( glua_pushBoolean(v), glua_tblKey(k) )
  #define glua_setTable_int(k,v)     glua_setTable_num(k,v)
  #define glua_setTable_uint(k,v)    glua_setTable_num(k,v)
  //
  #define glua_setTable_num_i(i,v)    ( glua_pushNumber(v), glua_tblKey_i(i) )
  #define glua_setTable_str_i(i,s)    ( glua_pushString(s), glua_tblKey_i(i) )
  #define glua_setTable_int_i(i,v)     glua_setTable_num_i(i,v)
  #define glua_setTable_uint_i(i,v)    glua_setTable_num_i(i,v)
  //
#else   // C++ can use inline funcs & parameter overloading :) 
  //
  } // extern "C"

  #ifdef _HAS_VA_ARGS
    //
    inline void _glua_tblkey( lua_State* L, int i )                { _glua_tblkey_i(L,i); }
    inline void _glua_tblkey( lua_State* L, const char* k )        { _glua_tblkey_k(L,k); }
    inline void _glua_tblkey( lua_State* L, int i, const char* k ) { _glua_tblkey_ijk(L,i,0,k); }
    inline void _glua_tblkey( lua_State* L, int i, int j )         { _glua_tblkey_ijk(L,i,j,NULL); }
    //
    // Note: Only for apps that want to do "glua_pushXxx(), glua_tblKey()" themselves:
    //
    #define glua_tblKey( ... )  ( _glua_tblkey( _lua_, __VA_ARGS__ ), _retcount_-- )
    //
  #else   // MSC 6.0
    //
    inline void glua_tblKey( int i )                { _glua_tblkey_i(_getL(),i); }
    inline void glua_tblKey( const char* k )        { _glua_tblkey_k(_getL(),k); }
    inline void glua_tblKey( int i, const char* k ) { _glua_tblkey_ik(_getL(),i,k); }
    inline void glua_tblKey( int i, int j )         { _glua_tblkey_ij(_getL(),i,j); }
    //
  #endif

  #ifdef _HAS_VA_ARGS
    // Note: 'int' funcs not needed, they'll use 'num' directly.
    //
    inline void _glua_setTable_num( lua_State* L, const char* k, double v )         { _glua_pushnumber(L,v); _glua_tblkey(L,k); }
    inline void _glua_setTable_num( lua_State* L, int i, double v )                 { _glua_pushnumber(L,v); _glua_tblkey(L,i); }
    inline void _glua_setTable_num( lua_State* L, int i, const char* k, double v )  { _glua_pushnumber(L,v); _glua_tblkey(L,i,k); }
    inline void _glua_setTable_num( lua_State* L, int i, int j, double v )          { _glua_pushnumber(L,v); _glua_tblkey(L,i,j); }
    //
    inline void _glua_setTable_str( lua_State* L, const char* k, const char* s )    { _glua_pushstring(L,s); _glua_tblkey(L,k); }
    inline void _glua_setTable_str( lua_State* L, int i, const char* s )            { _glua_pushstring(L,s); _glua_tblkey(L,i); }
    inline void _glua_setTable_str( lua_State* L, int i, const char* k, const char* s )  { _glua_pushstring(L,s); _glua_tblkey(L,i,k); }
    inline void _glua_setTable_str( lua_State* L, int i, int j, const char* s )     { _glua_pushstring(L,s); _glua_tblkey(L,i,j); }
    //
    inline void _glua_setTable_ud( lua_State* L, const char* k, void* p, int tag )  { _glua_pushuserdata(L,tag,p,0,NULL); _glua_tblkey(L,k); }
    inline void _glua_setTable_ud( lua_State* L, int i, void* p, int tag )          { _glua_pushuserdata(L,tag,p,0,NULL); _glua_tblkey(L,i); }
    inline void _glua_setTable_ud( lua_State* L, int i, const char* k, void* p, int tag )  { _glua_pushuserdata(L,tag,p,0,NULL); _glua_tblkey(L,i,k); }
    inline void _glua_setTable_ud( lua_State* L, int i, int j, void* p, int tag )   { _glua_pushuserdata(L,tag,p,0,NULL); _glua_tblkey(L,i,j); }
    //
    // Note: Here we can use 'bool' (not 'int_bool') since we're definately in C++.
    //
    inline void _glua_setTable_bool( lua_State* L, const char* k, bool b )          { _glua_pushboolean(L,b); _glua_tblkey(L,k); }
    inline void _glua_setTable_bool( lua_State* L, int i, bool b )                  { _glua_pushboolean(L,b); _glua_tblkey(L,i); }
    inline void _glua_setTable_bool( lua_State* L, int i, const char* k, bool b )   { _glua_pushboolean(L,b); _glua_tblkey(L,i,k); }
    inline void _glua_setTable_bool( lua_State* L, int i, int j, bool b )           { _glua_pushboolean(L,b); _glua_tblkey(L,i,j); }

    // Note: support for variable arglist in preprocessor required!
    //       (at least gcc has this, but is it standard...?)
    //
    #define glua_setTable_int( ... ) _glua_setTable_num(_lua_, __VA_ARGS__)   // (int->num)
    #define glua_setTable_uint( ... ) _glua_setTable_num(_lua_, __VA_ARGS__)   // (uint->num)
    #define glua_setTable_num( ... ) _glua_setTable_num(_lua_, __VA_ARGS__)
    #define glua_setTable_str( ... ) _glua_setTable_str(_lua_, __VA_ARGS__)
    #define glua_setTable_ud( ... )  _glua_setTable_ud(_lua_, __VA_ARGS__)
    #define glua_setTable_bool( ... ) _glua_setTable_bool(_lua_, __VA_ARGS__)
    //
  #else   // MSC (no __VA_ARGS__ support)
    //
    // Using '_getL()' incurs a slight runtime penalty (thread specific data).
    // That's why we don't use it for other compilers.
    //
    inline void glua_setTable_num( const char* k, double v )         { lua_State* L=_getL(); _glua_pushnumber(L,v); _glua_tblkey_k(L,k); }
    inline void glua_setTable_num( int i, double v )                 { lua_State* L=_getL(); _glua_pushnumber(L,v); _glua_tblkey_i(L,i); }
    inline void glua_setTable_num( int i, const char* k, double v )  { lua_State* L=_getL(); _glua_pushnumber(L,v); _glua_tblkey_ik(L,i,k); }
    inline void glua_setTable_num( int i, int j, double v )          { lua_State* L=_getL(); _glua_pushnumber(L,v); _glua_tblkey_ij(L,i,j); }
    //
    inline void glua_setTable_int( const char* k, int v )			 { glua_setTable_num(k,v); }
    inline void glua_setTable_int( int i, int v )                    { glua_setTable_num(i,v); }
    inline void glua_setTable_int( int i, const char* k, int v )	 { glua_setTable_num(i,k,v); }
    inline void glua_setTable_int( int i, int j, int v )			 { glua_setTable_num(i,j,v); }
	//
    inline void glua_setTable_uint( const char* k, uint v )			 { glua_setTable_num(k,v); }
    inline void glua_setTable_uint( int i, uint v )                  { glua_setTable_num(i,v); }
    inline void glua_setTable_uint( int i, const char* k, uint v )	 { glua_setTable_num(i,k,v); }
    inline void glua_setTable_uint( int i, int j, uint v )			 { glua_setTable_num(i,j,v); }
	//
    inline void glua_setTable_str( const char* k, const char* s )    { lua_State* L=_getL(); _glua_pushstring(L,s); _glua_tblkey_k(L,k); }
    inline void glua_setTable_str( int i, const char* s )            { lua_State* L=_getL(); _glua_pushstring(L,s); _glua_tblkey_i(L,i); }
    inline void glua_setTable_str( int i, const char* k, const char* s )  { lua_State* L=_getL(); _glua_pushstring(L,s); _glua_tblkey_ik(L,i,k); }
    inline void glua_setTable_str( int i, int j, const char* s )     { lua_State* L=_getL(); _glua_pushstring(L,s); _glua_tblkey_ij(L,i,j); }
    //
    inline void glua_setTable_ud( const char* k, void* p, int tag )  { lua_State* L=_getL(); _glua_pushuserdata(L,tag,p,0,NULL); _glua_tblkey_k(L,k); }
    inline void glua_setTable_ud( int i, void* p, int tag )          { lua_State* L=_getL(); _glua_pushuserdata(L,tag,p,0,NULL); _glua_tblkey_i(L,i); }
    inline void glua_setTable_ud( int i, const char* k, void* p, int tag )  { lua_State* L=_getL(); _glua_pushuserdata(L,tag,p,0,NULL); _glua_tblkey_ik(L,i,k); }
    inline void glua_setTable_ud( int i, int j, void* p, int tag )   { lua_State* L=_getL(); _glua_pushuserdata(L,tag,p,0,NULL); _glua_tblkey_ij(L,i,j); }
    //
    inline void glua_setTable_bool( const char* k, bool b )          { lua_State* L=_getL(); _glua_pushboolean(L,b); _glua_tblkey_k(L,k); }
    inline void glua_setTable_bool( int i, bool b )                  { lua_State* L=_getL(); _glua_pushboolean(L,b); _glua_tblkey_i(L,i); }
    inline void glua_setTable_bool( int i, const char* k, bool b )   { lua_State* L=_getL(); _glua_pushboolean(L,b); _glua_tblkey_ik(L,i,k); }
    inline void glua_setTable_bool( int i, int j, bool b )           { lua_State* L=_getL(); _glua_pushboolean(L,b); _glua_tblkey_ij(L,i,j); }
    //
  #endif
  
  extern "C" {
#endif  // C++

//-----
// Functions for reading in table fields from an input parameter table.
//
// NOTE: These calls need to be made _immediately_after_ a 'getTable()'
//       that has 'opened' the table for access. Once other 'get' functions
//       are called, these are no longer valid!
//
// Subkeys (s.a. "obj.color.red") can be placed directly within the 'key'
// string itself or provided in separate 'index' and 'key' entries which
// is practical to browse through arrays.
//
void _glua_usekey_i( lua_State* L, int argn, int i );
void _glua_usekey_k( lua_State* L, int argn, const char* k );
void _glua_usekey_ijk( lua_State* L, int argn, int i, int j, const char* k );
//
#define _glua_usekey_ik(L,a,i,k) _glua_usekey_ijk( L, a, i,0, k )
#define _glua_usekey_ij(L,a,i,j) _glua_usekey_ijk( L, a, i,j, NULL )

#ifndef __cplusplus   // ANSI C only gets the basic forms:
  //
  #define glua_useKey(k)   _glua_usekey_k(_lua_,_tbl_argn_,k)
  #define glua_useKey_i(i) _glua_usekey_i(_lua_,_tbl_argn_,i)
  //
  #define glua_getTable_int(k)    ( glua_useKey(k), _glua_getInteger(_lua_,0) )
  #define glua_getTable_num(k)    ( glua_useKey(k), _glua_getNumber(_lua_,0) )
  #define glua_getTable_str(k)    ( glua_useKey(k), _glua_getString(_lua_,0) )
  #define glua_getTable_ud(k,tag) ( glua_useKey(k), _glua_getUserdata(_lua_,0,tag) )
  #define glua_getTable_bool(k)   ( glua_useKey(k), _glua_getBoolean(_lua_,0) )
  #define glua_getTable_uint(k)   ( glua_useKey(k), _glua_getUnsigned(_lua_,0) )
  //
  #define glua_getTable_int_i(i)    ( glua_useKey_i(i), _glua_getInteger(_lua_,0) )
  #define glua_getTable_num_i(i)    ( glua_useKey_i(i), _glua_getNumber(_lua_,0) )
  #define glua_getTable_str_i(i)    ( glua_useKey_i(i), _glua_getString(_lua_,0) )
  #define glua_getTable_ud_i(i,tag) ( glua_useKey_i(i), _glua_getUserdata(_lua_,0,tag) )
  #define glua_getTable_bool_i(i)   ( glua_useKey_i(i), _glua_getBoolean(_lua_,0) )
  #define glua_getTable_uint_i(i)   ( glua_useKey_i(i), _glua_getUnsigned(_lua_,0) )
  //
#else   // C++
  } // extern "C"

  #ifdef _HAS_VA_ARGS
    //
    inline void _glua_usekey( lua_State* L, int argn, int i )                { _glua_usekey_i(L,argn,i); }
    inline void _glua_usekey( lua_State* L, int argn, const char* k )        { _glua_usekey_k(L,argn,k); }
    inline void _glua_usekey( lua_State* L, int argn, int i, const char* k ) { _glua_usekey_ijk(L,argn,i,0,k); }
    inline void _glua_usekey( lua_State* L, int argn, int i, int j )         { _glua_usekey_ijk(L,argn,i,j,NULL); }
    //
    #define glua_useKey(...)    _glua_usekey( _lua_, _tbl_argn_, __VA_ARGS__ )
    #define glua_useKey_i(...)  _glua_usekey( _lua_, _tbl_argn_, __VA_ARGS__ )
    //
    inline int _glua_getTable_int( lua_State* L, int a, const char* k )        { _glua_usekey(L,a,k); return _glua_getInteger(L,0); }
    inline int _glua_getTable_int( lua_State* L, int a, int i )                { _glua_usekey(L,a,i); return _glua_getInteger(L,0); }
    inline int _glua_getTable_int( lua_State* L, int a, int i, const char* k ) { _glua_usekey(L,a,i,k); return _glua_getInteger(L,0); }
    inline int _glua_getTable_int( lua_State* L, int a, int i, int j )         { _glua_usekey(L,a,i,j); return _glua_getInteger(L,0); }
    //
    inline double _glua_getTable_num( lua_State* L, int a, const char* k )     { _glua_usekey(L,a,k); return _glua_getNumber(L,0); }
    inline double _glua_getTable_num( lua_State* L, int a, int i )             { _glua_usekey(L,a,i); return _glua_getNumber(L,0); }
    inline double _glua_getTable_num( lua_State* L, int a, int i, const char* k ) { _glua_usekey(L,a,i,k); return _glua_getNumber(L,0); }
    inline double _glua_getTable_num( lua_State* L, int a, int i, int j )      { _glua_usekey(L,a,i,j); return _glua_getNumber(L,0); }
    //
    inline const char* _glua_getTable_str( lua_State* L, int a, const char* k ) { _glua_usekey(L,a,k); return _glua_getString(L,0); }
    inline const char* _glua_getTable_str( lua_State* L, int a, int i )         { _glua_usekey(L,a,i); return _glua_getString(L,0); }
    inline const char* _glua_getTable_str( lua_State* L, int a, int i, const char* k ) { _glua_usekey(L,a,i,k); return _glua_getString(L,0); }
    inline const char* _glua_getTable_str( lua_State* L, int a, int i, int j )  { _glua_usekey(L,a,i,j); return _glua_getString(L,0); }
    //
    inline void* _glua_getTable_ud( lua_State* L, int a, const char* k, int tag ) { _glua_usekey(L,a,k); return _glua_getUserdata(L,0,tag); }
    inline void* _glua_getTable_ud( lua_State* L, int a, int i, int tag )         { _glua_usekey(L,a,i); return _glua_getUserdata(L,0,tag); }
    inline void* _glua_getTable_ud( lua_State* L, int a, int i, const char* k, int tag ) { _glua_usekey(L,a,i,k); return _glua_getUserdata(L,0,tag); }
    inline void* _glua_getTable_ud( lua_State* L, int a, int i, int j, int tag )  { _glua_usekey(L,a,i,j); return _glua_getUserdata(L,0,tag); }
    //
    inline bool _glua_getTable_bool( lua_State* L, int a, const char* k )     { _glua_usekey(L,a,k); return _glua_getBoolean(L,0); }
    inline bool _glua_getTable_bool( lua_State* L, int a, int i )             { _glua_usekey(L,a,i); return _glua_getBoolean(L,0); }
    inline bool _glua_getTable_bool( lua_State* L, int a, int i, const char* k ) { _glua_usekey(L,a,i,k); return _glua_getBoolean(L,0); }
    inline bool _glua_getTable_bool( lua_State* L, int a, int i, int j )      { _glua_usekey(L,a,i,j); return _glua_getBoolean(L,0); }
    //
    #define glua_getTable_int( ... ) _glua_getTable_int(_lua_, _tbl_argn_, __VA_ARGS__)
    #define glua_getTable_num( ... ) _glua_getTable_num(_lua_, _tbl_argn_, __VA_ARGS__)
    #define glua_getTable_str( ... ) _glua_getTable_str(_lua_, _tbl_argn_, __VA_ARGS__)
    #define glua_getTable_ud( ... )  _glua_getTable_ud(_lua_, _tbl_argn_, __VA_ARGS__)
    #define glua_getTable_bool( ... ) _glua_getTable_bool(_lua_, _tbl_argn_, __VA_ARGS__)
    #define glua_getTable_uint( ... ) (uint)_glua_getTable_int(_lua_, _tbl_argn_, __VA_ARGS__)
    //
  #else   // MSC 6.0
    //
    // These inline funcs return 'L' pointer for our use. 
    // Applications should regard them as 'void' functions, though.
    //
    // Note: MSC has '_tbl_argn_' zero, it'll be fetched from thread specific storage.
    //
    inline lua_State* glua_useKey( const char* key )  
        { lua_State* L=_getL(); _glua_usekey_k(L,0,key); return L; }
    //
    inline lua_State* glua_useKey( int i )
        { lua_State* L=_getL(); _glua_usekey_i(L,0,i); return L; }
    //
    inline lua_State* glua_useKey( int i, const char* k )
        { lua_State* L=_getL(); _glua_usekey_ik(L,0,i,k); return L; }
    //
    inline lua_State* glua_useKey( int i, int j )
        { lua_State* L=_getL(); _glua_usekey_ij(L,0,i,j); return L; }
    //
    #define glua_useKey_i(i) glua_useKey(i)
    //
    inline int glua_getTable_int( const char* k )        { return _glua_getInteger( glua_useKey(k), 0 ); }
    inline int glua_getTable_int( int i )                { return _glua_getInteger( glua_useKey(i), 0 ); }
    inline int glua_getTable_int( int i, const char* k ) { return _glua_getInteger( glua_useKey(i,k), 0 ); }
    inline int glua_getTable_int( int i, int j )         { return _glua_getInteger( glua_useKey(i,j), 0 ); }
    //
    inline double glua_getTable_num( const char* k )     { return _glua_getNumber( glua_useKey(k), 0 ); }
    inline double glua_getTable_num( int i )             { return _glua_getNumber( glua_useKey(i), 0 ); }
    inline double glua_getTable_num( int i, const char* k ) { return _glua_getNumber( glua_useKey(i,k), 0 ); }
    inline double glua_getTable_num( int i, int j )      { return _glua_getNumber( glua_useKey(i,j), 0 ); }
    //
    inline const char* glua_getTable_str( const char* k ) { return _glua_getString( glua_useKey(k), 0 ); }
    inline const char* glua_getTable_str( int i )         { return _glua_getString( glua_useKey(i), 0 ); }
    inline const char* glua_getTable_str( int i, const char* k ) { return _glua_getString( glua_useKey(i,k), 0 ); }
    inline const char* glua_getTable_str( int i, int j )  { return _glua_getString( glua_useKey(i,j), 0 ); }
    //
    inline void* glua_getTable_ud( const char* k, int tag ) { return _glua_getUserdata( glua_useKey(k), 0, tag ); }
    inline void* glua_getTable_ud( int i, int tag )         { return _glua_getUserdata( glua_useKey(i), 0, tag ); }
    inline void* glua_getTable_ud( int i, const char* k, int tag ) { return _glua_getUserdata( glua_useKey(i,k), 0, tag ); }
    inline void* glua_getTable_ud( int i, int j, int tag )  { return _glua_getUserdata( glua_useKey(i,j), 0, tag ); }
    //
    inline bool_int glua_getTable_bool( const char* k )     { return _glua_getBoolean( glua_useKey(k), 0 ); }
    inline bool_int glua_getTable_bool( int i )             { return _glua_getBoolean( glua_useKey(i), 0 ); }
    inline bool_int glua_getTable_bool( int i, const char* k ) { return _glua_getBoolean( glua_useKey(i,k), 0 ); }
    inline bool_int glua_getTable_bool( int i, int j )      { return _glua_getBoolean( glua_useKey(i,j), 0 ); }
    //
    inline uint glua_getTable_uint( const char* k )        { return _glua_getUnsigned( glua_useKey(k), 0 ); }
    inline uint glua_getTable_uint( int i )                { return _glua_getUnsigned( glua_useKey(i), 0 ); }
    inline uint glua_getTable_uint( int i, const char* k ) { return _glua_getUnsigned( glua_useKey(i,k), 0 ); }
    inline uint glua_getTable_uint( int i, int j )         { return _glua_getUnsigned( glua_useKey(i,j), 0 ); }
    //
  #endif
  extern "C" {
#endif  // C++

//---
// These are jointly for all compilers:
//
bool_int _glua_gettablekey( lua_State* L,
		    			    uint argn,
                            const char** key_ref, 
                            enum e_Glua_Type* type_ref );

#define /*bool_int*/ glua_getTableKey( key_ref, type_ref ) \
        _glua_gettablekey( _getL(), _getT(), (key_ref), (type_ref) )

#define /*uint*/ glua_getTableN() \
        (_tbl_argn_>0) ? _glua_getn( _getL(), _getT() ) : 0  // ('nil' table)


/*---------------------
// GLUA_FUNC & GLUA_END macros
//
// Usage:   GLUA_FUNC( BitAnd )
//          {   // <-- (allows matching braces as in normal functions)
//              int a = glua_getInteger(1);
//              int b = glua_getInteger(2);
//
//              glua_pushNumber( a & b );
//              // <-- (no 'return' required, is calculated automatically)
//          }
//          GLUA_END
//
// Note:    The braces in the above sample are just for the "looks" - they
//          are not strictly required since the actual braces of the function
//          are within the macros. However, this makes the code look more 
//          like 'normal' C functions.
//
----------------------*/

#define _glua_function(fname) \
    static int fn_##fname(lua_State* _lua_) \
    { \
    uint _retcount_= 0; \
    uint _tbl_argn_= 0; \
	_setL(_lua_); /*for MSC*/ \
	do {	// this 'do' allows 'break' to be used to get into 'GLUA_END'

#define GLUA_FUNC( fname ) \
    _glua_function(fname)

#define GLUA_END \
	} while(0); /* 'break' within function body jumps here */ \
    /* (dummy casts make sure we never get "unused variable" warnings) */ \
    (void)_tbl_argn_; \
    return _retcount_; \
    }   // <-- matches the brace within 'glua_function' macro

#define glua_Forward( fname ) \
    _retcount_= fn_##fname( _lua_ )


/*--- Dynamic extension support ------------------------------------------*/
//
// Note: We use 'uint' instead of 'enum' because this is the interface 
//		 to dynamic modules and size of 'enum' may vary between compilers.
//
typedef uint  luafunc_u_u( uint );
typedef int   luafunc_i_si( const char*, int );
typedef void  luafunc_v_L( lua_State* );
typedef void  luafunc_v_Lb( lua_State*, bool_int );
typedef void  luafunc_v_Ld( lua_State*, double );
typedef void  luafunc_v_Ls( lua_State*, const char* );
typedef void* luafunc_V_LiVzF( lua_State*, int, const void*, size_t, lua_CFunction );
typedef void  luafunc_v_Lf( lua_State*, lua_CFunction );
typedef int   luafunc_i_L( lua_State* );
typedef int   luafunc_i_Li( lua_State*, int );
typedef bool_int luafunc_b_Li( lua_State*, int );
typedef double luafunc_d_Li( lua_State*, int );
typedef const char* luafunc_s_Li( lua_State*, int );
typedef void  luafunc_v_Li( lua_State*, int );
typedef void  luafunc_v_Lii( lua_State*, int, int );
typedef void* luafunc_V_Li( lua_State*, int );
typedef enum_uint /*e_GluaType*/ luafunc_u_Lii( lua_State*, int, int );
typedef void  luafunc_v_LisF( lua_State*, int, const char*, lua_CFunction );
typedef void  luafunc_v_Lsu( lua_State*, const char*, uint );
typedef void  luafunc_v_LFu( lua_State*, lua_CFunction, uint );
typedef uint  luafunc_u_Li( lua_State*, int );
typedef void  luafunc_v_Lis( lua_State*, int, const char* );
typedef const char* luafunc_s_LiZ( lua_State*, int, size_t* );

//-----
// Function table used by the Lua host environment to pass some of
// its functions on to GluaX extension modules.
//
// NOTE: Existing entries MAY NOT BE REMOVED OR REORDERED since doing so would
//       mess up already compiled hosts & extension modules! New entries may
//       be added at the end if more interaction with the Lua engine is needed.
//
struct s_GluaFuncs
{
	// Note: These field names are the same as actual GluaX function names.
	//       This makes static/dynamic linkage macros more simple.
	//
    luafunc_v_Ls    *_glua_error;
    luafunc_u_u     *_glua_ver;
	luafunc_u_Lii	*_glua_type2;
	luafunc_b_Li	*_glua_toboolean;
    luafunc_d_Li    *_glua_tonumber_raw;
    luafunc_s_Li    *_glua_tostring_raw;  // for backwards compatibility
    luafunc_V_Li    *_glua_touserdata;    
    luafunc_v_Lb    *_glua_pushboolean;
    luafunc_v_Ld    *_glua_pushnumber;
    luafunc_v_Ls    *_glua_pushstring;  // for backwards compatibility
    luafunc_V_LiVzF *_glua_pushuserdata;
    luafunc_v_Lf    *_glua_pushcfunction;  // for backwards compatibility
    luafunc_v_L     *_glua_newtable;    // was: glua_pushtable()
    luafunc_v_L     *_glua_totable;
    luafunc_i_L     *_glua_gettop;
    luafunc_v_Li    *_glua_settop;
    luafunc_v_Ls    *_glua_setglobal;
    luafunc_b_Li    *_glua_checkstack;
    luafunc_i_L     *_glua_newtag;

    // Added 26-Sep-02 to support 'getTable()':
    //
    luafunc_v_Li    *_glua_fromtable;
    luafunc_b_Li    *_glua_next;
	luafunc_v_Li    *_glua_remove;
	luafunc_v_Li    *_glua_pushvalue;
    
	// Added 18-Nov-02 to support multithreading:
	//
	luafunc_i_si	*_glua_exec;

    // Added 2-Dec-02:
    //
    luafunc_i_Li    *_glua_getn;
	luafunc_v_Li	*_glua_concat;

	// Added 13-Dec-02:
	//
	luafunc_v_L		*_glua_getreg;
	luafunc_v_L		*_glua_setreg;
	luafunc_v_LisF	*_glua_settagmethod;

    // Added 14-Feb-03:
    //
	// Note: Gluaport functions needed for classic Lua C API compatibility,
    //       not GluaX itself.
	//
    luafunc_i_Li    *_gluaport_strlen;
    luafunc_v_Lsu   *_glua_pushlstring;   // now replaces 'glua_pushstring()'
    luafunc_v_LFu   *_glua_pushcclosure;  // now replaces 'glua_pushcfunction()'
	luafunc_u_Li	*_gluaport_tagofuserdata;
	luafunc_v_Li	*_gluaport_rawset;
	luafunc_v_Lii	*_gluaport_rawseti;
	luafunc_v_Li    *_glua_insert;
	luafunc_v_Lii   *_gluaport_call;
	luafunc_v_Li	*_gluaport_settable;
    luafunc_v_Lii   *_gluaport_settag;
	luafunc_v_Lis	*_gluaport_settagmethod;

    // Added 16-Mar-03:
    //
    luafunc_s_LiZ   *_glua_tolstring_raw;

    //...(can be continued)...
};

#ifdef GLUA_STATIC
  // 
  // This macro is used by 'gluahost.c' (or other host application) to fill in
  // the host-side 's_GluaFuncs' structure.
  //
  #define GLUA_LINKS \
	_glua_error,\
	_glua_ver,\
	_glua_type2,\
	_glua_toboolean,\
    _glua_tonumber_raw,\
    _glua_tostring_raw,\
    _glua_touserdata,\
	_glua_pushboolean,\
    _glua_pushnumber,\
    _glua_pushstring,\
    _glua_pushuserdata,\
    _glua_pushcfunction,\
    _glua_newtable,\
    _glua_totable,\
    _glua_gettop,\
    _glua_settop,\
    _glua_setglobal,\
    _glua_checkstack,\
    _glua_newtag,\
	_glua_fromtable,\
	_glua_next,\
	_glua_remove,\
	_glua_pushvalue,\
	_glua_exec,\
    _glua_getn,\
	_glua_concat,\
	_glua_getreg,\
	_glua_setreg,\
	_glua_settagmethod,\
	\
	_gluaport_strlen,\
	_glua_pushlstring,\
	_glua_pushcclosure,\
	_gluaport_tagofuserdata,\
	_gluaport_rawset,\
	_gluaport_rawseti,\
	_glua_insert,\
	_gluaport_call,\
	_gluaport_settable,\
	_gluaport_settag,\
	_gluaport_settagmethod,\
    \
    _glua_tolstring_raw,\
  //
  #define glua_link( L, funcs_s, gluamodule_ptr, filename, tbl_idx ) \
          (*gluamodule_ptr)( L, &(funcs_s), sizeof(funcs_s), filename, tbl_idx )
  //
#endif  // static

#ifdef PLATFORM_WINDOWS
  #ifdef GLUA_DYNAMIC
    #define GLUA_EXPORT  __declspec(dllexport)
  #else
    #define GLUA_EXPORT /* nothing (static link) */
  #endif
  #define GLUA_STDCALL __stdcall
  //
#else
  #define GLUA_EXPORT   /* nothing */
  #define GLUA_STDCALL  /* nothing */
#endif


// Prototype of the module initialisation ('GluaModule()') function:
//
// NOTE: We fake the 'lua_State*' pointer as a 'void*' for CVI 5.5 compiler
//       because the ODL Type Library would otherwise have problems with it.
//       Only 'GluaModule()' is affected by this since it is the only 
//       function we need to export.
//
#ifdef _CVI_	// LabWindows/CVI 5.5 compiler
  #define LuaState_PTR void*
#else
  #define LuaState_PTR lua_State*
#endif

typedef int (GLUA_STDCALL t_GluaModuleFunc)( LuaState_PTR lua,
											 const void* v_func_tbl,
											 uint func_tbl_size,
											 const char* my_dll_path,
											 int my_tbl_index );

int _glua_create_tbl( lua_State* lua, const char* table_name );

#if ((defined _MSC_VER) && (defined GLUA_DYNAMIC))
  //
  // MSC did not like the declaration below (without it, C++ compilations 
  // do name mangling)
  //
#else
  _EXTERN_C t_GluaModuleFunc GluaModule;   // Plain 'GluaModule()' always declared
#endif

// Return values:
//       0 = success :)
//      -1 = no function table (or NULL entry)
//      -2 = requires newer host
//      -3 = data alignment error (recompile with 4-byte or smaller alignment)
//      ...(for more codes, see 'linktohost()' in 'gluax.c')...
//

// 'GLUA_MODULE()' definition may be overriden in projects where several
// modules are statically linked (to avoid function name collisions).
//
#ifdef GLUA_DYNAMIC     // Ignore 'GLUA_MODULE_NAME' when doing dynamic linking.
  #undef GLUA_MODULE_NAME
#endif

#ifndef GLUA_MODULE_NAME
  #define GLUA_MODULE_NAME GluaModule  // default
#endif

#ifndef _SET_GLUAPORT_NAMESPACE   // defined in 'gluaport.h'
  #define _SET_GLUAPORT_NAMESPACE(var)  /*nothing*/
#endif

// Note: Call the Lua handle 'lua' to be nice to static users. :)
//
#define GLUA_MODULE( path_var, tbl_var ) \
	\
    _EXTERN_C /*avoid C++ name mangling*/ \
    int GLUA_EXPORT GLUA_STDCALL \
    GLUA_MODULE_NAME( LuaState_PTR lua, \
                const void* _ftbl_, \
                uint _ftbl_size_, \
                const char* path_var, \
                int tbl_var /* >0 (table) or 0 (global namespace) */ ) \
    { \
    (void)(path_var); /*need not be used*/ \
    _SET_GLUAPORT_NAMESPACE(tbl_var); \
	{ int _tmp2_= _glua_linktohost( lua, _ftbl_, _ftbl_size_ ); \
    \
	if (_tmp2_ < 0)  /* error? */ \
        return _tmp2_;  // (function table mismatch, newer gluahost needed!)
	// ...continues with { + function body...

#define GLUA_MODULE_END \
	} _SET_GLUAPORT_NAMESPACE(0); /*namespace no longer available*/ \
    return 0; /*init successful*/ }


/*---------------------
// GLUA_DECL macros
//
// Usage:   GLUA_DECL( tbl_index )
//              { 
//              glua_func(...),
//              ...
//              glua_const(...),
//              ...
//              glua_str(...)
//              }
//          GLUA_DECL_END
//
----------------------*/

#define GLUA_DECL( tbl_index ) \
  { int _mytbl_= (tbl_index); \
    static struct s_GluaItem _items_[]=   //... (list of items follows)

  // function entries:
  #define glua_func2( lua_name, func_name ) \
    { lua_name, fn_##func_name, NULL, NULL, 0.0 }

  #define glua_func(func_name) \
    glua_func2( #func_name, func_name )
    
  // string constants:
  #define glua_str2( lua_name, str_value ) \
    { lua_name, NULL, str_value, NULL, 0.0 }

  #define glua_str(str_name) \
    glua_str2( #str_name, str_name )

  // number constants:
  #define glua_const2( lua_name, num_value ) \
    { lua_name, NULL, NULL, NULL, num_value }

  #define glua_const( var_name ) \
    glua_const2( #var_name, var_name )

  // userdata type tags:
  #define glua_tag( name, tag_ref ) \
    { name, NULL, NULL, tag_ref, 0.0 }

  #define glua_tagmethod( tag, event_str, func_name ) \
    { NULL, fn_##func_name, event_str, &tag, 0.0 }

  #define glua_info( label, value ) \
    glua_str2( "_info." label, value )

// Note: Keep 'lua' parameter named as it is (plain "lua") to allow usage
//       from both statically linked extensions and those using 'GLUA_MODULE'.
//
#define GLUA_DECL_END \
   ; \
   _glua_register2( lua, _items_, sizeof(_items_)/sizeof(*_items_), _mytbl_ );\
   }


/*--- 'glua_50.c' etc. functions (implementation details) -----------*/
//
#ifdef GLUA_STATIC
  //
  int _glua_gc_hub( lua_State* );	   // external ref to 'Glua_GcHub()'

  #ifdef LUA_V40	  // Lua 4.0 static
    //
    void* Loc40_PushUserdata( lua_State*, const void*, int, lua_CFunction );
    //
  #else     // 5.0 and later
    //
    void* Loc50_PushUserdata( lua_State*, const void*, uint, int, lua_CFunction );
    void Loc50_SetTagMethod( lua_State*, int, const char*, bool_int );
    void* Loc50_ToUserdata( lua_State* L, int argn );
    int Loc50_Tag( lua_State* L, int argn );
	double Loc50_ConvToNumber( lua_State* L, int argn );
	void Loc50_SetTag( lua_State* L, int argn, int tag );

	#ifdef LUA50_BETA   // Only needed for 5.0 beta (not final):
      int Loc50_GetN( lua_State* L, int t );
    #endif
  #endif
#endif  // GLUA_STATIC


/*--- Compatibility with Glua 1.0 apps ------------------------*/
//
// Note: There's been so many changes that this compatibility may no longer work.
//
#ifdef GLUA10_COMPATIBLE
  //
  typedef struct s_GluaItem glua_fitem;
  typedef struct s_GluaItem glua_citem;

  #define glua_regFunctions(flist)\
    _glua_register(lua, flist, (sizeof(flist) / sizeof(glua_fitem)), NULL)

  #define glua_regConstants(clist)  glua_regFunctions(clist)

  #define glua_fn(fname)            glua_func(fname)
  #define glua_function(fname)      _glua_function(fname)  
  #define glua_getNone(argn)        _glua_getDone(_lua_,argn)
  //
  #define glua_newtag(tagname)		_glua_regtag( lua, tagname )
  //
#endif  // GLUA10_COMPATIBLE

#ifdef __cplusplus
  } // extern "C"
#endif

#endif  // GLUAX_H
