/*---------------------
// GLUAX_50.C   by asko.kauppi@fi.flextronics.com
//
// Code for STATICALLY compiling GluaX using the Lua 5.0 engine (or newer).
// You do not need this code for DYNAMICALLY compiled modules.
----------------------*/

#include "gluax.h"

#ifdef GLUA_DYNAMIC
  #error "You do not need this source for dynamic GluaX compilations."
#endif

#ifdef LUA_V40
  #error "This source is for Lua 5.0 (and newer)!"
#endif

//---
// The magic word is only used for making sure we don't step on any non-gluax
// userdata's toes.. There shouldn't even be such in the system, though.
//
#define UD_MAGIC_WORD  0xa0a0

//---
// This struct will be at the beginning of any userdata that GluaX creates.
//
struct s_Userdata  // (just after the Lua-internal fields)
{
#ifdef UD_MAGIC_WORD
    uint magic;
#endif

    int tag;	// tells the 'type' of this userdata (emulates 4.0 tags)

	void* ptr;	// application-provided pointer (if any)

	lua_CFunction gc_func;	// garbage collection (object specific)

	// application-specific data block follows if 'size' mode is used.
};

#ifdef UD_MAGIC_WORD
  #define APPROVE_UD(ud)  ASSUME( (ud)->magic == UD_MAGIC_WORD )
#else
  #define APPROVE_UD(ud)  /*no check*/
#endif


//---
// Return pointer to the application's real userdata block.
//
static void* Loc_AppUserdata( struct s_Userdata* ud )
{
    if (!ud) return NULL;
	
    APPROVE_UD(ud);
	
	if (ud->ptr)
	    return ud->ptr;		// == 'vp' (caller knows it already)

    // Pointer to the Lua-allocated data just after the prelude:
	//
	ASSUME( sizeof(char) == 1 );
		
	return ((char*)ud) + sizeof(*ud);
}


//---
// Pushes a reference to the tag type's 'metatable', creating
// such if not already there.
//
static bool_int // TRUE:pushed, FALSE:none
Loc_PushMetatableRef( lua_State* L, int tag, bool_int create_if_none )
{
bool_int pushed= TRUE;

	STACK_CHECK(L);
		{
		// Get 'registry["glua_meta_xx"]':
		//
        _glua_pushstring( L, "glua_meta_" );
        _glua_pushnumber( L, tag );
        _glua_concat( L, 2 );

		lua_pushvalue( L, -1 );	 // copy of the key (eaten by 'gettable')

		lua_gettable( L, LUA_REGISTRYINDEX );
        //
        // [-2]= "glua_meta_<tag>" string
        // [-1]= registry entry (or nil)

		if (lua_istable( L, -1 ))
			lua_remove( L, -2 );	// don't need 2nd key
		else
		if (create_if_none)
			{
			// Was not yet there, create an empty table.

			lua_remove( L, -1 );	// remove the 'nil'

			// [-1]= key name to the registry 

			lua_newtable( L );	// value

			// make copy of value (2nd ref) to [-3]:
			//
			lua_pushvalue( L, -1 );
			lua_insert( L, -3 );

            // [-3]= 2nd ref to [-1] table
            // [-2]= key name
            // [-1]= empty table

			lua_settable( L, LUA_REGISTRYINDEX );	// eats key + value

			// [-1]= 2nd ref to the created empty table
			}
		else	// (!create_if_none)
			{
			lua_pop( L, 2 );	// eat 'nil' & copy of the key
								// (pushes none to the caller)
			pushed= FALSE;
			}
		}
	STACK_END( pushed ? +1:0 )

	return pushed;
}


//---
// Events:	"add"
//			"sub"
//			"mul"
//			"div"
//			"pow"
//			"unm"
//			"lt"
//			"le"		(not in lua 4.0!)
//			"concat"
//			"index"		("gettable" in lua 4.0)
//			"newindex"	("settable" in lua 4.0)
//			"call"		("function" in lua 4.0)
//			"gc"
//
// Note: To help 'gluaport', we get the function from Lua stack (-1)
//		 and (optionally) push the old value back to the stack.
//
// Note2: We should NOT convert the lua stack entries for the C functions
//		 into function pointers! The application code may be using C
//		 'closures' that attach values to the functions themselves. With
//		 conversions to C pointers and back, these would be lost, right?!?
//
void Loc50_SetTagMethod( lua_State* L, int tag, const char* event,
									   bool_int push_old_func )
{
	ASSUME( lua_type(L,-1) == LUA_TFUNCTION );

	STACK_CHECK(L)
		{
		// Creates a new metatable if not already there:
		//
		Loc_PushMetatableRef( L, tag, TRUE );

		// push the event prefixed with "__":
		lua_pushstring( L, "__" );
		lua_pushstring( L, event );		// key
		lua_concat( L, 2 );

		// [-3]= new function
		// [-2]= table ref
		// [-1]= key name

		if (push_old_func)
			{
			lua_pushvalue( L, -1 );	 // copy of key name
			lua_gettable( L, -3 );	 // pops key, pushes old func

			// [-4]= new function
			// [-3]= table ref
			// [-2]= key
			// [-1]= old func

			// Now move the old func below other stuff:
			lua_insert( L, -4 );

			// [-4]= old function
			// [-3]= new function
			// [-2]= table ref
			// [-1]= key name 
			}

		// Copy func to top of stack:
		lua_pushvalue( L, -3 );
		lua_remove( L, -4 );
	
		// ([-4]= old function)
		// [-3]= table ref
		// [-2]= key name 
		// [-1]= new function

		lua_settable( L, -3 );	// (eats key + value)

		lua_remove( L, -1 );	// remove the table reference

		// ([-1]= old function)   (if 'push_old_func'==TRUE)
		}
	STACK_END( push_old_func ? 0:-1 )
}


//---
// Attach a metatable to a given userdata or table.
//
static void Loc_SetMetaTable( lua_State* L, int rel_index /*ud|tbl*/, int tag )
{
int abs_index= STACK_ABS( L, rel_index );
int tmp;

	STACK_CHECK(L)
		{
		// 1st: Try to find metatable specifically for this tag type.
		// 2nd: Use the common ('tag 0') metatable.
		//
		if (!Loc_PushMetatableRef( L, tag, FALSE ))
			{
			Loc_PushMetatableRef( L, 0, TRUE );  // create if none
			}

		tmp= lua_setmetatable( L, abs_index );

        ASSUME( tmp != 0 );   // 0 means setting failed!
		}
	STACK_END(0)
}


//---
void*
Loc50_PushUserdata( lua_State* L, const void* ptr, uint size,
                                  int tag, lua_CFunction gc_func )
{
struct s_Userdata* ud;
		
    // We let Lua allocate its data tables for the userdata and also request 
    // space for our 'prelude' and (optional) application block.
    //
    STACK_CHECK(L)
        {
	    ud= (struct s_Userdata*)
            lua_newuserdata( L, sizeof(struct s_Userdata) + size );
        }
    STACK_END(+1)

	ASSUME_NONZERO( ud );

    #ifdef UD_MAGIC_WORD
      ud->magic= UD_MAGIC_WORD;
    #endif

	ud->tag= tag;
	ud->ptr= (size>0) ? NULL : (void*)ptr;  // dynamic pointer?
	ud->gc_func= gc_func;	// object specific cleanup func

    // Associate a metatable with the userdata. 
	//
	// First check 'registry["glua_meta_<xx>"]' where 'xx' is the
	// tag number. This allows for tag method emulation (like Lua 4.0).
	//
	// If no special metatable for this tag, use 'xx'=0 (which always
	// exists and only contains the '__gc' method pointing to 'glua_GcHub()'.
	//
	Loc_SetMetaTable( L, -1, tag );

    return Loc_AppUserdata( ud );
}

//----
void* Loc50_ToUserdata( lua_State* L, int argn )
{
struct s_Userdata* ud;
	
    ud= (struct s_Userdata*)lua_touserdata( L, argn );

    if (ud) APPROVE_UD(ud);

	return Loc_AppUserdata(ud);
}


//---
int Loc50_Tag( lua_State* L, int argn )
{
struct s_Userdata* ud;

	ud= (struct s_Userdata*)lua_touserdata( L, argn );

	if (!ud) return 0;		// No userdata, no tag!

    APPROVE_UD(ud);

	return ud->tag;		// We placed it there at creation
}


//---
// Modify '[argn]' in location, converting numeric strings to
// actual numbers (and leaving other strings alone).
//
double Loc50_ConvToNumber( lua_State* L, int argn )
{
int argn_abs= STACK_ABS(L,argn);
double val= 0.0;	// default (for non-convertible types)

	STACK_CHECK(L)
		{
		switch( lua_type(L, argn) )
			{
			case LUA_TSTRING:
                {
                char* end;

                strtod( lua_tostring(L,argn), &end );

                if (*end == '\0')   // was numeric?
					{
					val= lua_tonumber(L,argn);

					lua_pushnumber( L, val );
					lua_replace(L, argn_abs );
					}
                }
				break;

			case LUA_TNUMBER:
				// No conversion necessary, already a number.
				val= lua_tonumber(L,argn);
				break;

			default:	// Non-convertible types (let them be)
				break;
			}
		}
	STACK_END(0)

	return val;
}


//---
void Loc50_SetTag( lua_State* L, int argn, int tag )
{

    switch( lua_type(L,argn) )
        {
        case LUA_TTABLE:
            Loc_SetMetaTable( L, argn, tag );
            break;
        
        case LUA_TLIGHTUSERDATA:
        case LUA_TUSERDATA:
            {
            struct s_Userdata* ud=
                (struct s_Userdata*)lua_touserdata( L, argn );

            if (ud)
                {
                APPROVE_UD(ud);

                Loc_SetMetaTable( L, argn, tag );

                ud->tag= tag;
                }
            }
            break;
        
        default:
            ASSUME(0);  // Cannot set tag for other types!
            break;
        }
}


//---
// REMOVE THIS IN LUA 5.0 FINAL (which has 'lua_getn()')
//
#ifdef LUA50_BETA
  static void Loc_GetTableItem( lua_State* L, const char* tbl_name,
                                            const char* item_name )
  {
	STACK_CHECK(L)
		{
		lua_getglobal( L, tbl_name );

		if (lua_istable(L,-1))
			{
			lua_pushstring( L, item_name );
			lua_gettable( L, -2 );	// pops key, pushes value (or 'nil')

			lua_remove( L, -2 );	// remove the table ref
			}
        else 
        if (!lua_isnil(L,-1))
            {
            ASSUME_NEVER(); // Other than table or 'nil'.
            }
		}
	STACK_END(+1)
  }
  //
  int Loc50_GetN( lua_State* L, int index )
  {
  int abs_index= STACK_ABS( L, index );
  int val;

	STACK_CHECK(L)
		{
		// Let's try calling "table.getn()" Lua function via the API.
		// (there is no direct C linkage to 'aux_getn()' in 5.0 beta!)
		//
		Loc_GetTableItem( L, "table", "getn" );

		ASSUME( lua_iscfunction(L,-1) );

		// Push the parameter (rather - make a copy of it)
		//
		lua_pushvalue( L, abs_index );

		// Make the call
		//
		lua_call( L, 1, 1 );	// eats function ref + params

		// Should have pushed the result (int) to stack.

		val= (int)lua_tonumber( L, -1 );

		lua_remove( L, -1 );	// take away the returned int
		}
	STACK_END(0)

	return val;
  }
#endif
