/*
* This 'loadlib.c' has been modified by Adrian Perez & Asko Kauppi to
* contain native OS X support (without the need of the 'dlcompat' library)
* and to allow usage of the dynamic link functions at the C level.
*
* typedef void* h_LoadLib;
*
* h_LoadLib loadlib_open( const char* path );
* void loadlib_close( h_LoadLib handle );
* void* loadlib_func( h_LoadLib handle, const char* name );
* const char* loadlib_error( char* buf, size_t bufsize );
*
* Defines:
*   -DUSE_LOADLIB=0     disables the dynamic linkage feature
*   -DUSE_LOADLIB=1     enabled (default)
*   -DUSE_LOADLIB=2     link in just functions, not the Lua glue
*/
#include "loadlib.h"

/*
** $Id: loadlib.c,v 1.1 2003/12/22 15:29:27 tulrich Exp $
** Bare-bones dynamic library loader for Lua
** See Copyright Notice in lua.h
*
* This  Lua library  exports a  single function,  called loadlib,  which is
* called from Lua  as loadlib(lib,init), where lib is the  full name of the
* library to be  loaded (including the complete path) and  init is the name
* of a function  to be called after the library  is loaded. Typically, this
* function will register other functions,  thus making the complete library
* available  to Lua.  The init  function is  *not* automatically  called by
* loadlib. Instead,  loadlib returns  the init function  as a  Lua function
* that the client  can call when it  thinks is appropriate. In  the case of
* errors,  loadlib  returns  nil  and two  strings  describing  the  error.
* The  first string  is  supplied by  the operating  system;  it should  be
* informative and useful  for error messages. The second  string is "open",
* "init", or  "absent" to identify  the error and is  meant to be  used for
* making  decisions without  having to  look into  the first  string (whose
* format is system-dependent).
*
* This bare-bones loadlib  function is supposed to be used  as a foundation
* for  more sophisticated  dynamic library  loaders, possibly  still called
* loadlib and probably written in Lua, that  will be smart enough to try to
* find the library in different directories and also perhaps guess the name
* of the init function.
*
* This module contains  an implementation of loadlib for  Unix systems that
* have dlfcn, an implementation for Windows,  and a stub for other systems.
* See  the list  at  the end  of  this  file for  some  links to  available
* implementations of dlfcn  and interfaces to other  native dynamic loaders
* on top of which loadlib could be implemented.
*
*/
#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"

#ifndef USE_LOADLIB
#define USE_LOADLIB 1
#endif

#ifndef USE_DLOPEN
#define USE_DLOPEN 0
#endif

/* =======================================================
**   Dynamic linkage (platform specific)
** =======================================================
*/
#if USE_LOADLIB

/*--------------------- Linux & most Unixes (dlfcn) --------------------*/

#if defined(linux) || defined(sun) || defined(sgi) || defined(BSD) || \
	defined(__FreeBSD__) || defined(__NetBSD__) || defined(bsdi) || \
	defined(__OpenBSD__) || USE_DLOPEN
#define LOADLIB
/*
* The dlfcn interface is available in Linux, SunOS, Solaris, IRIX, FreeBSD,
* NetBSD, AIX 4.2, HPUX 11, and  probably most other Unix flavors, at least
* as an emulation layer on top of native functions.
*/

# include <dlfcn.h>
# include <stdlib.h>

/* OpenBSD apparently doesn't define RTLD_NOW */
# ifndef RTLD_NOW
#  define RTLD_NOW RTLD_LAZY
# endif

  LOADLIB_API h_LoadLib loadlib_open( const char* path )    
    { return dlopen( path, RTLD_NOW /*|RTLD_LOCAL*/ ); }

  LOADLIB_API void loadlib_close( h_LoadLib handle )        
    { dlclose( handle ); }

  LOADLIB_API void* loadlib_func( h_LoadLib handle, const char* name )
  {
    void* result = dlsym( handle, name );
    if (result == NULL) {
      /* a.out systems (e.g. OpenBSD) have another problem where dlsym
       * requires us to prepend "_" when looking for a symbol.  So if our
       * first attempt failed, try again with a pre-pending "_". */
      char* buffer = malloc(strlen(name) + 2);
      if (buffer == NULL) return NULL;
      strcpy(buffer, "_");
      strcat(buffer, name);
      result = dlsym(handle, buffer);
      free(buffer);
    }
    return result;
  }

  LOADLIB_API const char* loadlib_error( char* buf, size_t bufsize )
    { return dlerror(); }

/*--------------------- MS Windows --------------------*/

#elif defined(_WIN32) || defined(__WIN32__) || defined(WIN32)
#define LOADLIB

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

  LOADLIB_API h_LoadLib loadlib_open( const char* path )	
    { return /*HINSTANCE*/ LoadLibrary(path); }

  LOADLIB_API void loadlib_close( h_LoadLib handle )		
    { FreeLibrary( (HINSTANCE)handle ); }

  LOADLIB_API void* loadlib_func( h_LoadLib handle, const char* name )
    { return GetProcAddress( handle, name ); }

  LOADLIB_API const char* loadlib_error( char* buf, size_t bufsize )
    {
    static char buf2[200];   /* (in case 'buf'==NULL) */
    int error= GetLastError();

    if ((!buf) || (bufsize<100))
        buf= buf2, bufsize= sizeof(buf2);
    
    if ( FormatMessage(FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
                        0, error, 0, buf, bufsize, 0 ) )
        return buf;

    sprintf( buf, "system error %d\n", error );

    return buf;
    }

/*--------------------- Apple OS X / Darwin --------------------*/

#elif defined(__APPLE__) && defined(__MACH__)
#define LOADLIB

/*
 * Small (and fast) wrapper around the Dyld API to dynamically load
 * modules. Due to limitations in OSX's public APIs, only modules
 * with a MH_BUNDLE header can be loaded; you can create one of
 * these by passing the '-bundle' option to Apple's GCC when linking
 * the module.
 *
 * At this moment, this module loader does nothing more than providing
 * a clearer API to the obscure one that Dyld includes (splitted over
 * three different sets of functions, argh!).
 *
 * 2003-11-26 AP: Improved now that Apple has published detailed docs
 *		on the Dyld API (previous version is comented out, and was made
 *		by reading the "dyld.h" header and guessing what each funtion
 *		does: it worked, but the new approach is the right one).
 */
#include <mach-o/dyld.h>
#include <string.h>   /* strlen() */
  
#if 0
	/* Not needed anymore (see below). */
  struct s_LoadLib
  {
    NSModule module;
		NSObjectFileImage image;
  };
#endif

  LOADLIB_API h_LoadLib loadlib_open( const char *path )
  {
  	NSObjectFileImageReturnCode ret;
  	NSObjectFileImage nsimg;
  	NSModule nsmod;
#if 0
  	struct s_LoadLib* h; /* Not needed anymore. */
#endif

    if (!path) return NULL;

		/* Load image from file:
     */
		ret = NSCreateObjectFileImageFromFile(path, &nsimg);

    if (ret != NSObjectFileImageSuccess)
        return NULL;    /* Unable to load! */
	
		/* Link into the rest of the executable:
     */
		/* According to Apple's documentation, we can destroy the object image
		 * after linking it: although we instruct Dyld to do that, Dyld will
		 * keep it in memory as long as it is linked in the executable (i.e:
		 * until NSUnLinkModule() is called). 
		 *
		 * This way wa can just return the NSModule after releasing the image,
		 * instead of having the s_LoadLib struct (We only need the NSModule
		 * later in order to unload the module).
		 */
		nsmod = NSLinkModule(nsimg, path,
			NSLINKMODULE_OPTION_PRIVATE | NSLINKMODULE_OPTION_RETURN_ON_ERROR);
		NSDestroyObjectFileImage(nsimg);
		if (!nsmod)
			return NULL;    /* Unable to link! */

    /* Success :)
     */	
#if 0
		h = malloc( sizeof(struct s_LoadLib) );

		h->image  = nsimg;
		h->module = nsmod;

		return (h_LoadLib)h; 
#endif

		return (h_LoadLib) nsmod;
  }
  
  LOADLIB_API void loadlib_close( h_LoadLib handle )
  {
#if 0
		/* This is old (and odd), unneeded code. */
 		struct s_LoadLib* h= (struct s_LoadLib*)handle;
	
		NSUnLinkModule(h->module, NSUNLINKMODULE_OPTION_NONE);
		NSDestroyObjectFileImage(h->image);
    free(h);
#endif
		NSUnLinkModule((NSModule) handle,
			NSUNLINKMODULE_OPTION_NONE);
  }
  
  LOADLIB_API void* loadlib_func( h_LoadLib handle, const char *name )
  {
#if 0
  	struct s_LoadLib* h= (struct s_LoadLib*)handle; /* Not needed anymore. */
#endif
  	NSSymbol nss;
  	char real_name[ strlen(name) + 2 ]; 	/* '_'+name */

    real_name[0]='_';
    strcpy( real_name+1, name );
    
#if 0
		nss = NSLookupSymbolInModule(h->module, real_name);
#endif
		nss = NSLookupSymbolInModule((NSModule) handle, real_name);

    return (nss) ? NSAddressOfSymbol(nss) : NULL;
  }
  
  LOADLIB_API const char* loadlib_error( char* buf, size_t bufsize )
  {
    return "loadlib can't load bundle";    /* no explanations for now */
  }

/*--------------------- Others.. --------------------*/

#else
/*
* write an implementation for your system  here and send it to us, together
* with preprocessing symbols that identify your system.
*/

#endif  /* platforms */
#endif  /* USE_LOADLIB */


/* =======================================================
**   Lua interface
** =======================================================
*/
#if (USE_LOADLIB < 2)
static int loadlib(lua_State *L)
{
#ifdef LOADLIB
  const char *path=luaL_checkstring(L,1);
  const char *init=luaL_checkstring(L,2);
  h_LoadLib lib=loadlib_open(path);
  char buf[128];
  
    if (lib!=NULL)
        {
        lua_CFunction f=(lua_CFunction) loadlib_func(lib,init);
        if (f!=NULL)
            {
            lua_pushlightuserdata(L,lib);
            lua_pushcclosure(L,f,1);
            return 1;
            }
        }

    lua_pushnil(L);
    lua_pushstring(L,loadlib_error(buf,sizeof(buf)));
    lua_pushstring(L,(lib) ? "init" : "open");
    if (lib) loadlib_close(lib);
    
#else	/* Fallback for other systems */
    lua_pushnil(L);
    lua_pushliteral(L,"`loadlib' not supported");
    lua_pushliteral(L,"absent");
#endif
    return 3;
}

LUALIB_API int luaopen_loadlib (lua_State *L)
{
 lua_register(L,"loadlib",loadlib);
 return 0;
}
#endif  /* USE_LOADLIB < 2 */

/*
* Here are some links to available implementations of dlfcn and
* interfaces to other native dynamic loaders on top of which loadlib
* could be implemented. Please send contributions and corrections to us.
*
* AIX
* Starting with AIX 4.2, dlfcn is included in the base OS.
* There is also an emulation package available.
* http://www.faqs.org/faqs/aix-faq/part4/section-21.html
*
* HPUX 
* HPUX 11 has dlfcn. For HPUX 10 use shl_*.
* http://www.geda.seul.org/mailinglist/geda-dev37/msg00094.html
* http://www.stat.umn.edu/~luke/xls/projects/dlbasics/dlbasics.html
*
* Macintosh, Windows
* http://www.stat.umn.edu/~luke/xls/projects/dlbasics/dlbasics.html
*
* GLIB has wrapper code for BeOS, OS2, Unix and Windows
* http://cvs.gnome.org/lxr/source/glib/gmodule/
*
*/
