/*
* lmd5.c
* MD5 library for Lua 5.0 based on Rivest's API
* Luiz Henrique de Figueiredo <lhf@tecgraf.puc-rio.br>
* 29 Aug 2003 08:03:44
* This code is hereby placed in the public domain.
*/

/* luacheia */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>

#ifdef USE_DEUTSCH
#define AUTHOR "Deutsch"
#define MD5_CTX			md5_state_t
#define MD5Init(c)		md5_init(c)
#define MD5Update(c,s,n)	md5_append(c,(const md5_byte_t *)s,n)
#define MD5Final(digest,c)	md5_finish(c,digest)
#else
#ifdef USE_PLUMB
#define AUTHOR "Plumb"
#else
#define AUTHOR "Rivest"
#include "md5global.h"
#endif
#endif
#include "md5.h"

#include "lua.h"
#include "lauxlib.h"

#define MYNAME		"md5"
#define MYVERSION	MYNAME " library for " LUA_VERSION " / Aug 2003 / "\
			"using " AUTHOR "'s library"
#define MYTYPE		MYNAME " context"

static MD5_CTX *getcontext(lua_State *L, int i)
{
 if (luaL_checkudata(L,i,MYTYPE)==NULL) luaL_typerror(L,i,MYTYPE);
 return lua_touserdata(L,i);
}

static int Lnew(lua_State *L)			/** new() */
{
 void *p=lua_newuserdata(L,sizeof(MD5_CTX));
 MD5_CTX *c=p;
 MD5Init(c);
 luaL_getmetatable(L,MYTYPE);
 lua_setmetatable(L,-2);
 return 1;
}

static int Lreset(lua_State *L)			/** reset(c) */
{
 MD5_CTX *c=getcontext(L,1);
 MD5Init(c);
 return 0;
}

static int Lupdate(lua_State *L)		/** update(c,s) */
{
 MD5_CTX *c=getcontext(L,1);
 const char *s=luaL_checkstring(L,2);
 MD5Update(c,s,lua_strlen(L,2));
 return 0;
}

static int Ldigest(lua_State *L)		/** digest(c or s,[raw]) */
{
 unsigned char digest[16];
 if (lua_isuserdata(L,1))
 {
  MD5_CTX *c=getcontext(L,1);
  MD5Final(digest,c);
 }
 else
 {
  const char *s=luaL_checkstring(L,1);
  MD5_CTX c;
  MD5Init(&c);
  MD5Update(&c,s,lua_strlen(L,1));
  MD5Final(digest,&c);
 }
 if (lua_toboolean(L,2))
  lua_pushlstring(L,digest,sizeof(digest));
 else
 {
  char hex[2*16+1];
  int i;
  for (i=0; i<16; i++) sprintf(hex+2*i,"%02x",digest[i]);
  lua_pushlstring(L,hex,sizeof(hex)-1);
 }
 return 1;
}

static int Ltostring(lua_State *L)		/** tostring(c) */
{
 MD5_CTX *c=getcontext(L,1);
 char s[64];
 sprintf(s,MYTYPE " %p",c);
 lua_pushstring(L,s);
 return 1;
}

static const luaL_reg R[] =
{
	{ "digest",	Ldigest	},
	{ "new",	Lnew	},
	{ "reset",	Lreset	},
	{ "tostring",	Ltostring},
	{ "update",	Lupdate	},
	{ NULL,		NULL	}
};

LUAMODULE_API int luaopen_md5(lua_State *L)
{
 luaL_openlib(L,MYNAME,R,0);
 lua_pushliteral(L,"version");			/** version */
 lua_pushliteral(L,MYVERSION);
 lua_settable(L,-3);
 lua_pushliteral(L,"metatable");		/** metatable */
 luaL_newmetatable(L,MYTYPE);
 lua_pushliteral(L,"__index");
 lua_pushvalue(L,-4);
 lua_settable(L,-3);
 lua_pushliteral(L,"__tostring");
 lua_pushcfunction(L,Ltostring);
 lua_settable(L,-3);
 lua_settable(L,-3);
 return 1;
}


LUAMODULE_API const char* luaLM_version()
{
 return LUA_VERSION;
}


LUAMODULE_API int luaLM_import(lua_State* L)
{
	return luaopen_md5(L);
}

