/*========================================================*/
/* PCRE (Perl compatible regexp) for Lua 5                */
/* Martin Spernau, martin at traumwind dot de             */
/* 2003-03-02                                             */
/*========================================================*/
/* PCRE copyright:                                        */
/* Regular expression support is provided by the          */
/* PCRE library package, which is open source             */
/* software, written by Philip Hazel, and copyright       */
/* by the University of Cambridge, England.               */
/* ftp://ftp.csx.cam.ac.uk/pub/software/programming/pcre/ */
/*========================================================*/
/* this Lua extension was built to be used                */
/* with the 'loadmodule' by Thatcher Ulrich               */
/* http://lua-users.org/wiki/LuaBinaryModules             */
/*========================================================*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifndef LUAMODULE_API
# if defined(_WIN32) || defined(WIN32)
#  define LUAMODULE_API __declspec(dllexport)
# else
#  define LUAMODULE_API
# endif
#endif

#ifdef __cplusplus
extern "C" {
#endif
    
#include <stdlib.h>

#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>

    
#ifdef __cplusplus
extern "C" {
#endif
int lpcre_initialize(lua_State *L);
#ifdef __cplusplus
}
#endif

#include "pcre.h"
#include "internal.h"


static int wrap_pcre_version(lua_State *L)
{
  lua_pushstring(L, pcre_version());
  return 1;
}

static int wrap_pcre_match(lua_State *L)
{
        const char * _regex;
        const char * _searchstr;
        pcre * _re;
        const char * _error;
        int _erroffset;
        int _ovectorsize = 30;
        int* _ovector = malloc(sizeof(int) * (_ovectorsize));
        int _rc, _i;
        char *substring_start;
        int substring_length;
        
        // get the regex string (first arg)
        luaL_checkstring(L,1);
        _regex = (char*)lua_tostring(L, 1);
        
        // get the string to search in
        luaL_checkstring(L,2);
        _searchstr = (char*)lua_tostring(L, 2);
        
        /* Compile the regular expression in the first argument */
        _re = pcre_compile(
          _regex,              /* the pattern */
          0,                   /* default options */
          &_error,             /* for error message */
          &_erroffset,         /* for error offset */
          NULL);               /* use default character tables */
        
        /* error checking */
        if (_re == NULL)
        /* Compilation failed: print the error message and exit */
        {
          lua_pushfstring (L, 
            "PCRE compilation failed at offset %d: %s", 
            _erroffset, _error);
          lua_error(L);
        }

        /* Compilation succeeded: match the subject in the second argument */
        do
        {
          _rc = pcre_exec(
          _re,                     /* the compiled pattern */
          NULL,                    /* no extra data - we didn't study the pattern */
          _searchstr,              /* the subject string */
          (int)strlen(_searchstr), /* the length of the subject */
          0,                       /* start at offset 0 in the subject */
          0,                       /* default options */
          _ovector,                /* output vector for substring information */
          _ovectorsize);           /* number of elements in the output vector */
          
          if (_rc == 0)
          {
            _ovectorsize *= 2;
            _ovector = realloc(_ovector, sizeof(int)*_ovectorsize);
          }  
          /* match succeded, but no room for submatches */
          /* -> we grow the ovector and try again */
        } 
        while (_rc == 0);
        
        /* match failed report */
        if (_rc < 0) 
        {
          if (_rc == -1) 
          { /* this is the internal PCRE error */
            lua_pushnil(L); /* failed */
            lua_pushstring(L, "no match");
            return 2;
          }
          else 
          {
            lua_pushnil(L); /* failed */
            lua_pushfstring (L, 
              "PCRE error #%d", 
              _rc); /* report actual error_num */
            return 2;
          }
        }

        lua_pushnumber(L, _rc);
        /* match succeded, return numer of submatches */
        
        lua_newtable(L); 
        /* table for submatches */
        lua_newtable(L); 
        /* table for extrainfo */        
        
		for (_i = 0; _i < _rc; _i++)
		{
			substring_start = (char*)(_searchstr + _ovector[2 * _i]);
			substring_length = _ovector[2 * _i + 1] - _ovector[2*_i];

			lua_pushnumber(L, _i+1); /* _i+1 as we are dealing with Lua */
            /* key = index of submatch */
			lua_pushlstring(L, substring_start, substring_length);
			/* value = the actual substring match */
            lua_settable (L, -4); /* put into submatch table !! */
			/* put it into table */

            /* extra info goes into second table */
            lua_pushnumber(L, _i+1);
            /* key = index of submatch */
            lua_newtable(L);
            /* value is a table with info */
            {
              /* start pos of match */
              lua_pushstring(L, "start");
              lua_pushnumber(L, _ovector[2*_i]+1); 
              /* we return Lua-style, counting from 1, not 0 */
              lua_settable(L,-3);
              
              /* length of match */
              lua_pushstring(L, "length");
              lua_pushnumber(L, substring_length);
              lua_settable(L,-3);
			}
            lua_settable (L, -3); /* put into submatch table !! */
			/* put it into table */
			
		}
		
		pcre_free(_re);
		/* free compiled regex */
		
		return 3; 
		/* number of submatches, submatch table and info table */
}

static const struct luaL_reg pcrelib[] = {
  {"match",    wrap_pcre_match},
  {"version", wrap_pcre_version},
  { NULL, NULL }
  };

int lpcre_initialize(lua_State *L)
{
  luaL_openlib (L, "pcre", pcrelib, 0);
  return 0;
} 

/*=========================================*/
/* now comes the special 'loadmodule' code */
/*=========================================*/

LUAMODULE_API
int	luaLM_import(lua_State* L)
{
	return lpcre_initialize(L);
}
/* Lua version. */
LUAMODULE_API
const char* luaLM_version(void)
{
	return LUA_VERSION;
}

