/////////////////////////////////////////////////////////////////////////////
// Purpose:       Stuff to setup wxLua in a lua_State
// Author:        J Winwood
// Created:       14/11/2001
// Modifications: Type checking added when accessing user data items
//                Support for wxWindows 2.3
//                Thanks to Francis Irving for separating out into
//                Library and Standalone modules.
//                The number of non-wxObject class encapsulations was getting
//                out of hand! Reimplemented using preprocessor macros.
// Copyright:     (c) 2001-2002 Lomtick Software. All rights reserved.
// Licence:       wxWindows licence
/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
    #pragma implementation "wxLua.cpp"
    #pragma interface "wxLua.cpp"
#endif

#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "wx/datetime.h"
    
#include "wxLuaInternals.h"
#include "wxLuaCallback.h"
#include "wxLuaHtmlWindow.h"

#ifdef FindText
# undef FindText
#endif

#if defined(__WXGTK__)
# include "../Import/wxLuaWrapGTK.hpp"
#elif  defined(__WXMSW__)
# include "../Import/wxLuaWrapWin.hpp"
#endif

#if defined(__WXGTK__) || defined(__WXMOTIF__)
    #include "../wxLua.xpm"
#endif

bool		 s_typesRegistered = false;
int          g_startTag;            // The first and last wxLua-allocated lua tag
int          g_lastTag;
int          g_functionTag;         // The lua tag for function objects.
int          s_wxLuaTable;          // The lua tag for the wxLua table
int          s_wxNull;              // The lua tag for wxNull (used for NULL pointers

wxHashTable *g_pDerivedList    = NULL; // The list of derived objects.
wxHashTable *g_pTrackedList    = NULL; // The list of tracked objects.
wxList		*g_pAppHandlerList = NULL; // The list of event objects associated with the application object

static int LUACALL LuaTableErrorHandler(lua_State *L)
{
    terror(L, "Cannot modify read-only wxLua table");
    return 0;
}

static const char wxLuaReferences[]     = "wxLuaReferences";
static const char wxNull[]              = "wxNull";

// Register (and re-register if required) all the types, classes and objects
// contained within the generated binding
void LUACALL RegisterWxLua(lua_State *L, bool registerTypes)
{
	if (!registerTypes && !s_typesRegistered)
		terror(L, "wxLua: First time registration must register types");

	// create references table in registry
    tpushliteralstring(L, wxLuaReferences);
    lua_newtable(L);
    lua_rawset(L, LUA_REGISTRYINDEX);
    
    // create a tag for the wxLua table
	if (registerTypes)
        s_wxLuaTable = tnewtag(L);
        
    // create the wxLua table
    lua_newtable(L);
    int tableOffset = lua_gettop(L);
    
    // set the table tag
    tsettag (L, s_wxLuaTable);    
    tsettagmethod (L, s_wxLuaTable, "__newindex", LuaTableErrorHandler);
        
    // register all out classes etc. in the wxLua table
	registerGeneratedClasses(L, tableOffset, registerTypes);

	// register wxNull
	if (registerTypes)
        s_wxNull = tnewtag(L);    

    tpushliteralstring(L, wxNull);
    tpushusertag(L, NULL, s_wxNull);
    lua_rawset(L, tableOffset);
        
	registerOthers(L, registerTypes);

	// luacheia: put in a re-writable entry for "wx.go", to be
	// registered as a function by our loader.
	lua_pushstring(L, "go");
	lua_pushnumber(L, 0);
	lua_rawset(L, tableOffset);

	// luacheia: save the original print function.
	lua_pushstring(L, "_old_print");
	lua_getglobal(L, "print");
	lua_rawset(L, tableOffset);
    
    // create a global called 'wx'
    lua_pushliteral(L, "wx");
    lua_pushvalue(L, tableOffset);
    lua_rawset(L, LUA_GLOBALSINDEX);
    lua_pop(L, 1);
    
	s_typesRegistered = true;
    
    wxIcon *pIcon = new wxIcon(wxICON(LUA));
    addToTrackedMemoryList(pIcon);
    pushuserdatatype(L, s_wxIcon, pIcon);   
    lua_setglobal(L, "wxLuaEditorIcon");
}


// Free up the tracked resources list.
void LUACALL CleanupWxLua(lua_State *L, bool closeLua /* = true*/)
{
    if (g_pDerivedList != NULL)
    {
		g_pDerivedList->DeleteContents(true);
        delete g_pDerivedList;
        g_pDerivedList = NULL;
    }

    if (g_pTrackedList != NULL)
    {
		g_pTrackedList->DeleteContents(true);
        delete g_pTrackedList;
        g_pTrackedList = NULL;
    }

    if (closeLua)
    {
        tpushliteralstring(L, wxLuaReferences);
        lua_pushnil(L);
        lua_rawset(L, LUA_REGISTRYINDEX);
        
        lua_close(L);
    }
    
    // Handle the special case of event handlers attached to the application object
    if (g_pAppHandlerList != NULL)
    {
        wxNode *pNode = g_pAppHandlerList->GetFirst();
        // For all nodes in the list
        while (pNode != NULL)
        {
            LuaCallback *pCallback = (LuaCallback *) pNode->GetData();
            // Flag that Lua has gone and there is no
            // point removing the reference to the Lua function to call
            // this also stops the Lua function from being called.
           if (pCallback != NULL)
                pCallback->ClearState();
            pNode = pNode->GetNext();
        }

        delete g_pAppHandlerList;
        g_pAppHandlerList = NULL;
    }
}

// The SmartStringArray provides an automatic destructor for 
// character string arrays. Used to make the binding easier to
// generated
SmartStringArray::SmartStringArray() 
{ 
    m_choices = NULL; 
}
    
SmartStringArray::~SmartStringArray() 
{ 
    if (m_choices != NULL) 
        delete[] m_choices;  
}
    
void SmartStringArray::operator =(wxString *choices) 
{ 
    m_choices = choices; 
}

// The SmartStringArray provides an automatic destructor for 
// integer arrays. Used to make the binding easier to
// generate.
SmartIntArray::SmartIntArray() 
{ 
    m_choices = NULL; 
}
    
SmartIntArray::~SmartIntArray() 
{ 
    if (m_choices != NULL) 
        delete[] m_choices;  
}
    
void SmartIntArray::operator =(int *choices) 
{ 
    m_choices = choices; 
}

int LUACALL tinsert(lua_State *L, int iParam)
{
    if (lua_isnil(L, iParam))
    {
        lua_pop(L, 1);
        return LUA_REFNIL;
    }
    
    int iCount;
    
    lua_pushvalue(L, iParam);                // get value to store
    int nTop = lua_gettop(L);

    tpushliteralstring(L, wxLuaReferences); 
    lua_rawget(L, LUA_REGISTRYINDEX);        // pop key, push result

    lua_pushliteral(L, "n");                 // push 'n' as key
    lua_rawget(L, -2);                       // pop key, push result
    if (lua_isnil(L, -1))                    // not used before?
        iCount = 0;                          // 
    else
        iCount = (int) lua_tonumber(L, -1);  // get result value           
    lua_pop(L, 1);                           // pop result
    
    iCount++;                                // next unallocated index

    lua_pushvalue(L, nTop);                  // push value to store
    lua_rawseti(L, -2, iCount);              // store value, pop value

    lua_pushliteral(L, "n");                 // push key
    lua_pushnumber(L, iCount);               // push value
    lua_rawset(L, -3);                       // pop key, value

    lua_pop(L, 3);                           // restore stack

    return iCount;
}

bool LUACALL tget(lua_State *L, int iIndex)
{
    if (iIndex == LUA_REFNIL)
    {
        lua_pushnil(L);
        return true;
    }
    
    int  iCount;
    bool ret = false;                        // false indicates index out of range

    tpushliteralstring(L, wxLuaReferences); 
    lua_rawget(L, LUA_REGISTRYINDEX);        // pop key, push result (refs table)

    lua_pushliteral(L, "n");                 // push 'n' as key
    lua_rawget(L, -2);                       // pop key, push result
    iCount = (int) lua_tonumber(L, -1);      // get result value           
    lua_pop(L, 1);                           // pop result

    if (iIndex > 0 && iIndex <= iCount)      // ensure in range
    {
        lua_rawgeti(L, -1, iIndex);          // push result
        ret = true;
    }
    else
        lua_pushnil(L);                      // push result

    lua_remove(L, -2);                       // balance stack

    return ret;
}

bool LUACALL tremove(lua_State *L, int iIndex)
{    
    if (iIndex == LUA_REFNIL)
        return true;
    
    int  iCount;
    bool ret = false;						 // false indicates index out of range

    tpushliteralstring(L, wxLuaReferences); 
    lua_rawget(L, LUA_REGISTRYINDEX);        // pop key, push result (refs table)

    lua_pushliteral(L, "n");				 // push 'n' as key
    lua_rawget(L, -2);						 // pop key, push result
    iCount = (int) lua_tonumber(L, -1);		 // get result value           
    lua_pop(L, 1);							 // pop result

    if (iIndex > 0 && iIndex <= iCount)		 // ensure in range
    {
        lua_pushnil(L);						 // push nil as value
        lua_rawseti(L, -2, iIndex);			 // set table, pop value
        ret = true;
    }
    lua_pop(L, 1);							 // clean up stack
    return ret;
}

int LUACALL tgetn (lua_State *L) 
{    
    tpushliteralstring(L, wxLuaReferences); 
    lua_rawget(L, LUA_REGISTRYINDEX);        // pop key, push result (refs table)
    
    lua_pushliteral(L, "n");
    lua_rawget(L, -2);
    
    int n;
    if (lua_isnumber(L, -1))          
    {
        n = (int) lua_tonumber(L, -1);
    }
    else
    {  
        /* must count elements */
        n = 0;
        for (;;) 
        {
            lua_rawgeti(L, -2, ++n);
            bool isNil = lua_isnil(L, -1);
            lua_pop(L, 1);
            if (isNil)
                break;
        }
        --n;
    }
    lua_pop(L, 2);
    return n;
}

void LUACALL terror (lua_State *L, const char *errorMsg) 
{    
    lua_pushstring(L, errorMsg);
    lua_error(L);
}

void LUACALL tpushusertag(lua_State *L, const void *u, int tag)
{
    const void **ptr = (const void **) lua_newuserdata(L, sizeof(void *));
    if (ptr != NULL)
    {
        *ptr = u;
        if (tag != TLUA_NOTAG && tget(L, tag))
        {
            if (lua_setmetatable(L, -2) == 0)
                terror(L, "wxLua: Unable to set metatable");
        }
    }
    else
        terror(L, "wxLua: Out of memory");
}

void LUACALL tsettag(lua_State *L, int tag)
{
    if (tag != TLUA_NOTAG && tget(L, tag))
    {
        if (!lua_setmetatable(L, -2))
            terror(L, "wxLua: Unable to set metatable");
    }
}

int LUACALL ttag (lua_State *L, int index)
{
    int tag = TLUA_NOTAG;
    if (lua_getmetatable(L, index) != 0)
    {
        lua_pushliteral(L, "tag");
        lua_rawget(L, -2);
        if (lua_isnumber(L, -1))
        {
            tag = (int) lua_tonumber(L, -1);
        }
        lua_pop(L, 2);           
    }
    return tag;
}

void * LUACALL ttouserdata (lua_State *L, int index, bool reset /* = false*/)
{
    void *pdata = NULL;
    void **ptr = (void **) lua_touserdata(L, index);
    if (ptr != NULL)
    {
        pdata = *ptr;
        if (reset)
            *ptr = NULL;
    }
    return pdata;
}

int LUACALL tnewtag(lua_State *L)
{
    lua_newtable(L);
    lua_pushvalue(L, -1);
    int tag = tinsert(L, -1);
    lua_pushliteral(L, "tag");
    lua_pushnumber(L, tag);
    lua_rawset(L, -3);
    lua_pop(L, 1);
    return tag;
}

bool LUACALL tsettagmethod(lua_State *L, int tag, const char *method, lua_CFunction func, void *pClass /* = NULL*/)
{    
    if (tag != TLUA_NOTAG && tget(L, tag))
    {
        if (pClass != NULL)
        {
            lua_pushstring(L, method);
            lua_pushlightuserdata(L,  pClass);
            lua_pushcclosure(L, func, 1);                
        }
        else
        {
            lua_pushstring(L, method);
            lua_pushcclosure(L, func, 0);
        }
        lua_rawset(L, -3);
        lua_pop(L, 1);
        return true;
    }
    return false;
}

// The wxLuaObject class wraps a Lua object reference in
// a wxWindows object derived class so that a Lua object
// can be used for user data, and also with a simple
// extension by the input/output mechanism of the 
// wxValidator classes.

IMPLEMENT_DYNAMIC_CLASS(wxLuaObject, wxObject)

// Default constructor
wxLuaObject::wxLuaObject()
{
    // reset all the flags related to wxVariant
    // interfaces
    m_allocatedBool   = false;
    m_allocatedInt    = false;
    m_allocatedString = false;
    m_allocatedArray  = false;
    
    // nothing allocated yet
    m_luaState        = NULL;
    m_iReference      = LUA_NOREF;
}

// Constructor that is passed a lua state and a parameter index.

wxLuaObject::wxLuaObject(lua_State *L, int iParam, int iRef) : m_luaState(L)
{
    // reset all the flags related to wxVariant
    // interfaces
    m_allocatedBool   = false;
    m_allocatedInt    = false;
    m_allocatedString = false;
    m_allocatedArray  = false;

    // set up the reference
    m_iReference = tinsert(L, iParam);
}

wxLuaObject::~wxLuaObject()
{
    // If a refererence exists, remove it
    if (m_iReference != LUA_NOREF && m_luaState != NULL)
        tremove(m_luaState, m_iReference);
}

// Return a flag value that indicated whether the
// object is being used by a wxValidator class
// (else 0).
int  wxLuaObject::GetAllocationFlags() const
{
    int flags = 0;
    if (m_allocatedBool)
        flags = 1;
    if (m_allocatedInt)
        flags += 2;
    if (m_allocatedString)
        flags += 4;
    if (m_allocatedArray)
        flags += 8;
    return flags;
}

// Allow the flag value to be manipulated.
void wxLuaObject::ModifyAllocationFlags(int iValue)
{
    if (iValue & 1)
        m_allocatedBool = false;
    if (iValue & 2)
        m_allocatedInt  = false;
    if (iValue & 4)
        m_allocatedString = false;
    if (iValue & 8)
        m_allocatedArray = false;
    
    if (iValue & 16)
        m_allocatedBool = true;
    if (iValue & 32)
        m_allocatedInt = true;
    if (iValue & 64)
        m_allocatedString = true;
    if (iValue & 128)
        m_allocatedArray = true;
}

// Get the value of the reference object (or
// a proxy if the object has been aliased for a
// wxValidator class.
bool wxLuaObject::GetObject()
{
    if (m_allocatedBool)
    {       
        lua_pushnumber(m_luaState, m_bool);
        return true;
    }
    else
    if (m_allocatedInt)
    {
        lua_pushnumber(m_luaState, m_int);
        return true;
    }
    else
    if (m_allocatedString)
    {
        lua_pushstring(m_luaState, m_string.c_str());
        return true;
    }
    else
    if (m_allocatedArray)
    {
        lua_newtable(m_luaState);
        
        size_t idx;
        for (idx = 0; idx < m_arrayInt.Count(); ++idx)
        {
            lua_pushnumber(m_luaState, m_arrayInt.Item(idx));
            lua_rawseti(m_luaState, -2, idx);
        }
        return true;
    }
    return (m_iReference != LUA_NOREF && 
            tget(m_luaState, m_iReference) != 0);
}

// Set the reference and lua state up from a supplied
// parameter.
void wxLuaObject::SetObject(lua_State *L, int iParam)
{
    m_luaState = L;
    SetObject(iParam);
}

// Remove any existing reference and allocated another 
void wxLuaObject::SetObject(int iParam)
{
    if (m_iReference != LUA_NOREF)
        tremove(m_luaState, m_iReference);

    m_iReference = LUA_NOREF;    
    
    if (m_luaState != NULL)
    {
        m_iReference = tinsert(m_luaState, iParam);
    }
}

lua_State *wxLuaObject::GetLuaState() const
{
    return m_luaState;
}

// The following methods are used by the wxValidator interface
bool *wxLuaObject::GetBoolPtr()
{
    if (m_luaState != NULL && m_iReference != LUA_NOREF && GetObject())
    {
        m_bool          = (lua_tonumber(m_luaState, -1) != 0);
        m_allocatedBool = true;
    }    
    return &m_bool;
}

int *wxLuaObject::GetIntPtr()
{
    if (m_luaState != NULL && m_iReference != LUA_NOREF && GetObject())
    {
        m_int          = (int) lua_tonumber(m_luaState, -1);
        m_allocatedInt = true;
    }    
    return &m_int;
}

wxString *wxLuaObject::GetStringPtr()
{
    if (m_luaState != NULL && m_iReference != LUA_NOREF && GetObject())
    {
        m_string          = lua_tostring(m_luaState, -1);
        m_allocatedString = true;
    }    
    return &m_string;
}

wxArrayInt *wxLuaObject::GetArrayPtr()
{
    if (m_luaState != NULL && m_iReference != LUA_NOREF && GetObject())
    {
        if (lua_istable(m_luaState, -1))
        {
            int idx, nItems = luaL_getn(m_luaState, -1);
            if (nItems > 0)
                for (idx = 1; idx <= nItems; ++idx)
                {
                    lua_rawgeti(m_luaState, -1, idx);
                    int iValue = (int) getnumbertype(m_luaState, -1);
                    m_arrayInt.Add(iValue);
                    lua_pop(m_luaState, 1);
                }
        }
        lua_pop(m_luaState, 1);
        m_allocatedArray = true;
    }
    return &m_arrayInt;
}

#if 0
CheckStack::CheckStack(lua_State *lua_state, const char *msg)
{
    m_luastate = lua_state;
    m_msg      = msg;
    if (msg == NULL)
        m_msg = "";
    else
        m_msg = msg;
    m_top = lua_gettop(m_luastate);
}

CheckStack::~CheckStack()
{
    TestStack(m_msg);
}

void CheckStack::TestStack(const char *msg)
{
    if (msg == NULL)
        msg = m_msg;

    char buffer[256];
    wsprintf(buffer, "Stack state in %s: in %u out %u\r\n", msg, m_top, lua_gettop(m_luastate));
    OutputDebugString(buffer);
}

void CheckStack::DumpGlobals(lua_State *lua_state)
{
    lua_pushvalue(lua_state, LUA_GLOBALSINDEX);
    DumpTable(lua_state, lua_gettop(lua_state));
}

void CheckStack::DumpTable(lua_State *lua_state, const char *name)
{
    lua_pushstring(lua_state, name);
    lua_gettable(lua_state, LUA_GLOBALSINDEX);
    DumpTable(lua_state, lua_gettop(lua_state));
}

const char *CheckStack::GetTableInfo(lua_State *lua_state, int index, char *buffer)
{
    int         nTag   = ttag(lua_state, index);
    int         nItems = luaL_getn(lua_state, index); 
    const void *pItem  = lua_topointer(lua_state, index);
    
    if (nTag != TLUA_NOTAG)
    {
        if (nItems == 0)
            sprintf(buffer, "%p (tag %u)", pItem, nTag);
        else
            sprintf(buffer, "%p (tag %u, approx. %u items)", pItem, nTag, nItems);
    }
    else
    {
        if (nItems == 0)
            sprintf(buffer, "%p (no tag)", pItem);
        else
            sprintf(buffer, "%p (no tag, approx %u items)", pItem, nItems);        
    }
    return buffer;
}

void CheckStack::DumpType(lua_State *lua_state, const char *name, int index, char *outputbuffer)
{
        const char *type = NULL;
        const char *value = NULL;
    
        char buffer[256];
    
        switch (lua_type(lua_state, index))
        {
            default:
                sprintf(buffer, "%u", lua_type(lua_state, index));
                type = buffer;
                value = "?";
                break;
            
            case LUA_TNIL:
                type  = "Nil";
                value = "nil";
                break;

            case LUA_TNUMBER:
                type = "Number";
                value = MakeNumber(lua_tonumber(lua_state, index), buffer);
                break;

            case LUA_TSTRING:
                type  = "String";
                value = lua_tostring(lua_state, index);
                break;

            case LUA_TTABLE:
                type  = "Table";
                value = GetTableInfo(lua_state, index, buffer);
                break;

            case LUA_TFUNCTION:
                if (lua_iscfunction(lua_state, index))
                {
                    type  = "C Function";
                    sprintf(buffer, "%p", lua_tocfunction(lua_state, index));
                }
                else
                {
                    type = "Lua Function";
                    sprintf(buffer, "%p", lua_topointer(lua_state, index));
                }
                value = buffer;
                break;

            case LUA_TUSERDATA:
                type  = "User Data";
                value = GetUserDataInfo(lua_state, index, buffer);
                break;
    }
    sprintf(outputbuffer, "%s - type: %s, value: %s\n", name, type, value); 
}
  
const char *CheckStack::GetUserDataInfo(lua_State *lua_state, int index, char *buffer)
{
    const char *pTagName = NULL;
    int         nTag     = ttag(lua_state, index);
    
    if (nTag != TLUA_NOTAG)
        pTagName = GetLuaTagName(nTag);
    
    if (pTagName != NULL)
        sprintf(buffer, "%p (%s)", lua_touserdata(lua_state, index), pTagName);
    else
        sprintf(buffer, "%p (%u)", lua_touserdata(lua_state, index), nTag);
    return buffer;
}

const char *CheckStack::MakeNumber(double dnum, char *buffer)
{
    long num = (long) dnum;
    if ((double) num == dnum)
    {
        if (num >= 0)
            sprintf(buffer, "%lu (0x%lx)", num, num);
        else
            sprintf(buffer, "%ld (0x%lx)", num, num);
    }
    else
        sprintf(buffer, "%g", dnum);
    return buffer;
}

void CheckStack::DumpTable(lua_State *lua_state, int index)
{
    lua_pushnil(lua_state);
    while(lua_next(lua_state, index) != 0)
    {
        char outputbuffer[512];
        DumpType(lua_state, "index", -2, outputbuffer);
        DumpType(lua_state, "value", -1, outputbuffer + strlen(outputbuffer)); 
        OutputDebugString(outputbuffer);
        lua_pop(lua_state, 1);
    }
    lua_pop(lua_state, 1);
}
#endif

// The following functions support memory allocated for objects in
// the Lua system. 

void LUACALL removeTrackedMemory(void *pObject, bool fDelete)
{
    if (g_pTrackedList != NULL)
    {
        wxObject *pDeleteObject = g_pTrackedList->Delete((int) pObject);
        if (fDelete && pDeleteObject != NULL)
            delete pDeleteObject;
    }
}
    
void LUACALL addToTrackedMemoryList(wxObject *pObject)
{
    if (g_pTrackedList == NULL)
        g_pTrackedList = new wxHashTable(wxKEY_INTEGER);

    if (g_pTrackedList != NULL)
        g_pTrackedList->Put((int)pObject, pObject);
}

// These macros implement classes that encapsulate non-wxObject derived
// classes so that they can be tracked by the tracked memory system.

IMPLEMENT_ENCAPSULATION(wxPoint)
IMPLEMENT_ENCAPSULATION(wxSize)
IMPLEMENT_ENCAPSULATION(wxRect)
IMPLEMENT_ENCAPSULATION(wxFont)
IMPLEMENT_ENCAPSULATION(wxColour)
IMPLEMENT_ENCAPSULATION(wxColourData)
IMPLEMENT_ENCAPSULATION(wxTimeSpan)
IMPLEMENT_ENCAPSULATION(wxDateSpan)
IMPLEMENT_ENCAPSULATION(wxDateTime)
IMPLEMENT_ENCAPSULATION(wxAcceleratorEntry)
IMPLEMENT_ENCAPSULATION(wxListItemAttr)
IMPLEMENT_ENCAPSULATION(wxTreeItemId)
IMPLEMENT_ENCAPSULATION(wxFontEnumerator)
IMPLEMENT_ENCAPSULATION(wxFile)
IMPLEMENT_ENCAPSULATION(wxDir)
IMPLEMENT_ENCAPSULATION(wxDataFormat)
IMPLEMENT_ENCAPSULATION(wxDataObjectSimple)
IMPLEMENT_ENCAPSULATION(wxDataObjectComposite)
IMPLEMENT_ENCAPSULATION(wxDropSource)
IMPLEMENT_ENCAPSULATION(wxCalendarDateAttr)
IMPLEMENT_ENCAPSULATION(wxXmlProperty)
IMPLEMENT_ENCAPSULATION(wxXmlNode)
IMPLEMENT_ENCAPSULATION(wxRegEx)

// Pointers and counts to the information generated by the binder
// These are initialized in registerGeneratedClasses()
static size_t       classCount      = 0;
static WXLUACLASS  *pClassList      = NULL;
static size_t       defineCount     = 0;
static WXLUADEFINE *pDefineList     = NULL;
static size_t       eventCount      = 0;
static WXLUAEVENT  *pEventList      = NULL;
static size_t       objectCount     = 0;
static WXLUAOBJECT *pObjectList     = NULL;

// Function to compare the class tag of two classes
static int classListCompareFn(const void *p1, const void *p2)
{
    return (*((const WXLUACLASS *) p1)->class_tag) - (*((const WXLUACLASS *) p2)->class_tag);
}

// Find the WXLUACLASS that corresponds to a given tag.
// If found see if the required tag is derived from the
// tag of the class found.
static bool LUACALL IsDerivedTag(int iTag, int iParamTag)
{
    WXLUACLASS classItem;
    classItem.class_tag = &iTag;
    // this relies on LUA allocating tags in ascending order of definition
    // if LUA stops doing this, then the search may break.
    const WXLUACLASS *pLuaClass = (WXLUACLASS *) bsearch(&classItem, 
                                                         pClassList, 
                                                         classCount,
                                                         sizeof(WXLUACLASS),
                                                         classListCompareFn);

    if (pLuaClass != NULL)
        while (pLuaClass->baseclassindex != -1)
        {
            pLuaClass = pClassList + pLuaClass->baseclassindex;
            if (*pLuaClass->class_tag == iParamTag)
                return true;
        }
    return false;
}

// Return a pointer to the object that corresponds to the specified stack
// index and tag
void * LUACALL getuserdatatype(lua_State *L, int iParam, int iParamTag)
{    
    int   iTag;
    void *pReturn = NULL;
    if (!lua_isuserdata(L, iParam))
        goto error;

    if (lua_islightuserdata(L, iParam))
        goto error;
    
    iTag = ttag(L, iParam);
    if (g_functionTag == iTag)
    {
        wxLuaFunction *pFunction = (wxLuaFunction *) ttouserdata(L, iParam);
        pReturn = pFunction->GetObject();
    }
    else
    if (g_startTag <= iTag   && 
        g_lastTag  >= iTag   && 
        ((iTag == iParamTag) || IsDerivedTag(iTag, iParamTag)))
            pReturn = ttouserdata(L, iParam);
    else
    if (s_wxNull != iTag)
        goto error;

    return pReturn;

error:
    terror(L, wxString::Format("wxLua: Unexpected user defined data type for parameter %d.", iParam));
    return NULL;
}

// Push a data type onto the stack and set its tag
void LUACALL pushuserdatatype(lua_State *L, int iTag, const void *data)
{
    if (data != NULL)
        tpushusertag(L, data, iTag);
    else
        lua_pushnil(L);
}

// Validate that the object at the stack index specified is a string object 
// or that the object can be converted to a string.
// Return the string.
const char * LUACALL getstringtype(lua_State *L, int iParam)
{
    switch(lua_type(L, iParam))
    {
        case LUA_TNIL:
        case LUA_TSTRING:
        case LUA_TNUMBER:
            break;
        default:
            terror(L, wxString::Format("wxLua: Expected string for parameter %d.", iParam));
            break;
    }
    
	return lua_tostring(L, iParam);
}

// Validate that the object at the stack index specified is a boolean object 
// or that the object can be converted to a boolean.
// Return the boolean.
bool LUACALL getbooleantype(lua_State *L, int iParam)
{
    switch(lua_type(L, iParam))
    {
        case LUA_TNIL:
        case LUA_TNUMBER:
		case LUA_TBOOLEAN:
            /*OK*/
            break;
        default:
            terror(L, wxString::Format("wxLua: Expected boolean for parameter %d.", iParam));
            break;
    }
    
	int num = (int) lua_toboolean(L, iParam);

//  This test is not reliable FALSE == 0 but TRUE !=0 not TRUE == 1
//  The paramter can be the result of evaluating an expression as well as one of the two
//  boolean constants.    
//  if (!(num == TRUE) || !(num == FALSE))
//		terror(L, wxString::Format("wxLua: Expected 'TRUE' or 'FALSE' for parameter %d.", iParam));
    
	return num ? true : false;
}

// Validate that the object at the stack index specified is a enumeration object
// or that the object can be converted to a enumeration.
// Return the enumeration value.
long LUACALL getenumerationtype(lua_State *L, int iParam)
{
    switch(lua_type(L, iParam))
    {
        case LUA_TNIL:
        case LUA_TSTRING:
        case LUA_TNUMBER:
            /*OK*/
            break;
        default:
            terror(L, wxString::Format("wxLua: Expected enum number for parameter %d.", iParam));
            break;
    }
    
	return (long) lua_tonumber(L, iParam);
}

// Validate that the object at the stack index specified is a enumeration object
// or that the object can be converted to a enumeration.
// Return the enumeration value.
double LUACALL getnumbertype(lua_State *L, int iParam)
{
    switch(lua_type(L, iParam))
    {
        case LUA_TNIL:
        case LUA_TSTRING:
        case LUA_TNUMBER:
		case LUA_TBOOLEAN:
            /*OK*/
            break;
        default:
            terror(L, wxString::Format("wxLua: Expected number for parameter %d.", iParam));
            break;
    }

	return lua_tonumber(L, iParam);
}

// Validate that the object at the stack index specified is a table object
// Convert the table to an array of wxStrings. This assumes that each 
// numeric entry in the table is a string object or can be converted to a string
// Return a pointer to the array of wxStrings.
wxString * LUACALL tostringarray(lua_State *L, int iParam, int &count)
{
    int       idx;
    wxString *pItems = NULL;
    
    count = 0;
    
    if (lua_istable(L, iParam))
    {
        int nItems = luaL_getn(L, iParam);
        if (nItems > 0)
            pItems = new wxString[nItems];
        if (pItems != NULL)
        {
            for (idx = 1; idx <= nItems; ++idx)
            {
                lua_rawgeti(L, iParam, idx);
                const char *pString = getstringtype(L, -1);
                if (pString != NULL)
                    pItems[count++] = wxString(pString);
                lua_pop(L, 1);
            }
        }
    }
    return pItems;
}

#if wxCHECK_VERSION(2, 3, 0)	
// Convert a wxArrayString object to an table of strings 
int LUACALL tostringtable(lua_State *L, wxArrayString &files)
{
    lua_newtable(L);
    
    size_t idx;
    for (idx = 0; idx < files.Count(); ++idx)
    {
        const wxString &item = files.Item(idx);
        lua_pushstring(L, item.c_str());
        lua_rawseti(L, -2, idx + 1);
    }    
    return idx;
}
#endif

// Validate that the object at the stack index specified is a table object
// Convert the table to an array of ints. This assumes that each 
// numeric entry in the table is a numeric object or can be converted to a number
// Return a pointer to the array of ints.
int * LUACALL tointarray(lua_State *L, int iParam, int &count)
{
    int  idx;
    int *pItems = NULL;
    
    count = 0;
    
    if (lua_istable(L, iParam))
    {
        int nItems = luaL_getn(L, iParam);
        if (nItems > 0)
            pItems = new int[nItems];
        if (pItems != NULL)
        {
            for (idx = 1; idx <= nItems; ++idx)
            {
                lua_rawgeti(L, iParam, idx);
                int iValue = (int) getnumbertype(L, -1);
                pItems[count++] = iValue;
                lua_pop(L, 1);
            }
        }
    }
    return pItems;
}

const char ** LUACALL tochararray(lua_State *L, int iParam, int &count)
{
    int          idx;
    const char **pItems = NULL;
    
    count = 0;
    
    if (lua_istable(L, iParam))
    {
        int nItems = luaL_getn(L, iParam);
        if (nItems > 0)
            pItems = new const char *[nItems];
        if (pItems != NULL)
        {
            for (idx = 1; idx <= nItems; ++idx)
            {
                lua_rawgeti(L, iParam, idx);
                const char *pValue = getstringtype(L, -1);
                pItems[count++] = pValue;
                lua_pop(L, 1);
            }
        }
    }
    return pItems;
}

// Lookup a lua method called cpIndex in the WXLUACLASS array pClass
// Ensure that the lookup is of the required type. If the lua method
// can not be found on the current class recurse through base classes
// Return a pointer to the WXLUAMETHOD that corresponds to the method name,
// else return NULL;
WXLUAMETHOD * LUACALL getLuaMethod(WXLUACLASS *pClass, const char *cpIndex)
{    
    int iMethod;
	// find a method in the class, recurse through classes from which this class is derived.
    for (iMethod = 0; iMethod < pClass->num_methods; ++iMethod)
    {
        WXLUAMETHOD *pMethod = pClass->methods + iMethod;
        if ((pMethod->type == LuaMethod) && (strcmp(pMethod->name, cpIndex) == 0))
            return pMethod;
    }
    
    if (pClass->baseclassindex != -1)
        return getLuaMethod(pClassList + pClass->baseclassindex, cpIndex);
    
    return NULL;
}

// Function to compare to events by eventType
static int eventListCompareFn(const void *p1, const void *p2)
{
#if wxCHECK_VERSION(2, 3, 0)
    return (*((const WXLUAEVENT *) p1)->eventType) - (*((const WXLUAEVENT *) p2)->eventType);
#else
    return (((const WXLUAEVENT *) p1)->eventType) - (((const WXLUAEVENT *) p2)->eventType);
#endif
}

// Register the classes, definitions, objects and pointers generated by the binding
// Sort the event list into order for faster event handler processing.
void LUACALL registerGeneratedClasses(lua_State *L, int tableOffset, bool registerTypes)
{
	static const luaL_reg funcTable[] =
	{
		{"__gc",           garbageCollect  },
		{"__index",        getTableFunc   },
		{"__newindex",   setTableFunc   }
	};

    unsigned iClass;
	if (registerTypes)
	    g_startTag = tnewtag(L);
    
	int iTag = g_startTag;
	// install the classes, functions and methods
    pClassList = GetClassList(classCount);
    for (iClass = 0; iClass < classCount; ++iClass, 
		iTag = registerTypes ? tnewtag(L) : iTag + 1)
    {
        WXLUACLASS *pClass = pClassList + iClass;

        *pClass->class_tag = iTag;
        
        for (unsigned iFunction = 0; iFunction < sizeof(funcTable)/sizeof(funcTable[0]); iFunction++) 
        {
            tsettagmethod (L, iTag, funcTable[iFunction].name, funcTable[iFunction].func, (void *) pClass);
        }

        for (int iMethod = 0; iMethod < pClass->num_methods; ++iMethod)
        {
            WXLUAMETHOD *pMethod = pClass->methods + iMethod;
            if (pMethod->type == LuaConstructor || pMethod->type == LuaGlobal)
            {
                lua_pushstring(L, pMethod->name);
                lua_pushcfunction(L, pMethod->func);
                lua_rawset(L, tableOffset);
            }
        }
    }
    g_lastTag = iTag;

	// for backward compatibility
	lua_pushliteral(L, "FALSE");
	lua_pushboolean(L, false);
	lua_rawset(L, tableOffset);

	lua_pushliteral(L, "TRUE");
	lua_pushboolean(L, true);
	lua_rawset(L, tableOffset);

	// install the definitions and strings
    pDefineList = GetDefineList(defineCount);
    for (unsigned iDefine = 0; iDefine < defineCount; ++iDefine)
    {
        WXLUADEFINE *pDefine = pDefineList + iDefine;
        lua_pushstring(L, pDefine->name);
		if (pDefine->isString)
			lua_pushstring(L, pDefine->strValue);
		else
			lua_pushnumber(L, pDefine->value);
        lua_rawset(L, tableOffset);
    }

	// install the objects and pointers
    pObjectList = GetObjectList(objectCount);
    for (unsigned iObject = 0; iObject < objectCount; ++iObject)
    {
        WXLUAOBJECT *pObject = pObjectList + iObject;
        lua_pushstring(L, pObject->objName);
		if (pObject->objPtr != 0)
			tpushusertag(L, pObject->objPtr, *pObject->objClassTag);
		else
			tpushusertag(L, *pObject->pObjPtr, *pObject->objClassTag);
        lua_rawset(L, tableOffset);
    }

    // register all the wxWindows event types
    pEventList = GetEventList(eventCount);
    for (unsigned iEvent = 0; iEvent < eventCount; ++iEvent)
    {
        WXLUAEVENT *pEvent = pEventList + iEvent;
        lua_pushstring(L, pEvent->eventName);
#if wxCHECK_VERSION(2, 3, 0)
        lua_pushnumber(L, *pEvent->eventType);
#else
        lua_pushnumber(L, pEvent->eventType);
#endif
        lua_rawset(L, tableOffset);
    }
    
	// sort the event list into order for faster lookup.
	qsort(pEventList, 
          eventCount,
          sizeof(WXLUAEVENT),
          eventListCompareFn);
}

// Register any other objects that are not generated by the binding.
void LUACALL registerOthers(lua_State *L, bool registerTypes)
{
	// register our 'function' object handlers
	if (registerTypes)
        g_functionTag = tnewtag(L);
    
    tsettagmethod (L, g_functionTag, "__gc",    functionGarbageCollect);
    tsettagmethod (L, g_functionTag, "__call",   callFunction);
}

// A function object needs to be deleted because the object is being
// garbage collected.
int LUACALL functionGarbageCollect(lua_State *L)
{
    if (lua_isuserdata(L, 1) && (lua_islightuserdata(L, 1) == 0) && (ttag(L, 1) == g_functionTag))
    {
        wxLuaFunction *pFunction = (wxLuaFunction *)ttouserdata(L, 1, true);
        if (pFunction != NULL)
            delete pFunction;
    }
    return 0;
}

// if the class defines a gc function, then call it.
int LUACALL garbageCollect(lua_State *L)
{
    int retVal = 0;
	WXLUACLASS *pClass = (WXLUACLASS *)lua_touserdata(L, lua_upvalueindex(1));
    
    if ((pClass != NULL) && lua_isuserdata(L, 1) && (lua_islightuserdata(L, 1) == 0) && (ttag(L, 1) == *pClass->class_tag))
    {    
        if (g_pDerivedList != NULL)
        {
            wxHashTable *pHashTable = (wxHashTable *) g_pDerivedList->Delete((int) ttouserdata(L, 1, true));
            if (pHashTable != NULL)
                delete pHashTable;
        }
        
        int iMethod;
        for (iMethod = 0; iMethod < pClass->num_methods; ++iMethod)
        {
            WXLUAMETHOD *pMethod = pClass->methods + iMethod;
            if (pMethod->type == LuaDelete)
            {
                retVal = (*pMethod->func)(L);
                break;
            }
        }
    }
    return retVal;
}

// Find a derived method given an object and and a method name.
// If the method can be found, push it on to the Lua stack and return 1
// otherwise return 0 and nothing is pushed.
lua_State * LUACALL getDerivedMethod(void *pObject, const char *method)
{
    lua_State *luaState = NULL;
    if (g_pDerivedList != NULL)
    {
        wxHashTable *pHashTable = (wxHashTable *) g_pDerivedList->Get((int) pObject);
        if (pHashTable != NULL)
        {             
            wxLuaObject *pLuaObject = (wxLuaObject *) pHashTable->Get(wxString(method));
            if (pLuaObject != NULL && pLuaObject->GetObject())
                luaState = pLuaObject->GetLuaState();
        }
    }
    return luaState;
}

static const char baseString[] = "base_";

// Called by LUA to find the method that corresponds to
// a given method name. The class to lookup is in an
// upvalue. (gettable tag method).
// If the method is a function method push the create a
// wxLua function object onto the lua stack, setting its tag
// so that the function tag of the object will be called to
// invoke the function.
// todo: Handling of properties
int LUACALL getTableFunc(lua_State *L)
{
    bool        fFound  = false;
    int         result  = 0;
	WXLUACLASS *pClass  = (WXLUACLASS *)lua_touserdata(L, lua_upvalueindex(1));
	const char *cpIndex = "{unknown}";

    if ((pClass != NULL) && lua_isuserdata(L, 1) && (lua_islightuserdata(L, 1) == 0) && (ttag(L, 1) == *pClass->class_tag))
    {
        void *pObject = ttouserdata(L, 1);
        cpIndex = lua_tostring(L, 2);
        
        bool fCallbase = (memcmp(cpIndex, baseString, sizeof(baseString)-1) == 0);        
        if (fCallbase)
            cpIndex += sizeof(baseString)-1;
        else
        {
            if (getDerivedMethod(pObject, cpIndex) != NULL)
            {
                fFound = true;
                result = 1;
            }
        }
        
        if (!fFound)
        {
            WXLUAMETHOD *pMethod = getLuaMethod(pClass, cpIndex);
            if (pMethod != NULL)
            {
                fFound = true;
                wxLuaFunction *pFunction = new wxLuaFunction(pMethod, pClass, pObject);
                if (pFunction != NULL)
                {
                    tpushusertag(L, pFunction, g_functionTag);
                    result = 1;
                    fFound = true;
                }
            }
        }
    }
    
    if (!fFound)
	    terror(L, wxString::Format("wxLua: Attempt to call an invalid method '%s'.", cpIndex));
    
	return result;
}

// Called by LUA to find the method that corresponds to
// a given method name. The class to lookup is in an
// upvalue. (settable tag method).
// todo: Handling of properties
int LUACALL setTableFunc(lua_State *L)
{
	WXLUACLASS *pClass = (WXLUACLASS *)lua_touserdata(L, lua_upvalueindex(1));
	const char *cpIndex = "{unknown}";

    if ((pClass != NULL) && lua_isuserdata(L, 1) && (lua_islightuserdata(L, 1) == 0) && (ttag(L, 1) == *pClass->class_tag))
    {
        cpIndex = lua_tostring(L, 2);
        void *pObject = ttouserdata(L, 1);
        
        if (g_pDerivedList == NULL)
            g_pDerivedList = new wxHashTable(wxKEY_INTEGER);

        if (g_pDerivedList != NULL)
        {   
            bool         fCreated   = false;
            wxHashTable *pHashTable = (wxHashTable *) g_pDerivedList->Get((int) pObject);
            if (pHashTable == NULL)
            {                
                pHashTable = new wxHashTable(wxKEY_STRING);
                if (pHashTable != NULL)
                {
                    pHashTable->DeleteContents(true);
                
                    g_pDerivedList->Put((int) pObject, pHashTable);
                    fCreated = true;
                }
            }
            
            if (pHashTable != NULL)
            {
                wxString index(cpIndex);
                if (!fCreated)
                {
                    wxLuaObject *pObject = (wxLuaObject *)pHashTable->Delete(index);
                    if (pObject != NULL)
                        delete pObject;
                }
                pHashTable->Put(index, new wxLuaObject(L, 3));
            }
        }
    }
    
    return 0;
}

// Handler for the 'function' tag method.
int LUACALL callFunction(lua_State *L)
{
    if (lua_isuserdata(L, 1) && (lua_islightuserdata(L, 1) == 0) && (ttag(L, 1) == g_functionTag))
    {
        wxLuaFunction *pFunction = (wxLuaFunction *)ttouserdata(L, 1);
        return pFunction->CallMethod(L);
    }  
    return 0;
}

// Lookup and return the 'name' of a Lua object from its tag.
const char * LUACALL GetLuaTagName(int nTag)
{
    if (nTag == s_wxNull)
        return wxNull;
    if (nTag == g_functionTag)
        return "wxLuaFunction";

    if (pClassList != NULL && nTag >= g_startTag && nTag <= g_lastTag)
    {
        size_t iClass = 0;
        for (iClass = 0; iClass < classCount; ++iClass)
        {
            WXLUACLASS *pClass = pClassList + iClass;
    
            if (*pClass->class_tag == nTag)
                return pClass->name;
        }
    }
    return NULL;
}

// Connect an event to a handler. This Lua 'C' function supports
// function calls with either three or four parameters. These parameters
// are:         The class (which must be derived from wxEvtHandler),
//              The event type
// (Optional)   The ID of the object the event is for
//              A Lua function to call to handle the event.
//              The Lua function gets called with a single parameter 
//              which is a reference to the event object
//              associated with the event.
int LUACALL wxEvtHandler_ConnectEvent(lua_State *L)
{
    int          id;
    double       eventType = 0;
    LuaCallback *pCallback = NULL;

    int nParams = lua_gettop(L);

    wxEvtHandler *evtHandler = (wxEvtHandler *) getuserdatatype(L, 1, s_wxEvtHandler);

    switch (nParams)
    {
        case 4:
            {
                if (!lua_isfunction(L, 4))
				{
                    terror(L, "wxLua: ConnectEvent: Callback function is not a valid Lua function.");
					return 0;
				}

                if (lua_isnil(L, 3))
				{
	                terror(L, "wxLua: ConnectEvent: Invalid event type.");
					return 0;
				}
                else
                    eventType = lua_tonumber(L, 3);

                id  = (int) lua_tonumber(L, 2);
                if (lua_isnil(L, 2))
				{
	                terror(L, "wxLua: ConnectEvent: Invalid identifier.");
					return 0;
				}
				
				// Create and connect the callback
                pCallback = new LuaCallback(L, 4, id, (int) eventType, evtHandler);
            }
            break;

        case 3:
            {
                if (!lua_isfunction(L, 3))
				{
                    terror(L, "wxLua: ConnectEvent: Callback function is not a valid Lua function.");
					return 0;
				}

				if (evtHandler->IsKindOf(CLASSINFO(wxWindow)))
				{
					id = ((wxWindow *)evtHandler)->GetId();
				}
				else
				{
	                terror(L, "wxLua: ConnectEvent: Unexpected event object type.");
					return 0;
				}

                if (lua_isnil(L, 2))
				{
	                terror(L, "wxLua: ConnectEvent: Invalid event type.");
					return 0;
				}
                else
                {                    
                    eventType = lua_tonumber(L, 2);
					// Create and connect the callback
                    pCallback = new LuaCallback(L, 3, id, (int) eventType, evtHandler);
	            }
            }
            break;

        default:
            {
                terror(L, "wxLua: ConnectEvent: Incorrect number of parameters.");
            }
            break;
    }
	return 0;
}

// Attempt to cast an object reference (the first parameter) to another type.
// The type requested is specified by the second parameter. Presumably the 
// type requested will be derived from the supplied object, otherwise 
// bad things will happen.
int LUACALL wxObject_DynamicCast(lua_State *L)
{
    int         iResult   = 0;
    const char *className = lua_tostring(L, 2);
	lua_pushvalue(L, 1);
    if (className != NULL)
    {
		wxObject *pObject = ( wxObject * ) ttouserdata(L, 1 );
        unsigned iClass;
        for (iClass = 0; iClass < classCount; ++iClass)
        {
            WXLUACLASS *pClass = pClassList + iClass;
            if (strcmp(pClass->name, className) == 0)
            {
                if (pClass->pClassInfo != NULL)
                {
					if (pObject->IsKindOf(pClass->pClassInfo))
					{
						tsettag(L, *pClass->class_tag);
						iResult = 1;
					}
					else
						terror(L, "wxLua: DynamicCast: Unable to cast to a class of this type.");                
                }                
                break;
            }       
        }        
        if (iResult == 0)
            terror(L, "wxLua: DynamicCast: Class not found.");                
    }
    else
        terror(L, "wxLua: DynamicCast: Incorrect number of parameters.");        
    return iResult;
}

// Encapsulate a lua function reference for use by a event callback
LuaCallback::LuaCallback( lua_State    *state, 
						  int		    theRoutine,
						  int		    id, 
						  wxEventType   eventType, 
						  wxEvtHandler *pEvtHandler ) : m_state(state)
{ 
    m_routine      = tinsert(state, theRoutine);

    pEvtHandler->Connect( id, 
                          eventType,
						  (wxObjectEventFunction) &LuaCallback::EventHandler,
                          this);

	// Need to track events attached to an application object
	if (pEvtHandler->IsKindOf(CLASSINFO(wxApp)))
	{
		if (g_pAppHandlerList == NULL)
			g_pAppHandlerList = new wxList;

		if (g_pAppHandlerList != NULL)
			g_pAppHandlerList->Append(this);
	}
}

LuaCallback::~LuaCallback()
{
	// Remove the reference to the Lua function that we are
	// going to call
	if (m_state != NULL)
		tremove(m_state, m_routine);
}

// This function is called with "this" being of the type which the event is arriving
// at.  The user data is used to route it to the correct place.

void LuaCallback::EventHandler(wxEvent& event)
{
    LuaCallback *theCallback = (LuaCallback *) event.m_callbackUserData;    
    theCallback->CallFunction(&event);
}

// Call a lua function to handle an event. The lua function will receive
// a single parameter, the type of event.
void LuaCallback::CallFunction(wxEvent *pEvent)
{
	// Cannot call it if Lua is gone.
	if (m_state == NULL)
		return;

	int eventClassTag = s_wxEvent;

    WXLUAEVENT eventItem;
#if wxCHECK_VERSION(2, 3, 0)
    int        eventType = pEvent->GetEventType(); 
    eventItem.eventType = &eventType;
#else
    eventItem.eventType = pEvent->GetEventType();
#endif

    WXLUAEVENT *pLuaEvent = (WXLUAEVENT *) bsearch(&eventItem, 
                                                    pEventList, 
                                                    eventCount,
                                                    sizeof(WXLUAEVENT),
                                                    eventListCompareFn);

    if (pLuaEvent != NULL)
        eventClassTag = *pLuaEvent->eventClassTag;

    int oldTop = lua_gettop(m_state);
    if (tget(m_state, m_routine))
    {
        lua_pushvalue(m_state, LUA_GLOBALSINDEX);
        if (lua_setfenv(m_state, -2) != 0)
        {
            tpushusertag(m_state, pEvent, eventClassTag);
            LuaCall(m_state, 1, true);
        }
        else
            terror(m_state, "wxLua: CallFunction: function is not a Lua function.");
    }
    else
        terror(m_state, "wxLua: CallFunction: function has been garbage collected.");
    lua_settop(m_state, oldTop);
}

int LUACALL CreateStandaloneBitmaps(lua_State *luaState)
{
    
#include "../Examples/bitmaps/new.xpm"
#include "../Examples/bitmaps/open.xpm"
#include "../Examples/bitmaps/save.xpm"
#include "../Examples/bitmaps/saveall.xpm"
#include "../Examples/bitmaps/cut.xpm"
#include "../Examples/bitmaps/copy.xpm"
#include "../Examples/bitmaps/paste.xpm"
#include "../Examples/bitmaps/help.xpm"
    
    wxBitmap *returns;
    
    returns = new wxBitmap(new_xpm);
    addToTrackedMemoryList(returns);
    pushuserdatatype(luaState, s_wxBitmap, returns );   
    lua_setglobal(luaState, "newBitmap");
    
    returns = new wxBitmap(open_xpm);
    addToTrackedMemoryList(returns);
    pushuserdatatype(luaState, s_wxBitmap, returns );   
    lua_setglobal(luaState, "openBitmap");
    
    returns = new wxBitmap(save_xpm);
    addToTrackedMemoryList(returns);
    pushuserdatatype(luaState, s_wxBitmap, returns );   
    lua_setglobal(luaState, "saveBitmap");

    returns = new wxBitmap(saveall_xpm);
    addToTrackedMemoryList(returns);
    pushuserdatatype(luaState, s_wxBitmap, returns );   
    lua_setglobal(luaState, "saveAllBitmap");

    returns = new wxBitmap(cut_xpm);
    addToTrackedMemoryList(returns);
    pushuserdatatype(luaState, s_wxBitmap, returns );   
    lua_setglobal(luaState, "cutBitmap");
    
    returns = new wxBitmap(copy_xpm);
    addToTrackedMemoryList(returns);
    pushuserdatatype(luaState, s_wxBitmap, returns );   
    lua_setglobal(luaState, "copyBitmap");
    
    returns = new wxBitmap(paste_xpm);
    addToTrackedMemoryList(returns);
    pushuserdatatype(luaState, s_wxBitmap, returns );   
    lua_setglobal(luaState, "pasteBitmap");
    
    returns = new wxBitmap(help_xpm);
    addToTrackedMemoryList(returns);
    pushuserdatatype(luaState, s_wxBitmap, returns );   
    lua_setglobal(luaState, "helpBitmap");
    
    return 0;
}

