/////////////////////////////////////////////////////////////////////////////
// Purpose:     Implements the client end of wxLua debugging session
// Author:      J. Winwood
// Created:     May 2002
// RCS-ID:      $Id: wxLuaDebugTarget.cpp,v 1.1 2003/07/11 16:54:37 tulrich Exp $
// Copyright:   (c) 2002 Lomtick Software. All rights reserved.
// Licence:     wxWindows licence
/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
    #pragma implementation "wxLuaDebugTarget.h"
#endif

#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/wx.h"
#endif

#include "wxLuaDebugTarget.h"

static const char *debugName = "__wxLuaDebug__";
static const int  connectTimeout = 200;		// 20 seconds

//----------------------------------------------------------------------------

LuaDebugTarget::LuaDebugTarget(lua_State  *luaState, 
                               const char *serverName, 
                               int         portNumber) :
    m_luaState(luaState),
    m_portNumber(portNumber),
    m_serverName(serverName),
#if wxCHECK_VERSION(2, 3, 3)
    m_debugCondition(m_debugMutex),
#endif
    m_forceBreak(false),
    m_resetRequested(false),
    m_fConnected(false),
    m_fRunning(false),
    m_fStopped(false),
    m_fErrorsSeen(false),
    m_nFramesUntilBreak(0),
#if wxCHECK_VERSION(2, 3, 3)
    m_runCondition(m_runMutex),
#endif
	m_pThread(NULL)
{
    lua_pushlightuserdata(luaState, this);
    lua_setglobal(luaState, debugName);

    lua_sethook(luaState, LuaDebugHook, LUA_MASKCALL | LUA_MASKLINE | LUA_MASKRET, 0);
    
    lua_pushcfunction(m_luaState, LuaPrint);
	lua_setglobal(m_luaState, "print");    
    
	wxGetBaseApp()->EnterLuaCriticalSection();
}

LuaDebugTarget::~LuaDebugTarget()
{
	if (m_pThread != NULL)
		delete m_pThread;
    
	wxGetBaseApp()->LeaveLuaCriticalSection();
}

void LuaDebugTarget::DisplayError(const wxString &strError)
{
	NotifyError(strError);
}

bool LuaDebugTarget::WaitForConnect() const
{
	for (int idx = 0; idx < connectTimeout; ++idx)
	{
		if (m_fConnected)
			break;

		wxUsleep(100);
	}
	return m_fConnected;
}

bool LuaDebugTarget::Run()
{
	// Assume something is going to go wrong
	m_fErrorsSeen = true;

	m_pThread = new LuaThread(this);
	// Start the thread
    if (m_pThread != NULL						 &&
		m_pThread->Create() == wxTHREAD_NO_ERROR && 
        m_pThread->Run()    == wxTHREAD_NO_ERROR)        
    {
		// Wait for the connection to the server to complete
		if (!WaitForConnect())
		{
			wxMessageBox("Unable to connect to server", "wxLua client", wxOK | wxCENTRE, NULL);
		}
		else
		{
			// OK, now we can start running.
			m_runCondition.Wait();

			m_fRunning    = true;
			m_fErrorsSeen = false;
		    
			size_t idx;
			for (idx = 0; idx < m_bufferList.Count(); ++idx)
			{
				int rc = LUA_ERRERR;

				const wxLuaBuffer *item = m_bufferList.Item(idx);

                rc = LuaDoBuffer(m_luaState,
                                 item->GetBuffer(), 
                                 item->GetBufferLength(), 
                                 item->GetFileName());
                
				delete item;
		        
				switch (rc)
				{
					case LUA_ERRRUN:                    
						NotifyError("wxLua: Error while running chunk");
						m_fErrorsSeen = true;
						break;
					case LUA_ERRSYNTAX:
						NotifyError("wxLua: Syntax error during pre-compilation");
						m_fErrorsSeen = true;
						break;
					case LUA_ERRMEM:
						NotifyError("wxLua: Memory allocation error");
						m_fErrorsSeen = true;
						break;
					case LUA_ERRERR:
						NotifyError("wxLua: Generic error or an error occurred while running the error handler");
						m_fErrorsSeen = true;
						break;
					case LUA_ERRFILE:
						NotifyError("wxLua: Error occurred while opening file");
						m_fErrorsSeen = true;
						break;
					default:
						break;
				}

				if (m_fErrorsSeen)
					break;
			}
		}
	}
    return !m_fErrorsSeen;
}

void LuaDebugTarget::Stop()
{
    NotifyExit();        
    if (m_fConnected)
    {
        m_clientSocket.Shutdown(SD_BOTH);
        wxUsleep(100);
        m_clientSocket.Close();
    }
	m_pThread->Wait();
}

void *LuaDebugTarget::LuaThread::Entry()
{
	m_pTarget->ThreadFunction();
	return 0;
}

void LuaDebugTarget::LuaThread::OnExit()
{
}

bool LuaDebugTarget::LuaThread::TestDestroy()
{
	return wxThread::TestDestroy();
}

void LuaDebugTarget::ThreadFunction()
{
    bool fThreadRunning = false;
    try
    {
        m_clientSocket.Connect(m_serverName, m_portNumber);
        
        m_fConnected    = true;
        fThreadRunning = true;
    }
    catch(...)
    {
    }        
    
    while (fThreadRunning && !m_pThread->TestDestroy())
    {
        try
        {
            debugCommands debugCommand = (debugCommands) ReadByte(&m_clientSocket);
            switch(debugCommand)
            {
                case CMD_ADD_BREAKPOINT:
                    {
                        wxString fileName   = ReadString(&m_clientSocket);
                        int      lineNumber = ReadInt(&m_clientSocket);
                   
                        AddBreakPoint(fileName.c_str(), lineNumber);
                    }
                    break;
                
                case CMD_REMOVE_BREAKPOINT:
                    {
                        wxString fileName   = ReadString(&m_clientSocket);
                        int      lineNumber = ReadInt(&m_clientSocket);

                        RemoveBreakPoint(fileName.c_str(), lineNumber);
                    }
                    break;
                
                case CMD_CLEAR_ALL_BREAKPOINTS:
                    ClearAllBreakPoints();
                    break;
                
                case CMD_RUN_BUFFER:
                    {
                        wxString fileName = ReadString(&m_clientSocket);
                        wxString buffer   = ReadString(&m_clientSocket);

                        Run(fileName.c_str(), buffer.c_str());
                    }
                    break;
                
                case CMD_DEBUG_STEP:
					Step();
                    break;
                
                case CMD_DEBUG_STEPOVER:
                    StepOver();
                    break;
                
                case CMD_DEBUG_STEPOUT:
                    StepOut();
                    break;
                
                case CMD_DEBUG_CONTINUE:
                    Continue();
                    break;
                
                case CMD_DEBUG_BREAK:
                    Break();
                    break;

                case CMD_ENUMERATE_STACK:
                    EnumerateStack();
                    break;
                
                case CMD_ENUMERATE_STACK_ENTRY:
                    {
                        int stackRef = ReadInt(&m_clientSocket);
                        EnumerateStackEntry(stackRef);
                    }
                    break;
                
                case CMD_ENUMERATE_TABLE_REF:
                    {
                        int tableRef  = ReadInt(&m_clientSocket);
                        int index     = ReadInt(&m_clientSocket);
						int itemNode  = ReadInt(&m_clientSocket);
                        EnumerateTable(tableRef, index, itemNode);
                    }
                    break;
                    
                case CMD_RESET:
                    Reset();
                    break;
				
				case CMD_EVALUATE_EXPR:
					{
						int		 exprRef  = ReadInt(&m_clientSocket);
						wxString buffer   = ReadString(&m_clientSocket);
						EvaluateExpr(exprRef, buffer);
					}
					break;
                    
                case CMD_CLEAR_DEBUG_REFERENCES:
                    {
                        size_t idx, idxMax = m_references.GetCount();
                        for (idx = 0; idx < idxMax; ++idx)
                        {
                            int iItem = m_references.Item(idx);
                            tremove(m_luaState, iItem);
                        }
                    }
                    break;
            }
        }
        catch(...)
        {
            fThreadRunning = false;
        }
    }    
}

bool LuaDebugTarget::AddBreakPoint(const char *fileName, int lineNumber)
{                        
    wxCriticalSectionLocker locker(m_breakPointListCriticalSection);
    
    wxString breakPoint;
    breakPoint = breakPoint.Format("%u:%s", lineNumber, fileName);
        
    m_breakPointList.Add(breakPoint);
    
    return false;
}

bool LuaDebugTarget::RemoveBreakPoint(const char *fileName, int lineNumber)
{
    wxCriticalSectionLocker locker(m_breakPointListCriticalSection);

    wxString breakPoint;
    breakPoint = breakPoint.Format("%u:%s", lineNumber, fileName);
    
    m_breakPointList.Remove(breakPoint);
    
    return true;
}

bool LuaDebugTarget::ClearAllBreakPoints()
{
    wxCriticalSectionLocker locker(m_breakPointListCriticalSection);
    m_breakPointList.Clear();
    return true;
}

bool LuaDebugTarget::Run(const char *fileName, const char *buffer)
{
    wxLuaBuffer *pBuffer = new wxLuaBuffer(fileName, 
                                           buffer);
    if (pBuffer != NULL)
    {
        m_bufferList.Add(pBuffer);        
        return true;
    }
    return false;
}

bool LuaDebugTarget::Step()
{
    m_nextOperation = DEBUG_STEP;    
    if (!m_fRunning)
        m_runCondition.Signal();
    else 
    if (m_fStopped)
        m_debugCondition.Signal();
    return true;
}

bool LuaDebugTarget::StepOver()
{
    m_nFramesUntilBreak = 0;
    m_nextOperation = DEBUG_STEPOVER;    
    if (!m_fRunning)
        m_runCondition.Signal();
    else 
    if (m_fStopped)
        m_debugCondition.Signal();
    return true;
}

bool LuaDebugTarget::StepOut()
{
    m_nFramesUntilBreak = 1;    
    m_nextOperation = DEBUG_STEPOVER;    
    if (!m_fRunning)
        m_runCondition.Signal();
    else 
    if (m_fStopped)
        m_debugCondition.Signal();
    return true;
}

bool LuaDebugTarget::Continue()
{
    m_nextOperation = DEBUG_GO;    
    if (!m_fRunning)
        m_runCondition.Signal();
    else 
    if (m_fStopped)
        m_debugCondition.Signal();
    return true;
}

bool LuaDebugTarget::Break()
{
    m_forceBreak = TRUE;    
    return true;
}

bool LuaDebugTarget::Reset()
{
    m_forceBreak     = TRUE;    
    m_resetRequested = TRUE;
    if (!m_fRunning)
        m_runCondition.Signal();
    else 
    if (m_fStopped)
        m_debugCondition.Signal();
    return true;
}

bool LuaDebugTarget::EnumerateStack()
{    
    wxDebugData *pSortedList = new wxDebugData();
    if (pSortedList != NULL)
    {
        int       nIndex = 0;
        bool      fFirst = true;
        lua_Debug luaDebug;        
        
		wxGetBaseApp()->EnterLuaCriticalSection();
        
        while (lua_getstack(m_luaState, nIndex, &luaDebug) != 0)
        {
            if (lua_getinfo(m_luaState, "Sln", &luaDebug))
            {
                char buffer[256];
                buffer[0] = 0;
                // skip stack frames that do not have line number
                int  currentLine = luaDebug.currentline;
                if (!fFirst || currentLine != -1)
                {
                    if (currentLine == -1)
                        currentLine = 0;
                    
                    if (luaDebug.name != NULL)
                        sprintf(buffer, "function %s line %u", luaDebug.name, currentLine);
                    else
                        sprintf(buffer, "line %u", currentLine);
    
                    DebugData *item = new DebugData(buffer, "", "", nIndex, 0);
                    if (item != NULL)
                        pSortedList->Add(item);
                    
                    fFirst = false;
                }
            }
            
            ++nIndex;
        }
		wxGetBaseApp()->LeaveLuaCriticalSection();
    }
    return NotifyStackEnumeration(pSortedList);
}

bool LuaDebugTarget::EnumerateStackEntry(int stackRef)
{    
    wxDebugData *pSortedList = new wxDebugData;
    if (pSortedList != NULL)
    {
		wxGetBaseApp()->EnterLuaCriticalSection();
		{
			lua_Debug luaDebug;    
			if (lua_getstack(m_luaState, stackRef, &luaDebug) != 0)
			{
				int         idx  = 1;
				const char *name = lua_getlocal(m_luaState, &luaDebug, idx);
				while (name != NULL) 
				{
					bool        fIsTable = false;
					const char *type     = NULL;
					const char *value    = NULL;
					char buffer[256];
    
					switch (lua_type(m_luaState, -1))
					{
						case LUA_TNIL:
							type  = "Nil";
							value = "nil";
							break;

						case LUA_TBOOLEAN:
							type = "Boolean";
							value = MakeBoolean(lua_toboolean(m_luaState, -1));
							break;
                    
						case LUA_TNUMBER:
							type = "Number";
							value = MakeNumber(lua_tonumber(m_luaState, -1), buffer);
							break;
    
						case LUA_TSTRING:
							type  = "String";
							value = lua_tostring(m_luaState, -1);
							break;
    
						case LUA_TTABLE:
							type  = "Table";
							value = GetTableInfo(-1, buffer);
							fIsTable = true;
							break;
    
						case LUA_TFUNCTION:
							if (lua_iscfunction(m_luaState, -1))
							{
								type  = "C Function";
								sprintf(buffer, "%p", lua_tocfunction(m_luaState, -1));
							}
							else
							{
								type = "Lua Function";
								sprintf(buffer, "%p", lua_topointer(m_luaState, -1));
							}
							value = buffer;
							break;
    
						case LUA_TUSERDATA:
							type  = "User Data";
							value = GetUserDataInfo(-1, buffer);
							break;
                    
						case LUA_TLIGHTUSERDATA:
							type  = "Light User Data";
							value = GetUserDataInfo(-1, buffer, false);
							break;
                    
						case LUA_TTHREAD:
							type = "Thread";
							sprintf(buffer, "%p", lua_topointer(m_luaState, -1));
							value = buffer;
							break;
					}
    
					int nRef = LUA_NOREF;
    
					if (lua_istable(m_luaState, -1))
					{
						lua_pushvalue(m_luaState, -1);
						nRef = tinsert(m_luaState, -1);
						m_references.Add(nRef);
					}
    
					DebugData *item = new DebugData(name, type, value, nRef, 0);
					if (item != NULL)
						pSortedList->Add(item);
                
					name = lua_getlocal(m_luaState, &luaDebug, ++idx);
				}
			}    
		}
		wxGetBaseApp()->LeaveLuaCriticalSection();
    }
    return NotifyStackEntryEnumeration(stackRef, pSortedList);
}

bool LuaDebugTarget::EnumerateTable(int tableRef, int nIndex, int nItemNode)
{
    wxDebugData *pSortedList = new wxDebugData();
    if (pSortedList != NULL)
    {
		wxGetBaseApp()->EnterLuaCriticalSection();
		{
			if (tableRef == -1)
			{
				lua_pushvalue(m_luaState, LUA_GLOBALSINDEX);
				int nRef = tinsert(m_luaState, -1);
				pSortedList->Add(new DebugData("Globals", 
											   "", 
											   "", 
											   nRef, 
											   0));
				m_references.Add(nRef);
			}
			else
			{
				if (tget(m_luaState, tableRef) != 0)
				{        
					int nTop = lua_gettop(m_luaState);
    
					// start iterating
					lua_pushnil(m_luaState);    
					while (lua_next(m_luaState, nTop) != 0)
					{
						bool        fIsTable = false;
						const char *type     = NULL;
						const char *value    = NULL;
						const char *name     = NULL;
    
						char namebuffer[256];
    
						// get the index
						switch (lua_type(m_luaState, -2))
						{
							case LUA_TNIL:
								name = "Nil";
								break;
                        
							case LUA_TBOOLEAN:
								name = MakeBoolean(lua_toboolean(m_luaState, -2));
								break;
    
							case LUA_TNUMBER:
								name = MakeNumber(lua_tonumber(m_luaState, -2), namebuffer);
								break;
    
							case LUA_TSTRING:
								name = lua_tostring(m_luaState, -2);
								break;
    
							case LUA_TTABLE:
								name = GetTableInfo(-2, namebuffer);
								break;
    
							case LUA_TFUNCTION:
								if (lua_iscfunction(m_luaState, -2))
									sprintf(namebuffer, "%p", lua_tocfunction(m_luaState, -2));
								else
									sprintf(namebuffer, "%p", lua_topointer(m_luaState, -2));
								name = namebuffer;
								break;
    
							case LUA_TUSERDATA:
								name = GetUserDataInfo(-2, namebuffer);
								break;
                        
							case LUA_TLIGHTUSERDATA:
								name = GetUserDataInfo(-2, namebuffer, false);
								break;
                        
							case LUA_TTHREAD:
								sprintf(namebuffer, "%p", lua_topointer(m_luaState, -2));
								name = namebuffer;
								break;
						}
    
						char valuebuffer[256];
    
						// get the value
						switch (lua_type(m_luaState, -1))
						{
							case LUA_TNIL:
								type  = "Nil";
								value = "nil";
								break;

							case LUA_TBOOLEAN:
								value = MakeBoolean(lua_toboolean(m_luaState, -1));
								type = "Boolean";
								break;
                        
							case LUA_TNUMBER:
								value = MakeNumber(lua_tonumber(m_luaState, -1), valuebuffer);
								type = "Number";
								break;
    
							case LUA_TSTRING:
								type  = "String";
								value = lua_tostring(m_luaState, -1);
								break;
    
							case LUA_TTABLE:
								type  = "Table";
								value = GetTableInfo(-1, valuebuffer);
								fIsTable = true;
								break;
    
							case LUA_TFUNCTION:
								if (lua_iscfunction(m_luaState, -1))
								{
									type  = "C Function";
									sprintf(valuebuffer, "%p", lua_tocfunction(m_luaState, -1));
								}
								else
								{
									type = "Lua Function";
									sprintf(valuebuffer, "%p", lua_topointer(m_luaState, -1));
								}
								value = valuebuffer;
								break;
    
							case LUA_TUSERDATA:
								type  = "User Data";
								value = GetUserDataInfo(-1, valuebuffer);
								break;
                        
							case LUA_TLIGHTUSERDATA:
								type  = "Light User Data";
								value = GetUserDataInfo(-1, valuebuffer, false);
								break;
                        
							case LUA_TTHREAD:
								type = "Thread";
								sprintf(valuebuffer, "%p", lua_topointer(m_luaState, -1));
								value = valuebuffer;
								break;
						}
    
						int nRef = LUA_NOREF;
						if (fIsTable)
						{
							nRef = tinsert(m_luaState, -1);
							m_references.Add(nRef);
						}
						else
							lua_pop(m_luaState, 1);

						DebugData *item = new DebugData(name, type, value, nRef, nIndex);
						if (item != NULL)
							pSortedList->Add(item);    
					}
    
					// remove reference
					lua_pop(m_luaState, 1);
				}
			}
		}
		wxGetBaseApp()->LeaveLuaCriticalSection();
	}
    return NotifyTableEnumeration(nItemNode, pSortedList);
}

bool LuaDebugTarget::EvaluateExpr(int exprRef, const wxString &strExpr)
{
    wxString strResult("Error");
    int      nReference = LUA_NOREF;
    
    wxGetBaseApp()->EnterLuaCriticalSection();
	{
		if (strpbrk(strExpr.c_str(), " ~=<>+-*/%(){}[]:;,.\"'") != NULL)
		{        
			// an expression
			int nOldTop = lua_gettop(m_luaState);
                
            int nResult = luaL_loadbuffer(m_luaState, strExpr, strExpr.Length(), "debug");
            if (nResult == 0)
                nResult = lua_pcall(m_luaState, 0, LUA_MULTRET, 0);  /* call main */
                    
			switch (nResult)
			{
				default:
					if (lua_gettop(m_luaState) == nOldTop)
					{
						lua_pushliteral(m_luaState, "OK");
					}
					break;

				case LUA_ERRRUN:
					lua_pushliteral(m_luaState, "Error while running the chunk");
					break;

				case LUA_ERRSYNTAX:
					lua_pushliteral(m_luaState, "Syntax error during pre-compilation");
					break;

				case LUA_ERRMEM:
					lua_pushliteral(m_luaState, "Memory allocation error");
					break;

				case LUA_ERRERR:
					lua_pushliteral(m_luaState, "Unexpected error occurred while running the error message handler");
					break;
			}
        
			nReference = tinsert(m_luaState, -1);
			lua_settop(m_luaState, nOldTop);
		}
		else
		{
			 lua_Debug ar;
			 int	   iLevel = 0;
			 bool      fFound = false;
        
			 while (lua_getstack(m_luaState, iLevel++, &ar) != 0)
			 {
				int	   iIndex = 0;
				const char *name;
				while ((name = lua_getlocal(m_luaState, &ar, ++iIndex)) != NULL) 
				{
					if (strExpr == name)
					{
						nReference = tinsert(m_luaState, -1);
						fFound = true;
						break;
					}
                    else
                        lua_pop(m_luaState, 1);
				}
            
				if (fFound)
					break;
			 }

			 if (!fFound)
			 {
				  int nOldTop = lua_gettop(m_luaState);
				  lua_pushvalue(m_luaState, LUA_GLOBALSINDEX);
				  lua_pushnil(m_luaState);
				  while (lua_next(m_luaState, -2) != 0)
				  {
					  if (lua_type(m_luaState, -2) == LUA_TSTRING)
					  {
						  const char *name = lua_tostring(m_luaState, -2);
						  if (strExpr == name)
						  {
							  nReference = tinsert(m_luaState, -1); // reference value and pop it.
							  lua_pop(m_luaState, 1);	// pop index
							  fFound = true;
							  break;
						  }
						  else	
							  lua_pop(m_luaState, 1);  // removes `value'; 
					  }
					  else
						  lua_pop(m_luaState, 1);  // removes `value'; 
				  }
				  lua_settop(m_luaState, nOldTop); // the table of globals.
			 }
		}

		if (nReference != LUA_NOREF && tget(m_luaState, nReference) != 0)
		{           
			char buffer[256];
        
			tremove(m_luaState, nReference);
			switch (lua_type(m_luaState, -1))
			{
				case LUA_TNIL:
					strResult = "Nil";
					break;

				case LUA_TBOOLEAN:
					strResult = MakeBoolean(lua_toboolean(m_luaState, -1));
					break;

				case LUA_TNUMBER:
					strResult = MakeNumber(lua_tonumber(m_luaState, -1), buffer);
					break;

				case LUA_TSTRING:
					strResult = wxString::Format("\"%s\"", lua_tostring(m_luaState, -1));
					break;

				case LUA_TTABLE:                
					strResult = wxString::Format("table %s", GetTableInfo(-1, buffer));
					break;

				case LUA_TFUNCTION:
					if (lua_iscfunction(m_luaState, -1))
						strResult = wxString::Format("C-function %p", lua_tocfunction(m_luaState, -1));
					else
						strResult = wxString::Format("lua function %p", lua_topointer(m_luaState, -1));
					break;

				case LUA_TUSERDATA:
					strResult = wxString::Format("user data %s", GetUserDataInfo(-1, buffer));
					break;

				case LUA_TLIGHTUSERDATA:
					strResult = wxString::Format("light user data %s", GetUserDataInfo(-1, buffer, false));
					break;

				case LUA_TTHREAD:
					strResult = wxString::Format("thread %p", lua_topointer(m_luaState, -1));
					break;
                
				default:
					break;
			}

			lua_pop(m_luaState, 1);
		}
    }
    wxGetBaseApp()->LeaveLuaCriticalSection();
    
    return NotifyEvaluateExpr(exprRef, strResult);
}

const char *LuaDebugTarget::GetTableInfo(int index, char *buffer)
{
    int         nItems = luaL_getn(m_luaState, index); 
    const void *pItem  = lua_topointer(m_luaState, index);
    
    if (nItems == 0)
        sprintf(buffer, "%p", pItem);
    else
        sprintf(buffer, "%p (approx %u items)", pItem, nItems);        
    
    return buffer;
}

const char *LuaDebugTarget::GetUserDataInfo(int index, char *buffer, bool full)
{
    if (full)
    {
        int     nTag = ttag(m_luaState, index);
    
        const char *pTagName = NULL;
        if (nTag != TLUA_NOTAG)
            pTagName = GetLuaTagName(nTag);
    
        if (pTagName != NULL)
            sprintf(buffer, "%p (%s)", lua_touserdata(m_luaState, index), pTagName);
        else
            sprintf(buffer, "%p (%u)", lua_touserdata(m_luaState, index), nTag);
    }
    else
        sprintf(buffer, "%p", lua_touserdata(m_luaState, index));
        
    return buffer;
}

const char *LuaDebugTarget::MakeNumber(double dnum, char *buff)
{
    long num = (long) dnum;
    if ((double) num == dnum)
    {
        if (num >= 0)
            sprintf(buff, "%lu (0x%lx)", num, num);
        else
            sprintf(buff, "%ld (0x%lx)", num, num);
    }
    else
        sprintf(buff, "%g", dnum);
    return buff;
}

const char *LuaDebugTarget::MakeBoolean(int num)
{
    if (num) 
        return "True";
    return "False";
}

bool LuaDebugTarget::NotifyBreak(const char *fileName, int lineNumber)
{    
    if (WaitForConnect())
    {    
        WriteByte(&m_clientSocket, EVENT_DEBUG_BREAK);
        WriteString(&m_clientSocket, fileName);
        WriteInt(&m_clientSocket, lineNumber);
        return true;
    }
    return false;
}

bool LuaDebugTarget::NotifyPrint(const char *errorMsg)
{
    if (WaitForConnect())
    {
        WriteByte(&m_clientSocket, EVENT_DEBUG_ERROR);
        WriteString(&m_clientSocket, errorMsg);
        return true;
    }
    return false;
}

bool LuaDebugTarget::NotifyError(const char *errorMsg)
{
    if (WaitForConnect())
    {
        WriteByte(&m_clientSocket, EVENT_DEBUG_ERROR);
        WriteString(&m_clientSocket, errorMsg);
        return true;
    }
    else
        wxMessageBox(errorMsg, "wxLua client error", wxOK | wxCENTRE, NULL);

    m_fErrorsSeen = true;
    return false;
}

bool LuaDebugTarget::NotifyExit()
{
    if (WaitForConnect())
    {
        WriteByte(&m_clientSocket, EVENT_DEBUG_EXIT);
        return true;
    }
    return false;
}

bool LuaDebugTarget::NotifyStackEnumeration(wxDebugData *pDebugData)
{
    if (WaitForConnect())
    {
        WriteByte(&m_clientSocket, EVENT_DEBUG_STACK_ENUM);
        WriteDebugData(&m_clientSocket, pDebugData);
        return true;
    }
	delete pDebugData;
    return false;
}

bool LuaDebugTarget::NotifyStackEntryEnumeration(int          entryRef, 
                                                 wxDebugData *pDebugData)
{
    if (WaitForConnect())
    {
        WriteByte(&m_clientSocket, EVENT_DEBUG_STACK_ENTRY_ENUM);
        WriteInt(&m_clientSocket, entryRef);
        WriteDebugData(&m_clientSocket, pDebugData);
        return true;
    }
	delete pDebugData;
    return false;
}

bool LuaDebugTarget::NotifyTableEnumeration(int          itemNode,
                                            wxDebugData *pDebugData)
{
    if (WaitForConnect())
    {
        WriteByte(&m_clientSocket, EVENT_DEBUG_TABLE_ENUM);
		WriteInt(&m_clientSocket, itemNode);
        WriteDebugData(&m_clientSocket, pDebugData);
        return true;
    }
	delete pDebugData;
    return false;
}

bool LuaDebugTarget::NotifyEvaluateExpr(int				exprRef,
                                        const wxString &strResult)
{
    if (WaitForConnect())
    {
        WriteByte(&m_clientSocket, EVENT_DEBUG_EVALUATE_EXPR);
        WriteInt(&m_clientSocket, exprRef);
        WriteString(&m_clientSocket, strResult);
        return true;
    }
    return false;
}

const char *LuaDebugTarget::GetDebugInfo(int &lineNumber) const
{
	lua_Debug luaDebug;
	lua_getstack(m_luaState, 0, &luaDebug);
	lua_getinfo(m_luaState, "Sln", &luaDebug);
	lineNumber = luaDebug.currentline - 1;
	return luaDebug.source;
}

bool LuaDebugTarget::DebugHook(int event)
{    
    bool fWait = false;
    m_fStopped = true;
    if (m_forceBreak)
    {
        if (m_resetRequested)
		{
		    NotifyExit();        
            wxExit();
		}

		int         lineNumber = 0;
		const char *fileName   = GetDebugInfo(lineNumber);
        
        if (NotifyBreak(fileName, lineNumber))
            fWait = true;
    }
    else
    {
        if (event == LUA_HOOKCALL) // call
            m_nFramesUntilBreak++;        
        else
        if (event == LUA_HOOKRET || event == LUA_HOOKTAILRET) // return
        {
            if (m_nFramesUntilBreak > 0)
                m_nFramesUntilBreak--;        
        }
        else
        if (event == LUA_HOOKLINE) // line
        {
            switch (m_nextOperation)
            {               
                case DEBUG_STEP:
                    {
						int         lineNumber = 0;
						const char *fileName   = GetDebugInfo(lineNumber);
                        
                        if (NotifyBreak(fileName, lineNumber))
                            fWait = true;
                    }
                    break;
                    
                case DEBUG_STEPOVER:
                    if (m_nFramesUntilBreak == 0)
                    {
						int         lineNumber = 0;
						const char *fileName   = GetDebugInfo(lineNumber);
                        
                        if (NotifyBreak(fileName, lineNumber))
                            fWait = true;
                    }
                    break;
                    
                case DEBUG_GO:
                default:                    
                    {
						int         lineNumber = 0;
						const char *fileName   = GetDebugInfo(lineNumber);
                        
                        if (AtBreakPoint(fileName, lineNumber) && 
                            NotifyBreak(fileName, lineNumber))
                            fWait = true;
                    }
                    break;
            }
        }
    }
	
    if (fWait)
    {
		// release the critical section so
		// the other thread can access LUA
		wxGetBaseApp()->LeaveLuaCriticalSection();
		// Wait for a command
        m_debugCondition.Wait();        
		// acquire the critical section again
		wxGetBaseApp()->EnterLuaCriticalSection();
    }
    m_fStopped = false;
    return fWait;
}

void LUACALL LuaDebugTarget::LuaDebugHook(lua_State *luaState, lua_Debug *pLuaDebug)
{
    lua_getglobal(luaState, debugName);
    LuaDebugTarget *pTarget = (LuaDebugTarget *) lua_touserdata(luaState, -1);
    lua_pop(luaState, 1);

    if (pTarget != NULL)
        pTarget->DebugHook(pLuaDebug->event);
}

bool LuaDebugTarget::AtBreakPoint(const char *fileName,
                                  int         lineNumber) const
{
    wxCriticalSectionLocker locker(m_breakPointListCriticalSection);

    wxString breakPoint;
    breakPoint = breakPoint.Format("%u:%s", lineNumber, fileName);
        
    return (m_breakPointList.Index(breakPoint) != wxNOT_FOUND);
}

int LUACALL LuaDebugTarget::LuaPrint (lua_State *L) 
{
	int		 idx;
	wxString stream;
    int n = lua_gettop(L);  /* number of arguments */
    lua_getglobal(L, "tostring");
    for (idx = 1; idx <= n;  idx++) 
    {
        lua_pushvalue(L, -1);  /* function to be called */
        lua_pushvalue(L, idx);   /* value to print */
        lua_call(L, 1, 1);
        const char *s = lua_tostring(L, -1);  /* get result */
        if (s == NULL)
            return luaL_error(L, "`tostring' must return a string to `print'");
		if (idx > 1) 
			stream.Append("\t");
		stream.Append(s);
        lua_pop(L, 1);  /* pop result */
    }

    lua_getglobal(L, debugName);
    LuaDebugTarget *pTarget = (LuaDebugTarget *) lua_touserdata(L, -1);
    lua_pop(L, 1);

    if (pTarget != NULL)
        pTarget->NotifyPrint(stream.c_str());
    
	return 0;
}

