--
-- SC_MAIN.LUA
--
-- Lua front-end to SCons build engine ('www.scons.org').
--
-- This library presents a subset (with some lua-specific extensions) of the
-- SCons build engine. The primary intention is to be a Lua-scriptable "make
-- replacement" but the code can be used for any file dependency tasks, even
-- embedded within Lua applications.
--
-- The front-end can also be used together with other build engines (s.a. Jam)
-- by providing an appropriate hook function. The engine needs to be able to
-- scan #include dependencies and to decide, which targets need rebuilding and
-- which do not.
--
-- Also, a plain-and-simple "always build all" hook is included for reference,
-- and for people who don't want/need to install a real building engine.
--
-- Note: This script could be splitted into several files, but that would
--       make its launching harder for users (the main file would need to be
--       told where to find its siblings somehow) so we keep it monolithic 
--       for now.
--
--       Local function declarations are not used, in order to make this run
--       on Lua 4.0 as well.
--
-- Needs:   Lua 4.0 (or 5.0) (www.lua.org)
--          SCons v.0.11     (www.scons.org)    (optional)
--          Python v.1.5.2   (www.python.org)   (optional)
--          
-- License: GPL (see license.txt)
--
-- To-do:   ...
--
-- History:
--
-- 14-May-03/ak: Made compatible with older Lua 4.0, too (was quite a job!).
--               Changed license to GPL.
-- 6-May-03/ak:  Added 'sources=' for giving default sources (s.a. gluax_obj)
-- 16-Apr-03/ak: Test drove on Linux, some changes..
-- 7-Apr-03/ak:  Wrote some documentation. Program almost working...
--

--local m     -- namespace table
m=nil

if (USES) then   -- loading via 'USES' (GluaX)
    m= assert( arg[1] )
else            -- standalone via 'dofile', set up to "scons.*"
    m= {}
    scons= m
end

TRUE= true or "TRUE"    -- 'TRUE' is better for debugging than '1'
FALSE= false or nil

GLOBALS= getfenv and getfenv() or        -- 5.0 final
         getglobals and getglobals() or  -- 5.0 beta
         globals()                       -- 4.0

if (string) then
    _,_,tmp= string.find( _VERSION, "Lua (%d*)" )
else
    _,_,tmp= strfind( _VERSION, "Lua (%d*)" )
end
LUA_VER= tonumber(tmp)

if (LUA_VER >= 5.0) then
    ASSUME= ASSUME or assert
else
    ASSUME= function(v) assert(v); return v; end
end

--
if (LUA_VER < 5.0) then	    -- Lua 4.0 compatibility
    --
    ASSUME( not io )
    io= { write=write }

    io.open= function(fn,mode)
        --
        local obj={ h= openfile(fn,mode) }
        if (not obj.h) then return nil; end

        obj.write= function(obj,str) write(obj.h,str); end
        obj.close= function(obj) closefile(obj.h); end
        return obj
    end

    ASSUME( not string )
    string= { rep=strrep, sub=strsub, find=strfind }

    ASSUME( not table )
    table= { getn=getn, insert=tinsert, foreach=foreach }

    ASSUME( not os )
    os= { getenv=getenv, execute=execute, remove=remove }
end

ASSUME( io.write and io.open and 
        string.rep and string.sub and string.find and
        table.getn and table.insert and table.foreach and
        os.getenv and os.execute and os.remove )


--== DEBUGGING FUNCTIONS ==--

--
function WARN( str )
    print( "\nWARNING: "..str )
end

-----
-- void= Loc_Dump( tbl [,lev] )
--
function Loc_Dump( tbl, lev )
    --
    local items= FALSE
    local pre,pre2

    lev= lev or 0

    if (not tbl) then
        print "(nil)"
        return
    end

    pre= string.rep("  ", lev)

    for key,val in tbl do
        --
        items= TRUE

        pre2= pre..key.." : "

        if (type(val)=='table') then
            print( pre2.."table" )
            if (lev >= 0) then  -- recurse
                Loc_Dump( val, lev+1 )
            end
        elseif (type(val)=='string') then
            print( pre2.."\t'"..val.."'" )
        else
            print( pre2,val )
        end
    end

    if (not items) then
        print( pre.."(empty)" )
    end
end

dump= dump or Loc_Dump


--== TABLE FUNCTIONS ==--

-----
-- tbl2= Loc_TblCopy( tbl [,skip_func] )
--
function Loc_TblCopy( tbl1, skip_func )
    --
    local tbl2= {}

    for k,v in tbl1 do
        --
        if (skip_func and skip_func(k,v)) then
            -- Skip this field
        else
            tbl2[k]= v
        end
    end
    
    return tbl2
end


-----
-- bool= Loc_CoveringContents( tbl1, tbl2 )
--
-- Returns 'TRUE' if 'tbl1' covers all entries of 'tbl2' (with identical contents). 
-- This means that 'tbl2' is a subset of 'tbl1' entries.
--
function Loc_CoveringContents( tbl1, tbl2 )
    --
    tbl1.n= nil
    tbl2.n= nil
    
    for k,v in tbl2 do
        --
        if (tbl1[k] ~= v) then
            
            if ((type(tbl1[k])~="table") or (type(v)~="table")) then
                --
                return FALSE    -- some entry didn't exist or was different!
            end
            
            -- Subtables need to be dived into (different refs doesn't mean
            -- different contents).
            --
            if (not Loc_CoveringContents( tbl1[k], v )) then
                return FALSE
            end
        end
    end    
    
    return TRUE     -- covered it all!
end


-----
-- bool= Loc_IdenticalContents( tbl1, tbl2 )
--
-- Returns 'TRUE' if the two tables are content-wise identical, 'FALSE' if not.
--
function Loc_IdenticalContents( tbl1, tbl2 )
    --
    return (Loc_CoveringContents( tbl1, tbl2 ) and Loc_CoveringContents( tbl2, tbl1 ))
end


-----
-- 
function Loc_IsEmptyTable(tbl)

    if (tbl) then
        for _,_ in tbl do return FALSE; end
    end
        
    return TRUE   -- was empty
end


-----
-- keyed_tbl, indexed_tbl= Loc_TableSplit( tbl )
--
-- Splits a table to those items with a numeric (integer) key, and those
-- with other (string) keys.
--
-- Note: Returns 'nil' instead of empty tables.
--
function Loc_TableSplit( tbl )
    --
    local keyed_tbl= {}
    local indexed_tbl= {}
    
    if (not tbl) then return nil; end
    
    for k,v in tbl do
        --
        if (type(k) == "number") then
            indexed_tbl[k]= v
        elseif (k=="n") then
            -- ignore (Lua 4.0 automatic 'n' field)
        else
            keyed_tbl[k]= v
        end
    end

    if (Loc_IsEmptyTable(keyed_tbl)) then keyed_tbl= nil; end
    if (Loc_IsEmptyTable(indexed_tbl)) then indexed_tbl= nil; end

    return keyed_tbl, indexed_tbl
end


-----
-- tbl= Loc_TableFlatten( tbl )
--
-- Returns a table without subtables (flattens contents into the main table).
--
-- Note: Tables are expected to be just containers (index values don't matter).
--
function Loc_TableFlatten( tbl )
    --
    local ret= {}

    if (not tbl) then return nil; end

    for k,v in tbl do
        --
        if (type(v) == "table") then
            --
            local w= Loc_TableFlatten(v)

            for _,v2 in w do            
                table.insert( ret, v2 )
            end
    	elseif (k=="n") then    -- skip Lua 4.0 'n' fields
	        -- 
        elseif (not v) then     -- skip 'nil's and 'false's
            -- 
        else
            table.insert( ret, v )
        end
    end

    ret.n= nil  -- for Lua 4.0
    return ret
end

ASSUME( table.getn( Loc_TableFlatten( { 1,2,{3,4;n=2},5} ) ) == 5 )

-----
--
function Loc_IsFlatTableWithoutNuls( tbl )
    --
    if (tbl) then
        for _,val in tbl do
	    if ((not val) or (type(val)=="table")) then
                return FALSE
            end
        end
    end
    
    return TRUE
end


--== FILE SYSTEM FUNCTIONS ==--

-----
-- bool= Loc_IsSlashTerminated( str )
--
function Loc_IsSlashTerminated( str )
    --
    local c= string.sub( str, -1 )
    
    return ((c=="/") or (c=="\\"))
end

-----
-- str= Loc_MakeSlashTerminated( str )
--
function Loc_MakeSlashTerminated( str )
    --
    return Loc_IsSlashTerminated(str) and str or (str.."/")
end

-----
-- str= Loc_JoinPathAndFile( path, file )
--
function Loc_JoinPathAndFile( path, file )
    --
    return Loc_MakeSlashTerminated(path) .. file
end

-----
-- [tbl]= Loc_OsPathAsTable( [path] )
--
-- Returns the OS search path in table form (each directory as a separate entry).
--
-- Note: Each entry is slash terminated, to make concatenation with filenames easy.
--
function Loc_OsPathAsTable( path )
    --
    local path, head, tail
    local tbl= {}
    local sep= (Loc_PlatformDetect() == "win32") and ";" or ":"
    
    path= path or os.getenv( "PATH" )
    if (not path) then return nil; end

    -- Split the path at ';' (windows) or ':' (posix)
    --
    while( path and (path~="") ) do
        --
        _,_,head,tail= string.find( path, "(.-)"..sep.."(.*)" )

        table.insert( tbl, Loc_MakeSlashTerminated( head or path ) )
        
        path= tail
    end
    
    return tbl
end


-----
-- bool= Loc_FileExists( fn_str )
--
function Loc_FileExists( fn )
    --
    local fh= io.open(fn,"rt")
    
    if (fh) then
        fh:close(fh)
        return TRUE
    else
        return FALSE
    end
end


-----
-- fn[,ext]= Loc_StripSuffix( fn_str )
--
function Loc_StripSuffix( str )
    --
    local fn,ext
    
    if (str) then
        _,_,fn,ext= string.find( str, "(.*)%.(.*)" )

        if (fn) then
            return fn,ext
        end
    end

    return str
end

ASSUME( Loc_StripSuffix( "a.b.c.d" ) == "a.b.c" )
ASSUME( Loc_StripSuffix( "no_sfx" ) == "no_sfx" )


-----
-- path_str, filename_str= Loc_PathSplit( path_and_filename )
--
function Loc_PathSplit( str )
    --
    local path,fn
    
    _,_,path,fn= string.find( str, "(.*[/\\])(.*)" )

    if (path) then
        if (fn=="") then fn=nil; end
        return path,fn
    else
        return nil,str  -- just filename
    end
end

local p,r
p,r= Loc_PathSplit( "a/b/c/d" )
ASSUME( (p=="a/b/c/") and (r=="d") )

p,r= Loc_PathSplit( "a\\b\\c\\d" )
ASSUME( (p=="a\\b\\c\\") and (r=="d") )

p,r= Loc_PathSplit( "a\\b\\c\\d\\" )
ASSUME( (p=="a\\b\\c\\d\\") and (r==nil) )

p,r= Loc_PathSplit( "a_b_c_d" )
ASSUME( (p==nil) and (r=="a_b_c_d") )


--== SYSTEM INFORMATION ==--

-----
-- str= Loc_PlatformDetect()
--
-- Returns: "win32"   for any Windows OS
--          "posix"   for Linux & other Posix systems
--          "darwin"  for OS X
--          ...
--
-- Note: The OS is analyzed based on environment variables, file system etc.
--       that are available from pure Lua (without extensions).
--
function Loc_PlatformDetect( )
    --
    local tmp
    
    if (os.getenv("WINDIR")) then
        return "win32"
    end

    if (os.getenv("OSTYPE") == "darwin") then
        return "darwin" -- OS X
    end
    
    return "posix"  -- (default)
end


-----
-- str= Loc_CompilerDetect()
--
-- Returns: "msc" for MS Visual C++
--          "gcc" for Gnu GCC
--          'nil' for others (use default compiler)
--
-- Note: The compiler to be used can be analyzed from Lua by looking at the
--       PATH and trying to find 'cl.exe' (MSC) or 'gcc.exe'/'gcc' on it.
--       This is not bullet-proof but may be enough to give a default.
--
function Loc_CompilerDetect( )
    --
    if (Loc_PlatformDetect() ~= "win32") then
        return nil   -- Use default compiler
    end

    local msc_exe, gcc_exe= "cl.exe", "gcc.exe"
    local tbl

    tbl= Loc_OsPathAsTable()
    
    for _,dir in tbl do     -- One level at a time
        --
        if (msc_exe and Loc_FileExists( dir..msc_exe )) then
            return "msc"    -- Visual C++ found first!
        end
        if (gcc_exe and Loc_FileExists( dir..gcc_exe )) then
            return "gcc"    -- Gnu GCC found!
        end
        
        -- Note: SCons supports more compilers than this (s.a. Intel) and you
        --       can use them with SCons/Lua just as well.
    end

    return nil  -- No compiler found?!?    
end    


--== DEPENDENCY COLLECTION ==--

-----
-- Apply 'srcdir', 'objdir' or 'outdir' to node filenames:
--
-- 'val':       The value to be applied to (tbl, int or str)
-- 'basedir':   Current subpath where we're processing ('nil' for project root)
-- 'def_dir':   Default 'srcdir', 'objdir' or 'outdir' setting
-- 'autotarget': 'TRUE' if 'val' has been implicitly named
--
function Loc_ApplyDefaultDir( val, basedir, def_dir, autotarget, is_srcdir )
    --
    if (not val) then return nil; end

    if (type(val) == "table") then
        --
        local ret= {}
        
        for _,v in val do
            table.insert( ret, Loc_ApplyDefaultDir( v, basedir, def_dir, autotarget, is_srcdir ) )
        end
        
        return ret
    end
    
    if (type(val) == "number") then   -- Node id's
        return val
    end
    
    ASSUME( type(val) == "string" )   -- Filenames

    -- Rules: 
    --
    -- if not 'autotarget' and starts with '#' or '/':
    --      leave as is
    -- else
    --      if 'def_dir', apply it
    --      else apply 'basedir'
    --
    if (not autotarget) then
        --
        local c= string.sub( val, 1,1 )
    
        if ((c == '#') or (c == '/')) then
            return val
        end
    end

    if (def_dir) then
        --
        -- For 'objdir' and 'outdir', use just the filename (don't put stuff
        -- into subdirectories of the output folder).
        --
        -- For 'srcdir', use the subdirs as well (we're not generating anything
        -- there).
        --
        if (is_srcdir) then
            return Loc_JoinPathAndFile( def_dir, val )
        else
            local fn
            _,fn= Loc_PathSplit( val )     -- just the filename
        
            return Loc_JoinPathAndFile( def_dir, fn )
        end
    elseif (basedir) then
        return Loc_JoinPathAndFile( "#"..basedir, val )
    else
        return val
    end
end

-----
--
function Loc_AddPrefixes( val, prefix )
    --
    if (not val) then return nil; end

    if (type(val) == "table") then
        --
        local ret= {}
        
        for _,v in val do
            table.insert( ret, Loc_AddPrefixes( v, prefix ) )
        end
        
        return ret
    end

    ASSUME( type(val)=="string" )

    return prefix..val
end


-----
-- env_id= Loc_StoreEnvironment( env )
--
-- Returns an id number of an environment stored for output, or a previously 
-- stored that is content-wise identical with ours. 
--
-- Note: SCons/Lua specific fields are _not_ stored since they wouldn't be
--       passed to the build engine anyways.
--
function Loc_StoreEnvironment( env )
    --
    m._env_stack= m._env_stack or {}

    local stack= m._env_stack
    local cc= env.compiler
    
    -----
    -- Make a copy of the contents, and strip down extension fields. Also member 
    -- functions are skipped, which enhances speed and makes debug logs easier to read.
    --
    skip_these= { outdir=1, objdir=1, srcdir=1, sources=1 }  -- not 'incdir' (handled later)
    
    env= Loc_TblCopy( env, function( key, val )  -- skip if...
                             return ((type(val)=="function") or skip_these[key])
                           end )
        
    -----
    -- Merge 'incdir' to 'CPPPATH' (user may have used either one, or both).
    --
    if (env.incdir) then
        --
        env.CPPPATH= Loc_TableFlatten{ env.CPPPATH, env.incdir }
        env.incdir= nil
    end

    -----
    -- Apply 'optimize': "speed"= more optimization (gcc '-O2')
    --                   "size"=  (gcc '-Os')
    --                   "both"=  basic optimization (gcc '-O1', both size & speed)
    --                   "max"=   most optimization (gcc '-O3')
    --
    --                   "debug"= no optimizations, debug compile
    --
    if (env.optimize) then
        --
        local v= env.optimize
        local flag= nil

        if (v=="speed") then
            flag= (cc=="msc") and "/O2" or "-O2"
            
        elseif (v=="size") then
            flag= (cc=="msc") and "/O1" or "-Os"
            
        elseif ((v=="both") or (v==TRUE)) then
            flag= (cc=="msc") and "/O2" or "-O1"    -- MSC has no 'both' flag?
            
        elseif (v=="max") then
            flag= (cc=="msc") and "/Ox" or "-O3"
            
        elseif (v=="debug") then
            flag= (cc=="msc") and "/Zi" or "-g"
        else
            WARN ("Unknown setting: optimize='"..v.."' (ignored)" )
        end

        if (flag) then
            env.CPPFLAGS= Loc_TableFlatten{ env.CPPFLAGS, flag }
        end

        env.optimize= nil
    end

    -----
    -- Merge 'DEFS' to 'CPPFLAGS' (user may have used either one, or both).
    --
    if (env.DEFS) then

        env.CPPFLAGS= Loc_TableFlatten{ env.CPPFLAGS, 
                        Loc_AddPrefixes( env.DEFS, (cc=="msc") and "/D" or "-D" ) }
        env.DEFS= nil
    end

    -----
    -- 'CPPFLAGS_POST' come after any other CPPFLAGS (or defines)
    --
    -- Note: This has only visual impact (in compilation logs), not functional.
    --
    if (env.CPPFLAGS_POST) then

        env.CPPFLAGS= Loc_TableFlatten{ env.CPPFLAGS, env.CPPFLAGS_POST }
        env.CPPFLAGS_POST= nil
    end

    -----
    -- 'libs' (having possibly both path & filename) is split to 'LIBPATH' and 'LIBS'
    --
    if (env.libs) then
        for _,v in Loc_TableFlatten(env.libs) do
            local path,fn= Loc_PathSplit(v)
            Env_Append( env, { LIBS=fn, LIBPATH=path } )
        end
        env.libs= nil
    end

    -----
    -- Apply 'basedir' to "LIBPATH" etc. 
    --
    -- Note: This could be done better & perhaps in a better place, too?
    --
    local basedir= m._basedir
    local basedir_nohash= basedir
    
    if (basedir and (string.sub(basedir,1,1)=='#')) then    -- no-hash version
        basedir_nohash= string.sub(basedir,2)
    end
    
    env.CPPPATH= Loc_ApplyDefaultDir( env.CPPPATH, basedir, nil, FALSE )
    env.LIBPATH= Loc_ApplyDefaultDir( env.LIBPATH, basedir_nohash, nil, FALSE )
    --...    
    
    -----
    -- Already stored?
    --
    for i=1,table.getn(stack) do
        --
        if (Loc_IdenticalContents( env, stack[i] )) then
            return i    -- Found an identical environment!
        end
    end
    
    -- Our environment was unique - store it.
    --
    table.insert( stack, env )

    local id= table.getn(stack)
        
    if (SC_DEBUG) then
        print( "\n======= ENV "..id..": " )
        dump(env)
    end
    
    return id
end

-----
-- node_id|tbl= Loc_StoreNode( env, target_str, nodetype_str, src_tbl|str|int )
--
-- Returns either an index number (1..N) or a table of index numbers, usable
-- as sources to other target nodes.
--
function Loc_StoreNode( env, target_str, nodetype, sources, recursion_bool )
    --
    m._node_stack= m._node_stack or {}
    local stack= m._node_stack

    if (type(sources) ~= "table") then
        sources= {sources}
    end

    ASSUME( Loc_IsFlatTableWithoutNuls(sources) )
    
    -- "SharedObject" and "StaticObject" create N target files out of N sources
    -- (libraries and programs always create just a single target).
    --
    if (string.find( nodetype, "Object" )) then
        --
        if (sources[2]) then
            --
            if (target_str) then    -- Should NOT have a named target!
                --
                WARN( "Target name '"..target_str.."' ignored! (multiple objects)" )
            end

            local ret_tbl= {}
            
            sources.n= nil	-- (for Lua 4.0)
            
            for _,v in sources do
                table.insert( ret_tbl, Loc_StoreNode( env, nil, nodetype, v, TRUE ) )
            end
            
            return ret_tbl  -- A table of target id's
        end
    end

    -----
    -- Here we're creating just one target, one id.
    --
    
    -----
    -- Check whether any of the 'sources' is a library. If it is, add 
    -- appropriate 'LIBS' and 'LIBPATH' in the environment.
    --
    local libs= nil
    
    sources.n= nil	-- (lua 4.0)
    
    for _,v in sources do
        --
        if (type(v)=="number") then
            local node= m._node_stack[v]
            
            if (string.find( node.type, "Library" )) then
                libs= libs or {}
                table.insert( libs, node.target )   -- go on, may have others..
            end
        end
    end
    
    if (libs) then  -- Make a copy of 'env' only when we actually will modify it
        env= env:CopyAppend{ libs=libs }
    end
    
    -----
    -- Store a snapshot of the environment. If there already is an identical
    -- environment stored, we'll get the id to it.
    --
    local env_id= Loc_StoreEnvironment(env)

    -----
    -- Use default target names if the user didn't define any.
    --
    local autotarget= (not target_str)
    
    if (autotarget) then
        --
        target_str= Loc_StripSuffix( sources[1] )  -- source without ".c"

        if (sources[2]) then
            --
            WARN( "Target name missing for multiple sources! "..
                  "(guessing '"..target_str.."'...)" )
        end
    end

    -----
    -- Apply environment directories:
    --
    local basedir= m._basedir
    local lookup= { StaticObject=1, SharedObject=1 }
    
    if (lookup[nodetype]) then
        target_str= Loc_ApplyDefaultDir( target_str, basedir, env.objdir, autotarget )
    else
        target_str= Loc_ApplyDefaultDir( target_str, basedir, env.outdir, autotarget )
    end

    sources= Loc_ApplyDefaultDir( sources, basedir, env.srcdir, FALSE, TRUE )

    -----
    -- Expand implicit object targets (unnamed targets when doing 'Program' or
    -- 'Library' directly from sources) to nodes of their own. This is needed
    -- in order to apply 'objdir' to the intermediate targets.
    -- 
    lookup= { Program=1, StaticLibrary=1, SharedLibrary=1 }
    
    if (lookup[nodetype]) then  -- Note: not for 'Install', 'Alias', etc..
        --
        -- Any string sources will be replaced by a node reference instead.
        --
        local obj_type= string.find( nodetype, "Shared" ) and "SharedObject" 
                                                           or "StaticObject"
        for i=1,table.getn(sources) do
            --
            if (type(sources[i]) == "string") then
                --
                sources[i]= Loc_StoreNode( env, nil, obj_type, sources[i] )
            end
        end
    end
    
    -- Actually store the node:
    --
    local node= { type= nodetype,
                  target= target_str,  -- never 'nil'
                  sources= sources,    -- placed as table even if just one
                  env_id= env_id }

    table.insert( stack, node )

    local id= table.getn(stack)
    stack.n= nil    -- for Lua 4.0

    if (SC_DEBUG) then
        print( "\n-------- Node "..id..": " )
        dump(node)
    end
    
    return id
end


--== SCONS ENVIRONMENT OBJECT ==--

-----
-- env:Replace{ [fields] }
--
function Env_Replace( env, fields_tbl )

    if (not fields_tbl) then return env; end
    
    for k,v in fields_tbl do
        env[k]= v   --Loc_SetEnv(k,v)  -- overwrite previous keys
    end
    
    return env
end

-----
-- env2= env:Append{ [fields] }
--
function Env_Append( env, fields_tbl )
    
    if (not fields_tbl) then return env; end

    for k,v in fields_tbl do
        --
        if (not env[k]) then
            env[k]= v   -- no earlier field
        else
            -- Takes automatically care of str->table conversion and flattening:
            --
            env[k]= Loc_TableFlatten{ env[k], v }
        end                
    end
    
    return env
end


-----
-- env2= env:Copy{ [new_fields] }
--
function Env_Copy( env, fields_tbl )
    --
    return Env_Replace( Loc_TblCopy(env), fields_tbl )
end

-----
-- env2= env:CopyAppend{ [new_fields] }
--
function Env_CopyAppend( env, fields_tbl )
    --
    return Env_Append( Loc_TblCopy(env), fields_tbl )
end

-----
-- h_obj= env:Node( nodetype_str,   -- "StaticObject"/"Program"/...
--                  { [target=filename], 
--                    [keyname=value],  (environment overrides)
--                    ..sources.. } )
function Env_Node( env, nodetype_str, tbl )
    --
    local flags_tbl, sources_tbl, env_stored
    local target= tbl.target

    keyed_tbl, indexed_tbl= Loc_TableSplit( tbl )
    --
    -- 'keyed_tbl':   local overrides of environment parameters.
    -- 'indexed_tbl': sources, node id's, tables of these

    -- Remove subtable levels & apply default sources
    --
    if (not env.sources) then
        indexed_tbl= Loc_TableFlatten( indexed_tbl )
    else
        indexed_tbl= Loc_TableFlatten{ indexed_tbl, env.sources }
    end
    
    if (keyed_tbl) then
        keyed_tbl.target= nil
    
        if (Loc_IsEmptyTable(keyed_tbl)) then   
            keyed_tbl= nil
        end
    end
    
    if (keyed_tbl) then  -- need to make a local copy
        --
        local tbl1={}   -- collection for 'copy' fields
        local tbl2={}   -- collection for 'append' fields
        
        -- Note: Some fields are better to be 'overwriting' environment's 
        --       fields (s.a. 'srcdir' etc) while others (most, s.a. 'DEFS')
        --       are better appended to existing contents.
        --
        local lookup= { srcdir=1, objdir=1 }    -- 'copy' keys
        
        for key,val in keyed_tbl do
            if (lookup[key]) then 
                tbl1[key]= val
            else
                tbl2[key]= val
            end
        end

        env= env:Copy( tbl1 )
        env:Append( tbl2 )
    end

    -----
    -- Store the environment + node in the global state...
    --
    -- Note: Depending on 'nodetype', we get either a single integer id, or
    --       a table of ids (both can be used as sources to other nodes).
    --
    ASSUME( Loc_IsFlatTableWithoutNuls(indexed_tbl) )

    node_ids= Loc_StoreNode( env, target, nodetype_str, indexed_tbl )
    
    return node_ids
end

-----
-- h_obj= env:StaticObject{ [target=filename], 
--                          [keyname=value],  (environment overrides)
--                          ..sources.. }
function Env_StaticObject( env, tbl )
    --
    return Env_Node( env, "StaticObject", tbl )
end

--
function Env_SharedObject( env, tbl )
    --
    return Env_Node( env, "SharedObject", tbl )
end

--
function Env_StaticLibrary( env, tbl )
    --
    return Env_Node( env, "StaticLibrary", tbl )
end
    
--
function Env_SharedLibrary( env, tbl )
    --
    return Env_Node( env, "SharedLibrary", tbl )
end

--
function Env_Program( env, tbl )
    --
    return Env_Node( env, "Program", tbl )
end

--
function Env_Install( env, tbl )
    --
    if (not tbl.target) then
        tbl.target= env.outdir
    end

    return Env_Node( env, "Install", tbl )
end

--
function Env_Alias( env, tbl )
    --
    return Env_Node( env, "Alias", tbl )
end

-----
-- bool= Platform( match_str )
-- str= Platform( )
--
-- Returns: "win32"   for any Windows OS
--          "posix"   for Linux & other Posix systems
--          "darwin"  for OS X
--
function Env_Platform( env, match_str )
    
    if (match_str) then
        return (env.platform == match_str)    -- TRUE/FALSE
    else
        return env.platform
    end
end

-----
-- bool= Compiler( match_str )
-- str= Compiler()
--
-- Returns: "msc"   for MS Visual C++
--          "gcc"   for Gnu GCC (incl. Mingw32)
--           nil    for others/unknown
--
function Env_Compiler( env, match_str )
    
    if (match_str) then
        return (env.compiler == match_str)    -- TRUE/FALSE
    else
        return env.compiler
    end
end
    

-----
-- env_obj= Environment{ }  (constructor)
--
-- Fields:
--      .compiler
--      .platform
--      .srcdir
--      .incdir
--      .objdir
--      .outdir
--      .DEFS
--      ...     (any SCons fields s.a. CPPPATH etc.)
--
-- Methods:
--      :Copy
--      :Append
--      :CopyAppend
--      :Replace
--
--      :StaticObject (Object)
--      :SharedObject
--      :StaticLibrary (Library)
--      :SharedLibrary
--      :Program
--      :Install
--      :Alias
--
--      :Compiler
--      :Platform
--
function m.Environment( fields_tbl )

    if (not m._default_env) then    -- first time?
        --
        local env= {}

        -----    
        -- Method pointers:
        --
        env.Copy=           ASSUME( Env_Copy )
        env.Append=         ASSUME( Env_Append )
        env.CopyAppend=     ASSUME( Env_CopyAppend )
        env.Replace=        ASSUME( Env_Replace )
    
        env.SharedObject=   ASSUME( Env_SharedObject )
        env.StaticObject=   ASSUME( Env_StaticObject )
        env.Object=         env.StaticObject

        env.SharedLibrary=  ASSUME( Env_SharedLibrary )
        env.StaticLibrary=  ASSUME( Env_StaticLibrary )
        env.Library=        env.StaticLibrary

        env.Program=        ASSUME( Env_Program )
        env.Install=        ASSUME( Env_Install )
        env.Alias=          ASSUME( Env_Alias )
        
        env.Compiler=       ASSUME( Env_Compiler )
        env.Platform=       ASSUME( Env_Platform )

        -----
        -- Default data fields:
        --
        env.platform= Loc_PlatformDetect()
        env.compiler= Loc_CompilerDetect()
        
        m._default_env= env
    end

    -----
    -- User provided data fields:
    --
    return Env_Copy( m._default_env, fields_tbl )
end



--== SCRIPT GENERATION (SCons/Python) ==--

--
-- Default output hook, creating a Python/SCons script and running it.
--
-- Requires:    Python 1.5.2 (or newer)
--              SCons 0.11 (or newer)
--

-----
-- 'val' is either a filename, integer (node number) or a table of these. 
--
function Loc_PythonizeValues( val, prefix, quote )

    if (not val) then return nil; end

    if (type(val) == "table") then  -- Table to be expanded
        --
        local ret= "["

        val= Loc_TableFlatten(val)

        if (Loc_IsEmptyTable(val)) then
            return nil  -- nothing to output
        end

        for _,v in val do
            ret= ret.. Loc_PythonizeValues(v,prefix,quote) ..", "
        end
        
        -- Note: Need to remove last comma in case subtables are handled
        --       (we don't want empty commas in the middle of the array)
        --
        return string.sub( ret, 1,-3 ).."]"
    end

    if (type(val) == "number") then     -- Node indices
        prefix= "node"
        quote= FALSE
    else
        ASSUME( type(val) == "string" )
    end

    local p= prefix or ""
    local q= quote and "'" or ""

    return (q..p..val..q)
end

ASSUME( Loc_PythonizeValues( {"a","b"}, nil, TRUE ) == "['a', 'b']" )

-----
--
function Loc_OutputPyEnv( env_id, env )
    --
    local ret

    ret= "\nenv"..env_id.."= env0.Copy( \n"

    for k,v in env do
        --
        -- 'v' is either a string, or a table of strings.
        --
        if ((k=="compiler") or (k=="platform")) then  
            -- skip these
        else
            v= Loc_PythonizeValues( v, nil, TRUE )
            
            if (v) then
                ret= ret.."\t"..k.."= "..v..",\n"
            end
        end
    end
    
    return (ret.."\t )\n")
end

-----
--
function Loc_OutputPyNode( node_id, node )
    --
    local ret
    
    ret= "node"..node_id.."= env"..(node.env_id).."."..(node.type).."( "

    if (node.target) then   -- named target
        --
        ret= ret.."'"..(node.target).."', "
    end

    -- Simplify Python output if table only has one entry:
    --
    local sources= node.sources

    if (table.getn(sources)==1) then
        sources= sources[1]
    end
    
    local tmp= Loc_PythonizeValues( sources, nil, TRUE )
    
    ret= ret..tmp.." )\n\n"
         
    return ret
end

-----
-- str= Loc_Pythonize( env_stack, node_stack )
--
-- Creates the temporary Python script to be output in a temp file.
--
function Loc_Pythonize( env_stack, node_stack )
    --
    local py, tools, compiler, platform

    -----
    -- Note: We try to define the compiler to use (msc/gcc/other) from within 
    --       the SCons file (or makefile) even though the computer would
    --       have other compilers installed as well. This is different approach 
    --       from that of SCons/Python (where default compilers are used) and
    --       it seems to be a bit tricky to get it work the way we want.
    --
    --       For example, defining 'tools' separately for each environment did
    --       not seem to work right.. (so we do it once for all)
    --    
    compiler= env_stack[1].compiler    -- must be common to all environments!
    platform= env_stack[1].platform
    
    if (compiler == "msc") then      
        tools= "['msvc','mslink','mslib','masm']"
        --
    elseif (compiler == "gcc") then
        if (platform == "win32") then
            tools= "['gcc','g++','gnulink','mingw']"
        else
            tools= "['gcc','g++','gnulink']"    -- Linux etc.
        end
    end

    -----
    -- Header:
    --
    -- Note: If we don't declare "ENV=os.environ", SCons won't be able to
    --       find tools along the path, or pass environment variable values
    --       to them. If you'd like to disable this, use 'ENV=""' as parameter 
    --       to the environments you create.
    --
    py= "\n# ( Generated by SCons/Lua )\n\n"..
        "import os\n\n"..
        "env0= Environment( ENV=os.environ"
   
    if (tools) then
        py= py..", tools="..tools
    end

    -- This fixes a bug in SCons 0.13 that tries to feed '-shared' to OS X
    --    
    if (Loc_PlatformDetect() == "darwin") then
        py= py..", SHLINKFLAGS='$LINKFLAGS -dynamiclib'"

        -- ..and also changes default shared-lib suffix from ".so" to ".dylib"
        py= py..", SHLIBSUFFIX='.dylib'"
    end

    py= py.." )\n\n"
    
    -----
    -- Environments:
    --    
    py= py.."\n".."#--------\n"..
                  "# Environments:\n"..
                  "#\n"

    for i=1,table.getn(env_stack) do
        py= py..Loc_OutputPyEnv( i, env_stack[i] )
    end

    -----    
    -- Nodes:
    --    
    py= py.."\n".."#--------\n"..
                  "# Nodes:\n"..
                  "#\n"

    for j=1,table.getn(node_stack) do
        py= py..Loc_OutputPyNode( j, node_stack[j] )
    end

    return py
end


-----
-- Loc_PyTempFile( fn_str, contents_str )
--
--local
function Loc_PyTempFile( fn, str )
    --
    local fh= io.open( fn, "wt" )
      
    if (not fh) then
       error( "Unable to create temporary file '"..fn.."'!" )
    end
      
    fh:write( str )
    fh:close()
end


-----
-- rc_int [,err_str]= SConsHook( env_tbl, node_tbl, opts_tbl )
--
-- This is the default "output filter" and can be replaced by a custom hook.
-- It creates a temporary Python file, and runs it for the specified target(s).
--
-- 'opts_tbl':
--      .targets= "target1" | { "target1", "target2, ... }
--      .clean= TRUE|FALSE
--
function SConsHook( env_stack, node_stack, opts_tbl )
    --
    local targets= opts_tbl.targets
    local targets_str= ""
    local scons_flags= ""
    local temp_py= "~scons.tmp"    -- dependencies file
    local err= nil
    
    ASSUME( env_stack and (table.getn(env_stack) > 0) )
    ASSUME( node_stack and (table.getn(node_stack) > 0) )
    ASSUME( opts_tbl )

    -----
    -- Make a "target1 target2 ..." string:
    --
    if (targets) then
        --
        if (type(targets) ~= "table") then
            --
            targets_str= targets
        else
            for i=1,table.getn(targets) do
                --
                targets_str= targets_str..((i==1) and "" or " ")..targets[i]
            end
        end
    end
    
    -----
    -- Pythonizing...
    --
    local py_str= Loc_Pythonize( m._env_stack, m._node_stack )
    
    Loc_PyTempFile( temp_py, py_str )
        
    -----
    -- Running SCons...
    --
    scons_flags= "-f "..temp_py
    
    if (opts_tbl.clean) then
        scons_flags= scons_flags.." -c"
    end

    local scons_cmd= ( "scons "..scons_flags.." "..targets_str )

    --print( scons_cmd.."\n" )

    local rc= os.execute( scons_cmd )

    if (rc~=0) then
      err= ("Running '"..scons_cmd.."' failed!")
    end
     
    -----
    -- In case of failures, leave the tempfile for user study:
    --
    if (rc==0) then
      os.remove( temp_py )
    end

    return rc, err
end


--== SCRIPT GENERATION (plain 'build all' directly to compilers) ==--

-- This is a stand-alone hook needing no external build system, but also not
-- providing any dependency checks. That is, all targets are built from scratch
-- each time.

-----
-- rc_int [,err_str]= PlainHook( env_stack, node_stack, opts_tbl )
--
-- 'opts_tbl':
--      .targets= "target1" | { "target1", "target2, ... }
--      .clean= TRUE|FALSE
--
function PlainHook( env_stack, node_stack, opts_tbl )
    --
    dump( env_stack )
    dump( node_stack )
    
    error "NOT IMPLEMENTED YET!"
end


--== MAIN LEVEL ==--

-----
-- any= SConscript( filename_str [, arg1 [, ...] ] )
-- nil= SConscript( )
--
-- Returns anything that the called subscript returned.
-- Call with no params to clear current dependency tree (= reset the system).
--
-- Note: If 'filename' has only path (terminates with a slash) default 
--       filenames "SConscript.lua" and "SCons.lua" are used (in this order).
--
function m.SConscript( filename, ... )
    --
    local old_basedir= m._basedir
    local subpath
    local ret
    local args_tbl

    if (not filename) then   -- reset dependencies?
        m._env_stack= nil
        m._node_stack= nil
        return nil
        --
    elseif (type(filename) == "table") then  -- Using as 'SConscript{ ... }'
        --
        -- Note: We could support 'SConscript{ ... }' as well (putting all
        --       params into a single table) if the users find this annoying.
        --
        error "Use 'SConscript()' with paranthesis, not braces."
    else
        args_tbl= arg
    end

    if (old_basedir) then
        filename= old_basedir..filename
    end
    
    subpath,_= Loc_PathSplit( filename )
    
    if (subpath == filename) then   -- only pathname provided
        --
        filename= subpath.."SConscript.lua"     -- 1st default
        
        if (not Loc_FileExists( filename )) then
            --
            filename= subpath.."SCons.lua"      -- 2nd default
        end
    end

    if (subpath) then
        m._basedir= subpath
    end

    -- Store arguments in global 'arg' (for the subscript to read):
    --
    GLOBALS.arg= args_tbl

    print ( "Processing '"..filename.."'..." )
    
    ret= dofile( filename )
        
    m._basedir= old_basedir
    
    return ret
end


-----
-- rc, err= Build{ ..targets.. [,clean=TRUE] [,hook=func] }
--
-- Call this after all 'SConscript()'s to actually build the targets.
--
--   .targets= string or table of the targets to build (or clean)
--             'nil' (or no parameter) builds (or cleans) all targets.
--   .clean=   'TRUE' for cleaning the targets, instead of building them.
--   .hook=    function to be used for performing the build (allows use of
--             "third party" build engines with the front-end).
--
-- Note: Other fields of 'opts_tbl' can be used to pass information on to the
--       output filter (s.a. cmdline flags etc.) but this makes the building
--       specific to a certain build engine.
--
function m.Build( opts_tbl )
    --
    local rc, err
    local hook_func
    
    opts_tbl= opts_tbl or {}

    hook_func= opts_tbl.hook
    opts_tbl.hook= nil

    print "\n-----"

    if (not hook_func) then    -- select default hook (SCons or plain)
        --
        rc, err= os.execute( "scons -v" )   -- prints version & returns 0

        if (rc==0) then
            print ""
            hook_func= ASSUME( SConsHook )
        else
            print "SCons not found - using plain (build all)."
            hook_func= ASSUME( PlainHook )
        end
    end
    
    rc, err= hook_func( m._env_stack, m._node_stack, opts_tbl )  

    if (rc==0) then
      print "\nSuccess! :)"
    else
      print "\nBuild FAILED!"
    end
    
    return rc, err
end


-----
-- v1, v2, ...= m.Import( arg )
--
-- Returns the parameters provided for a subscript in the 'SConscript()' call.
--
function m.Import( arg_ )
    --
    ASSUME( arg_ )  -- If you want to import, you should have something to
                    -- import from...
    
    if (LUA_VER >= 5.0) then
        return unpack(arg_)   -- Lua 5.0..
    else
        return arg_[1], arg_[2], arg_[3], arg_[4], arg_[5], arg_[6], arg_[7], arg_[8], arg_[9]
    end
end

-----
-- Some functions are provided as a convenience for the SCons application scripts
-- as well:
--
m._FileExists= Loc_FileExists
m._MakeSlashTerminated= Loc_MakeSlashTerminated

-- Loaded ok:
return TRUE

